/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


// Constants
const CLASS_ID = Components.ID("{05aac0f1-503c-11de-8a39-0800200c9a66}");
const CLASS_NAME = "GTDInbox Preference Manager";
const CONTRACT_ID = "@gtdinbox.com/pref-manager;1";
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const nsIJSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON)

/**
 * Pref manager service class. Implemented as an XPCOM component to ensure only
 * one instance is ever in existence, it provides an interface to all preference
 * management required by the extension.
 *
 * @class giPrefManager
 */
function PrefManager() {
    // Required for pure JS XPCOM
    this.wrappedJSObject = this;

    // Initialize pref manager
    this._prefBranch = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch("extensions.gtdinbox.");
    this._prefBranch.QueryInterface(Ci.nsIPrefBranch2);
    this._changed = false;
    this._checkedPluginInit = {};

    this._saveTimer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    this.notify = this.save; // Alias them so that "this" can be used as callback

    // Call load to load prefs out of Firefox into memory
    this.load();
}

PrefManager.prototype = {
    /**
     * Loads the settings stored in Firefox pref manager and unserializes the
     * stored JSON.
     * @function ?
     */
    load: function() {
        // Grab and unserialize account pref data
        this._prefs = this._jsonRead('account_prefs', {});

        // Populate _accounts from _prefs
        this._accounts = [];
        for(var account in this._prefs) this._accounts.push(account);

        // Grab and unserialize account default prefs (for creating new accounts and reseting existing ones)
        this._defaults = this._jsonRead('account_default', {});

        // Grab and unserialize plugin version data
        this._pluginVersions = this._jsonRead('plugin_versions', {});
    },

    /**
     * Saves the in-memory prefs to the FF pref branch
     * @function ?
     */
    save: function() {
        this._prefBranch.setCharPref('account_prefs', nsIJSON.encode(this._prefs));
        this._prefBranch.setCharPref('account_default', nsIJSON.encode(this._defaults));

        this._saveTimer.cancel();
        this._changed = false;
    },

    /**
     * Resets all preferences across all accounts
     * @function ?
     */
    resetAll: function() {
        this._prefs = {};
        this._accounts = [];
        this._markChanged();
    },

    /**
     * Returns the data for the given pref path and account
     * @function {Object} ?
     * @param {String} account - The account name that the pref string is scoped to
     * @param {String} pref - The pref path
     * @return Pref data
     */
    getPref: function(account, pref) {
        // Initialize account if it hasn't yet been
        this._initializeAccount(account);

        // Split path on dots and add account in
        var prefPath = pref ? pref.split('.') : [];
        prefPath.unshift(account);

        // Find sought after pref, returning "undefined" if the path is invalid
        var wantedPref = this._prefs;
        while(prefPath.length) {
            var next = prefPath.shift();
            if(typeof wantedPref[next] == 'undefined') return undefined;
            wantedPref = wantedPref[next];
        }
        return wantedPref;
    },

    /**
     * Wrapper around getPref when getting plugin prefs
     * @function {Object} ?
     * @param {String} account - The account name that the pref string is scoped to
     * @param {String} plugin - The plugin name
     * @param {String} pref - The pref path
     * @return Pref data
     */
    getPluginPref: function(account, plugin, pref) {
        this._initializePluginPrefs(account, plugin);
        var prefPath = pref ? pref.split('.') : [];
        prefPath.unshift(plugin);
        return this.getPref(account, prefPath.join('.'));
    },

    /**
     * Sets the preference for the given account to the given value
     * @function ?
     * @param {String} account - The account name that the pref string is scoped to
     * @param {String} pref - The pref path
     * @param {Object} value - The value to set the preference to
     * @param {Boolean} instantSave - If true, save immediately. 
     */
    setPref: function(account, pref, value, instantSave) {
        // Split path on dots and add account in
        var prefPath = pref ? pref.split('.') : [];
        prefPath.unshift(account);

        // Set pref to given value, building tree as we go along
        var base = this._prefs;
        while(prefPath.length > 1) {
            var currentStep = prefPath.shift();
            if(typeof base[currentStep] == 'undefined') base[currentStep] = {};

            base = base[currentStep];
        }
        base[prefPath[0]] = value;
        if( typeof base[prefPath[0]]!="undefined" && (value===null || value===undefined) ) {
            // It is deleting an existing value, fully remove it
            delete base[prefPath[0]];
        }
        
        if( instantSave ) {
            this.save();
        } else {
            this._markChanged();
        }
    },

    /**
     * Wrapper around setPref when setting plugin prefs
     * @function ?
     * @param {String} account - The account name that the pref string is scoped to
     * @param {String} plugin - The plugin name
     * @param {String} pref - The pref path
     * @param {Object} value - The value to set the preference to
     * @param {Boolean} instantSave - If true, save immediately. 
     */
    setPluginPref: function(account, plugin, pref, value, instantSave) {
        this._initializePluginPrefs(account, plugin);
        var prefPath = pref ? pref.split('.') : [];
        prefPath.unshift(plugin);
        return this.setPref(account, prefPath.join('.'), value, instantSave);
    },
    
    /**
     * Return an array of known accounts (email addresses)
     * @function {String[]} ?
     * @return Array of email accounts with preferences
     */
    getAccounts: function() {
        var accounts = [];
        for( a in this._prefs ) {
            if( this._prefs.hasOwnProperty(a) ) {
                accounts.push(a);
            }
        }
        return accounts;
    },

    /**
     * Resets the prefs for the given plugin and account back to the defaults
     * @function ?
     * @param {String} account - The account name that the plugin is scoped to
     * @param {String} plugin - The plugin name
     */
    resetPluginPrefs: function(account, plugin) {
        var defaults = nsIJSON.decode(nsIJSON.encode(this._defaults[plugin]));
        this.setPref(account, plugin, defaults);
    },

    /**
     * Ensure the current prefs have all major key/values found in the defaults.
     * It is non-destructive: it does not alter any values found in the current prefs.
     * @function {} ?
     * @param {String} account - The account name that the plugin is scoped to
     * @param {String} plugin - The plugin name
     */
    extendPluginPrefsWithDefaults: function(account, plugin) {
        var defaults = nsIJSON.decode(nsIJSON.encode(this._defaults[plugin]));
        var current = this.getPref(account, plugin);
        this._extendPluginPrefsWithDefaults(defaults, current);
        //this.setPref(account, plugin, current); // Unnecessary step - should be all pointing to same object
    },
    /**
     * Extend the pref object for the account with any missing keys found in defaultPrefs (i.e. new items)
     * @function {}
     * @param {Object} defaultPrefs - The default preferences for the plugin
     * @param {Object} accountPrefs - The preferences for the plugin on a given account. Get modified.
     */
    _extendPluginPrefsWithDefaults: function(defaultPrefs, accountPrefs) {
        for( p in defaultPrefs ) {
            if( defaultPrefs.hasOwnProperty(p) ) {
                if( accountPrefs[p] ) {
                    // Exists, go recurse deeper
                    if( defaultPrefs[p].constructor && defaultPrefs[p].constructor==Object ) {
                        this._extendPluginPrefsWithDefaults(defaultPrefs[p], accountPrefs[p]);
                    }
                } else {
                    // Does not exist; add it and move on (no need to test children)
                    accountPrefs[p] = defaultPrefs[p];
                }
            }
        }
    },

    /**
     * Returns the defaults for the given plugin
     * @function {Object} ?
     * @param {String} plugin - The plugin name
     * @return The defaults
     */
    getPluginDefaults: function(plugin) {
        return this._defaults[plugin];
    },

    /**
     * Sets the default object for the given plugin
     * @function ?
     * @param {String} plugin - The plugin name
     * @param {Object} defaults - The default data
     */
    setPluginDefaults: function(plugin, defaults) {
        this._defaults[plugin] = defaults;
        this._markChanged();
    },

    /**
     * Returns the version of the given plugin currently on record
     * @function {String} ?
     * @param {String} plugin - The id of the plugin
     * @return The version string
     */
    getPluginVersion: function(plugin) {
        return this._pluginVersions[plugin];
    },

    /**
     * Sets the version of the given plugin to the given value and saves it
     * @function ?
     * @param {String} plugin - The id of the plugin
     * @param {String} newVersion - The new version string
     */
    setPluginVersion: function(plugin, newVersion) {
        this._pluginVersions[plugin] = newVersion;
        this._prefBranch.setCharPref('plugin_versions', nsIJSON.encode(this._pluginVersions));
    },

    /**
     * Populates the account with the default plugin settings if it is completely
     * empty
     * @function ?
     * @param {String} account - The account name
     */
    _initializeAccount: function(account) {
        if(typeof this._prefs[account] != 'undefined') return;
        this.setPref(account, null, nsIJSON.decode(nsIJSON.encode(this._defaults)));
        this._accounts.push(account);
    },

    /**
     * Populates the plugin settings from the defaults if not yet initialized
     * @function ?
     */
    _initializePluginPrefs: function(account, plugin) {
        if(this._checkedPluginInit[account+'|'+plugin]) return;

        // Retrieve data and if it is undefined, populate the account settings from the defaults
        var data = this.getPref(account, plugin);
        if(data === undefined) {
            var defaults = nsIJSON.decode(nsIJSON.encode(this._defaults[plugin]));
            this.setPref(account, plugin, defaults);
        }

        this._checkedPluginInit[account+'|'+plugin] = true;
    },

    /**
     * Marks the prefs as being changed and schedules a save for 15 seconds from now
     * @function ?
     */
    _markChanged: function() {
        this._changed = true;
        this._saveTimer.cancel();
        this._saveTimer.initWithCallback(this, 15*1000, Ci.nsITimer.TYPE_ONE_SHOT);
    },

    /**
     * Reads the given pref from the pref branch and decodes it, returning the
     * default if it fails
     * @function {Object} ?
     * @param {String} pref - The pref path to read from
     * @param {Object} defaultValue - The object to return if reading fails
     * @return The pref value
     */
    _jsonRead: function(pref, defaultValue) {
        var prefData = defaultValue;
        try {
            var prefString = this._prefBranch.getCharPref(pref);
            var prefData = nsIJSON.decode(prefString);
        } catch(e) {}
        return prefData;
    },

    QueryInterface: function(aIID) {
        if (!aIID.equals(Ci.nsISupports)) throw Cr.NS_ERROR_NO_INTERFACE;
        return this;
    }
};

// Other methods
function log(msg) {
    Cc['@mozilla.org/consoleservice;1'].getService(Ci.nsIConsoleService).logStringMessage(msg);
}

// Factory Definition
var gPrefManagerSingleton = null;
var PrefManagerFactory = {
    createInstance: function (aOuter, aIID) {
        if (aOuter != null) throw Cr.NS_ERROR_NO_AGGREGATION;

        if(!gPrefManagerSingleton) gPrefManagerSingleton = new PrefManager();
        return gPrefManagerSingleton.QueryInterface(aIID);
    }
};

// Module definition
var PrefManagerModule = {
    registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
        aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
        aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
    },

    unregisterSelf: function(aCompMgr, aLocation, aType) {
        aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
        aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);
    },

    getClassObject: function(aCompMgr, aCID, aIID) {
        if (!aIID.equals(Components.interfaces.nsIFactory)) throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        if (aCID.equals(CLASS_ID)) return PrefManagerFactory;

        throw Components.results.NS_ERROR_NO_INTERFACE;
    },

    canUnload: function(aCompMgr) { return true; }
};
function NSGetModule(aCompMgr, aFileSpec) { return PrefManagerModule; }