/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The extensible Nav Box
 *
 * @class ? @extends giUIObject
 */
var boxBox = giUIObject.extend({
    // Properties {
    
	
	/** @property {Element} ? The root element */
	el: null,
    
    
	/** @property {Element} ? The header element */
	elHeader: null,
    
    /**
     * The 'New' icon for the blog, in the header
     * @property {Element}
     */
    _elHeaderBlogNew: null,
    
    /**
     * A collection of elements: {root:, title:, content:}
     * @property {private Object}
     */
    _navBox: null,
	
    
    /**
     * An array of sections
     * @property {private boxSection[]}
     */
    _sections: null,
    
    
    /**
     * True if loaded
     * @property {private boolean}
     */
    _isLoaded: false,
    
    /**
     * An array of addSection requests, if made before we are loaded
     * Format: [[boxSection, boolean]]
     * @property {private [][]}
     */
    _sectionQueue: null,
    
    
    //}

    // Methods {
    /**
     * Connects the element to the given boxMain object's page. Called
     * automatically by {@link boxMain.getUIObject}.
     * @function ?
     * @param {boxMain} main - The boxMain object for the page
     */
    connect: function(main) {
        giLogger.log('Connecting boxBox object to view', 'boxBox');
        
        this.base(main);
        this._sections = [];
        this._sectionQueue = [];
        // Construct the nav box and attach it:
        this._createNavBox();
               
    },
    
    /**
     * Try to create the nav box
     * @function {} ?
     * @param {optional int} id - The ID of the navbox created; if created but not retrieved
     */
    _createNavBox: function(id, count) {
        var result = this._main.gmail.page.addNavModule(id);
        
        if( result.navBox ) {
            // Success
            this._navBox = result.navBox;
            this._UIReady();
        } else {
            // Failed
            if( !count ) count = 1;
            var me = this;
            if( count < 10 ) setTimeout( function(){me._createNavBox(result.id, count++)}, 500 );
        }
    },
    
    /**
     * In response to offline mode, potentially defer the load
     * @function {}
     */
    _UIReady: function() {
        this.el = this._navBox.root;
        this.el.className += " gtdi gtdi-box";
        
        // Set the title:
        this.elHeader = this._navBox.title;
        this.elHeader.innerHTML = "<img src='chrome://gtdinbox/skin/box/i/gtdi-logo-mini.png'/>"
        
        // Add the Offline Warning
        var offlineSection = this._main.getUIObject('offlineWarning');
        this.addSection(offlineSection)
        
        // Add the links (this is key to all boxes)
        var linksSection = this._main.getUIObject('links');
        this.addSection(linksSection)
        
        
        // Connect listener for the New icon
        linksSection.boxNews.addEventListener(boxNews.BLOG_UPDATED, this._blogUpdated, this);
        this._elHeaderBlogNew = this.el.ownerDocument.createElement("IMG");
        this._elHeaderBlogNew.className = 'gtdi-news-new';
        this._elHeaderBlogNew.style.display = "none";
        this._elHeaderBlogNew.src = "chrome://gtdinbox/skin/box/newitem.png";
        this.elHeader.appendChild( this._elHeaderBlogNew );
        
        this._isLoaded = true;
        
        // See if any sections queued:
        for( var i = 0; i < this._sectionQueue.length; i++ ) {
            this.addSection( this._sectionQueue[i][0], this._sectionQueue[i][1] );
        }
        this._sectionQueue.length = 0;
    },
    
    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting boxBox from view', 'boxBox');

        this.el.parentNode.removeChild(this.el);
        this.el = null;
        this.elHeader = null;

        for( var i = 0; i < this._sections.length; i++ ){
            this._sections[i].disconnect();
        }
        this._sections.length = 0;

        this.base();
    },
    
    
    
    /**
     * Add a section object to the list of sections; and add it to the UI
     * @function {} ?
     * @param {boxSection} section - The section object
     * @param {Boolean} insertBeforeLinks - If true, insert before the links section (no matter what is before it)
     */
    addSection: function(section, insertBeforeLinks) {
        
        if( !this._isLoaded ) {
            // Queue it up:
            this._sectionQueue.push([section, insertBeforeLinks]);
            return;
        }
        
        if( insertBeforeLinks && this._sections.length>0 && this._sections[this._sections.length-1] instanceof boxLinks ) {
            // Insert before the last item:
            this._sections.splice(this._sections.length-1, 0, section);
            this._navBox.content.insertBefore( section.el, this._sections[this._sections.length-1].el );
        } else {
            // Just add it
            this._sections.push(section);
            this._navBox.content.appendChild( section.el );
        }
    },

    /**
     * Notification the blog was updated. Update new symbol.
     * @function {} ?
     */
    _blogUpdated: function(evt) {
        this._elHeaderBlogNew.style.display = (evt.newItems>0)? '' : 'none';
        this._elHeaderBlogNew.title = giI18N.getString('News.blog.new.title').replace('%1', evt.newItems);
    },
    
    /**
     * Whether the UI object can be cached by {@link boxMain.getUIObject}
     * @function {Boolean} ?
     * @return Whether the UI can be cached
     */
    isCacheable: function() {
        return true;
    },
    
    
    //}
});

