/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * News box for Blog and Twitter
 *
 * @class ? @extends giUIObject
 */
var boxNews = giUIObject.extend({
    // Properties {
    
	/**
     * The blog/twitter box
     * @property {Element}
     */
	el: null,
    
    /**
     * List of entries
     * @property {Element}
     */
	_elBlogEntries: null,
    
    /**
     * The date of last update
     * @property {Element}
     */
	_elTwitterDate: null,
	
    /**
     * The RSS object for the blog
     * @property {giRss}
     */
	_blogRss: null,
    
    /**
     * The RSS object for twitter
     * @property {giRss}
     */
	_twitterRss: null,
    
    /**
     * Interval for refreshing
     * @property {int}
     */
    _refreshInterval: null,
    
    /** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
    
    /**
     * True if has started the loading process, or is loaded
     * @property {Boolean}
     */
    _isInitiated: false,
	
    //}

	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} main - The gtdMainPagePlugin object for the page
	*/
	connect: function(main) {
		giLogger.log('Connecting boxNews object to view', 'boxNews');
		
		this.base(main);
		this._eventsContainer = new giEventsContainer(this);
		
				
		var doc = main.gmail.canvasDocument;
        
        // Create news box:
		this.el = doc.createElement("DIV");
        this.el.className = 'gtdi-news';
        
        // Create blog section:
        var elBlog = doc.createElement("DIV");
        elBlog.className = 'gtdi-news-blog';
        this.el.appendChild( elBlog );
        
        var elBlogTitle = doc.createElement("A");
        elBlogTitle.className = 'gtdi-news-blog-title';
        elBlogTitle.href = "http://blog.gtdinbox.com";
        elBlogTitle.innerHTML = giI18N.getString('News.blog.title');
        elBlog.appendChild( elBlogTitle );
        
        // Create blog entries container:
        this._elBlogEntries = doc.createElement("DIV");
        this._elBlogEntries.className = 'gtdi-news-blog-entries';
        this._elBlogEntries.innerHTML = "<img src='chrome://gtdinbox/skin/indicator_medium.gif'/>";
        elBlog.appendChild( this._elBlogEntries );
        
        // Create blog subscribe link:
        var elSubscribe = doc.createElement("A");
        elSubscribe.className = 'gtdi-news-blog-subscribe';
        elSubscribe.href = 'http://feedburner.google.com/fb/a/mailverify?uri=GtdinboxBlog&email='+this._main.gmail.environment.getAccount();
        elSubscribe.setAttribute('target', '_blank');
        elSubscribe.innerHTML = giI18N.getString('News.blog.subscribe');
        elBlog.appendChild( elSubscribe );
        
        // Create twitter section:
        var elTwitter = doc.createElement("DIV");
        elTwitter.className = 'gtdi-news-twitter';
        this.el.appendChild( elTwitter );
        
        var elTwitterTitle = doc.createElement("A");
        elTwitterTitle.className = 'gtdi-news-twitter-title';
        elTwitterTitle.href = "http://www.twitter.com/gtdinbox";
        elTwitterTitle.innerHTML = "<img src='http://www.twitter.com/favicon.ico'/> "+giI18N.getString('News.twitter.title')+':';
        elTwitter.appendChild( elTwitterTitle );
        
        var elTwitterDate = doc.createElement("DIV");
        elTwitterDate.className = 'gtdi-news-twitter-date-container';
        elTwitterDate.innerHTML = giI18N.getString('News.twitter.lastUpdate')+" ";
        elTwitter.appendChild( elTwitterDate );
        this._elTwitterDate = doc.createElement("SPAN");
        this._elTwitterDate.className = 'gtdi-news-twitter-date';
        elTwitterDate.appendChild( this._elTwitterDate );
        
                
        // Add Click Handler:
        this._eventsContainer.observe(this.el, 'mousedown', this._click, false);
        
        // Set up RSS:
        this._blogRss = new giRss();
        this._blogRss.addEventListener(giRss.LOADED, this._blogLoaded, this);
        this._twitterRss = new giRss();
        this._twitterRss.addEventListener(giRss.LOADED, this._twitterLoaded, this);
        
        
        // Load blog:
        var me = this;
        // Load RSS - give a delay so as not to interrupt Gmail load
        setTimeout( function(){me.refresh()}, 1000*15);
                
        // Set it on a timer:
        this._refreshInterval = setInterval(function(){me.refresh()}, (2*60*60*1000));  // 2 hours
	},
	
    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting boxNews from view', 'boxNews');

        
        if( this.el && this.el.parentNode ) this.el.parentNode.removeChild( this.el );
        
        // Remove elements:
        for( p in this ) {
            if( this.hasOwnProperty(p) && this[p].nodeName && this[p].parentNode ) {
                this[p] = null;
            }
        }
        
        this._blogRss.unload();
        this._blogRss = null;
        this._twitterRss.unload();
        this._twitterRss = null;
        
        clearInterval(this._refreshInterval);
        
        this._eventsContainer.reset();
        this._eventsContainer = null;
        this.base();
    },
    
    /**
     * Refresh the blog entries and update twitter date
     * @function {} ?
     */
    refresh: function() {
        this._isInitiated = true;
        this._blogRss.load('http://feeds.feedburner.com/ProductiveFirefox');
        this._twitterRss.load('http://twitter.com/statuses/user_timeline/14171757.rss');
    },
    
    /**
     * Return true if loading has at least commenced
     * @function {Boolean} ?
     * @return True if the blog is loaded, or has started to load
     */
    isInitiated: function() {
        return this._isInitiated;
    },
    
    /**
     * Notification blog entries downloaded
     * @function {} ?
     */
    _blogLoaded: function(evt) {
        
                
        // Clear out existing:
        this._elBlogEntries.innerHTML = "";
        
        // Get last update:
        var tsLastUpdate = this._main.prefs.getPref('news.lastview') || 0;
        var newItems = 0;
                
        // Process last 3 entries:
        var doc = this._main.gmail.canvasDocument;
        var rss = evt.target;
        for( var i = 0; (i < 3 && i < rss.entries.length); i++ ) {
            var elLi = doc.createElement("DIV");
            elLi.className = 'gtdi-news-blog-lineitem';
            elLi.innerHTML = "<a href=\""+rss.entries[i].link+"\" target='_blank'>"+rss.entries[i].title+"</a>";
            elLi.title = giUtil.decodeEntities(rss.entries[i].contentSnippet);
            this._elBlogEntries.appendChild( elLi );
            
            var d = new Date(rss.entries[i].publishedDate);
            if( d.valueOf() > tsLastUpdate ) {
                // Beyond last blog view, mark it so
                giUIObject.addClassName(elLi, 'gtdi-news-blog-lineitem-new');
                newItems++;
            }
        }
                
        // Dispatch notification
        this.dispatchEvent({name: boxNews.BLOG_UPDATED, newItems: newItems});
    },
    
    /**
     * Notification twitter data downlaoded
     * @function {} ?
     */
    _twitterLoaded: function(evt) {
        var rss = evt.target;
        var dateTokens = rss.entries[0].publishedDate.split(" ");
        dateTokens.length = 3;
        this._elTwitterDate.innerHTML = dateTokens.join(" ");
    },
    
    /**
     * Notification something was clicked.
     * If line item of blog, update last click info and reflect in UI
     * @function {} ?
     */
    _click: function(evt) {
        try {
            
            var el = evt.target;
            if( el.nodeName=='A' && (el.className=='gtdi-news-blog-title' || el.parentNode.className.indexOf('gtdi-news-blog-lineitem')>-1) ) {
                this.unmarkNew();
            }

        } catch(e) {
            giLogger.warn(e, 'boxNews._click');
        }
    },
    
    
    unmarkNew: function() {
        var ts = (new Date()).valueOf();
        this._main.prefs.setPref('news.lastview', ts);
        
        // Unmark:
        var results = this.el.ownerDocument.evaluate( ".//*[contains(@class, 'gtdi-news-blog-lineitem-new')]", this.el, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        for( var i = 0; i < results.snapshotLength; i++ ) {
            giUIObject.removeClassName(results.snapshotItem(i), 'gtdi-news-blog-lineitem-new');
        }
        this.dispatchEvent({name: boxNews.BLOG_UPDATED, newItems: 0});
    },
    
    //}
},{
    
    /**
     * Notification the blog posts were updated
     * @event
     */
    BLOG_UPDATED: 'blog_updated',
    
});
boxNews.implement(giEventDispatcher); // Add event dispatching support