/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * A major section in the extensible Nav Box
 *
 * @class ? @extends giUIObject
 */
var boxSection = giUIObject.extend({
    // Properties {
    
	
    /**
     * The ID of the box. Used primarily by expander.
     * @property {String}
     */
    id: null,
    
	/** @property {Element} ? The root element */
	el: null,
    
  	/** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,

    /**
     * The element of the expander
     * @property {Element}
     */
    _elExpand: null,
        
    /**
     * An array of rows. Format: {id:, el:, elPopout:}
     * @property {private Object[]}
     */
    _rows: null,
    
    /**
     * True if the view is expanded (which it is by default, as the expander is not shown)
     * @property {private Boolean}
     */
    _expanded: true,
    
    /**
     * If popout in view, this is the element
     * @property {private Element}
     */
    _currentPopout: null,
    
    /**
     * The currently popped out row
     * @property {private Element}
     */
    _currentPopoutRow: null,
    
    //}

    // Methods {
    /**
     * Connects the element to the given boxMain object's page. Called
     * automatically by {@link boxMain.getUIObject}.
     * @function ?
     * @param {boxMain} main - The boxMain object for the page
     */
    connect: function(main) {
        giLogger.log('Connecting boxSection object to view', 'boxSection');
        this.base(main);
        this._rows = [];
        this._eventsContainer = new giEventsContainer(this);
        
        this.el = this._main.gmail.doc.createElement("DIV");
        this.el.className = "gtdi-box-section";
        
        this._eventsContainer.observe(this.el, 'mouseover', this._mouseover, false);
        this._eventsContainer.observe(this.el, 'click', this._click, false);
    },
    
    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting boxSection from view', 'boxSection');

        this.el.parentNode.removeChild(this.el);
        
        this._eventsContainer.reset();
        this._eventsContainer = null;

        
        if( this._currentPopout ) {
            if( this._currentPopout.parentNode ) this._currentPopout.parentNode.removeChild( this._currentPopout );
            this._currentPopout = null;
            this._currentPopoutRow = null;
        }

        this.base();
    },
    
    
    
    
    
    /**
     * Add a row to the section
     * @function {} ?
     * @param {String} id
     * @param {Element} content - The element to go into the row
     * @param {optional Element} popout - If present, this is the content for the popup
     * @param {optional Boolean} clickMode - If true, use click on/off instead of mouse over/out
     */
    _addRow: function(id, content, popout, clickMode) {
        // Create containing element:
        var el = this._main.gmail.doc.createElement("DIV");
        el.setAttribute('gtdi-box-row-id', id);
        el.className = 'gtdi-box-row';
        if( content ) el.appendChild( content );
        if( popout ) {
            popout.setAttribute('gtdi-box-row-id', id);
            giUIObject.addClassName(popout, 'gtdi-box-row-popout');
        }
        
        // Store data:
        var data = {id: id, el:el, elPopout: popout, clickMode: clickMode};
        this._rows.push( data );
        
        // Add to section:
        this.el.appendChild( el );
        
        
        return data;
    },


    
    
    /**
     * Add the expandable UI, and minimise
     * @function {Boolean} ?
     * @param {Boolean} v - If true, add expander. If false, remove expander. If undefined, return status.
     * @param {optional String} title - If expandable, provide title for expander row
     * @return If no params, return true/false is expandable
     */
    _isExpandable: function(v) {
        if( typeof v=='undefined' ) {
            return !!(this._elExpand);
        }
        
        if( v ) {
            if( !this._elExpand ) {
                // Create expander and attach it
                this._elExpand = this._main.gmail.doc.createElement("DIV");
                this._elExpand.className = 'gtdi-box-section-expand';
                this._elExpand.innerHTML = this.id? giI18N.getString('Box.expander.'+this.id) : "";
                if( this.el.firstChild ) {
                    this.el.insertBefore( this._elExpand, this.el.firstChild );
                } else {
                    this.el.appendChild( this._elExpand );
                }
                // Set up listener:
                this._eventsContainer.observe(this._elExpand, 'click', this._clickExpand, false);
                // Try to recall shown status to shown status:
                var expanded = this.id? this._main.prefs.getPref('expanded_sections.'+this.id) : true;
                if( expanded!==true && expanded!==false ) expanded = true;
                this.expanded(expanded);
            }
        } else {
            if( this._elExpand ) {
                this._elExpand.parentNode.removeChild( this._elExpand );
                this._elExpand = null;
                // Remove listener:
                this._eventsContainer.stopObserving(this._elExpand, 'click', this._clickExpand, false);
                // Ensure everything is shown:
                this.expanded(true);
            }
        }
    },
    
    /**
     * Expand or hide the expanded view of the section.
     * This is specific to each implementation of a section; and so should be handled by the subclass.
     * @function {Boolean} ?
     * @param {Boolean} v - If true, show as expanded. If false, hide. If undefined, return current status.
     * @return If no params, return true/false for expanded status.
     */
    expanded: function(v) {
        if( typeof v=='undefined' ) return this._expanded;
        
        if( v ) {
            this._expanded = true;
            giUIObject.addClassName(this._elExpand, 'gtdi-box-section-expanded');
            //this._elExpand.firstChild.src = "chrome://gtdinbox/skin/box/i/gtdi-up-arrow.png";
        } else {
            this._expanded = false;
            giUIObject.removeClassName(this._elExpand, 'gtdi-box-section-expanded');
            //this._elExpand.firstChild.src = "chrome://gtdinbox/skin/box/i/gtdi-dropdown-arrow.png";
        }
        
        // Remember status:
        if( this.id ) this._main.prefs.setPref('expanded_sections.'+this.id, v);
    },
    
    /**
     * Respond to a clicked expander.
     * Toggle expanded.
     * @function ?
     */
    _clickExpand: function(evt){
        try {
            this.expanded(!this.expanded());
        } catch(e) {
            giLogger.warn(e, 'boxSection._clickExpand');
        }
    },
    
    /**
     * Notification that an element was clicked
     * @function {} ?
     */
    _click: function(evt) {
        try {
            var el = evt.target;
            
            var elRow = null;
            while( el.parentNode && !elRow ) {
                if( el==this.el ) break;
                if( el.getAttribute('gtdi-box-row-id') ) elRow = el;
                el = el.parentNode;
            }
            if( elRow ) {
                // Find the data object for the row and see if it has a popout object.
                for( var i = 0; i < this._rows.length; i++ ) {
                    if( this._rows[i].el===elRow ) {
                        // Match
                        if( this._rows[i].elPopout && this._rows[i].clickMode==true ) {
                            if( this._currentPopout==this._rows[i].elPopout ) {
                                // Same item, close it
                                this._hidePopout();
                            } else {
                                // Another row, open it
                                this._showPopout(this._rows[i].elPopout, elRow, this._rows[i]);
                            }
                        }
                    }
                }
            }
        } catch(e) {
            giLogger.warn(e, 'boxSection._click');
        }
    },
    
    /**
     * Notification that an element within the el had a mousemove event
     * Respond appropriately.
     * @function {}
     */
    _mouseover: function(evt) {
        try {
            var el = evt.target;
            
            var elRow = null;
            while( el.parentNode && !elRow ) {
                if( el==this.el ) break;
                if( el.getAttribute('gtdi-box-row-id') ) elRow = el;
                el = el.parentNode;
            }
            
            if( elRow ) {
                // Mouse over into a row. Find the data object for the row and see if it has a popout object.
                for( var i = 0; i < this._rows.length; i++ ) {
                    if( this._rows[i].el===elRow ) {
                        // Match
                        if( this._rows[i].elPopout && !this._rows[i].clickMode ) {
                            
                            // It has a popout
                            if( this._popupTimeout ) {
                                // Stop the requested close
                                clearTimeout(this._popupTimeout);
                                this._popupTimeout = null;
                            }
                            
                            if( this._currentPopout!=this._rows[i].elPopout ) {
                                // It's a new item
                                if( this._currentPopout ) {
                                    // Queue it to show when (if) hide this item; makes it less twitchy
                                    this._nextPopout = {'currentPopout': this._rows[i].elPopout, 'currentPopoutRow': elRow, 'currentPopoutData': this._rows[i]};
                                    setTimeout(giUtil.bind(this, this._tryNextPopout), 150);
                                    return;
                                }
                                
                                this._showPopout(this._rows[i].elPopout, elRow, this._rows[i]);
                            }
                            
                        }
                        break;
                    }
                }
            }
        
        } catch(e) {
            giLogger.warn(e, 'boxSection._mouseover');
        }
    },
    
    _tryNextPopout: function() {
        
        if( this._nextPopout ) {
            if( this._currentPopout!=this._nextPopout.currentPopout ) {
                this._showPopout(this._nextPopout.currentPopout, this._nextPopout.currentPopoutRow, this._nextPopout.currentPopoutData);
            }
        }
        
    },
    
    _showPopout: function(popout, elRow, popoutData) {
        // Hide current item:
        this._hidePopout();
        
        this._currentPopout = popout;
        this._currentPopoutRow = elRow;
        
        giUIObject.addClassName(this._currentPopoutRow, 'gtdi-box-row-popped-out');
        
        
        // Attach it to the view:
        this._main.gmail.canvasDocument.body.appendChild( this._currentPopout );
                
        // Position it:
        var borderWidth = giUIObject.getBorderWidth(elRow);	// get width of el's parent (canvas?) border, which is not included in the cumulativeOffset
        var offset = giUIObject.cumulativeOffset(elRow);
        //offset[0] = offset[0] + borderWidth;
        offset[1] = offset[1] + parseInt(borderWidth[3]) + parseInt(borderWidth[1]);
        var dims = giUIObject.getDimensions(elRow);
        var xy = [offset[0] + dims.width, offset[1]]
        this._currentPopout.style.left = xy[0] + "px";
        this._currentPopout.style.top = xy[1] + "px";
        
        if( popoutData.clickMode==true ) {
            // Listen for clicks:
            this._eventsContainer.observe( this._main.gmail.canvasDocument.body, 'mousedown', this._clickWindow, false );
        } else {
            // Listen for the mouse out:
            this._eventsContainer.observe( this._main.gmail.canvasDocument.body, 'mousemove', this._mouseoverWindow, false );
        }
        
        
        // Fire event - cause it to be populated
        this.dispatchEvent({name: boxSection.POPOUT, id: elRow.getAttribute('gtdi-box-row-id'), elRow: elRow, elPopout: this._currentPopout});
        
        // Verify position in bounds:
        xy = giUIObject.boundsCheckXY(this._currentPopout, xy[0], xy[1]);
        this._currentPopout.style.left = xy[0] + "px";
        this._currentPopout.style.top = xy[1] + "px";
        this._currentPopout.style.maxHeight = xy[3] + "px";
        
    },
    
    /**
     * Notification that a window element had a mouseover event
     * If no longer over a row or a popup, hide it
     * @function {}
     */
    _mouseoverWindow: function(evt) {
    
        var el = evt.target;
        
        var elRow = null;
        var currentId = this._currentPopout.getAttribute('gtdi-box-row-id');
        while( el.parentNode && !elRow ) {
            if( el==this.el ) break;
            if( (el.hasAttribute('gtdi-box-row-id') && el.getAttribute('gtdi-box-row-id')==currentId) ) {
                // Over the current popout or row
                elRow = el;
                this._nextPopout = null;    // As we're back on current territory, give up attempt to switch to new row
            }
            if( this._nextPopout && this._nextPopout.currentPopoutRow==el ) {
                // Over the next popout row
                elRow = el;
            }
            el = el.parentNode;
        }
        if( elRow ){
            if( this._popupTimeout ) {
                // Stop the close
                clearTimeout(this._popupTimeout);
                this._popupTimeout = null;
            }
        } else {
            if( !this._popupTimeout ) this._popupTimeout = setTimeout(giUtil.bind(this, this._hidePopout), 150);
        }
    
    },
    
    /**
     * Notification that a window element was clicked.
     * If not on the current popout, close
     * @function {} ?
     */
    _clickWindow: function(evt) {
        try {
            var el = evt.target;
                
            var clickedWithin = false;
            while( el.parentNode  ) {
                if( el==this.el ) break;
                if( el==this._currentPopout ) {
                    // Clicked within
                    clickedWithin = true;
                    break;
                }
                el = el.parentNode;
            }
            
            if( !clickedWithin ) {
                // Hide it
                this._hidePopout();
            }
        } catch(e) {
            giLogger.warn(e, 'boxSection._clickWindow');
        }
    },
    
    /**
     * Hide the popout if it is showing
     * @function {} ?
     */
    _hidePopout: function() {
        if( this._currentPopout ) {
            
            if( this._popupTimeout ) {
                // Stop the close
                clearTimeout(this._popupTimeout);
                this._popupTimeout = null;
            }
            
            this._nextPopout = null;
            
            // Movement was not over the row or a popout
            giUIObject.removeClassName(this._currentPopoutRow, 'gtdi-box-row-popped-out');
            this.dispatchEvent({name: boxSection.POPOUT_HIDE, elPopout: this._currentPopout, elRow: this._currentPopoutRow})
            this._currentPopout.parentNode.removeChild( this._currentPopout );
            this._currentPopout = null;
            this._currentPopoutRow = null;
            this._eventsContainer.stopObserving( this._main.gmail.canvasDocument.body, 'mousemove', this._mouseoverWindow, false );
            this._eventsContainer.stopObserving( this._main.gmail.canvasDocument.body, 'mousedown', this._clickWindow, false );
            //var me = this;
            //setTimeout( function() {
                //me._eventsContainer.observe( me.el, 'mouseover', me._mouseover, false );
            //}, 50 );
        }
    },
       
    /**
     * Whether the UI object can be cached
     * @function {Boolean} ?
     * @return Whether the UI can be cached
     */
    isCacheable: function() {
        return true;
    },
        
    //}
} , {
    
    /**
     * Fired when the popout for a row is shown
     * @event
     */
    POPOUT: 'popout',
    
    /**
     * Fired when the popout for a row is closed
     * @event
     */
    POPOUT_HIDE: 'popout_hide',
    
} );
boxSection.implement(giEventDispatcher); // Add event dispatching support
