/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 *
 * TODO get new licence text from base machine
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Provides an abstraction layer for accessing stored data related to a specific
 * gmail account. This includes preferences and saved data.
 *
 * @class ?
 */
var giAccountData = giBase.extend({
    // Properties {


    /**
     * The e-mail address of the account
     * @property {String} ?
     */
    account: null,

    /**
     * The account preferences
     * @property {Object} ?
     */
    _prefs: null,

    /**
     * Whether the preferences have been changed since the last save
     * @property {Boolean} ?
     */
    _prefsChanged: false,

    /**
     * Periodic save timeout identifier
     * @property {int} ?
     */
    _periodicSaveTimeout: false,
        
    /**
     * True if currently writing to the server. Try to avoid multiple simultaneous calls. 
     * @property {Boolean} ?
     */
    _serverWriteLock: false,
    
    /**
     * Reference to all requests made so they can be cleaned up
     * @property {giAccountRequest[]} ?
     */
    _requests: null,
    
    //}

    // Methods {
    
    // General {
    /**
     * Shuts down the giAccountData object and saves everything to disk
     * @function ?
     */
    shutdown: function() {
        for( var i = 0; i < this._requests.length; i++ ) {
            this._requests[i].reset();
        }
        this.savePrefs(null, true);   // Save, but do not write to the server
        clearTimeout(this._periodicSaveTimeout);
    },
    
    /**
     * Find giAccountRequest for this evt.target and clean it up
     * @function {} ?
     */
    _clearOutRequest: function(evt) {
        for( var i = this._requests.length -1; i>=0; i-- ) {
            if( this._requests[i]===evt.target ) {
                this._requests[i].reset();
                this._requests.splice(i, 1);
            }
        }
    },


    /**
     * If there are unsaved prefs, return true
     * @function {Boolean} ?
     * @return True if unsaved preferences
     */
    isNotSaved: function() {
        return this._prefsChanged;
    },

    //}
    
    
    
    
    
    
    
    
    
    
    // Load Specific {

    /**
     * Called by {@link retrieveForAccount}, it sets the _account variable
     * @function ?
     * @param {String} account - The account e-mail address to load up settings for
     */
    _loadForAccount: function(account) {
        this.account = account;
    },
    
    /**
     * Start the load process
     * @function ?
     */
    load: function() {
        // Read local settings (it should have synced when Firefox opened)
    	this._loadForAccountUseLocal();
    },
    
    
    
    
    /**
     * Read in settings from Firefox preferences (which could have been just updated from server).
     * Migrate any updates in the code (which will be uploaded back to the server).
     * @function {}?
     */
    _loadForAccountUseLocal: function() {
        this._loadPrefs();

        // Test essential integrity (prior to migrate, as it will start reading/writing prefs)
        if( !this._prefs.mainBranch || !this._prefs.transactionBranch ) {
            this._prefs = giAccountData.getDefaultPrefs();
            this._prefsChanged = true;
            this._setSaveTimeout();
        }
        
        this._migratePrefs();
        
    },
    
    /**
     * Loads and unserializes the account's prefs
     * @function ?
     */
    _loadPrefs: function() {
        if(this._prefs != null) giLogger.error('_loadPrefs can only be called once', 'giAccountData');

        giLogger.log('Loading prefs...', 'giAccountData');
        var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
        try {
            var prefString = prefManager.getCharPref("extensions.gtdinbox."+this.account);
            this._prefs = giAccountData.JSON.decode(prefString);
        } catch(e) {
            // Prefs not yet set, so set this._prefs to default
            this._prefs = giAccountData.getDefaultPrefs();
            this._prefsChanged = true;
        }
        giLogger.log('Loaded prefs', 'giAccountData');

        if( this._prefsChanged ) {
            this._setSaveTimeout();
        }
    },
    
    
    /**
     * Performs migrations on the preferences if they are below the current version
     * @function ?
     */
    _migratePrefs: function() {
        /* Put in case statements for each released version and then it will perform
         * all migrations necessary after that version to get to the current version
         * so long as "break" statements are removed.
         */
        var initialVersion = this.getPref('version', true) || '1.0';
        switch(initialVersion) {
            case '1.0':
                this.setPref('components.traffic_watcher', {enabled:true});
                this.setPref('version', '1.0.1');
            case '1.0.1':
                var temp = ['next_action', 'action', 'waiting_on', 'some_day', 'finished'];
                this.setPref('labels.types.statuses.core', {});
                for(var i = 0; i < temp.length; i++) {
                    this.setPref('labels.types.statuses.core.'+temp[i],this._prefs.labels.types.statuses[temp[i]]);
                    this.setPref('labels.types.statuses.'+temp[i],null);
                }
                this.setPref('version', '1.0.2');
            case '1.0.2':
                if( !this._prefs.components ) this.setPref('components', {});
                var temp = ['compose_personal','search_button','dashboard'];
                for( var i = 0; i < temp.length; i++ ) {
                    if( !this._prefs[temp[i]] ) this.setPref(temp[i], {enabled:true});
                }
            	if( !this._prefs.label_box ) this.setPref('label_box', {expanded: ['statuses', 'projects', 'contexts', 'misc']});
            	if( !this._prefs.general ) this.setPref('general', {centre_popup:true});
                this.setPref('version', '1.0.3');
            case '1.0.3':
                this.setPref('components.traffic_watcher', null);
                this.setPref('version', '1.0.4');
            case '1.0.4':
                // Major structural change (in alpha testing), overwrite:
                this._prefs = giAccountData.getDefaultPrefs();
                this.setPref('version', '1.0.5');
            case '1.0.5':
                // Major structural change (in alpha testing), overwrite:
                this._prefs = giAccountData.getDefaultPrefs();
                this.setPref('version', '1.0.6');
            case '1.0.6':
                this.setPref('disabled', false);
                this.setPref('version', '1.0.7');
            case 'fallthrough':
                // Always called if a migration takes place - do final processing here
                this._prefsChanged = true;
                break;
        }
	
        if( this._prefsChanged ) {
            this._setSaveTimeout();
        }

        if(initialVersion != this.getPref('version')) {
            giLogger.log('Migrated prefs from "'+initialVersion+'" to "'+this.getPref('version')+'"', 'giAccountData');
        }
    },
    
    
    /**
     * Get the latest settings from the server, merge with local settings, and save locally.
     */
    syncWithServer: function(async) {
        // Try to retrieve latest settings
        var req = giAccountRequest.createReadSettingsRequest(this.account);
        this._requests.push(req);
        req.addEventListener(giAccountRequest.SUCCESS, this._syncWithServerCallback, this);
        req.addEventListener(giAccountRequest.FAIL, this._syncWithServerCallback, this);
        req.addEventListener(giAccountRequest.WRONG_PASSWORD, this._syncWithServerCallback, this);
        req.addEventListener(giAccountRequest.NO_ACCOUNT, this._syncWithServerCallback, this);
        req.run();
    },
    
    
    /**
     * Notification the ReadSettings request has run
     * @function {} ?
     */
    _syncWithServerCallback: function(evt) {

        switch(evt.name) {
            case giAccountRequest.SUCCESS:
                var latestPrefs = evt.result.settings;
                // Check our account had settings:
                if( latestPrefs ) {
                    // Read in local preferences:
                    this._loadForAccountUseLocal();
		    
                    // Merge the new copy with more recent changes in the local copy
                    var status = {prefsChanged: false};
                    this._mergePreferences(latestPrefs.mainBranch, this._prefs.mainBranch, latestPrefs.transactionBranch, this._prefs.transactionBranch, status);
                    
                    if( status.prefsChanged ) {
                        // Commit the merged object to our local store:
                        
                        var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
                        var prefString = giAccountData.JSON.encode(latestPrefs);
                        prefManager.setCharPref("extensions.gtdinbox."+this.account, prefString);
                        
                        // Update live preferences:
                        this._prefs = latestPrefs;
                        
                        // Save our merge now
                        this._prefsChanged = true;
                        this.savePrefs();
                    }
                } else {
                    // No settings on the server. Suggests it's broken. Upload current prefs.
                    // Read in local preferences:
                    this._loadForAccountUseLocal();
                    this._prefsChanged = true;
                    this.savePrefs();
                }
                break;
            case giAccountRequest.FAIL:
                // Could not read from server.
                break;
            case giAccountRequest.WRONG_PASSWORD:
                // Tell settings that the last password attempt failed
		
                this._loadForAccountUseLocal();
                this.setPref('wrong_password', true);
                this._prefsChanged = true;
                this.savePrefs(null, true);	// Note it is only a local save
                break;
            case giAccountRequest.NO_ACCOUNT:
                // Fail
                break;
        }
        this._clearOutRequest(evt);
    },
    
    //}






    // Pref Read/Write  {
    /**
     * Returns the value of the given preference
     * @function {Object} ?
     * @param {String} pref - The preference string
     * @param {boolean} suppressNotFoundError - Optional. Do not throw error if not found.
     * @return The value of the preference
     */
    getPref: function(pref, suppressNotFoundError) {
        var prefPath = pref.split('.');
        var wantedPref = this._prefs.mainBranch;
        while(prefPath.length) {
            var next = prefPath.shift();
            if(typeof wantedPref[next] != 'undefined') {
                wantedPref = wantedPref[next];
            } else {
            	if( suppressNotFoundError ) {
            		return null;
            	} else {
	                giLogger.error('No preference: "'+pref+'"', 'giAccountData');
                }
            }
        }

        return wantedPref;
    },

    /**
     * Sets the preference to the given value
     * @function ?
     * @param {String} pref - The preference string
     * @param {Object} value - The value to set the preference to
     */
    setPref: function(pref, value) {
        
        try {
            var prefPath = pref.split('.');
            var base = this._prefs.mainBranch;
            var transBase = this._prefs.transactionBranch;
            while(prefPath.length > 1) {
                var currentStep = prefPath.shift();
                // Ensure step exists; and are objects (otherwise would fail if property changed from value to an object)
                if(typeof base[currentStep] == 'undefined' || !(typeof base[currentStep]=='object' && !(base[currentStep] instanceof Array)) ) base[currentStep] = {};
                if(typeof transBase[currentStep] == 'undefined' || !(typeof transBase[currentStep]=='object' && !(transBase[currentStep] instanceof Array)) ) transBase[currentStep] = {};
    
                base = base[currentStep];
                transBase = transBase[currentStep];
            }
            if( value===null ) {
                // Delete the value
                delete base[prefPath[0]];
                transBase[prefPath[0]] = (new Date()).valueOf(); // A key with a timestamp but no pref is understood to be deleted (by merge routine)
            } else {
                base[prefPath[0]] = value;
                if( (typeof value=='object') && !(value instanceof Array) ) {
                    // It's an object (not an array) so overwrite all leaf values with transaction statements
                    transBase[prefPath[0]] = {};
                    this._recurseUpdateValueTransactions(value, transBase[prefPath[0]], (new Date()).valueOf());
                } else {
                    transBase[prefPath[0]] = (new Date()).valueOf();
                }
            }
            
            this._prefsChanged = true;
            this._setSaveTimeout();
        } catch(e) {
            giLogger.warn(e, 'giAccountData.setPref');
        }
    },

    /**
     * Apply a timestamp value to all leaf (non-object) properties on this object and its children.
     * @function {} ?
     * @param {Object} definition - The object with the correct structure (to be cloned into o)
     * @param {Object} o - The object to apply timestamps too
     * @param {int} timestamp - The timestamp
     */
    _recurseUpdateValueTransactions: function(definition, o, timestamp) {
        for( p in definition ) {
            if( definition.hasOwnProperty(p) ) {
                if( (typeof definition[p]=='object') && !(definition[p] instanceof Array) ) {
                    if( !o[p] ) o[p] = {};
                    this._recurseUpdateValueTransactions(definition[p], o[p], timestamp);
                } else {
                    o[p] = timestamp;
                }
            }
        }
    },

    //}
    
    
    
    
    
    
    
    
    
    
    // Save Specific {


    /**
     * Serializes the prefs and saves them
     * @function ?
     * @param {optional int} transactionId - An ID that will be passed back when events are fired; so external requests can follow call-paths more precisely
     * @param {optional boolean} doNotUpload
     */
    savePrefs: function(transactionId, doNotUpload) {
        
        if( this._periodicSaveTimeout ) clearTimeout(this._periodicSaveTimeout);
            
        if(this._prefsChanged) {
            
            var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
            
            // First, check another instance of giAccountData has not had a more recent update - if so; merge it into ours
            var currentPrefs = null;
            try {
                var prefString = prefManager.getCharPref("extensions.gtdinbox."+this.account);
                currentPrefs = giAccountData.JSON.decode(prefString);
            } catch(e) {}
            if( currentPrefs && currentPrefs.mainBranch && currentPrefs.transactionBranch ) {
                // Merge, so anything that is newer in the current values is put into this._prefs
                this._mergePreferences(this._prefs.mainBranch, currentPrefs.mainBranch, this._prefs.transactionBranch, currentPrefs.transactionBranch);
            }
            
            var prefString = giAccountData.JSON.encode(this._prefs);
            prefManager.setCharPref("extensions.gtdinbox."+this.account, prefString);
            
            
            if( !doNotUpload && !this._serverWriteLock  ) { // "DO NOT RUN"=="REALLY DON'T RUN" && 
                // Commit changes to server
                
                // First, retrieve the latest copy from the server
                var req = giAccountRequest.createReadSettingsRequest(this.account);
                req.transactionId = transactionId;
                this._requests.push(req);
                req.addEventListener(giAccountRequest.SUCCESS, this._saveGetLatestCallback, this);
                req.addEventListener(giAccountRequest.FAIL, this._saveGetLatestCallback, this);
                req.addEventListener(giAccountRequest.WRONG_PASSWORD, this._saveGetLatestCallback, this);
                req.addEventListener(giAccountRequest.NO_ACCOUNT, this._saveGetLatestCallback, this);
                req.run();
                
            }
        }
        this._prefsChanged = false;

    },
    
    /**
     * Notification that we have received the latest copy from the server.
     * @function {} ?
     */
    _saveGetLatestCallback: function(evt) {
        switch(evt.name) {
            case giAccountRequest.SUCCESS:
                // Compare this new data object with the current prefs; and upload any changes
                this._savePrefsWithLatestData(evt.result.settings, evt.target.transactionId);
                break;
            case giAccountRequest.WRONG_PASSWORD:
                this._synced = false;
                this.dispatchEvent({name:giAccountData.SAVE_SERVER_WRONG_PASSWORD, transactionId: evt.target.transactionId});
                this.dispatchEvent({name: giAccountData.SYNC_CHANGED});
                break;
            case giAccountRequest.NO_ACCOUNT:
                this._synced = false;
                this.dispatchEvent({name:giAccountData.SAVE_SERVER_NO_ACCOUNT, transactionId: evt.target.transactionId});
                this.dispatchEvent({name: giAccountData.SYNC_CHANGED});
                break;
            case giAccountRequest.FAIL:
                this._synced = false;
                this.dispatchEvent({name:giAccountData.SAVE_SERVER_FAIL, transactionId: evt.target.transactionId});
                this.dispatchEvent({name: giAccountData.SYNC_CHANGED});
                break;
        }
        this._clearOutRequest(evt);
    },
    
    /**
     * Take the most recent version of the date; compare timestamps to see what is newer in our local version (and overwrite); then reupload
     * @function {}?
     * @param {Object} latestPrefs - The latest preferences object
     * @param {optional int} transactionId - An ID that will be passed back when events are fired; so external requests can follow call-paths more precisely
     */
    _savePrefsWithLatestData: function(latestPrefs, transactionId) {
        // Pass through latest object
        
        // Check essential structure:
        if( !latestPrefs || !latestPrefs.mainBranch || !latestPrefs.transactionBranch) {
            latestPrefs = {mainBranch:{}, transactionBranch:{}};
        }
        
        // Compare:
        var status = {prefsChanged: false};
        this._mergePreferences(latestPrefs.mainBranch, this._prefs.mainBranch, latestPrefs.transactionBranch, this._prefs.transactionBranch, status);
        
        if( status.prefsChanged ) {
            // Write the modified latestPrefs
            var req = giAccountRequest.createWriteSettingsRequest(this.account, latestPrefs);
            req.transactionId = transactionId;
            req.latestPrefs = latestPrefs;
            this._requests.push(req);
            req.addEventListener(giAccountRequest.SUCCESS, this._savePrefsCallback, this);
            req.addEventListener(giAccountRequest.FAIL, this._savePrefsCallback, this);
            req.addEventListener(giAccountRequest.WRONG_PASSWORD, this._savePrefsCallback, this);
            req.addEventListener(giAccountRequest.NO_ACCOUNT, this._savePrefsCallback, this);
            req.run();
        } else {
            // No changes; save complete.
            this._serverWriteLock = false;
            if( !this._synced ) this.dispatchEvent({name: giAccountData.SYNC_CHANGED});
            this._synced = true;
            // TODO - To be truly accurate, this would be a SAVE_SERVER_NO_CHANGE event
            this.dispatchEvent({name:giAccountData.SAVE_SERVER_SUCCESS, transactionId: evt.target.transactionId});
        }
        
    },
    /**
     * Take the most latest version of the preferences; compare timestamps to see what is newer in our current/local version (and overwrite latest)
     * The idea is we will keep latestPrefs, discard currentPrefs; and latestPrefs will - after this - have all the most recent values/structure
     * It's also designed to handle minor structural changes
     * @function {}?
     * @param {Object} prefs - The latest preferences object
     * @param {Object} currentPrefs - The current preferences object
     * @param {Object} latestTrans - The latest transactions object
     * @param {Object} currentTrans - The current transactions object
     * @param {Object} status - Used for status messages. Format {prefsChanged: boolean}
     */
    _mergePreferences: function(latestPrefs, currentPrefs, latestTrans, currentTrans, status) {
        status = status || {};
        var seenLatestKeys = {};
        for( p in latestPrefs ) {
            if( latestPrefs.hasOwnProperty(p) ) {
                seenLatestKeys[p] = true;
                // See if we need to overwrite the latestPref with the currentPref (if it is more recent)
                if( typeof currentPrefs[p]=='undefined' || currentPrefs[p]===null ) {
                    // Value is deleted in Current. Do we need to delete from Latest?
                    if( typeof currentTrans[p]!='undefined' ) {
                        // When there is a Current timestamp but not a Current pref, the value was deleted. If Current timestamp more recent, also delete from Latest.
                        if( !latestTrans[p] || currentTrans[p]>latestTrans[p] ) {
                            delete latestPrefs[p];
                            if( typeof latestTrans[p]!='undefined' ) delete latestTrans[p];
                            status.prefsChanged = true;
                        }
                    }
                } else {
                    // Value found in Current
                    if( typeof latestPrefs[p]=='object' && !(latestPrefs[p] instanceof Array) ) {
                        // Latest is an object.
                        if( typeof currentPrefs[p]!='object' || currentPrefs[p] instanceof Array ) {
                            // Current is a leaf value. Structural change.
                            // Need to see if any of it's child values contain a timestamp greater than our own (or a timestamp)
                            var probe = {foundMoreRecentTimestamp: false};
                            this._mergePreferences_testNewChild(latestPrefs[p], latestTrans[p], currentTrans[p], probe);
                            if( !probe.foundMoreRecentTimestamp ) {
                                // Current value is more modern than the object in Latest; so discard object for value
                                latestPrefs[p] = currentPrefs[p];
                                latestTrans[p] = currentTrans[p];
                                status.prefsChanged = true;
                            }
                        } else {
                            // Both are objects. Recurse into. 
                            this._mergePreferences(latestPrefs[p], currentPrefs[p], latestTrans[p], currentTrans[p], status);
                        }
                    } else {
                        // Latest is a leaf value.
                        if( typeof currentPrefs[p]=='object' && !(currentPrefs[p] instanceof Array) ) {
                            // A leaf has turned into an object. Structural change. 
                            // Need to see if any of it's child values contain a timestamp greater than our own (or a timestamp)
                            var probe = {foundMoreRecentTimestamp: false};
                            this._mergePreferences_testNewChild(currentPrefs[p], currentTrans[p], latestTrans[p], probe);
                            if( probe.foundMoreRecentTimestamp ) {
                                // We should implement
                                latestPrefs[p] = currentPrefs[p];
                                latestTrans[p] = currentTrans[p];
                                status.prefsChanged = true;
                            }
                        } else {
                            // Both are leaf values. Compare time stamps.
                            if( currentTrans[p] && (!latestTrans[p] || currentTrans[p]>latestTrans[p]) ) {
                                latestPrefs[p] = currentPrefs[p];
                                latestTrans[p] = currentTrans[p];
                                status.prefsChanged = true;
                            }
                        }
                    }
                }
                
                
            }
        }
        // Now do a simpler test to see if anything new in current needs adding to latest
        for( p in currentPrefs ) {
            if( currentPrefs.hasOwnProperty(p) ) {
                if( !seenLatestKeys[p] ) {
                    // There is a key in current that is not in latest.
                    if( typeof latestTrans[p]!='undefined' && typeof latestTrans[p]!='object' && (!currentTrans[p] || typeof currentTrans[p]=='object' || latestTrans[p]>currentTrans[p])) {
                        // It was intentionally deleted - do not bring back
                    } else {
                        // Ok to add new property
                        latestPrefs[p] = currentPrefs[p];
                        latestTrans[p] = currentTrans[p];
                        status.prefsChanged = true;
                    }
                }
            }
        }
    },
    _mergePreferences_testNewChild: function(currentPrefs, currentTrans, latestTimestamp, probe) {
        for( p2 in currentPrefs ) {
            if( typeof currentPrefs[p2]=='object' && !(currentPrefs[p2] instanceof Array) ) {
                // It's an object
                this._mergePreferences_testNewChild(currentPrefs[p2], currentTrans[p2], latestTimestamp, probe);
            } else {
                // If this new key is more recent than our timestamp; we have proved this child should be used
                if( currentTrans[p2] && currentTrans[p2]>latestTimestamp ) {
                    probe.foundMoreRecentTimestamp = true;
                    return;
                }
            }
        }
    },
    
    
    /**
     * Notification of complete following our attempt to write settings to the server
     * @function {} ?
     */
    _savePrefsCallback: function(evt) {
	
	
        this._serverWriteLock = false;
        switch(evt.name) {
            case giAccountRequest.SUCCESS:
                // Update local pref store again:
                var latestPrefs = evt.target.latestPrefs;  // Already integrity checked
                // Have to resync in case any changes while we were writing
                this._mergePreferences(latestPrefs.mainBranch, this._prefs.mainBranch, latestPrefs.transactionBranch, this._prefs.transactionBranch);
                this._prefs = evt.target.latestPrefs;  
                var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
                var prefString = giAccountData.JSON.encode(this._prefs);
                prefManager.setCharPref("extensions.gtdinbox."+this.account, prefString);
                // Check sync
                if( !this._synced ) this.dispatchEvent({name: giAccountData.SYNC_CHANGED});
                this._synced = true;
                this.dispatchEvent({name:giAccountData.SAVE_SERVER_SUCCESS, transactionId: evt.target.transactionId});
                break;
            case giAccountRequest.FAIL:
                this._synced = false;
                this.dispatchEvent({name:giAccountData.SAVE_SERVER_FAIL, transactionId: evt.target.transactionId});
                this.dispatchEvent({name: giAccountData.SYNC_CHANGED}); 
                break;
            case giAccountRequest.WRONG_PASSWORD:
                this._synced = false;
                this.dispatchEvent({name:giAccountData.SAVE_SERVER_WRONG_PASSWORD, transactionId: evt.target.transactionId});
                this.dispatchEvent({name: giAccountData.SYNC_CHANGED}); 
                break;
            case giAccountRequest.NO_ACCOUNT:
                this._synced = false;
                this.dispatchEvent({name:giAccountData.SAVE_SERVER_NO_ACCOUNT, transactionId: evt.target.transactionId});
                this.dispatchEvent({name: giAccountData.SYNC_CHANGED}); 
                break;
        }
	
        this._clearOutRequest(evt);
    },
    

    /**
     * We try to reduce actual saves as they impact the server.
     * Attempt to defer a proper save.
     * This will still be called if .unload
     * @function ?
     */
    _setSaveTimeout: function() {
        if(!this._prefs) return;
        if( !this._periodicSaveTimeout ) {
            var me = this;
            clearTimeout(this._periodicSaveTimeout);
            this._periodicSaveTimeout = setTimeout(function() {
                try {
                    me.savePrefs();
                } catch(e) {
                    giLogger.warn(e, 'giAccountData._setSaveTimeout');
                }
            }, giAccountData.AUTO_SAVE_FREQUENCY);
        }
    },
    
    
    
    	
	
	
	
    //}
    //}
},{
    // Static Properties {
    
    /**
     * Time, in ms, after a .setPref call until it saves
     * Currently 5 mins
     * @property {static int}
     */
    AUTO_SAVE_FREQUENCY: 300000,
    
    /**
     * The default properties object
     * @property {static Object} ?
     */
    DEFAULT_PREFS: {
        mainBranch: {
            version: '1.0.7',
            plugins: {
                gtd: {
                    path: 'chrome://gtdinbox/content/js/gtd/gtdMainPagePlugin.js',
                    className: 'gtdMainPagePlugin',
                    disabled: false,
                    prefs: null,		
                },
            },
            disabled: false,
	    wrong_password: false,
        },
        transactionBranch: {
            
        },
    },

    /**
     * A reference to the nsIJSON interface
     * @property {static nsIJSON} ?
     */
    JSON: Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON),
    //}

    // Static Methods {
    /**
     * Clones the default prefs and returns the clone
     * @function {static Object} ?
     * @return The default prefs as an object
     */
    getDefaultPrefs: function() {
        var json = giAccountData.JSON;
        return json.decode(json.encode(giAccountData.DEFAULT_PREFS));
    },

    /**
     * Returns an account data object for the given account
     * @function {static giAccountData} ?
     * @param {String} account - The account e-mail address to load up settings for
     * @return An initialized giAccountData object
     */
    retrieveForAccount: function(account) {
        giLogger.log('Loading account data for "'+account+'"', 'giAccountData');
        var accountData = new giAccountData();
        accountData._requests = [];
        accountData._loadForAccount(account);
        return accountData;
    },
    //}
    
    // Static Properties {
	
	
    /**
     * Dispatched when ready to load GTDI
     * @event ?
     */
    LOADED: 'loaded',
    
    /**
     * Dispatched when confirm/lose connection with GTDInbox server
     * @event ?
     */
    SYNC_CHANGED: 'sync_changed',
    
    /**
     * Dispatch when saved to server
     * @event ?
     */
    SAVE_SERVER_SUCCESS: 'save_server_success',
    
    /**
     * Dispatch when saved to server, and password is incorrect
     * @event ?
     */
    SAVE_SERVER_WRONG_PASSWORD: 'save_server_wrong_password',
    
    /**
     * Dispatch when saved to server, and the user did not have an account
     * @event ?
     */
    SAVE_SERVER_NO_ACCOUNT: 'save_server_no_account',
    
    /**
     * Dispatch when saved to server, and it failed
     * @event ?
     */
    SAVE_SERVER_FAIL: 'save_server_fail',
    
	
    //}
});
giAccountData.implement(giEventDispatcher); // Add event dispatching support