/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * gtdRequest is an XMLHttpRequest wrapper for queries to GTDInbox servers. 
 *
 * @class ?
 */
var giAccountRequest = giBase.extend({
    // Properties {

    /**
     * The URL that the request will use
     * @property {String} ?
     */
    url: null,

    /**
     * The account used in the last connect attempt
     * @property {String} ?
     */
    account: null,

    /**
     * The password used in the last connect attempt
     * @property {String} ?
     */
    password: null,

    /**
     * The post data in a string
     * @property {String} ?
     */
    _postString: null,

    /**
     * The XMLHttpRequest object
     * @property {XMLHttpRequest} ?
     */
    _xmlRequest: null,
    //}

    /**
     * The constructor.
     * If one of the properties in params or postData is set to an array, that
     * property is added multiple times - once for each value in the array.
     * For the 'at' property, passing in a value of true will cause it to automatically
     * be calculated and added to the url params. The 'rt' property is optional
     * but if not set, it defaults to 'j' (which requires the least overhead to
     * process).
     * @constructor ?
     * @param {optional glGmail} gmail
     * @param {String} path - The path for the URL
     * @param {Object} urlParams - The URL params.
     * @param {optional Object} postData - The post data
     */
    constructor: function(path, urlParams, postData) {
        if(typeof urlParams == "undefined") giLogger.error('urlParams is undefined', 'giAccountRequest');

        // Generate URL including url parameters {
        var url = giAccountRequest.GTDI_URL;
        url += path;
        url += glRequest._convertToRequestString(urlParams);
        this.url = url;
        //}

        // Generate POST string {
        if(typeof postData != "undefined") {
            this._postString = glRequest._convertToRequestString(postData);
        }
        
        
        giLogger.log('constructed giAccountRequest: ' + url, 'giAccountRequest.constructor');
        //}
    },

    // Methods {
    /**
     * Runs the request either synchronously or asynchronously
     * @function ?
     * @param {optional Boolean} async - Whether or not to make the call asynchronously, defaults to true
     */
    run: function(async) {
        giLogger.log('run request for: ' + this.url, 'giAccountRequest.run');
        if(async == undefined) async = true;
        this.reset();

        this._xmlRequest = new XMLHttpRequest();

        var me = this;
        this._xmlRequest.onload = function(evt) {me._requestCallback(evt)};
        this._xmlRequest.onerror = function(evt) {me._requestError(evt)}

        if(this._postString == null) {
            giLogger.log('Making '+(async?'asynchronous':'synchronous')+' GET XMLHttpRequest to "'+this.url+'"', 'giAccountRequest');
            this._xmlRequest.open("GET", this.url, async);
            this._xmlRequest.send(null);
        } else {
            giLogger.log('Making '+(async?'asynchronous':'synchronous')+' POST XMLHttpRequest to "'+this.url+'"'+"\n"+'Post data: "'+this._postString+'"', 'giAccountRequest');
            this._xmlRequest.open("POST", this.url, async);
            this._xmlRequest.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
            this._xmlRequest.setRequestHeader("Content-length", this._postString.length); 
            this._xmlRequest.setRequestHeader("Connection", "close"); 
            this._xmlRequest.send(this._postString);
        }
        
    },

    /**
     * The request callback from the XMLHttpRequest, after reaching the server.
     * @function ?
     */
    _requestCallback: function(evt) {
        giLogger.log('request callback for: ' + this.url, 'giAccountRequest._requestCallback');
        giLogger.log("Response:\n"+this._xmlRequest.responseText, 'giAccountRequest._requestCallback');
        
        
        try {
            // Received data
            try {
                var evalJS = 'var result = (' + this._xmlRequest.responseText + ')';
                var sb = Components.utils.Sandbox(this.url);
                Components.utils.evalInSandbox(evalJS, sb);
            } catch(e) {
                this.dispatchEvent({name: giAccountRequest.FAIL});
                this._xmlRequest = null;
                return;
            }
            if(typeof sb.result != 'object' ) {
                this.dispatchEvent({name: giAccountRequest.FAIL});
                // There was a problem
            } else {
                // Received valid data
                if( sb.result.status=='ok' ) {
                    this.dispatchEvent({name: giAccountRequest.SUCCESS, result: sb.result});
                } else if( sb.result.status=='wrong_password' ) {
                    this.dispatchEvent({name: giAccountRequest.WRONG_PASSWORD});        
                } else if( sb.result.status=='no_account' ) {
                    this.dispatchEvent({name: giAccountRequest.NO_ACCOUNT});
                } else {
                    // Unknown
                    this.dispatchEvent({name: giAccountRequest.FAIL});
                }
            }       
            
            this._xmlRequest = null;
        } catch(e) {
            giLogger.warn(e, 'giAccountRequest._requestCallback');
            this.dispatchEvent({name: giAccountRequest.FAIL});
            this._xmlRequest = null;
        }
    },

    /**
     * The request callback from the XMLHttpRequest in case of error
     * @function ?
     */
    _requestError: function(evt) {
        giLogger.log('error for: ' + this.url, 'giAccountRequest._requestError');
        this.dispatchEvent({name: giAccountRequest.FAIL});
        this._xmlRequest = null;
    },

    /**
     * Cleans up the remnants of the previous run, preserving the URL and POST data
     * @function ?
     */
    reset: function() {
        giLogger.log('Reseting the giAccountRequest object', 'giAccountRequest');
        if(this._xmlRequest) this._xmlRequest.abort();
        this._xmlRequest = null;

    }
    //}
},{
    
    GTDI_URL: 'http://www.gtdinbox.com/',
    
    // Static Properties {
    /**
     * Dispatched when the request is successful
     * @event ?
     */
    SUCCESS: 'success',
    
    /**
     * Dispatched when the request failed (e.g. could not reach server)
     * @event ?
     */
    FAIL: 'fail',
    
    /**
     * Dispatched when the request failed due to a wrong password
     * @event ?
     */
    WRONG_PASSWORD: 'wrong_password',
    
    /**
     * Dispatched when the request failed due to a user not having an account
     * @event ?
     */
    NO_ACCOUNT: 'no_account',
    //}

    // Static Methods {
    

    /**
     * Create a request to find out if one or more accounts has an account and if the tested password is correct
     * Return JSON Format:
     * {<account>:{hasAccount:Boolean,correctPassword:Boolean}}
     * 
     * @function {static glRequest} ?
     * @param {optional glGmail} gmail - 
     * @param {Object} accounts - Format: {a0: <account>, p0: <password>, a0: ...}
     * @return The created request object
     */
    createHasAccountRequest: function(accounts) {
        return new giAccountRequest(
            'hasaccount',
            {},
            accounts
        );
    },

    /**
     * Creates and returns a request that adds the given label to the given
     * thread(s).
     * @function {static glRequest} ?
     * @param {String} label - The label to add
     * @param {mixed} threads - The thread or threads to add the label to
     * @return The created request object
     */
    createReadSettingsRequest: function(account) {
        return new giAccountRequest(
            'readsettings',
            {},
            {}
        );
    },
    
    
    /**
     * Creates and returns a request that adds the given label to the given
     * thread(s).
     * @function {static glRequest} ?
     * @param {String} label - The label to add
     * @param {mixed} threads - The thread or threads to add the label to
     * @return The created request object
     */
    createWriteSettingsRequest: function(account, prefs) {
        return new giAccountRequest(
            'writesettings',
            {},
            {}
        );
    },
    
    
    /**
     * Tell server to send the user their password. 
     * @function {static glRequest} ?
     * @param {String} label - The label to add
     * @param {mixed} threads - The thread or threads to add the label to
     * @return The created request object
     */
    createRequestPassword: function(account) {
        return new giAccountRequest(
            'requestpassword',
            {},
            {}
        );
    },
    

    //}
});
giAccountRequest.implement(giEventDispatcher); // Add event dispatching support
