/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Stephen Augenstein (perl.programmer@gmail.com)
 *
 * ***** END LICENSE BLOCK ***** */


/**
 * Base class for all main page plugins with UI objects to manage. Provides a
 * simple API for loading, initializing, and shutting down UI objects on page
 * unload.
 *
 * @class ? @extends giPlugin
 */
var giMainPageUIPlugin = giPlugin.extend({
    // Properties {
    /**
     * glGmail object
     * @property {glGmail} ?
     */
    gmail: null,

    /**
     * Reference to the plugin's pref wrapper object. All preference setting for
     * the plugin should go through this wrapper.
     * @property {giPluginPrefManager} ?
     */
    prefs: null,

    /**
     * The prefix for classes in the view, e.g. 'gtd'
     * @property {String}
     */
    _classPrefix: null,


    /**
     * A hash of all the UI objects
     * @property {Object} ?
     */
    _ui: null,
    //}

    // Methods {
    /**
     * Called by the plugin loader after the page base class has been succesfully
     * instantiated and plugins are ready for loading.
     * @function ?
     * @param {giPluginAccountData} accountData - Helper object for managing access to preferences and other account data
     * @param {Object} pageBase - The page base class, so {glGmail} for full pages for example
     * @param {String} classPrefix - The prefix for classes in the view, e.g. 'gtd'
     */
    load: function(accountData, pageBase, classPrefix) {
        this.prefs = accountData;
        this.gmail = pageBase;
        this._ui = {};
        this._classPrefix = classPrefix;
    },

    /**
     * Called on page unload, it cleans up everything
     * @function ?
     */
    unload: function() {
        for(var element in this._ui) {
            this._ui[element].disconnect();
        }
        this._ui = null;
    },

    /**
     * Returns the given UI object after initializing it.
     * @function {glUIObject} ?
     * @param {String} type - The UI object to return
     * @return The initialized UI object
     */
    getUIObject: function(type) {
        giLogger.log('Getting UI object '+type+'...', 'giUIPlugin');

        var object = null;
        if(!this._ui[type]) {
            // Not in cache so create and cache (or return if not supposed to cache)

            // Add class prefix to the begining of the type and raise the first character up
            // mainView -> gtdMainView
            theClass = this._classPrefix+type.substr(0,1).toUpperCase()+type.substr(1);

            giLogger.log('Creating object: "'+theClass+'"', 'giUIPlugin');

            try {
                object = new window[theClass]();
            } catch(e) {
                // Use giLogger.error because it gives a stack trace
                giLogger.error("Failure instantiating "+theClass+" UI object:\n"+e.message, 'giUIPlugin');
            }

            object.connect(this);

            // Is the ui object cachable - return if not
            if(object.isCacheable()) {
                // Cache the instantiated object
                this._ui[type] = object;
            }
        } else {
            // Get object from cache
            object = this._ui[type];
        }

        return object;
    },


    /**
     * Triggered by the {@link glPageWrapper.VIEW_CHANGED} event, it calls
     * getCurrentViewObject to make sure it gets connected.
     * Ths inheriting class is expected to add the event listener if it's needed
     * @param {Event} evt - The VIEW_CHANGED event
     */
    _viewChanged: function(evt) {
        
        // Clean out old view object (EVEN IF THE VIEW IS THE SAME!)
        if(typeof this._ui['_currentViewObject'] != 'undefined') {
            this._ui['_currentViewObject'].object.disconnect();
            delete this._ui['_currentViewObject'];
        }
        
        // Refresh the current view object:
        this.getCurrentViewObject();
    },


    /**
     * Returns the UI object for the current view, initializing it if it hasn't
     * already been loaded.
     * @function {giUIObject} ?
     * @return The view UI object
     */
    getCurrentViewObject: function() {
        if(!this.gmail.page.isUIReady()) return null;


        // In the cache, so return
        if(typeof this._ui['_currentViewObject'] != 'undefined') {
            return this._ui['_currentViewObject'].object;
        }

        // Get the UI object type for a call to getUIObject
        var uiObject = '';
        var viewType = this.gmail.page.getActiveViewType();
        switch(viewType) {
            case 'tl': uiObject = 'threadListView'; break;
            case 'cv': uiObject = 'conversationView'; break;
            case 'co': uiObject = 'composeView'; break;
            case 'ct': uiObject = 'contactsView'; break;
            case 's': uiObject = 'settingsView'; break;
        }

        if(uiObject == '') {
            return null;
        } else {
            try {
            this._ui['_currentViewObject'] = {
                object: this.getUIObject(uiObject),
                type: viewType
            };
            } catch(e) {
                giLogger.warn(e, 'giMainPageUIPlugin.getCurrentViewObject');
            }
        }

        return this._ui['_currentViewObject'].object;
    },

    //}
});