/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */


/**
 * Manages pages loads, page base class initialization, and plugin initialization.
 *
 * @class ?
 */
var giPageManager = giBase.extend(null ,{
    // Static Properties {
    /**
     * Stores all page instances currently loaded in the browser. Used to prevent
     * redundant instances and make garbage collection easier.
     * @property {static Array} ?
     */
    _pageInstances: [],

    /**
     * A hash of all the loaded files to prevent loading anything again.
     * @property {static Object} ?
     */
    _loadedFiles: {},
    //}

    // Static Methods {
    /**
     * Called by {giPageManager} once it's ready, it sets up page load event listeners.
     * @function {static} ?
     */
    initialize: function() {
        var appcontent = document.getElementById("appcontent");
        if( !appcontent ) {
            appcontent = window.document.getElementById("browser_content"); // Prism
        }
        appcontent.addEventListener("DOMContentLoaded", giPageManager._pageLoad, true);

        giPageManager._jsLoader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService(Components.interfaces.mozIJSSubScriptLoader);

        giLogger.log('Page watching hooked up', 'giPageManager');
    },

    /**
     * Called when a browser page is loaded. It tries to determine if it's a
     * gmail page load and then determines if there are plugins registered for
     * that type of gmail page load. If there are, it instantiates the page base
     * class and then loads all the plugins for that page type.
     * @function {static} ?
     * @param {Event} evt - The DOMContentLoaded event
     */
    _pageLoad: function(evt) {
        if(!giPageManager._validPageLoad(evt)) return;

        // Extract doc and URL
        var doc = evt.originalTarget;
        var url = (doc && doc.location && doc.location.href) ? doc.location.href : '';

        // Get page type for the page
        var pageType = giPageManager._getPageType(url);
        if(!pageType) return;


        // Initialize page base class if there are any plugins enabled
        var pageBase = null, account = null, pluginData = null;
        if(pageType == 'main') {
            account = doc.defaultView.wrappedJSObject.top.GLOBALS[10];

            pluginData = giPluginManager.getEnabledPlugins(account, pageType);
            if(pluginData.length == 0) return;

            pageBase = new glGmail(doc);
        } else {
            return;
        }

        // Initialize all plugins
        var plugins = [];
        for(var i = 0; i < pluginData.length; i++) {
            try {
                plugins.push(giPageManager._initializePlugin(account, pluginData[i], pageBase));
            } catch(e) {
                giLogger.warn('Failed to initialize "'+pluginData[i].plugin+'" for "'+pageType+'" page: '+e);
                
                var box = giPageManager.getPluginInstance(boxMain, pageBase);
                if( box && box.showLoadError ) {
                    // Add error
                    box.showLoadError(e);
                }
            }
        }

        // Start pageBase
        pageBase.connect();

        // Store data into page instances and hook up shutdown listener
        giPageManager._pageInstances.push({doc: doc, pageBase: pageBase, plugins: plugins});
        doc.defaultView.addEventListener('pagehide', giPageManager._pageUnload, true);
    },

    /**
     * Called when a browser page is unloaded. Sees if there is a pageInstance
     * object for the given document, and if there is, unloads all plugins and
     * the glGmail object.
     * @function {static} ?
     * @param {Event} evt - The pagehide event
     */
    _pageUnload: function(evt) {
        var doc = evt.target.ownerDocument != null ? evt.target.ownerDocument : evt.target;
        if(!!doc.defaultView.frameElement || doc.URL == "about:blank") return;

        doc.defaultView.removeEventListener('pagehide', giPageManager._pageUnload, true);

        for(var i = 0; i < giPageManager._pageInstances.length; i++) {
            var page = giPageManager._pageInstances[i];
            if(page.doc != doc) continue;

            // It's an active page so shut everything down and clean up
            giLogger.log('Unloading active page...', 'giPageManager');

            // Unload all plugins
            for(var j = 0; j < page.plugins.length; j++) {
                try {
                     page.plugins[j].unload();
                } catch(e) {
                    giLogger.warn("Error during plugin shutdown:\n"+e, 'giPageManager');
                }
            }

            // Unload page base class
            page.pageBase.unload(); // Should never throw exceptions, so no try...catch

            // Delete instance
            giPageManager._pageInstances.splice(i, 1);
            return;
        }
    },

    /**
     * Checks whether the page load event target is a valid page load. Must be as
     * quick as possible, as it is called every page load.
     * @function {static Boolean} ?
     * @param {Event} evt - The page load event
     * @return Whether the page load is valid
     */
    _validPageLoad: function(evt) {
        var doc = evt.originalTarget;
        var url = (doc && doc.location && doc.location.href) ? doc.location.href : '';

        // Is it roughly a gmail page?
        if(!/^http(s?):\/\/mail.google.com/.test(url)) return false;

        // Are we working with a real page load?
        if( typeof gBrowser=="undefined" ) { // Prism
            var browser = document.getElementById("browser_content");
            if( doc==browser.contentDocument ) {
                found = true;
            }
        } else {
            var browsers = gBrowser.browsers;
            var found = false;
            for(var i = browsers.length - 1; i>=0; i--) {
                if(doc == browsers[i].contentDocument) {
                    found = true;
                    break;
                }
            }
        }
        if(!found) return false;

        // Is this page already registered?
        for(var i = giPageManager._pageInstances.length - 1; i>=0; i--) {
            if(giPageManager._pageInstances[i].doc == doc) {
                giLogger.log('Page already registered', 'giPageManager');
                return false;
            }
        }

        return true;
    },

    /**
     * Tries to determine the page type from the given URL
     * @function {static String} ?
     * @param {String} url - The page URL
     * @return The calculated page type. null if not a valid page.
     */
    _getPageType: function(url) {
        // Check that base portion is okay
        if(!(new RegExp('^http(s?)://mail.google.com/mail/')).test(url) && !(new RegExp('^http(s?)://mail.google.com/a/[^/]+/')).test(url)) {
            return null;
        }

        // Does it have a view url parameter (common for popout pages)
        if(/[?&]view=/.test(url)) {
            return null;
        }

        return 'main';
    },

    /**
     * Initializes a plugin and returns an instance of it
     * @function {static giPlugin} ?
     * @param {String} account - The account of the person the page belongs to
     * @param {Object} data - Plugin data
     * @param {Object} pageBase - Instance of the "gl" namespace base API class for the page
     * @return Initialized plugin instance
     */
    _initializePlugin: function(account, data, pageBase) {
        if(!giPageManager._loadedPlugins) giPageManager._loadedPlugins = {};
        var key = data.type+'-'+data.plugin;
        var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);

        // Load script file if it hasn't yet been
        if(!giPageManager._loadedPlugins[key]) {
            for(var i = 0; i < data.files.length; i++) {
                giPageManager.loadFile(data.files[i]);
            }
            giPageManager._loadedPlugins[key] = true;
        }

        // Instantiate plugin and load
        var pluginInstance = new window[data.class]();
        pluginInstance.manifestURI = ios.newURI(data.manifest, null, null);
        pluginInstance.load(new giPluginPrefManager(account, data.plugin), pageBase);

        return pluginInstance;
    },
    
    /**
     * Add a single file into the environment
     * @function {} ?
     * @param {String} path - The file to load
     */
    loadFile: function(path) {
        if(!giPageManager._loadedFiles[path]) {
            giPageManager._jsLoader.loadSubScript(path);
            giPageManager._loadedFiles[path] = true;
        }
    },
    
    
    /**
     * Many plugins need access to the box plugin, so added this helper function
     * @function {giPlugin} ?
     * @param {Object} classDef - The definition class for the plugin
     * @param {Object} pageBase - Most likely glGmail instance
     * @return The instance/object of the requierd plugin class
     */
    getPluginInstance: function(classDef, pageBase) {
        for( var i = 0; i < giPageManager._pageInstances.length; i++ ) {
            if( giPageManager._pageInstances[i].pageBase==pageBase ) {
                for( var j = 0; j < giPageManager._pageInstances[i].plugins.length; j++ ) {
                    if( giPageManager._pageInstances[i].plugins[j] instanceof classDef ) {
                        return giPageManager._pageInstances[i].plugins[j];
                    }
                }
            }
        }
    },
    
    //}
});