/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */


/**
 * Manages plugin registration and initialization
 *
 * @class ?
 */
var giPluginManager = giBase.extend(null, {
    // Static Properties {
    /**
     * Hash of all the registered plugins with the plugin ids as the keys.
     * Contains the plugin name, description, and preference info.
     * @property {static Object} ?
     */
    _plugins: {},

    /**
     * Hash of all the registered plugin classes by page type.
     * @property {static Object} ?
     */
    _pagePlugins: {
        main: []
    },

    /**
     * Reference to the preference manager object
     * @property {giPrefManager} ?
     */
    _prefs: null,
    //}

    // Static Methods {
    /**
     * Perform general initialization
     * @function {static} ?
     */
    initialize: function() {
        // TODO: Check if plugin dir exists, and if it doesn't create it

        giPluginManager._prefs = Components.classes['@gtdinbox.com/pref-manager;1'].getService().wrappedJSObject;
        giPluginManager._ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    },

    /**
     * Returns the path to the given plugin's pref pane
     * @function {static String} ?
     * @param {String} plugin - The plugin id
     * @return The path to the pane
     */
    pluginPrefPane: function(plugin) {
        return giPluginManager._plugins[plugin].prefPane;
    },

    /**
     * Returns all enabled plugins for the given account and page type
     * @function {static Array} ?
     * @param {String} account - The account of the person the page belongs to
     * @param {String} pageType - The page type
     * @return Array of plugins
     */
    getEnabledPlugins: function(account, pageType) {
        // Get page plugins
        var pagePlugins = giPluginManager._pagePlugins[pageType];

        // Check each plugin going through account data to see if it's enabled
        var enabledPlugins = [];
        for(var i = 0; i < pagePlugins.length; i++) {
            // Do NOT modify pagePlugins, or you will permanently modify _pagePlugins[pageType]
            var enabled = this._prefs.getPluginPref(account, pagePlugins[i].plugin, 'enabled');
            if(enabled) enabledPlugins.push(pagePlugins[i]);
        }

        // List out the plugin names that are enabled for the account
        giLogger.log('Enabled plugins for "'+account+'" '+pageType+' page: ['+enabledPlugins.map(function(p) {return p.plugin}).toString()+']', 'giPluginManager');

        return enabledPlugins;
    },

    /**
     * Parses the given plugin manifest file and registers the plugin internally
     * @function {static} ?
     * @param {String} manifestFilename - The path to the plugin manifest file
     */
    registerPlugin: function(manifestFilename) {
        var d = giPluginManager._parseManifest(manifestFilename);

        // Store extracted manifest data
        for(var i = 0; i < d.pages.length; i++) {
            var page = d.pages[i];
            page.plugin = d.id; // Associate it with the plugin it came frome
            page.manifest = d.manifest; // Needed by giPageManager to initizlize the plugins
            giPluginManager._pagePlugins[page.type].push(page);
        }
        delete d.pages; // So it's not duplicated unnecessarily
        giPluginManager._plugins[d.id] = d;

        // Check version and update if necessarys
        var oldVersion = giPluginManager._prefs.getPluginVersion(d.id);
        if(oldVersion != d.version) {
            giLogger.log('version change for ' + d.id + ' from ' + oldVersion + ' to ' + d.version);
            // Update stored defaults
            var newDefaults = {};
            try  {
                // Permit to throw an error at any point (e.g. bad json, bad file) - it will not write preferences and it will not write new version; so we have the option to correct it and migrate it properly later
                
                // Load json decoder:
                var json = Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON);
                // Create an object for the defaults and write it to pref store:
                newDefaults = json.decode(giUtil.readFile(d.prefDefaults));
                if(typeof newDefaults['enabled'] == 'undefined') {
                    newDefaults.enabled = true;
                }
                this._prefs.setPluginDefaults(d.id, newDefaults);
                
                giLogger.log(d.id + ' version change - new defaults written to default-store');
                
                
                // For each account with preferences for this plugin,
                // ensure all default preference key/values are present, and optionally run migrate routine
                
                // Load migrator (but do not migrate yet):
                if( d.prefMigrator ) {
                    try {
                        giLogger.log(d.id + ' version change - load migrator, file: ' + d.prefMigrator);
                        giPageManager._jsLoader.loadSubScript(d.prefMigrator);
                        
                        giLogger.log(d.id + ' version change - load migrator, class: ' + d.prefMigratorClass);
                        var migrator = new window[d.prefMigratorClass]();
                        giLogger.log(d.id + ' version change - loaded migrator');
                    } catch(e) {
                        giLogger.warn(e, 'giPluginManager.registerPlugin - load migrator for ' + d.id);
                    }
                }
                
                var accounts = giPluginManager._prefs.getAccounts();
                for( var i = 0; i < accounts.length; i++ ) {
                    var account = accounts[i];
                    giLogger.log(d.id + ' version change - upgrading ' + account);
                    
                    // Sync the current prefs with the default prefs
                    giPluginManager._prefs.extendPluginPrefsWithDefaults(account, d.id);
                    
                    if( migrator ) {
                        giLogger.log(d.id + ' version change - migrate ' + account);
                        migrator.migrate(new giPluginPrefManager(account, d.id), oldVersion);
                    }
                    
                    giLogger.log(d.id + ' version change - upgraded ' + account);
                }
                
                if( migrator ) {                
                    // Clean up:
                    delete migrator;
                    window[d.prefMigratorClass] = undefined;
                }
                
                // Update plugin version
                this._prefs.setPluginVersion(d.id, d.version);
            
                giLogger.log(d.id + ' version change - complete');
            
                // Store changes
                giPluginManager._prefs.save();
            
            } catch (e) {
                giLogger.warn('Error loading "'+d.prefDefaults+'" defaults for '+d.id+':'+"\n"+e, 'giPluginManager');
            }
        }
    },
    


    /**
     * Registers all plugins found in the plugins folder that have a valid manifest
     * @function {static} ?
     */
    autoloadPlugins: function() {
        // TODO: Complete this method
        // Get a list of all folders in the plugin directory
        // For each folder
           // See if it has a manifest
           // Pass manifest to "registerPlugin"  (wrap in try/catch because registerPlugin throws exceptions)
    },

    // TODO: Install plugin (what does it get? - nsIFile?, nsIURI, string for file path to zip)




    /**
     * Parses the manifest file and returns the parsed data. Resolves all paths
     * in the manifest, so nothing needs to be done on the filenames post parse.
     * @function {static Object} ?
     * @param {String} manifestFilename - The path to the plugin manifest file
     * @return Manifest data
     */
    _parseManifest: function(manifestFilename) {
        // Create XML wrapper around manifest
        var doc = XML(giUtil.readFile(manifestFilename));
        var uri = giPluginManager._ios.newURI(manifestFilename, null, null);

        // Parse data into pluginData object
        var pluginData = {
            id: doc.id.toString(),
            version: doc.version.toString(),
            manifest: manifestFilename,
            description: doc.description.toString(),
            prefDefaults: uri.resolve(doc.prefs.@defaults.toString()),
            prefPane: uri.resolve(doc.prefs.@panel.toString()),
            prefMigrator: doc.prefs.@migrator.toString(),
            prefMigratorClass: doc.prefs.@migratorClass.toString(),
            pages: []
        };
        if( pluginData.prefMigrator ) pluginData.prefMigrator = uri.resolve(pluginData.prefMigrator);
        for each(var pageNode in doc.pages.page) {
            // Build pageData object
            var pageData = {
                type: pageNode.@type.toString(),
                class: pageNode.@class.toString(),
                files: []
            }
            for each(var fileNode in pageNode.file) {
                pageData.files.push(uri.resolve(fileNode.toString()));
            }

            // Add page data to pages array
            pluginData.pages.push(pageData);
        }

        // Validate parsed data (somewhat)
        if(!pluginData.id) throw new Error('Missing "id"');
        if(!pluginData.version) throw new Error('Missing "version"');
        if(!pluginData.description) throw new Error('Missing "description"');
        if(!pluginData.prefDefaults) throw new Error('Missing prefs element defaults attribute');
        if(!pluginData.prefPane) throw new Error('Missing prefs element panel attribute');
        if(pluginData.pages.length == 0) throw new Error('No registered pages');

        return pluginData;
    },

    /**
     * Loads and parses the plugin defaults file
     * @function {static Object} ?
     * @param {String} defaultsFilename - The path to the defaults file
     * @return Default data
     */
    /*
    _parseDefaults: function(defaultsFilename) {
        var defaults = {};
        try  {
            var fileData = giUtil.readFile(defaultsFilename);
            defaults = Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON).decode(fileData);
            if(typeof defaults['enabled'] == 'undefined') defaults.enabled = true;
        } catch (e) {
            giLogger.warn('Error loading "'+defaultsFilename+'" defaults:'+"\n"+e, 'giPluginManager');
        }
        return defaults;
    }
    */
    //}
});