/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The parent class for all "view" UI objects. It includes code for recording that
 * it has been connected and is automatically connected to the view when loaded
 * by {@link giUIPlugin.getUIObject}.
 *
 * @class {abstract} ?
 */
var giUIObject = giBase.extend({
    // Properties {
    /**
     * The giUIPlugin that this UI object is associated with
     * @property {giUIPlugin} ?
     */
    _main: null,

    /**
     * Whether or not it is connected. Useful for only running connect once or
     * re-connecting when it gets disconnected.
     * @property {Boolean} ?
     */
    _isConnected: null,
    //}

    /**
     * The constructor.
     * @constructor ?
     */
    constructor: function() {
        if(this.constructor === giUIObject) throw new Error('Cannot be instantiated');
    },

    // Methods {
    /**
     * Connects the element to the given giUIPlugin.
     * @function ?
     * @param {giUIPlugin} main - The giUIPlugin object for the plugin
     */
    connect: function(main) {
        this._isConnected = true;
        this._main = main;
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        this._isConnected = false;
        this._main = null;
    },

    /**
     * Whether the UI object can be cached.
     * Most things do not cache, beware the implications!
     * @function {Boolean} ?
     * @return Whether the UI can be cached
     */
    isCacheable: function() {
        return false;
    },

    /**
     * Returns whether or not connected
     * @function {Boolean} ?
     * @return Connection status
     */
    isConnected: function() {
        return this._isConnected;
    }
    //}
}, {
    // Static Methods {
    /**
     * Simulates the given mouse event on the given node
     * @function {static} ?
     * @param {Element} node - The node to simulate the event on
     * @param {String} eventType - The event type: mousedown, mouseup, ...
     * @param {optional int} x - The X screen/client coord (defaults to 50)
     * @param {optional int} y - The Y screen/client coord (defaults to 50)
     */
    simulateMouseEvent: function(node, eventType, x, y) {
        if( typeof x=='undefined' ) x = 50;
        if( typeof y=='undefined' ) y = 50;
        var event = node.ownerDocument.createEvent("MouseEvents");
        event.initMouseEvent(
            eventType,
            true, // can bubble
            true, // cancellable
            node.ownerDocument.defaultView,
            1, // clicks
            x, y, // screen coordinates
            x, y, // client coordinates
            false, false, false, false, // control/alt/shift/meta
            0, // button,
            (eventType == 'mouseover' || eventType == 'mouseout') ? node : null
        );
        node.dispatchEvent(event);
    },

    addClassName: function(el, cn) {
        if( giUIObject.hasClassName(el,cn) ) return;
        if( el.className ) cn = el.className + " " + cn;
        el.className = cn;
    },

    removeClassName: function(el, cn) {
        el.className = el.className.replace(cn, "");
        el.className = el.className.replace(/\s+$/, "");
    },

    hasClassName: function(el, cn) {
        var re = new RegExp("(^|\\s)" + cn + "(\\s|$)");
        return re.test( el.className );
    },

    createGmailButton: function(name, v, options, doc) {
        switch(name) {
            case "gmail-send-button":
            case "gmail-reply-button":
            case "gmail-header-button":
                var el = doc.createElement("DIV");
                el.className = "goog-imageless-button goog-inline-block";
                if( name=="gmail-send-button" ) el.className += " B8wEwc ipG21e";
                if( name=="gmail-reply-button" ) el.className += " B8wEwc ipG21e";
                el.setAttribute("role", "button");
                if( "gmail-header-button" ) {
                    el.setAttribute("tabindex", "0");
                } else {
                    el.setAttribute("tabindex", "1");
                }
                el.setAttribute("onfocus", "javascript:this.className+=' goog-imageless-button-focused'");
                el.setAttribute("onblur", "javascript:this.className=this.className.replace(' goog-imageless-button-focused', '')");
                if( options && options.collapseLeft ) el.className += " " + "goog-imageless-button-collapse-left";
                if( options && options.collapseRight ) el.className += " " + "goog-imageless-button-collapse-right";
                el.innerHTML = "<div class='goog-inline-block goog-imageless-button-outer-box'><div class='goog-inline-block goog-imageless-button-inner-box'><div class='goog-imageless-button-pos'><div class='goog-imageless-button-top-shadow'> </div><div class='goog-imageless-button-content'>" + v + "</div></div></div></div>"
                return el;
        }
    },

    /**
     * Get the offset from the top/left of the viewport
     * @function {int[]} ?
     * @param {Element} element - the el to get offset of
     * @param {optional Element} boundaryElement - Get cumulative offset from a given bounding point (rather than viewport)
     */
    cumulativeOffset: function(element, boundaryElement) {
        var valueT = 0, valueL = 0;
        do {
            valueT += element.offsetTop  || 0;
            valueL += element.offsetLeft || 0;
            element = element.offsetParent;
            if( boundaryElement && element==boundaryElement ) break;
        } while (element);
        return [valueL, valueT];
    },

    getDimensions: function(element) {
        var doc = element.ownerDocument;
        if( element.style.display!="none" ) return {width: element.offsetWidth, height: element.offsetHeight};
        return {width: 0, height: 0};
    },

    pageScrollOffset: function(doc){
		doc = doc || document;
		var deltaX =  window.pageXOffset
					|| doc.documentElement.scrollLeft
					|| doc.body.scrollLeft
					|| 0;
		var deltaY =  window.pageYOffset
					|| doc.documentElement.scrollTop
					|| doc.body.scrollTop
					|| 0;
		return [deltaX, deltaY];
	},

    viewport: function(doc, useScrollOffset, incScrollBars) {
        var win = doc.defaultView;
        var browserWidth = 0;
        var browserHeight = 0;
        if( incScrollBars ) {
            browserWidth = win.innerWidth;
            browserHeight = win.innerHeight;
        } else {
            browserWidth = doc.documentElement.clientWidth;
            browserHeight = doc.documentElement.clientHeight;
        }

        if( useScrollOffset ) {
            browserWidth -= win.pageXOffset;
            browserHeight -= win.pageYOffset;
        }

        return [browserWidth, browserHeight];
    },

    getBorderWidth: function(el) {
        // Return the width, in pixels, of the border
        var css = el.ownerDocument.defaultView.getComputedStyle(el,null);
        if( css.borderWidth ) {
            var width = css.borderWidth;
            return [width, width, width, width];
        }
        
        var widths = [];
        widths.push( css.borderTopWidth || '0px' );
        widths.push( css.borderRightWidth || '0px' );
        widths.push( css.borderBottomWidth || '0px' );
        widths.push( css.borderLeftWidth || '0px' );
        return widths;
        
        /*
        var borderWidth = el.style.borderWidth || (css? css.getPropertyValue('borderWidth'):null);
        if( !borderWidth ) borderWidth = el.style.border || (css? css.getPropertyValue('border'):null);
        var borderStyle = el.style.borderStyle || (css? css.getPropertyValue('borderStyle'):null);
        if( borderStyle && borderStyle.indexOf("none")==-1 && borderWidth && (borderWidth = borderWidth.match(/(\d)px/)) ) {
            borderWidth = parseInt(borderWidth[1]);
        } else {
            borderWidth = 0;
        }
        return borderWidth;
        */
    },
    
    getPadding: function(el) {
        var css = el.ownerDocument.defaultView.getComputedStyle(el,null);
        if( css.padding ) {
            var padding = css.padding;
            return [padding, padding, padding, padding];
        }
        
        var paddings = [];
        paddings.push( css.paddingTop || '0px' );
        paddings.push( css.paddingRight || '0px' );
        paddings.push( css.paddingBottom || '0px' );
        paddings.push( css.paddingLeft || '0px' );
        return paddings;
    },
    
    /**
	* Check we're in bounds with the proposed x/y
	* @function {int[]} ?
	* @param {Element} elPopup - The element being positioned
	* @param {optional int} x - The x position 
	* @param {optional int} y - The y position
	* @return Array of potentially modified x/y, and 3rd index with viewport-y
	*/
	boundsCheckXY: function(elPopup, x,y, ignoreX, ignoreY) {
		
		// Check right/bottom
		var popupDims = giUIObject.getDimensions(elPopup);
		var viewport = giUIObject.viewport(elPopup.ownerDocument);
		
        
        // Factor in scroll:
        var scroll = giUIObject.pageScrollOffset(elPopup.ownerDocument);
        viewport[0] += scroll[0];
        viewport[1] += scroll[1];
		
		if( !ignoreX ) if( (x + popupDims.width) > viewport[0] ) x = viewport[0] - popupDims.width - 18;
		if( !ignoreY ) if( (y + popupDims.height) > viewport[1] ) y = viewport[1] - popupDims.height;
		
		// Check not off left/top
		if( x<0 ) x = 0;
		if( y<0 ) y = 0;
		
		return [x,y, viewport[0], viewport[1]];
	},
    
    //}
});