/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Queue requests to manipulate the More Actions and Label menus to update threads.
 * Faster and more visually coherent than ajax calls; but more susceptable to changes.
 *
 * @class ?
 */

var glActionDOMBuffer = giBase.extend({

    _queue: null,
    
    _currentQItem: null,
    
    _queueInterval: null,
    
    _gmail: null, 
    
    constructor: function(gmail) {
        this._gmail = gmail;
        this._queue = [];
    },
    
    /**
    * Add a request to the queue
    * @function ?
    * @param {Function} fStart - The function that will kick start the queued item 
    */
    _addQueueItem: function(fStart) {
        var me = this;
        var qItem = {
            date: (new Date()).valueOf(),
            fStart: giUtil.bind(this, fStart),
        };
        qItem.release = function() {
            me._releaseQueueItem(qItem);
        };
        this._queue.push(qItem);
        this._tryNextQueue();
    },
    
    /**
    * Remove an item from the queue; and check for next queue item
    * @function ?
    * @param {Object} qItem - The queue object to remove
    */
    _releaseQueueItem: function(qItem) {
        for( var i = 0; i < this._queue.length; i++ ) {
            if( this._queue[i]===qItem ) {
                this._queue.splice(i,1);
                break;
            }
        }
        if( this._currentQItem==qItem ) this._currentQItem = null;
        this._tryNextQueue();
    },
    
    /**
    * See if there if an item on the queue waiting to start (and no queue item is running). Start it.
    * @function ?
    */
    _tryNextQueue: function() {
            
        if( this._queue.length==0 || this._queue[0]==this._currentQItem ) {	// No items, or in progress
            return;
        }
        
        this._currentQItem = this._queue[0];
        this._currentQItem.fStart(this._currentQItem);
        
        // Start interval
        if( !this._queueInterval ) {
            this._queueInterval = setInterval(giUtil.bind(this, this._queueIntervalFunction), 500);
        }
    },
    
    /**
    * Runs while ever there are items on the queue.
    * Look for aging actions (that may have got snagged) and remove them.
    * @function ?
    */
    _queueIntervalFunction: function() {
        // Clear anything over 10 seconds (do not release it; remove it)
        var d = (new Date()).valueOf() - (1000*10);
        for( var i = this._queue.length - 1; i >= 0; i--) {
            if( this._queue[i].date < d ) {
                if( this._queue[i]==this._currentQItem ) this._currentQItem = null;
                this._queue.splice(i,1);
            }
        }
        // Test to clear interval
        if( this._queue.length==0 ) {
            clearInterval(this._queueInterval);
            this._queueInterval = null;
        }
        this._tryNextQueue();
    },
    

    addLabelToThread: function(labelname, fCallback, _qItem) {
        try {
            /*
                Queue it up (we have to, as it's a single threaded process).
                Test to make sure the list of labels has been loaded.
                Scan the list of labels,
                        If it's checked, remove from queue and callback.
                        Otherwise, click it (mousedown+mouseup), and rescan list every 200ms (up to 2 seconds) until it's recorded as checked. 
            */


            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) { this.addLabelToThread(labelname, fCallback, qItem) })
                return;
            }
            

            // Scan the list of labels
            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            var elList = actionsHeader.labelsDropdownMenuLabels;
            var doc = elList.ownerDocument;
            

            for( var i = 0; i < elList.childNodes.length; i++ ) {
                var elLi = elList.childNodes[i];
                if( giUtil.stripTags(elLi.innerHTML)==labelname ) {	// Match
                    if( elLi.getAttribute("aria-checked")=="true" ) {	// Already on
                        if( fCallback ) fCallback();
                        _qItem.release();
                    } else {	// Turn it on; then test when it's activated.
                        giUIObject.simulateMouseEvent(elLi, 'mousedown');
                        giUIObject.simulateMouseEvent(elLi, 'mouseup');
                        var count = 0;
                        this._addLabelInterval = setInterval( giUtil.bind(this, function() {
                            if( elLi.getAttribute("aria-checked")=="true" || ++count>5 ) {
                                clearInterval( this._addLabelInterval );
                                setTimeout( giUtil.bind(this, function() {
                                    if( fCallback ) fCallback();
                                    _qItem.release();
                                }), 500 );
                            }
                        }), 200);
                    }
                    return;
                }
            }
            _qItem.release();	// Nothing found

        } catch(e) {
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.removeLabelFromThread', this._gmail);
        }
    },

    removeLabelFromThread: function(labelname, fCallback, _qItem) {
        try {
            /*
                Queue it up (we have to, as it's a single threaded process).
                Test to make sure the list of labels has been loaded.
                Scan the list of labels,
                        If it's unchecked, remove from queue and callback.
                        Otherwise, click it (mousedown+mouseup), and rescan list every 200ms (up to 2 seconds) until it's recorded as checked. 
            */



            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) { this.removeLabelFromThread(labelname, fCallback, qItem) })
                return;
            }


            // Scan the list of labels
            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            var elList = actionsHeader.labelsDropdownMenuLabels;
            for( var i = 0; i < elList.childNodes.length; i++ ) {
                var elLi = elList.childNodes[i];
                if( giUtil.stripTags(elLi.innerHTML)==labelname ) {	// Match
                    if( elLi.getAttribute("aria-checked")!="true" ) {	// Already off
                        if( fCallback ) fCallback();
                        _qItem.release();
                    } else {	// Turn it off; then test when it's activated.
                        giUIObject.simulateMouseEvent(elLi, 'mousedown');
                        giUIObject.simulateMouseEvent(elLi, 'mouseup');
                        var count = 0;
                        this._removeLabelInterval = setInterval( giUtil.bind(this, function() {
                            if( elLi.getAttribute("aria-checked")!="true" || ++count>5 ) {
                                clearInterval( this._removeLabelInterval );
                                setTimeout( giUtil.bind(this, function() {
                                    if( fCallback ) fCallback();
                                    _qItem.release();
                                }), 300 );
                            }
                        }), 200);
                    }
                    return;
                }
            }
            _qItem.release();	// Nothing found


        } catch(e) {
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.removeLabelFromThread', this._gmail);
        }
    },






    addStarToThread: function(fCallback, _qItem) {
        try {
            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) {this.addStarToThread(fCallback, qItem)} );
                return;
            }
            

            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            actionsHeader.clickMoreActionsItem( glActionsHeader.ADD_STAR );

            setTimeout( giUtil.bind(this, function() {
                if( fCallback ) fCallback();
                _qItem.release();
            }), 750 );

        } catch(e) {
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.addStarToThread', this._gmail);
        }
    },

    removeStarFromThread: function(fCallback, _qItem) {
        try {
            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) {this.removeStarFromThread(fCallback, qItem)} );
                return;
            }

            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            actionsHeader.clickMoreActionsItem( glActionsHeader.REMOVE_STAR );

            setTimeout( giUtil.bind(this, function() {
                if( fCallback ) fCallback();
                _qItem.release();
            }), 750 );

        } catch(e) {
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.removeStarThread', this._gmail);
        }
    },


    markThreadRead: function(fCallback, _qItem) {
        try {
            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) {this.markThreadRead(fCallback, qItem)} );
                return;
            }

            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            actionsHeader.clickMoreActionsItem( glActionsHeader.MARK_AS_READ );

            setTimeout( giUtil.bind(this, function() {
                if( fCallback ) fCallback();
                _qItem.release();
            }), 500 );
        } catch(e) {
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.markThreadRead', this._gmail);
        }
    },


    spamThread: function(fCallback, _qItem) {
        try {
            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) {this.spamThread(fCallback, qItem)} );
                return;
            }

            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            var el = actionsHeader.getButton(glActionsHeader.REPORT_SPAM_BUTTON);
            giUIObject.simulateMouseEvent(el, 'mousedown');
            giUIObject.simulateMouseEvent(el, 'mouseup');


            setTimeout( giUtil.bind(this, function() {
                if( fCallback ) fCallback();
                _qItem.release();
            }), 500 );

        } catch(e) {
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.spamThread', this._gmail);
        }
    },

    deleteThread: function(fCallback, _qItem) {
        try {
            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) {this.deleteThread(fCallback, qItem)} );
                return;
            }

            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            var el = actionsHeader.getButton(glActionsHeader.DELETE_BUTTON);
            giUIObject.simulateMouseEvent(el, 'mousedown');
            giUIObject.simulateMouseEvent(el, 'mouseup');


            setTimeout( giUtil.bind(this, function() {
                if( fCallback ) fCallback();
                _qItem.release();
            }), 500 );

        } catch(e) {
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.deleteThread', this._gmail);
        }
    },

    archiveThread: function(fCallback, _qItem) {
        try {
            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) {this.archiveThread(fCallback, qItem)} );
                return;
            }

            
            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            var el = actionsHeader.getButton(glActionsHeader.ARCHIVE_BUTTON);
            if( el ) {	
                giUIObject.simulateMouseEvent(el, 'mousedown');
                giUIObject.simulateMouseEvent(el, 'mouseup');
            } else {
                actionsHeader.clickMoreActionsItem( glActionsHeader.ARCHIVE );
            }
            
            setTimeout( giUtil.bind(this, function() {
                if( fCallback ) fCallback();
                _qItem.release();
            }), 500 );

        } catch(e) {
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.archiveThread', this._gmail);
        }
    },
    
    createLabel: function(fCallback, _qItem) {
        try {
            

            if( !_qItem ) {	// Create a proper queue item
                this._addQueueItem( function(qItem) { this.createLabel(fCallback, qItem) })
                return;
            }

            var actionsHeader = this._gmail.getCurrentViewObject().actionsHeader;
            var doc = this._gmail.canvasDocument;
            var elCreate = doc.evaluate("//div[@act='14']", actionsHeader.labelsDropdownMenu, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
            if( elCreate ) {
                giUIObject.simulateMouseEvent(elCreate, 'mousedown');
                giUIObject.simulateMouseEvent(elCreate, 'mouseup');
            }
            
            // Test for it's appearance //div[@class='az']/input[@class='qc']
            var d = {count: 0};
            d.interval = setInterval( function(){
                var elInput = actionsHeader.getCreateLabelInput();
                if( elInput || d.count++>5 ) {
                    clearInterval( d.interval );
                    if( fCallback ) fCallback();
                    _qItem.release();
                }
            }, 200);
            
        } catch(e) {
            if( fCallback ) fCallback();
            if( _qItem ) _qItem.release();
            giLogger.warn(e, 'glActionDOMBuffer.createLabel', this._gmail);
        }
    },
	
});