/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Manages running of actions and notifications of actions that have occurred.
 * Designed for actions that occur to threads.
 *
 * @class ?
 */
var glActionManager = giBase.extend({
    // Properties {
    /**
     * The {@link glActionBuffer} object that glActionManager uses as one of its backends
     * @property {glActionBuffer} ?
     */
    actionBuffer: null,

    /**
     * The {@link glActionDOMBuffer} object that glActionManager uses as one of its backends
     * @property {glActionDOMBuffer} ?
     */
    actionDOMBuffer: null,
	
    /** @property {Object[]} ? Recently (code) requested actions */
    _recentActions: null,

    /**
     * glGmail object
     * @property {glGmail} ?
     */
    _glGmail: null,
    //}

    // Methods {
    /**
     * Connects the action manager to the given glGmail object's page.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        this._glGmail = gmail;
        this.actionBuffer = new glActionBuffer(gmail);
        this.actionBufferSync = new glActionBuffer(gmail);
        this.actionBufferSync.async = false;
        this._recentActions = [];
        this.actionDOMBuffer = new glActionDOMBuffer(gmail);
        
        // Listen for environment generated actions (originally from trafficwatcher)
        this._glGmail.trafficWatcher.addEventListener(glTrafficWatcher.ACTION_REQUEST, this._environmentGeneratedAction, this);
        //gmail.environment.addEventListener(glEnvironment.ACTION_PERFORMED, this._environmentGeneratedAction, this);
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    shutdown: function() {
        this._glGmail = null;
        this.actionBuffer.shutdown();

        this._removeAllListeners();
    },

    /**
     * Performs the given action
     * @function ?
     * @paramset Action
     * @param {Object} action - The action constant for the action you want to perform
     * @param {glThread or String or Array} thread - The thread(s) you want to perform the action on
	 * @param {String} serverActionType - How to interact with server
     * @paramset Label
     * @param {Object} action - The action constant for the action you want to perform
     * @param {glThread or String or Array} thread - The thread(s) you want to perform the action on
	 * @param {String} serverActionType - How to interact with server
     * @param {String} label - The label to add or remove
     * @paramset Environment/Label
     * @param {Object} action - The action constant for the action you want to perform
     * @param 
     * @param {String} serverActionType - How to interact with server
     * @param {String} label - The label to create, rename or delete
     */
    performAction: function(action, threads, serverActionType, label) {
        // If threads is not an array, make it one
        if(typeof threads != 'array') threads = [threads];


		
        for(var i = 0; i < threads.length; i++) {
            // Perform action
            var threadId = (threads[i] instanceof glThread ? threads[i].threadId : threads[i]);
            if( this.actionBuffer[action['bufferFunc']] && (serverActionType==glActionManager.USE_AJAX || serverActionType==glActionManager.USE_AJAX_SYNC || !this.actionDOMBuffer[action['bufferFunc']]) ) {
                var args = [];
                if(threadId) args.push(threadId);
                if(label) args.push(label);
                if( serverActionType==glActionManager.USE_AJAX ) {
                    this.actionBuffer[action['bufferFunc']].apply(this.actionBuffer, args);
                } else {
                    this.actionBufferSync[action['bufferFunc']].apply(this.actionBufferSync, args);
                }
            } else if( serverActionType==glActionManager.USE_DOM ) {
                var args = [];
                if( label ) args.push(label);
                this.actionDOMBuffer[action['bufferFunc']].apply(this.actionDOMBuffer, args);
            }


            // Notify listeners that the action was performed
            if( serverActionType!=glActionManager.USE_DOM ) {
                var performedAction = {action: action, serverActionType: serverActionType};
                if( threadId ) {
                    performedAction.itemType = 'thread';
                    performedAction.item = threadId;
                } else {
                    performedAction.itemType = 'environment';
                }
                if(label) performedAction.label = label;
                if( threads[i] instanceof glThread ) performedAction.threadObject = threads[i];
                this._distributeAction(performedAction);
            }
            
            if( serverActionType==glActionManager.USE_AJAX || serverActionType==glActionManager.USE_AJAX_SYNC ) {
            	this._glGmail.environment.requestRefresh();
            }
        }
    },


    /**
     * Callback for the {@link glTrafficWatcher.ACTION_REQUEST} event. Notifies
     * listeners what event occurred.
     * @function ?
     * @param {Event} evt - The ACTION_REQUEST event
     * @return void
     */
    _environmentGeneratedAction: function(evt) {
        giLogger.log('Notified by environment that an action was performed', 'glActionManager');
        var actions = this._parseURLAndData(evt.url, evt.postData);
        for(var i = 0; i < actions.length; i++) {
            actions[i].environmentGenerated = true;
            this._distributeAction(actions[i]);
        }
    },

    /**
     * Takes the given action and affected threads and notifies all listeners
     * about what happened
     * @function ?
     * @param {Object} action - The action data
     * @... {Object} action - The action constant for the performed action
     * @... {String} label - The label that was added or removed if needed
     * @... {String} item - The affected item's id
     * @... {String} itemType - Either 'thread' or 'message'
     */
    _distributeAction: function(action) {
        giLogger.log('Notifying listeners that an action took place', 'glActionManager');
        if(action.itemType == 'thread') {
            // Notify listeners for this thread
            this.dispatchEvent({name: 'thread_action_'+action.item, action: action.action, serverActionType: action.serverActionType, threadObject: action.threadObject, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest, label: action.label});

            // Dispatch THREAD_CHANGED event
            this.dispatchEvent({name: glActionManager.THREAD_CHANGED, action: action.action, thread: action.item, serverActionType: action.serverActionType, threadObject: action.threadObject, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest, label: action.label});
        } else if( action.itemType=='environment' ) {
            var evt = {name: glActionManager.ENVIRONMENT_CHANGED, action: action.action, label: action.label, newLabel: action.newLabel, serverActionType: action.serverActionType, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest };
            this.dispatchEvent( evt ) ;
        } else {
            // Dispatch MESSAGE_CHANGED event
            var evt = {name: glActionManager.MESSAGE_CHANGED, action: action.action, message: action.item, serverActionType: action.serverActionType, threadObject: action.threadObject, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest};
            if(action.label) evt.label = action.label;
            this.dispatchEvent(evt);
        }

        // Dispatch generic action event
        this.dispatchEvent({name: action.action.eventName, item: action.item, itemType: action.itemType, serverActionType: action.serverActionType, threadObject: action.threadObject, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest, label: action.label, newLabel: action.newLabel});
    },

    /**
     * Parses the given request URL and post data to determine the action and affected threads
     * @function ?
     * @param {String} url - The action request URL
     * @param {String} postData - The request post data
     * @return An array of objects that can be passed to _distributeAction for event dispatching
     */
    _parseURLAndData: function(url, postData) {
        var match = /[?&]act=([^&]+)/.exec(url);
        if(!match) return [];

        giLogger.log('Action code is: '+match[1], 'glActionManager');

        var actions = [];
        var urlAction = match[1];

        // Extract affected items from post data
        postData = postData.replace(/^&+/, '').replace(/&+$/, ''); // Strip &s from beginning and end
        if( /cc_|nc_|dc_/i.test(urlAction) ) {
            // Performed environment action
            var itemType = 'environment';
            var items = null;
        } else {
            if(/&?t=/.test(postData)) {
                // Performed action on a thread
                var itemType = 'thread';
                var items = postData.split(/&?t=/);
            } else {
                // Performed action on message
                var itemType = 'message';
                var items = postData.split(/&?m=/);
            }
        }
        items.shift(); // Remove empty string from begining

        // Create simple helper function that merges in the affected item and type into the given action
        var addActionForItems = function(data) {
            if( items ) {
                for(var i = 0; i < items.length; i++) {
                    var action = {itemType: itemType, item: items[i]};
                    for( d in data ) action[d] = data[d];
                    actions.push(action);
                }
            } else {
                var action = {itemType: itemType};
                for( d in data ) action[d] = data[d];
                actions.push(action);
            }
        };

        // Determine action based on action code and process
        if(urlAction == 'arl') { // Multiple label modify
            // Calculate label actions - adds and removes
            var actionsTemp = [];
            var addLabelRegex = /[&?]acn=([^&]+)/g, removeLabelRegex = /[&?]rcn=([^&]+)/g, matches = null;
            url = decodeURIComponent(url);
            while((matches = addLabelRegex.exec(url)) != null) {
                actionsTemp.push({ action: glActionManager.ADD_LABEL, label: matches[1] });
            }
            while((matches = removeLabelRegex.exec(url)) != null) {
                actionsTemp.push({ action: glActionManager.REMOVE_LABEL, label: matches[1] });
            }

            // Go through actions and add
            for(var i = 0; i < actionsTemp.length; i++) {
                addActionForItems(actionsTemp[i]);
            }
        } else if(/^ac_/.test(urlAction)) { // Add label
            var label = decodeURIComponent(/^ac_(.*)/.exec(urlAction)[1]);
            addActionForItems({ action: glActionManager.ADD_LABEL, label: label});
        } else if(/^rc_/.test(urlAction)) { // Remove label
            var label = decodeURIComponent(/^rc_(.*)/.exec(urlAction)[1]);
            if(label == '^i') {
                addActionForItems({ action: glActionManager.ARCHIVE });
            } else {
                addActionForItems({ action: glActionManager.REMOVE_LABEL, label: label });
            }
        } else if(urlAction == 'sp') { // Report spam
            addActionForItems({ action: glActionManager.REPORT_SPAM });
        } else if(urlAction == 'us') { // UnReport spam
            addActionForItems({ action: glActionManager.NOT_SPAM });
        } else if(urlAction == 'st') { // Add star
            addActionForItems({ action: glActionManager.STAR });
        } else if(urlAction == 'xst') { // Remove star
            addActionForItems({ action: glActionManager.UNSTAR });
        } else if(urlAction == 'tr') { // Item deleted
            addActionForItems({ action: glActionManager.DELETE });
        } else if(urlAction == 'ib') { // Item moved to inbox
            addActionForItems({ action: glActionManager.MOVE_INBOX });
        } else if(urlAction.substr(0,3)=="cc_") { // Create label
            addActionForItems({ action: glActionManager.CREATE_LABEL, label: decodeURIComponent(urlAction.substr(3))});
        } else if(urlAction.substr(0,3)=="nc_") { // Rename a label
            var label = decodeURIComponent(urlAction.substr(3)).split("^");	// Format: [oldlabel, newlabel]
            addActionForItems({ action: glActionManager.RENAME_LABEL, label: label[0], newLabel: label[1]});
        } else if(urlAction.substr(0,3)=="dc_") { // Delete label
            addActionForItems({ action: glActionManager.DELETE_LABEL, label: decodeURIComponent(urlAction.substr(3))});
        }

        return actions;
    },
    //}
},{
    // Static Properties {
    /** @property {static Object} ? Action constant for adding a label */
    ADD_LABEL: { bufferFunc: 'addLabelToThread', eventName: 'label_added' },
    /** @property {static Object} ? Action constant for removing a label */
    REMOVE_LABEL: { bufferFunc: 'removeLabelFromThread', eventName: 'label_removed' },
    /** @property {static Object} ? Action constant for staring */
    STAR: { bufferFunc: 'addStarToThread', eventName: 'star_added' },
    /** @property {static Object} ? Action constant for unstaring */
    UNSTAR: { bufferFunc: 'removeStarFromThread', eventName: 'star_removed' },
    /** @property {static Object} ? Action constant for archiving */
    ARCHIVE: { bufferFunc: 'archiveThread', eventName: 'item_archived' },
    /** @property {static Object} ? Action constant for moving to inbox */
    MOVE_INBOX: { bufferFunc: 'moveInbox', eventName: 'item_move_inbox' },
    /** @property {static Object} ? Action constant for deleting */
    DELETE: { bufferFunc: 'deleteThread', eventName: 'item_deleted' },
    /** @property {static Object} ? Action constant for reporting spam */
    REPORT_SPAM: { bufferFunc: 'spamThread', eventName: 'item_reported_spam' },
    /** @property {static Object} ? Action constant for reporting spam */
    NOT_SPAM: { bufferFunc: 'notSpamThread', eventName: 'item_not_spam' },
    /** @property {static Object} ? Action constant for deleting */
    MARK_AS_READ: { bufferFunc: 'markThreadRead', eventName: 'item_marked_read' },
    /** @property {static Object} ? Action constant for creating a label */
    CREATE_LABEL: { bufferFunc: 'createLabel', eventName: 'label_created' },
    /** @property {static Object} ? Action constant for renaming a label */
    RENAME_LABEL: { bufferFunc: 'renameLabel', eventName: 'label_renamed' },
    /** @property {static Object} ? Action constant for deleting a label */
    DELETE_LABEL: { bufferFunc: 'deleteLabel', eventName: 'label_deleted' },

    /**
     * Dispatched when a environment is changed.
     * @event ?
     */
    ENVIRONMENT_CHANGED: 'environment_changed',

    /**
     * Dispatched when a thread is changed. Includes the action, which is a reference
     * to one of the action constants. i.e. <code>glActionManager.REMOVE_LABEL</code>.
     * It also includes the label if relevant and the thread id.
     * @event ?
     */
    THREAD_CHANGED: 'thread_changed',

    /**
     * Dispatched when a message is changed. Includes the action like {@link THREAD_CHANGED},
     * the message id, and the label if relevant.
     * @event ?
     */
    MESSAGE_CHANGED: 'message_changed',

    /** @property {static String} ? Tells performAction to use ajax if possible */
    USE_AJAX: 'ajax',
    
    /** @property {static String} ? Tells performAction to use ajax with sync if possible */
    USE_AJAX_SYNC: 'ajax_sync',
    
    /** @property {static String} ? Tells performAction to use dom menu clicking if possible */
    USE_DOM: 'dom',
    
    //}
});
glActionManager.implement(giEventDispatcher); // Add event dispatching support