/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Provides interfaces to the data that Gmail has. Label data, unread counts,
 * other e-mail addresses a user can send from, the thread id of the last sent
 * message, thread data for the threads in the current view, contacts. Basically
 * anything Gmail knows or can find out.
 *
 * @class ?
 */
/* Internal API Developers:
 * The API is structured in three parts: general environment data, special events
 * and event data, and traffic watcher data. General environment data is handled
 * by update and updateFromData and stored in _environmentData. It consists of
 * anything that is either constant, or doesn't need to be updated reqularly (yes
 * it's vague, but the other two parts make it easier to understand). Typically
 * it is accessed by calling something like getUserLabels or something. Special
 * events and event data consist of things like label name changes, sent e-mails,
 * and other like things. The way to tell if it falls in this category is to ask
 * whether the best way to convey that you have the information is by an event
 * and whether you're probably not going to save the information. This data is
 * always conveyed by event and rarely stored or cached. The final part, traffic
 * watcher data, consists of data that the traffic watcher grabs out of Gmail's
 * traffic and that is cached by glEnvironment. This includes things like thread
 * list data, conversation data that Gmail requests for improved performance, and
 * other things. The API for this can vary, but it is commonly accessed through
 * a general function like getThreadListData(). This way if there's a cache miss,
 * the function can still be ablt handle the data retrieval differently. Note that
 * just because the traffic watcher is being used to obtain the data doesn't mean
 * that the data can't be exposed through other parts of the API, like the action
 * detection.
 */
var glEnvironment = giBase.extend({
    // Properties {
    /**
     * The last environment data update date
     * @property {Date} ?
     */
    lastUpdate: null,

    /**
     * glGmail object
     * @property {glGmail} ?
     */
    _glGmail: null,

    /**
     * Stored environment data
     * @property {Object} ?
     */
    _environmentData: null,

    

    /**
    * The glThreadListCache instance. 
    * @property {glThreadListCache} ?
    */
    threadListCache: null,
    
    /** @property {Boolean} ? Refresh at the next request */
    _requestedRefresh: null,
	
	/**
	 * Data during a refresh; used to restore scrollTop after it shifts
	 * @property {Object} ? Format: {id: <int>, count: <int>, scrollTop: <int>}
	 */
	_requestedRefreshTimer: null,
    
    /** @property {giMultiToOne} ? */
    _multiToOne: null,
    
    /** @property {String} ? */
    lastSentThreadId: null,
	

    /**
     * Hash containing cached thread list data. The key is the URL that the request
     * was made to and the value is a hash containing the data (key "data"), and
     * a boolean flag (key "doDelete") that will trigger a delete of the data
     * the next time the view changes.
     * @property {Object} ?
     */
    //_threadListCache: null,	// TODO remove
    //}

    // Methods {
    /**
     * Connects the environment to the given glGmail object's page.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        this._glGmail = gmail;
        this._environmentData = {};
        //this._threadListCache = {};
		this._multiToOne = new giMultiToOne(5000);


        // Update from view data
        var pageViewData = gmail.page.getViewData();
        if( pageViewData ) {
            var viewData = glDataObject.createFromData(pageViewData, this._glGmail);
            this.updateFromData(viewData, true);
        }

        this._glGmail.trafficWatcher.addEventListener(glTrafficWatcher.SEND_MESSAGE_REQUEST_COMPLETED, this._handleSendMessageRequest, this);
		this._glGmail.actions.addEventListener(glActionManager.ENVIRONMENT_CHANGED, this._actionManagerEvent, this);
		this._glGmail.actions.addEventListener(glActionManager.ADD_LABEL.eventName, this._actionManagerEvent, this);
	
		this.threadListCache = new glThreadListCache(this._glGmail, viewData)

        this._connectToPageWrapper();
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    shutdown: function() {

        this._environmentData = null;

        this._glGmail = null;

        this._removeAllListeners();
    },

    // Environment Data Related Methods {
    /**
     * Performs a simple synchronous request to repopulate the environment data.
     * @function {public} ?
     */
    update: function() {
        var request = glRequest.createSimpleRequest(this._glGmail);

        request.run(false); // Run synchronously
        if(request.errors.length) {
            giLogger.error("Update request had errors:\n"+giLogger.dump(request.errors, true), 'glEnvironment');
        }

        this.updateFromData(request.data);
    },

    /**
     * Updates the environment data store with the given data.
     * If the data is different, it dispatches an ENVIRONMENT_CHANGED event.
     * @function ?
     * @param {glDataObject} data - A glDataObject object containing environment data
     * @param {Boolean} full - If true, always include complete analysis
     */
    updateFromData: function(data, full) {
        giLogger.log('Updating environment from data @ '+(new Date()), 'glEnvironment');

        this.lastUpdate = new Date();


        if(this.hasEventListener(glEnvironment.ENVIRONMENT_CHANGED) || full) {
            // Someone is listening for ENVIRONMENT_CHANGED, so we need to actually determine if the environment changed
            var changed = [];

            var compare = {
				userName: data.getUserName(),
                additionalAccounts: data.getAdditionalAccounts(),
                userLabels: data.getUserLabels(),
                internalLabels: data.getInternalLabels()
            };
            
            for(var set in compare) {
                if(compare[set] == null) continue;

                var currentData = this._environmentData[set];
                var newData = compare[set];

                // Compare currentData and newData - will probably need to be changed at some point to make it more generic, but it works for now
                var isNew = false;
                if(currentData && currentData.length == newData.length) {
                    for(var i = 0; i < newData.length; i++) {
                        for(var prop in newData[i]) {
                            if(currentData[i][prop] != newData[i][prop]) {
                                isNew = true;
                                break;
                            }
                        }
                        if(isNew) break;
                    }
                } else {
                    isNew = true;
                }
    
                if(isNew) {
                    changed.push(set);
                    this._environmentData[set] = compare[set];
                }
            }

            if(changed.length) {
                this.dispatchEvent({name: glEnvironment.ENVIRONMENT_CHANGED, changed: changed});
            }
        } else {
            // No one listening for ENVIRONMENT_CHANGED so don't bother doing calculations for it
			this._environmentData['userName'] = data.getUserName();
            this._environmentData['additionalAccounts'] = data.getAdditionalAccounts();
            this._environmentData['userLabels'] = data.getUserLabels();
            this._environmentData['internalLabels'] = data.getInternalLabels();
        }
		this.dispatchEvent({name: glEnvironment.ENVIRONMENT_DATA_RECEIVED});
    },

    /**
     * Returns the account for the current environment.
     * @function {String} ?
     * @return The e-mail address of the active account
     */
    getAccount: function() {
        if(typeof this._environmentData['account'] == 'undefined') {
            this._environmentData['account'] = this._glGmail.page.getGlobal(10);
        }
        return this._environmentData['account'];
    },
    
    /**
    * Return array of user-created labels
    * @function {} ?
    * @return Array of user-created labels
    */
    getUserLabels: function() {
    	return this._environmentData.userLabels;
    },
	
	/**
    * Return array of interal labels (inbox, etc.)
    * @function {} ?
    * @return Array of user-created labels
    */
    getInternalLabels: function() {
    	return this._environmentData.internalLabels;
    },
    
    /**
    * Return the user name
    * @function {String} ?
    * @return The user name
    */
    getUserName: function() {
    	return this._environmentData.userName;
    },
    

    /**
     * The IK value is some type of cookie value that Gmail requires in requests
     * to the server.
     * @function {String} ?
     * @return The IK value
     */
    getIKValue: function() {
        if(typeof this._environmentData['ikValue'] == 'undefined') {
            this._environmentData['ikValue'] = this._glGmail.page.getGlobal(9);
        }
        return this._environmentData['ikValue'];
    },

    /**
     * Returns the URL that requests should be made to. Used by {@link glRequest}
     * to build the request URL.
     * @function {String} ?
     * @return The request URL
     */
    getRequestURL: function() {
        if(typeof this._environmentData['requestURL'] == 'undefined') {
            this._environmentData['requestURL'] = this._glGmail.page.getGlobal(7);
        }
        return this._environmentData['requestURL'];
    },
    //}

    // glPageWrapper Related Methods {
    /**
     * Connects important event listeners to the page wrapper object.
     * @function ?
     */
    _connectToPageWrapper: function() {
        this._glGmail.page.addEventListener(glPageWrapper.VIEW_CHANGED, this._viewChanged, this);
    },

    /**
     * glPageWrapper view change callback
     * @function ?
     * @param {Event} evt - The {@link glPageWrapper.VIEW_CHANGED} event
     */
    _viewChanged: function(evt) {        
        if( this._requestedRefresh ) { // TODO Unsure about this; do we still want to refresh even after changing views?
        	// If can refresh UI, do so
        	if( this._glGmail.page.getActiveViewType()=="tl" ) {
        		this._requestRefresh();
        	}
        }
    },
    //}

    /**
    * Request a refresh at the earliest opportunity 
    * The Ajax requests - especially in the popup - do not cause Gmail to update; so we must ask it to otherwise users will doubt the integrity.
    * @function {} ?
    * @param {Boolean} immediate - If true, do right now. Otherwise defer.
    */
    requestRefresh: function(immediate) {
		if( immediate ) {
			this._multiToOne.remove("requestRefresh");
			this._requestRefresh();
		} else {
			this._multiToOne.process("requestRefresh", giUtil.bind(this, this._requestRefresh));
		}
    },
    
    /**
    * Actually attempt the refresh
    * For conversations, would it be possible to use location.hash to dip back to inbox and then back to the conversation? Would need lightbox as it can take a few secs - would need to detect change back to tl and then move back. It does work though for syncing labels. Could be useful after Create/Rename/Delete label event.
    * @function {} ?
    */
    _requestRefresh: function() {
		if( this._glGmail.page.getActiveViewType()=="tl" ) {
			this._requestedRefresh = false;
			var actionsHeader = this._glGmail.getCurrentViewObject().actionsHeader;
			var el = null;
			if( this._glGmail.doc.location.hash.indexOf("#inbox")>-1 ) {
				// Click Inbox link, why? Because in Multiple Inboxes, this is the only way it will refresh.
				var el = this._glGmail.getUIObject('navBar').getNavLink(glNavBar.INBOX);
			} else {
				var el = actionsHeader.getButton(glActionsHeader.REFRESH_BUTTON);
			}
			if( el ) {
				giUIObject.simulateMouseEvent(el, 'mousedown');
				giUIObject.simulateMouseEvent(el, 'mouseup');
				giUIObject.simulateMouseEvent(el, 'click');
			}
			this._requestedRefresh = false;
			
			// Start the timer that tests for the window scroll pos moving (as it will upon refresh)
			if( this._requestedRefreshTimer ) {
				clearTimeout(this._requestedRefreshTimer.id);
			}
			this._requestedRefreshTimer = {id: 0, count: 0, scrollTop: this._glGmail.canvasDocument.documentElement.scrollTop};
			var me = this;
			this._requestedRefreshTimer.id = setInterval( function(){me._requestRefreshInterval()}, 75 );
		} else {
			this._requestedRefresh = true;
		}
    },
	
	/** Notification the interval has fired after a refresh
	 * When spot the window has moved to 0, restore scrollTop
	 * @function {} ?
	 */
	_requestRefreshInterval: function() {
		// See if scroll position is now 0
		if( this._glGmail.canvasDocument.documentElement.scrollTop==0 ) {
			this._glGmail.canvasDocument.defaultView.scrollTo(0, this._requestedRefreshTimer.scrollTop);
			clearTimeout( this._requestedRefreshTimer.id )
			this._requestedRefreshTimer = null;
		} else if( this._requestedRefreshTimerCount.count++>=25 ) {
			// See if exceeded count (2 seconds):
			clearTimeout( this._requestedRefreshTimer.id )
			this._requestedRefreshTimer = null;
		}
	},

    // Traffic Watcher Related Methods {
    /**
     * Checks the internal thread list data cache to see if it has data matching
     * the request and returns it.
     * @function {Array} ?
     * @param {String} viewHash - The hash portion of the URL for the current view ('#search/Apple+computers' for example)
     * @param {int} startOffset - The start offset of the thread data
     * @param {int} endOffset - The end offset of the thread data
     * @return The threads that it found - empty array if none found
     */


    /**
     * Event handler for {@link glTrafficWatcher.SEND_MESSAGE_REQUEST_COMPLETED} events.
     * Dispatches the EMAIL_SENT event with the thread id of the sent e-mail if
     * it is in the response. Also updates the internal environment data from
     * the response.
     * @function ?
     * @param {Event} evt - The SEND_MESSAGE_REQUEST_COMPLETED event
     */
    _handleSendMessageRequest: function(evt) {
        // Is there a thread id for the sent e-mail?
        var actionResponse = evt.responseData.getActionResponseData();
        if(actionResponse && actionResponse.affectedThread) {
			this.lastSentThreadId = actionResponse.affectedThread;
            // Dispatch EMAIL_SENT event with the thread id
            this.dispatchEvent({name: glEnvironment.EMAIL_SENT, threadId: actionResponse.affectedThread});
        }

        // Update internal environment data from response
        this.updateFromData(evt.responseData);
    },
    
	
	

	/**
	 * Notification of an ActionManager event; see if it effects any environment stored data
	 * @function {} ?
	 */
	_actionManagerEvent: function(evt) {
				
		if( !evt.label ) return;
		
		var events = [];
		
		// See if a label change needs to be updated internally
		if( evt.action==glActionManager.CREATE_LABEL || evt.name==glActionManager.ADD_LABEL.eventName ) {
			var found = false;
			for( var i = 0; i < this._environmentData.userLabels.length; i++ ) {
				if( this._environmentData.userLabels[i].name==evt.label ) {
					var found = true;
					break;
				}
			}
			if( !found ) {
				this._addUserLabelInPlace({name:evt.label});
				events.push({name: glEnvironment.CREATE_LABEL, label:evt.label});
				events.push({name: glEnvironment.ENVIRONMENT_CHANGED});
			}
		}
		if( evt.action==glActionManager.RENAME_LABEL ) {
			var labelData = null;
			for( var i = 0; i < this._environmentData.userLabels.length; i++ ) {
				if( this._environmentData.userLabels[i].name==evt.label ) {
					labelData = this._environmentData.userLabels[i];
					labelData.name = evt.newLabel;
					// Remove it (and re-add sorted)
					this._environmentData.userLabels.splice(i, 1);
					break;
				}
			}
			if( labelData ) this._addUserLabelInPlace(labelData);
			events.push({name: glEnvironment.RENAME_LABEL, label:evt.label, newLabel: evt.newLabel});
			events.push({name: glEnvironment.ENVIRONMENT_CHANGED});
		}
		if( evt.action==glActionManager.DELETE_LABEL ) {
			for( var i = this._environmentData.userLabels.length -1; i>=0; i-- ) {
				if( this._environmentData.userLabels[i].name==evt.label ) {
					this._environmentData.userLabels.splice(i,1);
				}
			}
			events.push({name: glEnvironment.DELETE_LABEL, label:evt.label});
			events.push({name: glEnvironment.ENVIRONMENT_CHANGED});
		}
		
		if( events.length > 0 ) {
			try {
				for( var i = 0; i < events.length; i++ ) {
					this.dispatchEvent( events[i] );
				}
			} catch(e) {
				giLogger.warn(e, 'glEnvironment._actionManagerEvent');
			}
		}
		
		
	},
	
	/**
	 * Add the user label in the right place, as dictated by
	 * TODO - we do not have label 'count' data here, as we do if parsed from a server request. Can we extract it?
	 * @function {} ?
	 * @param {Object} newLabelData - The label data object {name: <label>}
	 */
	_addUserLabelInPlace: function(newLabelData) {
		var name = newLabelData.name.toLowerCase();
		var added = false;
		for( var i = 0; i < this._environmentData.userLabels.length; i++ ) {
			if(this._environmentData.userLabels[i].name.toLowerCase() > name) {
				// Insert at i
				this._environmentData.userLabels.splice(i, 0, newLabelData);
				added = true;
				break;
			}
		}
		if( !added ) {
			// Append
			this._environmentData.userLabels.push(newLabelData);
		}
	},
	
    //}
    //}
}, {
    // Static Properties {
    
	/**
     * Dispatched when a label is created
     * @event ?
     */
    CREATE_LABEL: 'create_label',
    
	/**
     * Dispatched when a label is renamed
     * @event ?
     */
    RENAME_LABEL: 'rename_label',
	
	/**
     * Dispatched when a label is deleted
     * @event ?
     */
    DELETE_LABEL: 'delete_label',
	
    /**
     * Dispatched when an e-mail is sent and includes the thread id of the sent e-mail
     * @event ?
     */
    EMAIL_SENT: 'email_sent',

    /**
     * Dispatched when an action is performed with the url and the post data
     * @event ?
     */
    ACTION_PERFORMED: 'action_performed',

    /**
     * Dispatched when environment data is changed in an update
     * @event ?
     */
    ENVIRONMENT_CHANGED: 'environment_changed',
	
	/**
	 * Dispatched when environment data is received, in updateFromData
	 * @event ?
	 */
	ENVIRONMENT_DATA_RECEIVED: 'environment_data_updated',
	
    //}
});
glEnvironment.implement(giEventDispatcher); // Add event dispatching support