/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The base object for a given browser tab with Gmail open. It coordinates all of
 * the gmail components and provides references to the document and windows for
 * the Gmail page.
 *
 * @class ?
 */
var glGmail = giBase.extend({
    // Properties {
    /**
     * The document object for the page
     * @property {Document} ?
     */
    doc: null,

    /**
     * A reference to the canvas iframe content document
     * @property {Document} ?
     */
    canvasDocument: null,

    /**
     * The window object for the page
     * @property {Window} ?
     */
    window: null,

    /**
     * The {@link glPageWrapper} object for the page
     * @property {glPageWrapper} ?
     */
    page: null,

	/**
     * The {@link glTrafficWatcher} instance. It will be null if it is not connected.
     * @property {glTrafficWatcher} ?
     */
    trafficWatcher: null,

    /**
     * The {@link glEnvironment} object for the page
     * @property {glEnvironment} ?
     */
    environment: null,

    /**
     * The {@link glActionManager} object for the page
     * @property {glActionManager} ?
     */
    actions: null,


    /**
     * A hash of all the UI objects
     * @property {Object} ?
     */
    _ui: null,
    //}

    /**
     * The constructor. Initializes all helper objects in preparation for page
     * connection. This allows plugins to hook event listeners in before anything
     * has happened.
     * @constructor ?
     * @param {Document} doc - The document element for the page to attach to
     */
    constructor: function(doc) {
        giLogger.log('glGmail object constructing', 'glGmail');

        // Initialize instance variables
        this.doc = doc;
        this.window = doc.defaultView;
        this._ui = {};

        // Set up page wrapper
        this.page = new glPageWrapper();
        this.page.addEventListener(glPageWrapper.UI_READY, this._baseUIReady, this);
		this.page.addEventListener(glPageWrapper.VIEW_CHANGED, this._baseViewChanged, this);

		
		// Set up traffic watcher:
		this.trafficWatcher = new glTrafficWatcher();
        this.trafficWatcher.connect(this);
		
		// Set up rest
        this.environment = new glEnvironment();
        this.actions = new glActionManager();
    },

    // Methods {
    /**
     * Connects the page wrapper to the page, check the page version, and then
     * continues on, connect the environment and action manager.
     * @function ?
     * @return void
     */
    connect: function() {
        giLogger.log('Connecting glGmail object to page', 'glGmail');

        // Connect page wrapper
        this.page.connect(this);

        // Check version and if it's not 2, throw an exception
        if(!/ui=2/.test(this.page.getGlobal(11))) { // Global 11 is the url for the javascript code for gmail
            giLogger.log('Global 11: '+this.page.getGlobal(11), 'glGmail');
            this.page.disconnect();
            throw new Error("glGmail only works with version 2 of GMail");
        }

        // Connect rest
		this.actions.connect(this);
        this.environment.connect(this);

        giLogger.log('Connection complete');
    },

    /**
     * Called on page unload, it cleans up everything
     * @function ?
     */
    unload: function() {
        for(var element in this._ui) {
            this._ui[element].disconnect();
            this._ui[element] = null;
        }

		this.trafficWatcher.shutdown();
        this.trafficWatcher = null;

        this.actions.shutdown();
        this.environment.shutdown();
        this.page.disconnect();

        this.canvasDocument = null;
        this.window = null;
        this.doc = null;
    },

    /**
     * Initializes a {@link glSearch} object and returns it
     * @function {glSearch} ?
     * @return The searcher object
     */
    getSearcher: function() {
        return new glSearch(this);
    },

    /**
     * Returns the given UI object after initializing it (if necessary).
     * For example, if you wanted a glActionsDropdown object, you would call
     * <code>getUIObject('actionsDropdown')</code>, dropping the gl and turning
     * the 'A' to an 'a'. If it has already been initialized/created already, it
     * returns the cached one (when possible).
     * @function {Object} ?
     * @param {String} type - The UI object to return
     * @return The initialized UI object
     */
    getUIObject: function(type) {
        giLogger.log('Getting '+type+'...', 'glGmail');

        var object = null;
        if(!this._ui[type]) {
            // Not in cache so create and cache (or return if not supposed to cache)

            // Add gl to the begining of the type and raise the first character up
            // mainView -> glMainView
            var theClass = 'gl'+type.substr(0,1).toUpperCase()+type.substr(1);

            giLogger.log('Creating object: "'+theClass+'"', 'glGmail');

            try {
                object = new window[theClass]();
            } catch(e) {
                // Use giLogger.error because it gives a stack trace
                giLogger.error("Failure instantiating "+theClass+" ui object:\n"+e.message, 'glGmail');
            }

            object.connect(this);

            // Is the ui object cachable - return if not
            if(object.isCacheable()) {
                // Cache the instantiated object
                this._ui[type] = object;
            }
        } else {
            // Get object from cache
            object = this._ui[type];
        }

        return object;
    },

    /**
     * Returns the UI object for the current view, initializing it if it hasn't
     * already been loaded.
     * @function {glUIView} ?
     * @return The view UI object
     */
    getCurrentViewObject: function() {
        if(!this.page.isUIReady()) return null;

        // In the cache, so return
        if(typeof this._ui['currentViewObject'] != 'undefined') {
            return this._ui['currentViewObject'];
        }

        // Get the UI object type for a call to getUIObject
        var uiObject = '';
        switch(this.page.getActiveViewType()) {
            case 'tl': uiObject = 'threadListView'; break;
            case 'cv': uiObject = 'conversationView'; break;
            case 'co': uiObject = 'composeView'; break;
            case 'ct': uiObject = 'contactsView'; break;
            case 's': uiObject = 'settingsView'; break;
        }

        if(uiObject == '') {
            return null;
        } else {
            // Use getUIObject so extending classes can override the view objects without overriding this method
            this._ui['currentViewObject'] = this.getUIObject(uiObject);
        }

        return this._ui['currentViewObject'];
    },

    /**
     * The UI_READY callback for the glGmail class. Currently just populates
     * the canvasDocument property.
     * @function ?
     * @param {Event} evt - The glPageWrapper.UI_READY event
     */
    _baseUIReady: function(evt) {
        giLogger.log('Base UI Ready, so retrieve canvas document', 'glGmail');
        this.canvasDocument = this.doc.getElementById("canvas_frame").contentDocument;

		// Attach offline detection:
		this.offline = new glOffline(this);

		this.lightbox = this.getUIObject('lightbox');
		giLogger.log('Base UI Ready... retrieved canvas document', 'glGmail');
		
		giLogger.outputWarningStack("glGmail._baseUIReady", this);
		
		// Add CSS
        this.page.addCSS("chrome://gtdinbox/skin/glGmail.css");
		this.page.addCSS("chrome://gtdinbox/skin/reset.css");
		
		
		// If not Installed, install now
		// Install routine will handle establishing a server account and showing welcome screen.
		// It will also disable GTDInbox if the user does not wish to use it.
		// TODO 
		
		// Otherwise, Check if last password attempted on a server sync was wrong.
		// If so, show the Change Password screen; which takes care of updating the password, getting the latest settings, merging them, and then asking the user to refresh Gmail.
		// TODO
	
    },

    /**
     * glPageWrapper view change callback. Currently just disconects and invalidates
     * the currentViewObject cache if it is populated.
     * @function ?
     * @param {Event} evt - The glPageWrapper.VIEW_CHANGED event
     */
    _baseViewChanged: function(evt) {
        // Disconnect and invalidate currentViewObject cache if populated
	
		
		giLogger.log("base view changed", "glGmail");
	
        if(typeof this._ui['currentViewObject'] != 'undefined') {
            if(this._ui['currentViewObject']) this._ui['currentViewObject'].disconnect();
            delete this._ui['currentViewObject'];
        }
	
    },
    //}
});