/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Monitor status of offline mode
 *
 * @class ?
 */

var glOffline = giBase.extend({
    // Properties {
    
    /**
     * glGmail object
     * @property {glGmail} ?
     */
    _gmail: null,
    
    /**
     * @property {private giEventsContainer}
     */
    _eventsContainer: null,
    
    /**
     * True if in unstable mode
     * @property {private boolean}
     */
    _isUnstable: false,
    
    /**
     * The offline icon element
     * @property {private Element}
     */
    _elIcon: null,

    //}

    // Methods {
    /**
     * Instantiate and get starting state
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    constructor: function(gmail) {
        giLogger.log('constructed glOffline', 'glOffline');
        this._gmail = gmail;
        this._eventsContainer = new giEventsContainer(this);
        
        // Find out our mode
        this._testState();
        
        // Listen for changes:
        var elHeader = this._gmail.canvasDocument.getElementById('guser');
        if( elHeader ) {
            this._eventsContainer.observe(elHeader, 'DOMAttrModified', this._headerChange, false);
        } else {
            throw new Error('Could not find Offline header');
        }
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    shutdown: function() {
        
        this._eventsContainer.reset();
        this._eventsContainer = null;
        
        this._elIcon = null;
        this._glGmail = null;
        this._removeAllListeners();
    },

    /**
     * Return true if in unstable mode
     * @function {} ?
     * @return True if Gmail in unstable mode
     */
	isUnstable: function() {
        return this._isUnstable;
    },
    
    
    /**
     * Notification that the header changed (including the Offline icon)
     * See if it affects us and update status
     * @function ?
     */
    _headerChange: function(evt) {
        try {
            if( evt.target.nodeName=="IMG" && evt.target.className.indexOf("bo")==0 && evt.target.className!=this._lastClassname) {
                giLogger.log('_headerChange event', 'glOffline');
                // It's the Offline icon
                this._testState();
                // Ensure we do not show it again (when rotating, it constantly fires)
                this._lastClassname = evt.target.className;
            }
        } catch(e) {
            giLogger.warn(e, 'glOffline._headerChange', this._gmail.fb);
        }
    },

    /**
     * Attempt to retrieve the Offline icon el, and add click handler
     * @function {} ?
     */
    _getIconEl: function() {
        if( !this._elIcon ) {
            var doc = this._gmail.canvasDocument;
            var elImg = doc.evaluate("//span/span/img[starts-with(@class,'bo')]", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
            this._elIcon = elImg? elImg : null;
        }
        return this._elIcon;
    },

    /**
     * Use the Offline icon CSS to detect what mode we're in.
     * Dispatch event if detect a change
     * @function {} ?
     */
    _testState: function() {
        
        var elIcon = this._getIconEl();
        
        
        var isUnstable = false;
        if( elIcon && elIcon.className ) {
            // Use the CSS to find it's mode:
            if( /bo (bw|bx)/.test(elIcon.className) ) {
                isUnstable = false;
            } else {
                isUnstable = true;
            }
        } else {
            // Probably not active
            isUnstable = false;
        }
        giLogger.log('testState, isUnstable: ' + isUnstable, 'glOffline._testState');
        
        if( isUnstable!=this._isUnstable ) {
            // Change, dispatch event
            this._isUnstable = isUnstable;
            if( isUnstable ) {
                this.dispatchEvent({name: glOffline.GO_UNSTABLE});
            } else {
                this.dispatchEvent({name: glOffline.GO_ONLINE});
            }
        }
        
    },

    //}
}, {
    // Static Properties {
    
	/**
     * Dispatched when detect a move online
     * @event ?
     */
    GO_ONLINE: 'go_online',
    
    /**
     * Dispatched when detect a move to unstable
     * @event ?
     */
    GO_UNSTABLE: 'go_unstable',
    
    //}
});
glOffline.implement(giEventDispatcher); // Add event dispatching support