/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Contains the functionality to securely interact with the underlying gmail page
 * by means of a hidden added element and events.
 *
 * @class ?
 */
var glPageWrapper = giBase.extend({
    // Properties {
    /**
     * The glGmail object for the page
     * @property {glGmail} ?
     */
    _glGmail: null,

    /**
     * Whether or not the UI is ready
     * @property {Boolean} ?
     */
    _uiIsReady: false,
    
    /**
     * As we cannot guarentee that a ui_ready event actually means it is ready; we manually queue events.
     * Also guarentees that ui_ready is fired before view_changed.
     * @property {Object} ? Format: {uiReady: <event>, viewChanged: [event], timer: <int>}
     */
    _loadEventQueue: null,
 
    /**
     * The injected element that is used for communication
     * @property {Element} ?
     */
    _injectedElem: null,

    /**
     * The sandbox object for the page
     * @property {Sandbox} ?
     */
    _sandbox: null,
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        giLogger.log('Page wrapper connected', 'glPageWrapper');
        this._glGmail = gmail;
        this._loadEventQueue = {};

        // Load injected code
        try {
            this._injectCode(giUtil.readFile('chrome://gtdinbox/content/js/gmail/injected-code.js'));
        } catch (e) {
            giLogger.error('Error loading injected-code.js file:'+"\n"+e, 'glPageWrapper');
        }

        // Get injected element and attach event listeners
        var me = this;
        this._injectedElem = this._glGmail.doc.getElementById('_glGmailInjected');
        this._injectedListener = function(e) { me._eventReceived(e); }; // So we can remove it later
        this._injectedElem.addEventListener('glGmailEventDispatched', this._injectedListener, false, true);
    },

    /**
     * Disconnects the page wrapper object from the page, removing event listeners
     * and the injected element.
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Page wrapper disconnected', 'glPageWrapper');
        
        if( this._loadEventQueue && this._loadEventQueue.timer ) clearTimeout( this._loadEventQueue.timer );
        
        this._injectedElem.removeEventListener('glGmailEventDispatched', this._injectedListener, true);
        this._injectedElem.parentNode.removeChild(this._injectedElem);
        this._injectedElem = null;
        this._injectedListener = null;
        this._glGmail = null;
    },

    /**
     * Returns the object at the given offset in the page GLOBALS object.
     * @function {Object} ?
     * @param {int} offset - The offset in the GLOBALS object
     * @return The value of the offset in GLOBALS
     */
    getGlobal: function(offset) {
        return this._runInPage('getGlobal', offset);
    },

    /**
     * Returns the VIEW_DATA object for the page
     * @function {Object} ?
     * @return The VIEW_DATA object
     */
    getViewData: function() {
        return this._runInPage('getViewData');
    },

    /**
     * Returns the current view type
     * @function {String} ?
     * @return The current view type as given by the gmonkey object ("tl" =>
     * theadlist, "cv" => conversation, "co" => compose, "ct" => contacts, or
     * "s" => settings)
     */
    getActiveViewType: function() {
        return this._runInPage('getActiveViewType');
    },
    
    /**
     * Retrieve the navPane element using the gmonkey element
     * @function {Element} ?
     * @return the nav pane element
     */
    getNavPaneElement: function() {
        
        // Mark the element
        this._runInPage('markNavPaneElement', "gtdi-nav-pane-element");
        
        var doc = this._glGmail.canvasDocument;
        var el = doc.evaluate("//*[contains(@class,'gtdi-nav-pane-element')]", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        
        if( el ) {
            giUIObject.removeClassName(el, 'gtdi-nav-pane-element');
        }
        return el;
    },
    
    /**
     * Returns a new nav module
     * @function {Object}
     * @param {optional int} id - If present, do not create navBox, but just retrieve it
     * @return A collection of elements: {id: <int>, navBox:{root:, title:, content:}}
     */
    addNavModule: function(id) {
        // Create unique id that will be used to retrieve it:
        if( !id ) {
            id = (new Date()).valueOf();
         
            // Create the box in the page:
            this._runInPage('addNavModule', "<span id='"+id+"-title'>&nbsp;</span>", "<span id='"+id+"-content'>&nbsp;</span>", null, id+'-root');
        }
        
        /*
         TODO
         The offline problem. #OFF1
         In offline mode, Gmail seems to defer loading the UI (but does not inform gmonkey).
         So, baseUIReady fires; but actually the new navbox is not connected to doc.body, so the xpath fails.
         
         Current solution: return the ID and enable this function to simply retrieve by ID, so boxBox can call again after a setTimeout
         Alternates: Figure out what is going on with the document; and get to the bottom of the real problem
            Or, break security rules and directly retrieve the navBox element that injected-code creates; rather than retrieving via xpath
        */
        
        // Now retrieve it from the environment:
        var doc = this._glGmail.canvasDocument;
        var navBox = {};
        try {
            navBox.root = doc.evaluate("//*[contains(@class,'"+id+"-root')]", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
            if( !navBox.root ) return {id: id}; // #OFF1
            navBox.title = doc.evaluate("//SPAN[@id='"+id+"-title']", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue.parentNode;
            navBox.content = doc.evaluate("//SPAN[@id='"+id+"-content']", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue.parentNode;
        } catch(e) {
            giLogger.warn(e, 'glPageWrapper.addNavModule');
            return {id: id};
        }
        
        // Clean up:
        navBox.title.innerHTML = "";
        navBox.content.innerHTML = "";
        navBox.root.className = navBox.root.className.replace(id+'-root', '');
        
        return {id: id, navBox: navBox};
    },

    /**
     * Injects the given CSS file into the current page
     * @function ?
     * @param {String} cssFile - The full path to the CSS file to inject into the page
     */
    addCSS: function(cssFile) {
        if(!this._uiIsReady) return;

        var doc = this._glGmail.canvasDocument;
        var elLink = doc.createElement("link");
        elLink.type = "text/css";
        elLink.href = cssFile;
        elLink.rel = "stylesheet";
        doc.getElementsByTagName("head")[0].appendChild(elLink);
    },

    /**
     * Returns the root node of the view. The class that it is looking for is
     * added by the injected code on view change.
     * @function {Element} ?
     * @return The root node of the view
     */
    getViewRoot: function() {
        var elems = this._glGmail.canvasDocument.getElementsByClassName('glViewRoot');
        if(!elems.length) {
            return null;
        } else {
            return elems[0];
        }
    },

    /**
     * Returns whether or not the UI is ready and all of the css and javascript
     * have loaded.
     * @function {Boolean} ?
     * @return Whether the UI is ready
     */
    isUIReady: function() {
        return this._uiIsReady;
    },

    /**
     * Runs the given function in the page using a glGmailRunFunction event and
     * returns the return value of the function.
     * @function {Object} ?
     * @param {String} func - The function to call
     * @param {Array} ...rest - The arguments to the function call
     * @return The value of the evaluated code
     */
    _runInPage: function(func) {
        giLogger.log('Running function "'+func+'"', 'glPageWrapper');

        // Set attribute for code
        this._injectedElem.setAttribute('func', func);
        this._injectedElem.setAttribute('params', arguments.length-1);
        for(var i = 1; i < arguments.length; i++) {
            this._injectedElem.setAttribute('param'+i, arguments[i]);
        }

        // Dispatch event
        var evt = this._glGmail.doc.createEvent('Events');
        evt.initEvent('glGmailRunFunction', true, false);
        this._injectedElem.dispatchEvent(evt);

        // Extract return value
        var returnValue = null;
        if(this._injectedElem.hasAttribute('return')) {
            // Return value is wrapped in object and placed in data property because native JSON can only handle objects and arrays - not bare strings and numbers
            var returnValueString = this._injectedElem.getAttribute('return');
            var json = Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON);
            returnValue = json.decode(returnValueString).data;
        }

        // Clean up attributes
        this._injectedElem.removeAttribute('func');
        this._injectedElem.removeAttribute('return');
        this._injectedElem.removeAttribute('params');
        for(var i = 1; i < arguments.length; i++) {
            this._injectedElem.removeAttribute('param'+i);
        }

        return returnValue;
    },

    /**
     * Callback for the glGmailEventDispatched event which is dispatched by the
     * injected code. The event name is turned into a real event and re-dispatched.
     * @function ?
     * @param {Event} evt - The glGmailEventDispatched event
     */
    _eventReceived: function(evt) {
        try {
            // Get event name
            var eventName = this._injectedElem.getAttribute('event');
    
            giLogger.log('Dispatching event "'+eventName+'"', 'glPageWrapper');
            var event = {name: eventName};
            switch(eventName) {
                case 'ui_ready':
                    if( !this._uiIsReady ) {
                        this._testUIIsReady(event);
                        return; // Processing moves elsewhere
                    }
                case 'view_changed':
                    if( !this._uiIsReady ) {
                        // Queue it up
                        if( !this._loadEventQueue.viewChanged ) this._loadEventQueue.viewChanged = [];
                        this._loadEventQueue.viewChanged.push(event);
                        return;
                    }
                    event.view = this.getActiveViewType();
                    break;
            }
            this.dispatchEvent(event);
        } catch(e) {
            giLogger.warn(e, 'glPageWrapper._eventReceived');
        }
    },
    
    /**
     * The Gmail offline mode delays loading of the HTML, which breaks it in multiple ways (navBox cannot be found, compose links, etc.)
     * The safest approach is to wait until we are confident the UI can be found via xpath statements, then proceed
     * @function {boolean} ?
     * @param {optional Event} evt - If present, store it under uiReady
     * @param {optional int} count - Used as a failsafe incase it can never load
     * @return True if ready
     */
    _testUIIsReady: function(evt, count) {
        
        if( !count ) {
            // Invoked by an event (rather than a setTimeout callback)
            count = 1;
            // Clear out existing setTimeout if it exists
            if( this._loadEventQueue && this._loadEventQueue.timer ) clearTimeout( this._loadEventQueue.timer );
        }
        
        if( this._glGmail.doc.getElementById('canvas_frame').contentDocument.body.innerHTML.indexOf('cf zt')>-1 ) { // TODO Give this multiple tests; where any 1 is appropriate, to protect against HTML changes
            // Loaded
            this._uiIsReady = true;
            // Execute queued up events:
            if( evt ) {
                // UI_READY event
                this.dispatchEvent(evt);
            } else {
                if( this._loadEventQueue.uiReady ) this.dispatchEvent(this._loadEventQueue.uiReady);
            }
            if( this._loadEventQueue.viewChanged ) {
                for( var i = 0; i < this._loadEventQueue.viewChanged.length; i++ ) {
                    this.dispatchEvent(this._loadEventQueue.viewChanged[i]);
                }
            }
            this._loadEventQueue = null;
        } else {
            // Store event
            if( evt ) this._loadEventQueue.uiReady = evt;
            // Try again
            if( count < 20 ) {
                var me = this;
                this._loadEventQueue.timer = setTimeout( function(){me._testUIIsReady(null, count++)}, 500 );
            } else {
                giLogger.warn('Could not load UI', 'glPageWrapper._testUIIsReady');
            }
        }
    },

    /**
     * Injects the given code into the page. Can be made public if a plugin
     * requests it, but it shouldn't be necessary in most cases.
     * @function ?
     * @param {String} code - The code to inject into the page
     */
    _injectCode: function(code) {
        
        // Create sandbox if not defined
        if(!this._sandbox) {
            this._sandbox = new Components.utils.Sandbox(this._glGmail.window);
            this._sandbox.window = this._glGmail.window;
            this._sandbox.document = this._glGmail.doc;
            this._sandbox.unsafeWindow = this._glGmail.window.wrappedJSObject;
            this._sandbox.log = giLogger.log;
            this._sandbox.__proto__ = this._glGmail.window;
        }

        // Wrap code in window scope
        var scriptToEval = "with (window?window:null) { " + code + " \n};";
        try {
            //giLogger.log("Injecting code:\n"+scriptToEval, 'glPageWrapper');
            Components.utils.evalInSandbox(scriptToEval, this._sandbox);
        } catch(e) {
            giLogger.error('Error injecting code:'+"\n"+e, 'glPageWrapper');
        }
    }
    //}
}, {
    // Static Properties {
    /**
     * Dispatched when the page is loaded and ready to interact with
     * @event ?
     */
    UI_READY: 'ui_ready',

    /**
     * Dispatched when gmonkey notifies us of a view change
     * @event ?
     */
    VIEW_CHANGED: 'view_changed',
    //}
});
glPageWrapper.implement(giEventDispatcher); // Add event dispatching support