/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * glRequest is an XMLHttpRequest wrapper for queries to Gmail servers. It is
 * used in making specialized requests to gmail and parsing the resulting data.
 * It can be used when updating label counts, getting available accounts, searching,
 * and getting other information. Responses are wrapped in a {@link glDataObject}
 * object for easy data access.
 *
 * @class ?
 */
var glRequest = giBase.extend({
    // Properties {
    /**
     * The glDataObject object containing the data received from the request
     * @property {glDataObject} ?
     */
    data: null,

    /**
     * A list of errors
     * @property {String[]} ?
     */
    errors: null,

    /**
     * The URL that the request will use
     * @property {String} ?
     */
    url: null,

    /**
    * Optional. Context data passed by caller into constructors; is returned in callbacks.
    * @property {Object} ?
    */
    _contextData: null,

    /**
     * The glGmail object for the page
     * @property {glGmail} ?
     */
    _glGmail: null,

    /**
     * The post data in a string
     * @property {String} ?
     */
    _postString: null,

    /**
     * The XMLHttpRequest object
     * @property {XMLHttpRequest} ?
     */
    _xmlRequest: null,
    //}

    /**
     * The constructor.
     * If one of the properties in params or postData is set to an array, that
     * property is added multiple times - once for each value in the array.
     * For the 'at' property, passing in a value of true will cause it to automatically
     * be calculated and added to the url params. The 'rt' property is optional
     * but if not set, it defaults to 'j' (which requires the least overhead to
     * process).
     * @constructor ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {Object} urlParams - The URL params.
     * @... {optional Boolean} at - Pass in as true to have the GMAIL_AT cookie data included in the URL
     * @param {optional Object} postData - The post data
     * @param {optional Object} contextData - Context data passed by caller into constructors; is returned in callbacks.
     */
    constructor: function(gmail, urlParams, postData, contextData) {
        if(typeof gmail == "undefined") giLogger.error('gmail is undefined', 'glRequest');
        if(typeof urlParams == "undefined") giLogger.error('urlParams is undefined', 'glRequest');
        this._glGmail = gmail;
        this._contextData = contextData;

        // Generate URL including url parameters {
        var ik = this._glGmail.environment.getIKValue(); // Some cookie like value needed in requests
        var protocol = this._glGmail.doc.location.protocol;
        var url = protocol + '//mail.google.com' + this._glGmail.environment.getRequestURL();
        url += "/?ui=2&ik=" + ik;
		
        // Get the GMAIL_AT data and include it if needed
        if(urlParams['at'] === true) {
			var cookie = this._glGmail.doc.cookie;
			var m = cookie.match(/GMAIL_AT\=([\d\w]+?);/i);
			if( m && m.length > 0 ) {
                urlParams['at'] = m[1];
            } else {
                delete urlParams['at'];
            }
        }

        // Default to a response type of js
        if(typeof urlParams['rt'] == 'undefined') urlParams['rt'] = 'j';

        urlParams.andySentIt = '1';

        url += glRequest._convertToRequestString(urlParams);
        this.url = url;
        //}

        // Generate POST string {
        if(typeof postData != "undefined") {
            this._postString = glRequest._convertToRequestString(postData);
        }
        
        //}
    },

    // Methods {
    /**
     * Runs the request either synchronously or asynchronously
     * @function ?
     * @param {optional Boolean} async - Whether or not to make the call asynchronously, defaults to true
     */
    run: function(async) {
        if(async == undefined) async = true;
        this._reset();

        this._xmlRequest = new XMLHttpRequest();
		var me = this;
        this._xmlRequest.onload = function() {me._requestCallback()};

        if(this._postString == null) {
            giLogger.log('Making '+(async?'asynchronous':'synchronous')+' GET XMLHttpRequest to "'+this.url+'"', 'glRequest');
            this._xmlRequest.open("GET", this.url, async);
            this._xmlRequest.send(null);
        } else {
            giLogger.log('Making '+(async?'asynchronous':'synchronous')+' POST XMLHttpRequest to "'+this.url+'"'+"\n"+'Post data: "'+this._postString+'"', 'glRequest');
            this._xmlRequest.open("POST", this.url, async);
            this._xmlRequest.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
            this._xmlRequest.setRequestHeader("Content-length", this._postString.length); // Gmail uses this header, so I'm setting it too
            this._xmlRequest.setRequestHeader("Connection", "close"); // Gmail uses this header, so I'm setting it too
            this._xmlRequest.send(this._postString);
        }
        
    },

    /**
     * The request callback from the XMLHttpRequest. Dispatches a
     * glRequest.COMPLETED event after parsing the data (check the errors property
     * to see if there were problems).
     * @function ?
     */
    _requestCallback: function() {
        //giLogger.log("Response:\n"+this._xmlRequest.responseText, 'glRequest');

        // Process response data
        try {
						
            var responseType = /[?&]rt=([^&]+)/.exec(this.url)[1];
            switch(responseType) {
                case 'j':
                    this.data = glDataObject.createFromJSResponse(this._xmlRequest.responseText, this.url, this._glGmail);
                    break;
                case 'h':
                    this.data = glDataObject.createFromHTMLResponse(this._xmlRequest.responseText, this.url, this._glGmail);
                    break;
                default:
                    throw new Error('Invalid response type: "'+responseType+'"');
            }
            this._xmlRequest = null;
        } catch(e) {
			giLogger.warn('Error in parse: '+this.url+'\n'+e, 'glRequest._requestCallback', this._glGmail);
            // Problem processing response to add to the errors array
            this.errors.push(e.toString());
        }

        // Notify listeners that the request is completed
		try {
			this.dispatchEvent({name: glRequest.COMPLETED, contextData: this._contextData});
		} catch(e) {
			giLogger.warn(e, 'glRequest._requestCallback COMPLETED', this._glGmail);
		}
    },

    /**
     * Cleans up the remnants of the previous run, preserving the URL and POST data
     * @function ?
     */
    _reset: function() {
		giLogger.log('Reseting the glRequest object', 'glRequest');
		if(this._xmlRequest) this._xmlRequest.abort();
		this._xmlRequest = null;
		
		this.data = null;
		this.errors = [];
    }
    //}
},{
    // Static Properties {
    /**
     * Dispatched when the request is completed
     * @event ?
     */
    COMPLETED: 'completed',
    //}

    // Static Methods {
    /**
     * Creates and returns a simple request that does not perform any searches (kind of).
     * Used by glEnvironment to update its cache.
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @return The created request object
     */
    createSimpleRequest: function(gmail) {
        return new glRequest(gmail, {view:'tl',search:'inbox',num:0});
    },

    /**
     * Creates and returns a request that adds the given label to the given
     * thread(s).
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {String} label - The label to add
     * @param {mixed} threads - The thread or threads to add the label to
     * @return The created request object
     */
    createAddLabelRequest: function(gmail, label, threads) {
        return new glRequest(
            gmail,
            { act: 'ac_'+label, at: true, search: 'inbox', view: 'up'},
            {t: threads}
        );
    },

    /**
     * Creates and returns a request that removes the given label from the given
     * thread(s).
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {String} label - The label to remove
     * @param {mixed} threads - The thread or threads to remove the label from
     * @return The created request object
     */
    createRemoveLabelRequest: function(gmail, label, threads) {
		return new glRequest(
			gmail,
			{ act: 'rc_'+label, at: true, search: 'inbox', view: 'up'},
			{t: threads}
		);
    },

    /**
     * Creates and returns a request that adds a star to the given thread(s).
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {mixed} threads - The thread or threads to add the star to
     * @return The created request object
     */
    createAddStarRequest: function(gmail, threads) {
        return new glRequest(
            gmail,
            { act: 'st', at: true, search: 'inbox', view: 'up'},
            {t: threads}
        );
    },

    /**
     * Creates and returns a request that removes a star from the given thread(s).
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {mixed} threads - The thread or threads to remove the star from
     * @return The created request object
     */
    createRemoveStarRequest: function(gmail, threads) {
        return new glRequest(
            gmail,
            { act: 'xst', at: true, search: 'inbox', view: 'up'},
            {t: threads}
        );
    },

    /**
     * Creates and returns a request that deletes the given thread(s)
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {mixed} threads - The thread or threads to delete
     * @return The created request object
     */
    createDeleteThreadRequest: function(gmail, threads) {
        return new glRequest(
            gmail,
            { act: 'tr', at: true, search: 'inbox', view: 'up' },
            { t: threads }
        );
    },

    /**
     * Creates and returns a request that spams the given thread(s)
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {mixed} threads - The thread or threads to report spam
     * @return The created request object
     */
    createSpamThreadRequest: function(gmail, threads) {
        return new glRequest(
            gmail,
            { act: 'sp', at: true, search: 'inbox', view: 'up' },
            { t: threads }
        );
    },

    /**
     * Creates and returns a request that unspams the given thread(s)
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {mixed} threads - The thread or threads to report spam
     * @return The created request object
     */
    createNotSpamThreadRequest: function(gmail, threads) {
        return new glRequest(
            gmail,
            { act: 'us', at: true, search: 'inbox', view: 'up' },
            { t: threads }
        );
    },
    
    /**
     * Creates and returns a request that marks the given thread(s) read
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {mixed} threads - The thread or threads to mark read
     * @return The created request object
     */
    createMarkThreadRequest: function(gmail, threads) {
        return new glRequest(
            gmail,
            { act: 'rd', at: true, search: 'inbox', view: 'up' },
            { t: threads }
        );
    },    
    
	/**
     * Creates and returns a request that creates a label
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {String} label - The label to create
     * @return The created request object
     */
    createCreateLabelRequest: function(gmail, label) {
		// TODO Why does TrafficWatcher refuse to pick this up? #TWLF1
        return new glRequest(
            gmail,
            { act: 'cc_'+label, at: true, view: 'up', rt: 'h' }
        );
    },
	
	/**
     * Creates and returns a request that renames a label
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {String} oldLabel - The label to rename
     * @param {String} label - The label to rename to
     * @return The created request object
     */
    createRenameLabelRequest: function(gmail, oldLabel, label) {
		// TODO Why does TrafficWatcher refuse to pick this up? #TWLF1
        return new glRequest(
            gmail,
            { act: 'nc_'+oldLabel+'^'+label, at: true, view: 'up', rt: 'h' }
        );
    },
	
	/**
     * Creates and returns a request that delete a label
     * @function {static glRequest} ?
     * @param {glGmail} gmail - The glGmail object for the page
     * @param {String} label - The label 
     * @return The created request object
     */
    createDeleteLabelRequest: function(gmail, label) {
		// TODO Why does TrafficWatcher refuse to pick this up? #TWLF1
        return new glRequest(
            gmail,
            { act: 'dc_'+label, at: true, view: 'up', rt: 'h' }
        );
    },

    /**
     * Utility function for turning an object into a request data string
     * @function {static String} ?
     * @param {Object} data - Request data - POST or url params
     * @return The request string
     */
    _convertToRequestString: function(data) {
        var str = "";
        for(var prop in data) {
            if(data[prop].constructor.name == 'Array') {
                for(var i = 0; i < data[prop].length; i++) {
                    str += "&"+encodeURIComponent(prop)+"="+encodeURIComponent(data[prop][i]);
                }
            } else {
                str += "&"+encodeURIComponent(prop)+"="+encodeURIComponent(data[prop]);
            }
        }
        return str;
    }
    //}
});
glRequest.implement(giEventDispatcher); // Add event dispatching support
