/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Provides an abstraction layer to GMail searching, with the ability to
 * build up complex search conditions and run the search in a variety of ways.
 *
 * @class ?
 */
var glSearch = giBase.extend({
    // Properties {
    /**
     * The glSearchConditions object for the query
     * @property {glSearchConditions} ?
     */
    conditions: null,

    /**
     * glGmail object
     * @property {glGmail} ?
     */
    _glGmail: null,

    /**
     * The glRequest object
     * @property {glRequest} ?
     */
    _request: null,
    
    /**
     * The type of search in the query. Defaults to 'search'
     * @property {String}
     */
    _searchType: null,
    
    //}

    /**
     * The constructor.
     * Creates a search object and initializes the conditions property.
     * @constructor ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    constructor: function(gmail) {
        this._glGmail = gmail;
        this.setSearchType();
        this.conditions = new glSearchConditions();
    },

    // Methods {
    /**
     * Adds the given search string to the query. Alias for {@link addCondition}.
     * @function {glSearch} ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - The condition to add the query
     * @paramset Input 2
     * @param {String} condition - The condition to add the query
     */
    find: function(condition) {
        return this.addCondition(condition);
    },

    /**
     * Add the given condition to the query
     * @function {glSearch} ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - The condition to add the query
     * @paramset Input 2
     * @param {String} condition - The condition to add the query
     * @return The object, so that it can be chained
     */
    addCondition: function(condition) {
        this.conditions.addCondition(condition);
        return this;
    },

    /**
     * Set the search type.
     * Typical values include 'search' and 'sent'. With 'sent', there is no query.
     * @function {} ?
     * @param {optional String} t - The type
     */
    setSearchType: function(t) {
        if( typeof t=="undefined" ) {
            t = 'search';
        }
        this._searchType = t;
    },

    /**
     * Runs the query, putting it in the searchbox on the page and submitting it
     * @function ?
     */
    run: function() {
        var searchBox = this._glGmail.getUIObject('searchBox');
        searchBox.setSearch(this.getQuery());
        searchBox.submit();
    },

    /**
     * Runs the query in the background
     * @function ?
     * @param {int} start - The start offset (defaults to 0)
     * @param {int} num - The number of rows to return (defaults to 25)
     * @param {optional Object} contextData - Data to be returned with the callback.
     */
    runBackground: function(start, num, contextData) {
        if(typeof start != 'number') start = 0;
        if(typeof num != 'number') num = 25;
        this._reset();

        if( this._searchType=='search' ) {
            this._request = new glRequest(this._glGmail, { start: start, num: num, search: 'query', view: 'tl', q: this.getQuery() }, null, contextData);
        } else {
            this._request = new glRequest(this._glGmail, { start: start, num: num, search: this._searchType, view: 'tl' }, null, contextData);
        }
        this._request.addEventListener(glRequest.COMPLETED, this._backgroundSearchCallback, this);
        this._request.run();
    },

    /**
     * Background search callback
     * @function ?
     */
    _backgroundSearchCallback: function(event) {
        giLogger.log('Search callback called', 'glSearch');
        this.dispatchEvent({name: glSearch.COMPLETED, contextData:event.contextData});
    },
    
    /**
     * Runs the query in the background, but sets the number of results returned
     * to 0 to speed up the request and reduce load on Gmail servers (hopefully).
     * @function ?
     */
    runCount: function() {
        this.runBackground(0, 0);
    },

    /**
     * Returns the number of results returned by the search. This is not the same
     * as the number of actual threads returned that we have data for, which can
     * be calculated by calling <code>getResults().length</code>.
     * @function {int} ?
     * @return The number of results in the search
     */
    getNumberResults: function() {
        if(!this._request) giLogger.error('this._request not defined', 'glSearch');

        if( this._request.data ) {
            var queryData = this._request.data.getQueryData();
            return queryData.results;
        } else {
            return 0;
        }
    },

    /**
     * Returns an array of the threads returned
     * @function {glThread[]} ?
     * @return The thread data from the search results
     */
    getResults: function() {
        if(!this._request) giLogger.error('this._request not defined', 'glSearch');

        if( this._request.data ) {
            var threads = this._request.data.getThreads();
            return threads;
        } else {
            return [];
        }
    },

    /**
     * Returns the query string
     * @function {String} ?
     * @return The query string
     */
    getQuery: function() {
        return this.conditions.toString();
    },

    /**
     * Returns whether or not the given thread matches the search
     * @function {Boolean} ?
     * @param {glThread} thread - The thread to check
     * @return Whether or not it's a match
     */
    matchThread: function(thread) {
        return this.conditions.matchThread(thread);
    },


    /**
     * Cleans up the remnants of the previous search run, preserving the conditions
     * that were built up.
     * @function ?
     */
    _reset: function() {
        giLogger.log('Reseting the glSearch object', 'glSearch');
        if(this._request) {
            this._request.removeEventListener(glRequest.COMPLETED, this._backgroundSearchCallback);
            this._request = null;
        }
    }
    //}
},{
    // Static Properties {
    /**
     * Dispatched when the search is completed
     * @event ?
     */
    COMPLETED: 'completed',
    //}

    // Static Methods {
    /**
     * Turns the given label into a query for that label
     * @function {static String} ?
     * @param {String} label - The label to search for
     * @return Search query string
     */
    l: function(label, inhibitEncode) {
        return 'label:'+label.toLowerCase().replace(/[\/\s\(\)\&]/g, '-');
    },
    
    /**
     * Extract the search terms from the URL
     * @function {String} ?
     * @param {glGmail} gmail - The Gmail object for the page
     * @return The search-terms portion of the URL
     */
    getQueryFromUrl: function(gmail) {
        var url = gmail.window.wrappedJSObject.top.location.href;
        url = url.replace(/\+/g, " ");
        //url = decodeURIComponent(url);
	
    	if( /(#search)/.test(url) ) {
            var match = url.match( /\/#search\/(.+?)($|\/\w+$)/ );
            if( match ) return decodeURIComponent(match[1]);
        } else if( /(#label)/.test(url) ) {
            var match = url.match( /#label\/(.+?)($|\/\w+$)/ );
            if( match ) return glSearch.l( decodeURIComponent(match[1]) );
        } else {
            // #inbox -> in:inbox, #sent -> in:sent, #all -> , #spam -> in:spam, #trash -> in:trash, #chats -> is:chat, #starred -> is:starred,
            if( /#inbox/.test(url) ) return "in:inbox";
            if( /#sent/.test(url) ) return "in:sent";
            if( /#spam/.test(url) ) return "in:spam";
            if( /#chats/.test(url) ) return "is:chat";
            if( /#starred/.test(url) ) return "is:starred";
        }	
        return "";
    },
    
    	/**
    	* Get query glSearchConditions that focuses a search on a specific user (who may have multiple email addresses)
    	* ((from:andy@gmail.com OR from:andy@gtdinbox.com) AND (to:andy@gmail.com OR to:andy@gtdinbox.com)) - personal emails
    	* ((from:andy@gmail.com OR from:andy@gtdinbox.com) OR (to:andy@gmail.com OR to:andy@gtdinbox.com)) - email concerning a user
    	* @function {static glSearchConditions} ?
    	* @paramset Single email
    	* @param {String} userids - The username or email address
    	* @param {String} fromOrTo - The join type (AND or OR)
    	* @paramset Single email
    	* @param {String[]} userids - The usernames or email addresses
    	* @param {String} fromOrTo - The join type (AND or OR)
    	* @return A self contained (bracketed) query string
    	*/
	restrictToUser: function(userids, fromOrTo) {	// create search block for a given user (set of names or emails)
            if( typeof userids=="string" ) userids = [userids];
            
            var cFrom = new glSearchOrConditions();
            var cTo = new glSearchOrConditions();
                    
            if( userids instanceof Array ) {
                for( var i = 0; i < userids.length; i++ ) {
                    cFrom.addCondition("from:" + userids[i]);
                    cTo.addCondition("to:" + userids[i]);
                }
            } else {
                for( e in userids ) {
                    cFrom.addCondition("from:" + e);
                    cTo.addCondition("to:" + e);
                }
            }

            var c = null;
            if( fromOrTo ) {
                c = new glSearchOrConditions();
            } else {
                c = new glSearchAndConditions();
            }
            c.addCondition(cFrom);
            c.addCondition(cTo);
            
            //var joinStr = fromOrTo? "OR" : "AND";
            //return "(" + "(" + fromStr.join(" OR ") + ") " + joinStr + " (" + toStr.join(" OR ") + ")" + ")";
            return c;
	},
	
	/**
	* Get a Gmail-suitable search string for a date
	* @function {static String} ?
	* @param {int} ms - The valueOf of a given date
	* @return A query string for a date
	*/
	searchDate: function(ms) {
            var d = new Date(ms);
            return d.getFullYear() + "/" + (d.getMonth()+1) + "/" + d.getDate();
	},
	
	/**
	* Combine labels into a glSearchConditions
	* @function {static glSearchConditions} ?
	* @param {String[]} labels - The labels
	* @param {String} joinType - AND or OR. Defaults to OR.
	* @param {boolean} incBrackets - If true, encapsulate in brackets
	* @return The query string
	*/
	joinLabels: function(labels, joinType, incBrackets) {
            // Convert to Gmail label search items:
            var c = null;
            if( !joinType || joinType.toLowerCase()=="or" ) {
                c = new glSearchOrConditions();
            } else {
                c = new glSearchAndConditions();
            }

            if(labels) for( var i = 0; i < labels.length; i++ ) c.addCondition(glSearch.l(labels[i]));
            
            return c;
            
	},
    //}
});
glSearch.implement(giEventDispatcher); // Add event dispatching support








/**
 * Container for search conditions
 *
 * @class ?
 */
var glSearchConditions = giBase.extend({
    // Properties {
    /**
    * Array of the search conditions
    * @property {Array} ?
    */
    conditions: null,

    /**
    * Cache of the query string
    * @property {String} ?
    */
    _queryString: null,
    
    /**
     * If true,
     * throw an error in matchThread if query is ambiguous for the level of detail available in the thread
     * @property {Boolean} ?
     */
    _strictMode: true,
	
    //}

    /**
     * The constructor.
     * @constructor ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - Optional conditions to initialize with
     * @paramset Input 2
     * @param {String} condition - Optional conditions to initialize with
    */
    constructor: function() {
        if(arguments.length) {
            this.conditions = arguments;
        } else {
            this.conditions = [];
        }
    },

    // Methods {
    
    /**
     * Update the strict mode, and propogate to children
     * @function {Boolean} ?
     * @param {Boolean} mode - True if strict mode
     * @param {Boolean} incChildren - Recurse into all children (note addCondition does not enforce).
     * @return the strictmode status
     */
    strictMode: function(mode, incChildren) {
        if( typeof mode!="undefined" ) this._strictMode = mode;
        if( incChildren ) {
            for( var i = 0; i < this.conditions.length; i++ ) {
                if( this.conditions[i] instanceof glSearchConditions ) {
                    this.conditions[i].strictMode( mode, incChildren );
                }
            }
        }
        return this._strictMode;
    },
    
    /**
     * Adds the given condition
     * @function ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - A condition to add to the current condition
     * @paramset Input 2
     * @param {String} condition - A condition to add to the current condition
     */
    addCondition: function(condition) {
        this.conditions.push(condition);
        this._queryString = null; // Invalidate cache
    },
    
    /**
     * Remove a given condition
     * @function ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - A condition to add to the current condition
     * @paramset Input 2
     * @param {String} condition - A condition to add to the current condition
     */
    removeCondition: function(condition) {
        for( var i = this.conditions.length - 1; i>=0; i-- ) {
            if( this.conditions[i]==condition ) {
                this.conditions.splice(i, 1);
                break;
            }
        }
    },

    /**
     * Converts the condition to a query string
     * @function {String} ?
     * @return The query string
     */
    toString: function() {
        if(this._queryString === null) {
            this._queryString = this._toQueryString();
        }
        return this._queryString;
    },

    /**
     * Returns whether or not the given thread matches the search
     * @function {Boolean} ?
     * @param {glThread} thread - The thread to check
     * @return Whether or not it's a match
     * @throws Impossible match error, if the search condition isn't supported
     */
    matchThread: function(thread, type) {
        // Type is a hidden parameter that determines the type of condition it is - 'and' vs 'or'
        if(typeof type == 'undefined') type = 'and';


        var pass = null;
        if( this.conditions.length==0 ) {
            // Edge case - pass all
            pass = true;
        } else {
            for(var i = 0; i < this.conditions.length; i++) {
                // Based on condition type, pass or fail, resetting pass variable afterwards
                if(pass !== null) {
                    if(type == 'and' && pass === false) return false;
                    if(type == 'or' && pass === true) return true;
                }
                pass = null;
                
                if(this.conditions[i] instanceof glSearchConditions) {
                    pass = this.conditions[i].matchThread(thread);
                } else if(typeof this.conditions[i] == 'string') { // No need to test 'to:' or 'has:attachment', etc. -> only revolves around things that can change. Those only change as new messages arrive (and therefore could be handled; but 'to:' is hard unless messages are available -> which they would be in that use-case)
                    var term = this.conditions[i];
                    // If a negating term, increase offset for substr
                    var negOffset = 0;
                    if( term.indexOf("-")==0 && term.indexOf(":")>0 ) {
                        negOffset = 1;
                    }
                    if(term=="") {
                        pass = true;
                    } else if(/^-?label:/.test(term)) {
                        // Must be able to handle the fact label search is not the same as a label (e.g P/GTDInbox -> P-GTDInbox); and there are close cases, e.g. P/GTDInbox and P/GTDInbox/Bugs, that can disrupt simple string tests
                        var re = RegExp('(^|[^A-Z0-9-])' + term.substr(6+negOffset) + '([^A-Z0-9-]|$)', 'i');
                        pass = re.test(thread.labels.toString().replace(/[\/\s\(\)]/g, '-')); // TODO Faster to create RE object or to do simple tests iterating over the labels array
                    } else if(/^-?has:attachment/.test(term)) {
                        pass = !!thread.attachmentsCSV;
                    } else if(/^-?in:inbox/.test(term)) {
                        pass = /\^i/.test(thread.labels.toString());
                    } else if(/^-?in:anywhere/.test(term)) {
                        pass = true;
                    } else if(/^-?in:trash/.test(term)) {
                        pass = /\^k/.test(thread.labels.toString());
                    } else if(/^-?in:spam/.test(term)) {
                        pass = /\^s/.test(thread.labels.toString());
                    } else if(/^-?is:starred/.test(term)) {
                        pass = thread.isStarred;
                    } else if(/^-?is:unread/.test(term)) {
                        pass = !thread.isRead;
                    } else if(/^-?is:read/.test(term)) {
                        pass = thread.isRead;
                    } else if(/^-?is:chat/.test(term)) {
                        pass = /\^b/.test(thread.labels.toString());
                    } else if(/^-?filename:/.test(term)) {
                        pass = (thread.attachmentsCSV && thread.attachmentsCSV.indexOf(term.substr(9+negOffset))>-1);
                    } else if(/^-?after:/.test(term)) {
                        var d = new Date(term.substr(6+negOffset));
                        var dThread = new Date(thread.timestampJS);
                        pass = (dThread.valueOf() > d.valueOf());
                    } else if(/^-?before:/.test(term)) {
                        var d = new Date(term.substr(7+negOffset));
                        var dThread = new Date(thread.timestampJS);
                        pass = (dThread.valueOf() < d.valueOf());
                    } else if(/^-?subject:/.test(term)) {
                        var term = term.substr(8+negOffset);
                        pass = this._matchThread_handleComplexTermValue(thread, "subject:", term, function(thread, value) {
                            if( m = /^\"([^"]+)\"$/.exec(value) ) value = m[1]; // Ensure quotes removed
                            return (thread.subjectSnippet.indexOf(value)>-1);
                        });
                    } else if(/^-?to:/.test(term)) {
                        // Right now, we're relaxed wrt to thread details - if search dictates terms that require threadDetails, we pass the terms if threadDetails doesn't exist. I.e. we do the best we can given the environment.
                        term = term.substr(3+negOffset);
                        if(thread.getThreadDetails(true)) {
                            pass = this._matchThread_handleComplexTermValue(thread, "to:", term, function(thread, value) {
                                return thread.hasContact( value, false, true );
                            });
                        } else if(!this._strictMode) {
                            pass = thread.hasContact( [term] );
                        }
                    } else if(/^-?from:/.test(term)) {
                        // Right now, we're relaxed wrt to thread details - if search dictates terms that require threadDetails, we pass the terms if threadDetails doesn't exist. I.e. we do the best we can given the environment.
                        term = term.substr(5+negOffset);
                        if(thread.getThreadDetails(true)) {
                            pass = this._matchThread_handleComplexTermValue(thread, "from:", term, function(thread, value) {
                                return thread.hasContact( value, true, false );
                            });
                        } else if(!this._strictMode) {
                            pass = thread.hasContact( [term] );
                        }
                    } else if( thread.getThreadDetails(true) || !this._strictMode ) {
                        // General string analysis
                        // Nb the expectation is that glSearchConditions has been constructed so that string terms are atomic (e.g. "a b" => "a", "b"). If in doubt, use .createFromSearchString
                        term = term.replace(/\"/g, ""); // Quotes have no use, as all our terms are atomic (quotes are only useful for grouping spaces into a single term).
                        if( m = /^\(([^)]+)\)$/.exec(term) ) term = m[1]; // It's wrapped in brackets, unnecessary
                        pass = thread.hasSearchTerm(term);
                    }
                    if( pass===null ) {
                        // No term found
                        throw new Error("Impossible match - could not match individual condition ("+this.conditions[i]+")");
                    }
                    if( negOffset==1 ) {
                        // Invert result for negation
                        pass = !pass;
                    }
                } else {
                    throw new Error('Impossible match - unknown condition type');
                }
            }
        }

        // Made it to the end, so the final test determines matching
        return pass;
    },
    /**
     * Used by .matchThread to expand a complex value on a term. E.g. subject:"Chat (with OR avec)"
     * If possible, resolve the complex value, otherwise create more nested glSearchCondition objects and recurse into them to find the answer
     *   e.g. Turn, subject:"Chat (with OR avec)" into a nested glSearchConditions that is equivelent to, subject:Chat AND (subject:with OR subject:avec); and test each term within that.
     * @function {Boolean} ?
     * @param {glThread} thread - The thread to test
     * @param {String} term - The term, e.g. subject:
     * @
     * @return True if there is an ultimate match for this complex term/value for the thread
     */
    _matchThread_handleComplexTermValue: function(thread, term, value, testFunction) {
        var pass = false;
        if( m = /^\"([^"]+)\"$/.exec(value) ) {
            // It's wrapped in quotes
            if( /\([^)]+\)/.test(value) || / OR /i.test(value) ) {
                // It contains brackets or OR; it could be very complex, so break it up. e.g. (Chat (with OR avec) "Chris L")
                var searchConditions = glSearchConditions.createFromSearchString(m[1], term); // Note the use of prefix. This would turn, subject:"Chat (with OR avec)" into, subject:Chat AND (subject:with OR subject:avec), in a structured object that can be tested.
                pass = searchConditions.matchThread(thread);    // Test each term.
            } else {
                // It's a simple quoted term - test it (all complex values reach this point eventually, after all nesting has been evaluated)
                pass = testFunction(thread, value);
            }
        } else if( m = /^\(([^)]+)\)$/.exec(value) ) {
            // It's wrapped in brackets
            var searchConditions = glSearchConditions.createFromSearchString(m[1], term);
            pass = searchConditions.matchThread(thread);    // Test each term.
        } else {
            // It's a simple term - test it (all complex values reach this point eventually, after all nesting has been evaluated)
            pass = testFunction(thread, value);
        }
        return pass;
    },

    /**
     * Return true if the matchThread function can run on this query.
     * It should be noted that as it requires a thread to be created, and throws an error, this is not a particularly efficient function. Use matchThread only where possible.
     * @function {Boolean} ?
     * @param {optioanl glThread} testThread - If it isn't provided, will create one
     * @param {optional glGmail} gmail - Needed if no testThread provided
     * @return True if can match thread
     */
    canMatchThreads: function(testThread, gmail) {
        if( !testThread ) {
            testThread = new glThread(gmail);
            testThread.labels = [];
            testThread.knownContacts = [];
        }
        try {
            this.matchThread(testThread);
        } catch(e) {
            return false;
        }
        return true;
    },

    /**
     * Converts the condition to a query string using the given seperator
     * @function {String} ?
     * @param {String} seperator - The optional seperator to place between conditions
     * @return The query string
     */
    _toQueryString: function(seperator) {
        var queryString = seperator ? "(" : "";

        for(var i = 0; i < this.conditions.length; i++) {
            if(!this.conditions[i]) continue;

            if(typeof this.conditions[i] == 'string') {
                queryString += this.conditions[i];
            } else if(this.conditions[i].toString) {
                queryString += this.conditions[i].toString();
            } else {
                giLogger.error('Invalid condition', 'glSearchConditions');
            }

            if(i < this.conditions.length-1) queryString += seperator ? seperator : " ";
        }

        if(seperator) queryString += ")";

        return queryString;
	},
    //}
},{
    
    /**
     * Take a general search string, and parse it into search-condition objects.
     * @function {glSearchConditions} ?
     * @param {String} str - The search string
     * @param {optional String} prefix - Prefix every term in the entire structure with this term (used during expansion by .matchThread)
     * @param {optional String{}} substitutions - The subsitutions object used within recursion
     * @return glSearchCondition object equivelent of the search string
     */
    createFromSearchString: function(str, prefix, substitutions) {
        if( typeof prefix!="string" ) prefix = "";
        
        // Create a placeholder glSearchConditions
        var searchConditions = null;
        
        
        // Pre-Process to temporarily remove spaces that need to be grouped:
        
        var substitutions = substitutions || {_idx:0};
        
        // Remove AND, as it serves no useful purpose (things are AND by default):
        str = str.replace(/ AND /gi, " ");
        
        //  Extract "":
        while( m = str.match( /(\"[^"]+\")/ ) ) {
            var code = String.fromCharCode(250) + substitutions._idx++;
            substitutions[code] = m[1];
            str = str.replace(m[1], code);
        }
        
        //  Extract :() - Bracket groups that are tied to a Term
        var bracketItems = glSearchConditions.extractBrackets(str, function(s, i) {
            // Ensure character before opening bracket is a :
            if( (i-1)>=0 && s[i-1]==":" ) return true;
            return false;
        });
        for( var i = 0; i < bracketItems.length; i++ ) {
            var code = String.fromCharCode(250) + substitutions._idx++;
            substitutions[code] = bracketItems[i];
            str = str.replace(bracketItems[i], code);
        }
        
        // Extract remaining loose () - Bracket groups containing terms
        var bracketItems = glSearchConditions.extractBrackets(str);
        for( var i = 0; i < bracketItems.length; i++ ) {
            var code = String.fromCharCode(251) + substitutions._idx++;
            substitutions[code] = bracketItems[i];
            str = str.replace(bracketItems[i], code);
        }
        
        
        // Now parse the terms:
        
        if( /\sOR\s/i.test(str) ) {  // It has major OR blocks, break them up and recursively construct their own glSearchConditions
            // Create OR container:
            searchConditions = new glSearchOrConditions();
            // Break into multiple searches:
            var tokens = str.split(/\sOR\s/i);
            // Create object for each search, and add to our OR container:
            for( var i = 0; i < tokens.length; i++ ) {
                var childSearchConditions = glSearchConditions.createFromSearchString(tokens[i], prefix, substitutions);
                if( childSearchConditions ) {
                    searchConditions.addCondition(childSearchConditions);
                }
            }
        } else {  // Tokenise terms by splitting on spaces, then process each term
            // Create AND container:
            searchConditions = new glSearchAndConditions();
            // Break into multiple terms:
            var tokens = str.split(" ");
            // Process each term, expand codeified terms; and if a loose (), transform into its own object
            for( var i = 0; i < tokens.length; i++ ) {
                var re = new RegExp('('+String.fromCharCode(251)+'\\d+)');
                if( m = re.exec(tokens[i]) ) {
                    // It needs expanding, and it's own object creating (without outer-most brackets)
                    tokens[i] = tokens[i].replace( m[1], substitutions[m[1]] );
                    tokens[i] = tokens[i].substr(1, tokens[i].length-2);
                    var childSearchConditions = glSearchConditions.createFromSearchString(tokens[i], prefix, substitutions);
                    if( childSearchConditions ) {
                        searchConditions.addCondition(childSearchConditions);
                    }
                } else {
                    // It's a simple string term:
                    var re = new RegExp('('+String.fromCharCode(250)+'\\d+)');
                    while( m = re.exec(tokens[i]) ) {
                        // Expand
                        tokens[i] = tokens[i].replace( m[1], substitutions[m[1]] );
                    }
                    searchConditions.addCondition(prefix+tokens[i]);
                }
            }
        }
        
        return searchConditions;
        
    },
    
    /**
     * Take a string, and extract all base parenthesis.
     * e.g. "a (b or (c or d)) e (f or g)" would return ["(b or (c or d))", "(f or g)"]
     * @function {String[]} ?
     * @param {String} str - The string to parse
     * @param {Function} openCondition - Called on the first encounter of a (, passing in the string and the index
     * @return Array of base parenthesis
     */
    extractBrackets: function(str, openCondition) {
        var container = {start:-1, end:-1, pdepth:0};
        var containedItems = [];
        for( var i = 0; i < str.length; i++ ) {
            if( str[i]=="(" ) {
                if( container.pdepth==0 ) {    // Start
                    // Ensure openCondition gives us permission to accept this bracket
                    if( !openCondition || openCondition(str, i) ) {
                        container.start = i;
                        container.pdepth = container.pdepth + 1;
                    }
                } else {  // Sub bracket
                    container.pdepth = container.pdepth + 1;
                }
            } else if( str[i]==")" && container.pdepth>0 ) {
                container.pdepth = container.pdepth - 1;
                if( container.pdepth==0 ) {    // End
                    container.end = i+1;
                    containedItems.push( str.substring(container.start, container.end) );
                }
            }
        }
        return containedItems;
    },
    
});

/**
 * Container for search conditions which joins them with an "AND"
 *
 * @class ? @extends glSearchConditions
 */
var glSearchAndConditions = glSearchConditions.extend({
    // Methods {
    /**
     * @function {Boolean} ?
     * @inheritdesc
     */
    matchThread: function(thread) {
        return this.base(thread, 'and');
    },

    /**
     * Converts the condition to a query string
     * @function {String} ?
     * @return The query string
     */
    toString: function() {
        if(this._queryString === null) {
            this._queryString = this._toQueryString(' AND ');
        }
        return this._queryString;
    },
    //}
});

/**
 * Container for search conditions which joins them with an "OR"
 *
 * @class ? @extends glSearchConditions
 */
var glSearchOrConditions = glSearchConditions.extend({
    // Methods {
    /**
     * @function {Boolean} ?
     * @inheritdesc
     */
    matchThread: function(thread) {
        return this.base(thread, 'or');
    },

    /**
     * Converts the condition to a query string
     * @function {String} ?
     * @return The query string
     */
    toString: function() {
        if(this._queryString === null) {
            this._queryString = this._toQueryString(' OR ');
        }
        return this._queryString;
    }
    //}
});
