/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Central cache for storing lists of threads.
 * Automatically populated when the view changes in Gmail (e.g. Inbox, Sent Items, Search Results).
 * TODO - make it possible to get threads even if there's a cache miss - download them, and issue 'update' events so a listener can check to see if their threads have arrived yet
 * TODO - Add garbage collection to actively delete old views
 *
 * @class ?
 */

var glThreadListCache = giBase.extend({

    /**
     * @property {glGmail}
     */
    _gmail: null,
    
    /**
     * Arrays of threads by view keys
     * @property {}
     */
    _views: null,
    
    /**
     * The VIEWDATA, if available
     * @property {glDataObject}
     */
    _fakeLoadData: false,
    
    /**
    * The Constructor
    * @function ?
    * @param {glGmail} gmail - the glgmail instance for the page
    * @param {optional glDataObject} viewData - the glDataObject for Gmail's VIEW_DATA - the initial view
    */
    constructor: function(gmail, viewData) {
        this._gmail = gmail;
        this._views = {};
        
        if( this._gmail.trafficWatcher ) {
            this._gmail.trafficWatcher.addEventListener(glTrafficWatcher.THREAD_LIST_REQUEST_COMPLETED, this._handleThreadListRequest, this);
        }
        
        this._gmail.environment.addEventListener(glEnvironment.RENAME_LABEL, this._modifyEnvironmentLabels, this);
        this._gmail.environment.addEventListener(glEnvironment.DELETE_LABEL, this._modifyEnvironmentLabels, this);
        
        this._fakeLoadData = viewData;
        
        this._gmail.page.addEventListener(glPageWrapper.UI_READY, this._baseUIReady, this);
            
    },
    
    
    
    /**
     * The UI_READY callback for the glGmail class. Currently just populates
     * the canvasDocument property.
     * @function ?
     * @param {Event} evt - The glPageWrapper.UI_READY event
     */
    _baseUIReady: function() {
        giLogger.log('LOAD BASEUIREADY');
        this._gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
        
        // Load default view:
        
        var url = glThreadListCache.locationKey( this._gmail.doc.location.hash );
        if( !url ) url = "#inbox";  // TODO This is a hack, as on some loads location.hash isn't populated. Do we need to slow it down? setTimeout? setTimeout would reduce clarity/logic-flow of code
        
        if( this._fakeLoadData ) {
            // Create fake url and populate _threadListCache for the inbox from the view data
            var threads = this._fakeLoadData.getThreads();
            this._views[url] = threads;
            this._fakeLoad = true;
        } else {
            // Use glSearch to grab the inbox items
            var search = null;
            if( url.indexOf('#search/')==0 ) {
                search = url.substr(8);
            } else if( url.indexOf('#label/')==0 ) {
                search = 'label:' + glSearch.l(url.substr(7));
            } else if( url=='#all' ) {
                search = '';
            } else if( url.indexOf('#')==0 ) {
                search = 'in:'+url.substr(1);
            }
            if( search!==null ) {
                var s = new glSearch(this._gmail);
                s.find(search);
                s.cacheUrl = url;
                s.addEventListener(glSearch.COMPLETED, this._baseSearchComplete, this);
                s.runBackground(0, 50);
            }
        }
    },
    
    /**
     * Notification that glSearch has returned initial view items
     * @function {} ?
     */
    _baseSearchComplete: function(evt) {
        var s = evt.target;
        this._views[s.cacheUrl] = s.getResults();
    },
    
    /**
     * Return an thread for a position in a given view.
     * Will create a proper (glThread) object from the simple thread object and cache it for faster future recall.
     * @function {glThread} ?
     * @param {String} key - The key for the view, typically location.hash
     * @param {int} start - The starting offset
     * @param {int} offset - The offset within the view
     * @return The glThread object at the offset
     */
    getThread: function(key, start, offset) {

        key = glThreadListCache.locationKey( key );
        
        if( !this._views[key] ) return null;
        
        
        
        var thread = this._views[key][start+offset];
        if( !thread ) return null;
        
        if( thread instanceof glThread ) {
            return thread;
        } else {
            // Load proper glThread and return it
            var properThread = glThread.createFromData(thread, this._gmail);
            this._views[key][start+offset] = properThread;  // TODO update all instances of thread ID in any view
            return properThread;    // Warning: may not be loaded.
        }
    },
    
    /**
     * Return an thread ID for a position in a given view.
     * @function {String} ?
     * @param {String} key - The key for the view, typically location.hash
     * @param {int} start - The starting offset
     * @param {int} offset - The offset within the view
     * @return The thread Id
     */
    getThreadId: function(key, start, offset) {
        key = glThreadListCache.locationKey( key );
        if( !this._views[key] ) return null;
        
        
        
        var thread = this._views[key][start+offset];
        if( !thread ) return null;
        
        return thread.threadId;
    },
    
    /**
     * Process all threads in the cache looking for the one with the ID
     * @function {glThread} ?
     * @param {String} threadId
     */
    getThreadById: function(threadId) {
        
        for( key in this._views ) {
            for( var i = 0; i < this._views[key].length; i++ ) {
                if( this._views[key][i].threadId==threadId ) {
                    // Found a match
                    if( this._views[key][i] instanceof glThread ) {
                        return this._views[key][i];
                    } else {
                        // Convert it
                        this._views[key][i] = glThread.createFromData( this._views[key][i], this._gmail );
                        return this._views[key][i];
                    }
                }
            }
        }
        
    },
    
    /**
     * Return an array of glThread objects for the range
     * @function {Object[]} ?
     * @param {String} key - The key for the view, typically location.hash
     * @param {int} start - The starting offset
     * @param {int} end - The ending offset
     * @return An array of glThread thread data
     */
    getThreads: function(key, start, end) {
        key = glThreadListCache.locationKey( key );
        if( !this._views[key] ) return [];
        if( end >= this._views[key].length ) end = this._views[key].length - 1;
        
        var arr = [];
        for( var i = start; i <= end; i++ ) {
            if( !(this._views[key][i] instanceof glThread) ) {
                // Convert it to a glThread; it's a simple object
                this._views[key][i] = glThread.createFromData( this._views[key][i], this._gmail );
            }
            if( this._views[key][i] ) arr.push( this._views[key][i] );
        }
        
        return arr;
    },
    
    /**
     * Event handler for {@link glTrafficWatcher.THREAD_LIST_REQUEST_RECEIVED}
     * events. Adds the response data to the thread list cache.
     * @function ?
     * @param {Event} evt - The THREAD_LIST_REQUEST_RECEIVED event
     */
    _handleThreadListRequest: function(evt) {
        
        
        // Extract key - take the search/cat/q values and transform to match the Gmail search format
        var key = null;
        var url = evt.url;
        if( m = url.match(/[?&]search=([^&]+)($|&)/) ) {
            var search = m[1];
            switch(search) {
                case 'cat':
                    // &search=cat&cat=<label> -> #label/<label>
                    if( m = url.match(/[?&]cat=([^&]+)($|&)/) ) {
                        key = '#label/' + decodeURIComponent(m[1]);
                    }
                    break;
                case 'query':
                    // &search=query&q=<query> -> #search/<query>
                    if( m = url.match(/[?&]q=([^&]+)($|&)/) ) {
                        key = '#search/' + decodeURIComponent(m[1]);
                    }
                    break;
                default:
                    // &search=inbox => #inbox
                    key = '#' + search;
                    break;
            }
        }
        if( !key ) return;
        
        
        
        // Extract start:
        var start = 0;
        if( m = url.match(/[?&]start=([^&]+)($|&)/) ) {
            start = m[1];
        }
        
        // Get the threads:
        var threads = evt.responseData.getThreads(true);
        
                
        // Add the threads to the cache:
        if( this._views[key] ) {
            // Already exists
            /*
            // TODO This protects any scans deeper into the current view. Is it actually useful? Nothing uses it. Kept for posterity.
            if( this._views[key].length>=(start+threads.length) ) {
                // Has messages beyond the range, are there new messages that would push them back?
                
                var found = false;
                for( var i = 0; (i < threads.length) && (i < 5); i++ ) {
                    for( var j = 0; j < this._views[key].length; j++ ) {
                        if( this._views[key][j].threadId==threads[i].threadId ) {
                            // Thread found, conclude all subsequent new threads will already exist
                            found = true;
                            break;
                        }
                    }
                    if( found ) {
                        break;
                    } else {
                        // Thread not found - therefore brand new - push a space back (so subsequent pages are accurate).
                        this._views[key].splice(0,0,null);
                    }
                }
            }
            */
            // Test to see if first item has changed
            if( start==0 && this._views[key][0] && threads[0].threadId!=this._views[key][0].threadId ) {
                // Nb the 'this._views[key][0]' condition is required for when the TrafficWatcher data is polluted with a random comma (,) stopping it being parsed. It seems to spawn a constant looping error.
                // First item has changed
                this.dispatchEvent({name:glThreadListCache.THREAD_LIST_CHANGED, key:key});
            }
            // Overwrite everything in range of the new threads (e.g. if start=50, and num=50, it will remove threads 50-100; and then re-add the updated threads)
            this._views[key].splice(start, threads.length);
            // Add new/updated thread objects at start point
            for( var i = threads.length - 1; i>=0; i-- ) {
                this._views[key].splice(start, 0, threads[i]);  // TODO is there a more efficient way? There is a shorter way, although whether it is faster is unclear: http://www.highdots.com/forums/javascript/splice-array-into-another-array-257292.html
            }
        } else {
            this._views[key] = threads;
        }
        
    },
    
    /**
     * Callback for the {@link glActionManager.THREAD_CHANGED} event. Updates all
     * stored threads from the event and removes any that no longer fit the search
     * criteria.
     * @function ?
     * @param {Event} evt - The thread changed event
     */
    _threadChanged: function(evt) {
        
        
        // Visit all affected threads:
        var threadMatch = false;
        var isGlThread = false;
        for( key in this._views ) {  
            var view = this._views[key];
            // TODO Can we do a precheck to see if a key contains this evt.thread? It would require having a hash of thread ids for a key; which isn't very efficient to build. We could build as we evaluate; so if a lot of thread action is happening, it only costs on the first check.
            for( var i = view.length - 1; i>=0; i-- ) {
                if( isGlThread = (view[i] instanceof glThread) ) {
                    threadMatch = view[i].hasId(evt.thread);
                } else {
                    threadMatch = view[i].threadId==evt.thread;
                }
                if( threadMatch ) {
                    // We have a match on this thread.
                    if( isGlThread = (view[i] instanceof glThread) ) {
                        view[i].updateFromEvent(evt);
                    }
                    
                    switch( evt.action ) {
                        case glActionManager.ARCHIVE:
                            if( key.indexOf('#inbox')>-1 || key.indexOf('in:inbox')>-1 ) {
                                view.splice(i,1); // Remove it
                                this.dispatchEvent({name:glThreadListCache.THREAD_LIST_CHANGED, key:key});
                            }
                            break;
                        case glActionManager.REPORT_SPAM:
                            if( key.indexOf('#spam')==-1 || key.indexOf('in:spam')==-1 ) {
                                view.splice(i,1); // Remove it
                                this.dispatchEvent({name:glThreadListCache.THREAD_LIST_CHANGED, key:key});
                            }
                            break;
                        case glActionManager.DELETE:
                            view.splice(i,1); // Remove it
                            this.dispatchEvent({name:glThreadListCache.THREAD_LIST_CHANGED, key:key});
                            break;
                        case glActionManager.ADD_LABEL:
                            if( key.indexOf('-label:' + evt.label)>-1 ) {
                                view.splice(i,1); // Remove it
                                this.dispatchEvent({name:glThreadListCache.THREAD_LIST_CHANGED, key:key});
                            } else {
                                if( !isGlThread ) {
                                    // Manually adjust simple thread object
                                    var found = false;
                                    for( var j = 0; j <= view[i].labels.length; j++ ) {
                                        if( view[i].labels[j]==evt.label ) {
                                            found = true;
                                            break;
                                        }
                                    }
                                    if( !found ) view[i].labels.push(evt.label);
                                }
                            }
                            break;
                        case glActionManager.REMOVE_LABEL:
                            if( (key.indexOf('label:' + evt.label)>-1) || (key.indexOf('#label/' + evt.label)>-1) ) {
                                view.splice(i,1); // Remove it
                                this.dispatchEvent({name:glThreadListCache.THREAD_LIST_CHANGED, key:key});
                            } else {
                                if( !isGlThread ) {
                                    // Manually adjust simple thread object:
                                    for( var j = view[i].labels.length -1; j>=0; j-- ) {
                                        if( view[i].labels[j]==evt.label ) {
                                            view[i].labels.splice(j,1);
                                        }
                                    }
                                }
                            }
                            break;
                    }
                }
            }
        }
        
    },
    
    
    /**
     * Notification that an environment label was renamed or removed.
     * Need to update labels in cached threads.
     * @function {} ?
     */
    _modifyEnvironmentLabels: function(evt) {
        
        // Clean up cached labels:
        // TODO Is this really necessary? If Google always reloads thread-list after a label rename/remove; no (as that will clear out the cache). The question is, does it? Watch inbox, preview item with a label, remove that label, listen for cache updates.
        
        var isDelete = (evt.name==glEnvironment.DELETE_LABEL)? true : false;
        
        for( key in this._views ) {  
            var view = this._views[key];
            for( var i = view.length - 1; i>=0; i-- ) {
                if( view[i] instanceof glThread ) {
                    view[i].updateFromEnvironmentEvent(evt);
                } else {
                    var index = view[i].labels.indexOf(evt.label);
                    if( isDelete ) {
                        // Delete it
                        if( index!=-1 ) view[i].labels.splice(index, 1);
                    } else {
                        // Rename it
                        if( index!=-1 ) view[i].labels.splice(index, 1, evt.newLabel);
                    }
                }
                
            }
        }
    },
    
    /**
     * Remove a specific thread ID from a cached view.
     * @function {} ?
     * @param {String} threadId - The ID of the thread to remove
     * @param {[]} view - Array of simple thread objects
     */
    /*
    _removeThread: function(threadId, view) {
        if( view ) {
            // TODO Can we do a quick test to see if this view contains the thread Id before we iterate over it? A seperate hash?
            for( var i = view.length-1; i>=0; i-- ) {
                if( view[i].threadId==threadId ) {
                    view.splice(i,1);
                }
            }
        } else {
            for( key in this._views ) {
                this._removeThread(threadId, this._views[key]);
            }
        }
    },
    */
    
},{
    // STATIC PROPERTIES {
      
    /**
     *  Notification that a thread list has changed (has a new message, or a message was removed)
     *  @Event
     */
    THREAD_LIST_CHANGED: 'thread_list_changed',
        
    //}
    
    // STATIC METHODS {
      
    /**
     * Take a location.hash and return the key
     * @function {String} ?
     * @return The key used to identify views
     */
    locationKey: function(locationHash) {
        if( !locationHash ) return "";
        if( m = locationHash.match(/(\/p\d+)$/) ) {
            // Remove pagination from URL:
            locationHash = locationHash.replace(m[1], '');
        }
        return locationHash;
    },
    
    //}
})
glThreadListCache.implement(giEventDispatcher); // Add event dispatching support