/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * A wrapper for the Compose (rich or plain) box in the 'Compose Mail' or 'Conversation' screens.
 *
 * @class ? @extends glUIView
 */
var glComposeBox = glUIWidget.extend({
    // Properties {
    
    
    /**
     * @property {glGmail}
     */
    _gmail: null,
    
    /**
     * @property {Element}
     */
    _container: null,
    
    /**
     * @property {String}
     */
    _cacheType: null,
    
    
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        giLogger.log('Connecting glComposeBox object to view', 'glComposeBox');
        this.base(gmail);

        this._gmail = gmail;
        this.setScope();
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting glComposeBox from view', 'glComposeBox');
        
        this._cacheType = null;
        this._container = null;
        this._gmail = null;
        this.base();
        
        giLogger.log('Disconnected glComposeBox from view', 'glComposeBox');
    },
    
    /**
     * The compose box is selected from within the container.
     * Narrow the scope to pick out individual compose boxes for replies.
     * An empty input resets it to the view root.
     * @function {} ?
     * @param {Element} container - The scope
     */
    setScope: function(container) {
        // Get the container:
        if( container ) {
            this._container = container;
        } else {
            this._container = this._gmail.page.getViewRoot();
        }
        
        // Reset the cache:
        this._cacheType = null;
    },
    
    /**
     * Extract the type (rich/plain) of the compose box
     * @function {String} ?
     * @param {optional Element} box - The box element.
     * @return The type
     */
    getType: function(box) {
        if( !this._cacheType ) {
            box = box || this.getComposeBox(this._container);
            if( box.nodeName=="IFRAME" ) {
                this._cacheType = "rich";
            } else {
                this._cacheType = "plain";
            }
        }
        return this._cacheType;
    },

    /**
     * Set the text of the visible (rich/plain) compose box.
     * @function {} ?
     * @param {String} text - The text to set the box to
     */
    setText: function(text) {
        var box = this.getComposeBox(this._container);
        var type = this.getType(box);
        
        if( type=="rich" ) {
            box.contentDocument.body.innerHTML = text;
        } else {
            box.value = text; 
        }
    },
    
    /**
     * Get the text of the visible (rich/plain) compose box.
     * @function {} ?
     */
    getText: function() {
        var box = this.getComposeBox(this._container);
        var type = this.getType(box);
        
        if( type=="rich" ) {
            return box.contentDocument.body.innerHTML;
        } else {
            return box.value; 
        }
    },
    
    /**
     * Retrieve the visible (rich/plain) compose box
     * @function {Element} ?
     * @param {Element} container - To narrow the scope for retrieving the box
     * @return The visible (either rich or plain) compose box
     */
    getComposeBox: function(container) {
        
        // Reset the cache:
        this._cacheType = null;
        
        // Set up variables
        var doc = this._gmail.canvasDocument;
        
        // Try to retrieve the rich text first:
        var elRich = doc.evaluate(".//iframe[contains(@class, 'editable')]", this._container, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        if( elRich ) {
            // Now confirm that it isn't hidden:
            var elVis = elRich;
            var vis = true;
            while( elVis.parentNode ) {
                if( elVis.style.display=="none" ) {
                    vis = false;
                    break;
                }
                elVis = elVis.parentNode;
            }
            if( vis ) {
                return doc.getElementById(elRich.id);
            }
        }
        
        // Try to retrieve the plain:
        var elPlain = doc.evaluate(".//textarea[contains(@class, 'dV')]", this._container, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        // Now confirm that it isn't hidden:
        var elVis = elPlain;
        var vis = true;
        while( elVis.parentNode ) {
            if( elVis.style.display=="none" ) {
                vis = false;
                break;
            }
            elVis = elVis.parentNode;
        }
        if( vis ) {
            return elPlain;
        } else {
            throw new Error('Cannot find compose box. It is possible that the rich iframe has not yet loaded.');
        }
    },


    /**
     * Whether the UI object can be cached by {@link glGmail.getUIObject}
     * @function {Boolean} ?
     * @return Whether the UI can be cached
     */
    isCacheable: function() {
        return false;
    },

    //}
});