/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The conversation view abstraction layer.
 *
 * @class ? @extends glUIView
 */
var glConversationReplyHandler = glUIWidget.extend({
    // Properties {
        
    /**
     * @property {Object[]}
     */
    replies: null,
    
    /**
     * @property {giEventsContainer}
     */
    _eventsContainer: null,
    
    /**
     * @property {Element}
     */
    _viewRoot: null,
    
    /**
     * Store show/hide status of replies to detect changes in refreshes
     * @property {Object}
     */
    _statusTracker: null,
	
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        giLogger.log('Connecting glConversationReplyHandler object to view', 'glConversationReplyHandler');
        this.base(gmail);


        this._eventsContainer = new giEventsContainer(this);
        
        this.replies = [];
        
        // Add listeners:
        this._viewRoot = this._glGmail.page.getViewRoot();
        this._eventsContainer.observe(this._viewRoot, "DOMNodeInserted", this._domChange, true);
        
        // Get initial view:
        this._refresh(true);
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting glConversationReplyHandler from view', 'glConversationReplyHandler');
        
        this._viewRoot = null;
        this.replies.length = 0;
        this._statusTracker = {};
        this._eventsContainer.reset();
        
        this.base();
    },



    _refresh: function() {
        var doc = this._glGmail.canvasDocument;
        
        // Clear history:
        this.replies.length = 0;
        var oldStatus = this._statusTracker;
        this._statusTracker = {};
        var updated = false;
        
        
        // Grab all outer Reply boxes in the Conversation (does not include hidden ones):
        var result = doc.evaluate(".//DIV[@class='gA gt']/DIV[@class='gB']", this._glGmail.page.getViewRoot(), null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        for( var i = 0; i < result.snapshotLength; i++ ) {
            
            // Store reference to it, and give it an addresseable ID (if it doesn't already have one)
            var elRoot = result.snapshotItem(i);
            var id = elRoot.getAttribute("gtdi-reply-id");
            if( !id ) {
                id = (new Date()).valueOf();
                elRoot.setAttribute("gtdi-reply-id", id);
            }
            var d = this.createReplyData(elRoot);
            d.showingFull = !!(d.elFull.innerHTML && d.elFull.firstChild.nodeName!="TEXTAREA");
            this.replies.push(d);
            
            // Track changes:
            this._statusTracker[id] = {'showingFull': d.showingFull};
            if( oldStatus ) {
                // Second or higher run; compare:
                if( oldStatus[id] ) {
                    // See if show/hide changed; then delete
                    if( oldStatus[id].showingFull!=this._statusTracker[id].showingFull ) {
                        // Changes showing full status:
                        if( this._statusTracker[id].showingFull ) {
                            // Started showing
                            this.dispatchEvent({name:glConversationReplyHandler.SHOW_REPLY_FULL, id: id});
                        } else {
                            // Hidden
                            this.dispatchEvent({name:glConversationReplyHandler.HIDE_REPLY_FULL, id: id});
                        }
                        updated = true;
                    }
                    delete oldStatus[id];
                } else {
                    // We have a new item
                    this.dispatchEvent({name:glConversationReplyHandler.ADD_REPLY_ROW, id: id});
                    updated = true;
                }
            }
        }
        
        if( oldStatus ) {
            // See if any items removed
            for( i in oldStatus ) {
                // This id was removed:
                this.dispatchEvent({name:glConversationReplyHandler.REMOVE_REPLY_ROW, id: i});
                updated = true;
            }
        }
        
        if( updated ) {
            this.dispatchEvent({name:glConversationReplyHandler.UPDATED});
        }
    },
    
    /**
     * Notification that the DOM was changed (add or remove)
     * If a relevant node, start a refresh
     * @function {} ?
     */
    _domChange: function(evt) {
        try {
            // Disconnect listener:
            this._eventsContainer.stopObserving(this._viewRoot, "DOMNodeInserted", this._domChange, true);
            
            // Set timeout to reconnect and refresh:
            var me = this;
            setTimeout( function(){me._domChangeTimeout()}, 500 );
            
        } catch(e) {
            giLogger.warn(e, 'glConversationReplyHandler._domChange', this._glGmail);
        }
    },
    
    _domChangeTimeout: function() {
        if( !this.isConnected() ) return;
        
        try {
            
            // Refresh view:
            this._refresh();
            
            // Reconnect listener:
            this._eventsContainer.observe(this._viewRoot, "DOMNodeInserted", this._domChange, true);
        } catch(e) {
            giLogger.warn(e, 'glConversationReplyHandler._domChangeTimeout', this._glGmail);
        }
    },

    /**
     * Take any element within a reply box and get the major portions of the reply
     * @function {Object} ?
     * @param {Element} el - Any element with the reply box
     * @return Format: {id:, elRoot:, elHeader:, elFull:}
     */
    createReplyData: function(el) {
        var elRoot = el;
        while( elRoot.parentNode && !elRoot.hasAttribute('gtdi-reply-id') ) elRoot = elRoot.parentNode;
        return {id: elRoot.getAttribute('gtdi-reply-id'), elRoot: elRoot, elHeader: elRoot.childNodes[0], elFull: elRoot.childNodes[1]};
    },
    
    
    /**
     * Retrieve the Send button for a reply, within the elRoot provided
     * @function {Element} ?
     * @param {Element} elRoot - The root of the reply
     * @return The Send button
     */
    getSendButton: function(elRoot) {
        var doc = this._glGmail.canvasDocument;
        return doc.evaluate("..//div[contains(@class, 'goog-container')]//div[@role='button'][1]", elRoot, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
    },
    
    /**
     * Retrieve the Send button for a reply, within the elRoot provided
     * @function {Element} ?
     * @param {Element} elRoot - The root of the reply
     * @return The Send button
     */
    getToField: function(elRoot) {
        var doc = this._ccbglGmail.canvasDocument;
        return doc.evaluate("..//textarea[@name='to']", elRoot, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
    },
    
    /**
     * Retrieve the Send button for a reply, within the elRoot provided
     * @function {Element} ?
     * @param {Element} elRoot - The root of the reply
     * @return The Send button
     */
    getCcField: function(elRoot) {
        var doc = this._glGmail.canvasDocument;
        return doc.evaluate("..//textarea[@name='cc']", elRoot, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
    },
    
    /**
     * Whether the UI object can be cached by {@link glGmail.getUIObject}
     * @function {Boolean} ?
     * @return Whether the UI can be cached
     */
    isCacheable: function() {
        return false;
    },
    

    //}
}, {
    
    
    /**
     * Fires when a reply box is added. e.g. Reply/Forward
     * @event
     */
    SHOW_REPLY_FULL: 'SHOW_REPLY_FULL',
    
    /**
     * Fires when a reply box is removed. e.g. Send/Discard.
     * @event
     */
    REMOVE_REPLY_FULL: 'REMOVE_REPLY_FULL',
    
    /**
     * Fires when a row is added (e.g. a new message, or an expand)
     * @event
     */
    ADD_REPLY_ROW: 'ADD_REPLY_ROW',
    
    /**
     * Fires when a reply row is removed (e.g. a message deleted)
     * @event
     */
    REMOVE_REPLY_ROW: 'REMOVE_REPLY_ROW',
    
    /**
     * Fires when the replies are updated for any reason (includes showing/contracting messages messages)
     * @event
     */
    UPDATED: 'updated',
    
});
glConversationReplyHandler.implement(giEventDispatcher); // Add event dispatching support

