/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The conversation view abstraction layer.
 *
 * @class ? @extends glUIView
 */
var glConversationView = glUIView.extend({
    // Properties {
    /**
     * The {@link glActionsHeader} object for the page
     * @property {glActionsHeader} ?
     */
    actionsHeader: null,
    
    
    /** @property {String} ? The thread ID of this conversation */
    threadId: null,
    
    /** @property {private glThread} ? Cached copy of thread associated with this conversation */
    _thread: null,
	
	
    /**
     * Cached reply handler for the conversation.
     * @property {private glConversationReplyHandler}
     */
	_replyHandler: null,
	
    
    /**
     * Cached title element (subject line)
     * @property {private Element}
     */
    _elTitle: null,
    
    /**
     * Cached main-pane element (contains email + sidebar)
     * @property {private Element}
     */
    _elMain: null,
    
    /**
     * Cached advertising box element
     * @property {private Element}
     */
    _elAdvertisingContainer: null,
	
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        giLogger.log('Connecting glConversationView object to view', 'glConversationView');
        this.base(gmail);

        this.actionsHeader = gmail.getUIObject('actionsHeader');
        
        var elRoot = this._glGmail.page.getViewRoot();
        var doc = gmail.canvasDocument;


        var m = gmail.canvasDocument.defaultView.window.top.location.href.match(/\/([^\/]+?)$/);
        if( m ) {
            if( m[1]=="#sent" ) {
                this.threadId = gmail.environment.lastSentThreadId;
            } else {
                this.threadId = m[1]? m[1] : null;
            }
        }
        
        // TODO For full data integrity, we should catch SEND_MESSAGE_REQUEST events from the TrafficWatcher and update the thread (so it's messages are correct). A potential alternative, depending on use case, is to listen to Reply rows being added.
        
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting glConversationView from view', 'glConversationView');
        
        //giLogger.warn('unloading glConversationView', 'glConversationView.disconnect');
        if( !this._glGmail ) return;    // TODO this function is being called twice on shutdown, need to work out why. 
        
    	this._glGmail.actions.removeEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
        
       
        if( this._replyHandler ) {
            this._replyHandler.disconnect();
            this._replyHandler = null;
        }
        
        this.threadId = null;
        this._thread = null;
        this.actionsHeader.disconnect();
        this.actionsHeader = null;
        
        
        // Clear all elements:
        for( p in this ) {
            if( this.hasOwnProperty(p) && this[p] && this[p].nodeName ) {
                this[p] = null;
            }
        }
        
        this.base();
    },

    /**
     * Return primed glRequestOriginalMsg object for the current thread
     * Recipient can add event listeners for COMPLETE and run it.
     * @function {glRequestOriginalMsg} ?
     */
    getOriginalMessageId: function() {
        return glRequestOriginalMsg.createRequest(this._glGmail, this.threadId);
    },
    
    /**
     * Return the reply handler for this conversation.
     * @function {glConversationReplyHandler} ?
     */
    getReplyHandler: function() {
        if( !this._replyHandler ) {
            this._replyHandler = this._glGmail.getUIObject('conversationReplyHandler');
        }
        return this._replyHandler;
    },
    
    /**
     * Retrieved (cacheable) element
     * @function {Element} ?
     * @return the email subject element
     */
    getTitle: function() {
        if( !this._elTitle ) {
            var doc = this._glGmail.canvasDocument;
            this._elTitle = doc.evaluate(".//h1[@class='ha']", this._glGmail.page.getViewRoot(), null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
            if( !this._elTitle ) this._elTitle = doc.evaluate(".//h1[@class='YfMhcb']", this._glGmail.page.getViewRoot(), null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        }
        return this._elTitle;
    },
    
    /**
     * Retrieved (cacheable) element
     * @function {Element} ?
     * @return the pane that contains the email content and sidebar
     */
    getMainPane: function() {
        if( !this._elMain ) {
            var doc = this._glGmail.canvasDocument;
            this._elMain = doc.evaluate("./div/table[contains(@class,'nH iY')]", this._glGmail.page.getViewRoot(), null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        }
        return this._elMain;
    },
    
    /**
     * Retrieved (cacheable) element
     * @function {Element} ?
     * @return the 
     */
    getAdvertisingContainer: function() {
        if( !this._elAdvertisingContainer ) {
            var doc = this._glGmail.canvasDocument;
            this._elAdvertisingContainer = doc.evaluate(".//td//div[@class='u5']", this._glGmail.page.getViewRoot(), null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        }
        return this._elAdvertisingContainer;
    },
    
    

    /**
    * Get the 'Newer' element - if it exists - from the navigation section
    * @function {Element} ?
    * @return The 'Newer' element
    */
    getMovePreviousElement: function() {
        var elNav = this._glGmail.canvasDocument.evaluate(".//div[@class=iG']/div[@class='h0']", this.actionsHeader.root, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        if( elNav.firstChild.nodeName=="SPAN" ) return elNav.firstChild; 	// <span><b><b><span>
    },

    /**
    * Get the 'Older' element - if it exists - from the navigation section
    * @function {Element} ?
    * @return The 'Older' element
    */
    getMoveNextElement: function() {
        var elNav = this._glGmail.canvasDocument.evaluate(".//div[@class='iG']/div[@class='h0']", this.actionsHeader.root, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        if( elNav.childNodes.length>0 && elNav.lastChild.nodeName=="SPAN" ) return elNav.lastChild; 	// <span><b><b><span>
    },

    /**
    * Get the thread associated with this view
    * @function {glThread} ?
    * @property {Boolean} doNotLoad - If true, and no thread, do not download one
    */
    getThread: function(doNotLoad) {
        if( !this._thread && !doNotLoad ) {
            if( this.threadId ) {
                this._thread = glThread.createFromId(this.threadId, this._glGmail);
                this._glGmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
            }
        }
        return this._thread;
    },
	
	
    /**
     * Callback for the {@link glActionManager.THREAD_CHANGED} event. 
     * @function ?
     * @param {Event} evt - The thread changed event
     */
    _threadChanged: function(evt) {
        // Find instances of the thread
        if( this._thread.threadId==evt.thread ) {
            this._thread.updateFromEvent(evt);
        }
    },
	

    
    /**
     * Retrieve an xpath iterator of label elements; each of which have the 'name' attribute that is the labelname.
     * @function {Iterator} ?
     * @param {Element} elH1 - The container
     * @return Xpath iterator
     */
    getHeaderLabels: function() {
        return this._glGmail.canvasDocument.evaluate(".//td/span[@class='hN']", this.getTitle(), null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
    },
    
    //}
});
