/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Abstraction layer for the labels box.
 *
 * @class ? @extends glUIWidget
 */
var glLabelsBox = glUIWidget.extend({
    // Properties {
    
    /** @property 
    /** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
    _eventsContainer: null,
    
    /** @property {private giMultiToOne} ? Reduce rapid function calls into a single one per Xms */
    _multiToOne: null,
    
    /** @property {Element} ? The container div for the labels box */
    _labelsBox: null,
    
    /** @property {Element} ? The main table for the labels box */
    _labelsBoxTable: null,
    
    /** @property {private Element} ? The element that doesn't change when the labels box refreshes */
    _observed: null,
    //}
	

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        giLogger.log('Connecting glLabelsBox object to view', 'glLabelsBox');

        this.base(gmail);
            
        this._eventsContainer = new giEventsContainer(this);
        this._multiToOne = new giMultiToOne(400);
    
    
        this._refresh();	
        giLogger.log('Connected glLabelsBox object to view', 'glLabelsBox');
		
    },

         

    /**
     * Event response for DOMNodeInserted on Labels box. Reduces rapid multiple calls to one and passes to _updatedBox1
     * @function {private}?
     * @param {Event} event - 
     */    
    _updatedBox: function(event) {
    	try {
            giLogger.log('glLabelsBox updated - domNodeInserted', 'glLabelsBox');
            if( this._observed ) this._eventsContainer.stopObserving( this._observed, "DOMNodeInserted", this._updatedBox, false);	// Stop observing to relieve pressure
            this._multiToOne.process("labelDomChange", giUtil.bind(this, this._refresh));	// Wait 400ms
        } catch(e) {
            giLogger.warn(e, 'glLabelsBox._updatedBox', this._glGmail);
        }
    },
    
    /**
     * Response to DOMNodeInserted. Reload Label box elements and broadcast event to listeners.
     * @function {private}?
     */
    _refresh: function() {
        giLogger.log('glLabelsBox updated (reduced)', 'glLabelsBox');

        // Invalidate Cache:
        this._labelsBox = null;
        this._labelsBoxTable = null;
        
        if(this.hasEventListener(glLabelsBox.LABELS_BOX_MODIFIED)) {
            this.dispatchEvent({name: glLabelsBox.LABELS_BOX_MODIFIED});
        }
        
        // Monitor future modifications:	
        // Refresh container
        this.getLabelsContainer();
        if( this._labelsBox ) {
            this._observed = this._labelsBox.parentNode.parentNode.parentNode;
            this._eventsContainer.observe( this._observed, "DOMNodeInserted", this._updatedBox, false);
        }
    },


    /**
    * Retrieve the labels container
    * @function {} ?
    */
    getLabelsContainer: function() {
        if( !this._labelsBox ) {
            var doc = this._glGmail.canvasDocument;
            var result = doc.evaluate( "//div[@class='nH']/div[@class='pP']/div[@class='pO']", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);
            if( !result.singleNodeValue ) result = doc.evaluate( "//div[@class='XoqCub XPj4ef']//div[@class='R7iiN c1norb']//div[@class='nb1Q2b']//div[@class='pvSW6e']", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);
            this._labelsBox = result.singleNodeValue;
        }
        return this._labelsBox;
    },
    
    /**
    * Retrieve the labels table
    * @function {} ?
    */
    getLabelsTable: function() {
        if( !this._labelsBoxTable ) {
            var doc = this._glGmail.canvasDocument;
            var result = doc.evaluate( "//div[@class='nH']/div[@class='pP']/div[@class='pO']//table", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);
            if( !result.singleNodeValue) result = doc.evaluate( "//div[@class='XoqCub XPj4ef']//div[@class='R7iiN c1norb']//div[@class='nb1Q2b']//div[@class='pvSW6e']//table[last()]", doc.body, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);
            this._labelsBoxTable = result.singleNodeValue;
        }
        return this._labelsBoxTable;
    },
    
    

    /**
     * Return label elements in an xpath iterator
     * @function ?
     */
    getLabelItems: function() {
        var doc = this._glGmail.canvasDocument;
        
        var result = doc.evaluate( "//a[@class='p9' and contains(@href,'#label')]", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        if( result.snapshotLength==0 ) result = doc.evaluate( "//div[@class='XoqCub XPj4ef']//div[@class='R7iiN c1norb']//div[@class='nb1Q2b']//div[@class='pvSW6e']//tr//a[@class='zD5BAe']", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        if( result.snapshotLength==0 ) result = doc.evaluate( "//div[@class='XoqCub XPj4ef']//div[@class='R7iiN c1norb']//div[@class='nb1Q2b']//div[@class='pvSW6e']//tr//span", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        return result;
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting glLabelsBox from view', 'glLabelsBox');
        this.labelsBox = null;
        this.labelsBoxTable = null;
        this._observed = null;
        this._eventsContainer.reset();
        this._multiToOne.reset();
	
        this.base();
    },
    //}
},{
    // Static Properties {

    /**
     * Dispatched when labels box is modified
     * @event ?
     */
    LABELS_BOX_MODIFIED: 'labels_box_modified',
    //}
});

glLabelsBox.implement(giEventDispatcher); // Add event dispatching support