/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The settings view abstraction layer.
 *
 * @class ? @extends giUIObject
 */
var glSettingsView = glUIView.extend({
    // Properties {
	
    /**
     * @property {Element}
     */
    _elRoot: null,
    
    /**
     * @property {giEventsContainer}
     */
    _eventsContainer: null,
    
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {glGmail} main - The glGmail object for the page
     */
    connect: function(main) {
        giLogger.log('Connecting glSettingsView object to view', 'glSettingsView');

        this.base(main);
        this._eventsContainer = new giEventsContainer(this);
        
        var doc = this._glGmail.canvasDocument;
        this._elRoot = this._glGmail.page.getViewRoot();
        
        var elTabs = doc.evaluate(".//div[contains(@class, 'nH f2')]/div[@class='nH fY']", this._elRoot, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        if( !elTabs ) {
            throw new Error('Could not find tabs header in Settings view');
        }
        
        // Attach click listener to tabs:
        this._eventsContainer.observe(elTabs, 'click', this._clickTabs, false);
        
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting glSettingsView from view', 'glSettingsView');
        
        this._eventsContainer.reset();
        this._eventsContainer = null;
        this.base();
    },
    
    /**
     * Retrieve tab elements for tab with given name
     * @function {Object} ?
     * @param {optional String} - name of tab, correlates to static properties (e.g. glSettingsView.LABELS_TAB)
     * @return Format: {tabName: , elTab:, elLink:, elContent:}
     */
    getTab: function(name) {
        
        
        var doc = this._glGmail.canvasDocument;
        if( !name ) {
            // Get the current tab
            var elTab = doc.evaluate(".//div[@class='fZ']/a", this._elRoot, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
            if( elTab ) {
                // Now extract the tail (in format, #settings/labels)
                var m = elTab.href.match(/settings\/(\w+)$/i);
                if( m ) {
                    name = m[1];
                }
            }
        }
        if( !name ) throw new Error('Could not find name of Settings tab');
        
        var tabData = {tabName: name};
        
        // Retrieve elTab:
        tabData.elLink = doc.evaluate(".//div[@class='fZ']/a[contains(@href, '#settings/"+name+"')]", this._elRoot, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        tabData.elTab = tabData.elLink.parentNode;
        
        // Find its index
        var el = tabData.elTab;
        var idx = 0;
        while( el.previousSibling ) {
            idx++;
            el = el.previousSibling;
        }
        
        // Use the index to find content node:
        var elContentContainer = tabData.elTab.parentNode.parentNode.childNodes[1];
        tabData.elContent = elContentContainer.childNodes[idx];
        
        return tabData;
    },
    
    /**
     * Notification that a tab changed.
     * Find out the selected tab using Xpath; extract the content; and fire an event
     * @function {}
     */
    _clickTabs: function() {
        try {
            // Get the selected tab:
            var evtData = this.getTab();
            evtData.name = glSettingsView.SELECT_TAB;
            this.dispatchEvent(evtData);
            
        } catch(e) {
            giLogger.warn(e, 'glSettingsView._click');
        }
    },
    

    //}
}, {
    
    /**
     * Fired when tab changes
     * @event
     */
    SELECT_TAB: 'select_tab',
    
    /**
     * The Labels tab (corresponds to href on tab)
     * @property {String}
     */
    LABELS_TAB: 'labels',
    
});
glSettingsView.implement(giEventDispatcher); // Add event dispatching support