/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The thread list view abstraction layer.
 *
 * @class ? @extends glUIView
 */
var glThreadListView = glUIView.extend({
    // Properties {
    /**
     * The {@link glActionsHeader} object for the page
     * @property {glActionsHeader} ?
     */
    actionsHeader: null,

    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        giLogger.log('Connecting glThreadListView object to view', 'glThreadListView');

        this.base(gmail);

        this.actionsHeader = gmail.getUIObject('actionsHeader');
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting glThreadListView from view', 'glThreadListView');
        this.actionsHeader.disconnect();
        this.base();
    },

    /**
     * Returns the starting offset, ending offset, and number of results in a hash
     * @function {Object} ?
     * @return Pagination information
     * @... {int} startingOffset - The starting thread offset
     * @... {int} endingOffset - The ending thread offset
     * @... {int} results - The number of threads in the result
     */
    getPaginationInfo: function() {
        // Get span that contains the info
        var result = this._glGmail.canvasDocument.evaluate(".//span/b", this.actionsHeader.pagination, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);
        var span = result.singleNodeValue.parentNode;
        var bNodes = span.getElementsByTagName('b');

        return {
            startingOffset: parseInt(bNodes[0].textContent),
            endingOffset: parseInt(bNodes[1].textContent),
            results: parseInt(bNodes[2].textContent),
        };
    },
    
    
    /**
     * Retrieve information about a thread row using just an el (from anywhere)
     * @function {Object} ?
     * @param {Element} el - An el within a threadlist, associated with a row
     * @return Format: {locationHash: the location hash, isMultiple: a multiple-inbox row, offset: the offset from the first row}
     */
    getOffsetInfo: function(el) {
        
        var data = {locationHash: null, isMultiple: false, offset: 0};
        // Get to the row:
        while( el.parentNode && el.nodeName!="TR" ) el = el.parentNode;
        if( el.nodeName=="TR" ) {
            
            // Now count back it's position
            while( el.previousSibling ) {
                data.offset++;
                el = el.previousSibling;
            }
            
        }
        
        
        // Now find the container:
        while( el.parentNode && !/nH (nK|nn)/.test(el.className) ) {
            el = el.parentNode;
        }
        if( el.className.indexOf('nH nK')>-1 ) {
            // It is a multiple inbox container
            data.isMultiple = true;
            // Now get the key:
            var elHeader = this._glGmail.canvasDocument.evaluate(".//div[@class='nJ A2']", el, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
            if( elHeader ) {
                var mLoc = elHeader.innerHTML.match(/\:(\s*)(.+)$/);
                data.locationHash = "#search/"+mLoc[2];
            } else {
                throw new Error('Could not find location header in Multiple Inbox container');
            }
        }else if( /nH (nn)/.test(el.className) ) {
            // It is a higher container, so not multiple inbox
            data.isMultiple = false;
            data.locationHash = this._glGmail.doc.location.hash;
        }
        
        
        return data;
    },

    /**
     * Returns the thread at the given offset
     * @function {glThread} ?
     * @param {Element} el - An el in the thread list, to identify it in multiple inboxes
     * @return The thread object - null if not available
     */
    getThread: function(el) {
        var paginationInfo = this.getPaginationInfo();
        var offsetInfo = this.getOffsetInfo(el);
        return this._glGmail.environment.threadListCache.getThread(offsetInfo.locationHash, paginationInfo.startingOffset - 1, offsetInfo.offset);
    },
    
    /**
     * Return an array of proper glThread objects for threads in the current view
     * @function {Object[]} ?
     * @param {optional Element} el - An el in the thread list, useful if want to focus on multiple inboxes.
     * @return An array of object/glThread thread data in the current view
     */
    getThreads: function(el) {
        var paginationInfo = this.getPaginationInfo();
        var locationHash = null;
        if( el ) {
            locationHash = this.getOffsetInfo(el).locationHash;
        } else {
            locationHash = this._glGmail.doc.location.hash;
        }
        return this._glGmail.environment.threadListCache.getThreads(locationHash, paginationInfo.startingOffset -1, paginationInfo.endingOffset -1);
    },
    
    /**
     * Returns the thread ID at the given offset
     * @function {String} ?
     * @param {Element} el - An el in the thread list
     * @return The thread ID - null if not available
     */
    getThreadId: function(el) {
        var paginationInfo = this.getPaginationInfo();
        var offsetInfo = this.getOffsetInfo(el);
        return this._glGmail.environment.threadListCache.getThreadId(offsetInfo.locationHash, paginationInfo.startingOffset - 1, offsetInfo.offset);
    },

    /**
     * @function {glThreadListCache} ?
     * @return The thread list cache
     */
    getThreadListCache: function() {
        return this._glGmail.environment.threadListCache;
    },


    /**
    * Return array of TABLE elements. The first is always the main thread list; remaining are 'multiple inbox' tables.
    * @function {Element[]} ?
    * @return Array of TABLE elements that represent thread lists
    */
    getThreadListTables: function() {
        var doc = this._glGmail.canvasDocument;
        var result = doc.evaluate( "//div[contains(@class,'glViewRoot')]//table[contains(@class, 'RnTrV')]", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        if( result.snapshotLength==0 ) result = doc.evaluate( "//div[contains(@class,'glViewRoot')]//table[contains(@class, 'cf zt')]", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);

        var tables = [];
        for( var i = 0; i < result.snapshotLength; i++ ) tables.push( result.snapshotItem(i) );
        return tables;
    },

    /**
    * Return xpath iterator of TD elements - the last in each row TR in the thread lists
    * @function ?
    * @return XPath iterator of TD elements
    */
    getLastTDInRows: function() {
        var doc = this._glGmail.canvasDocument;
        var result = doc.evaluate( "//div[contains(@class,'glViewRoot')]//tr[starts-with(@class,'zA')]//td[last()]", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        if( result.snapshotLength==0 ) result = doc.evaluate( "//div[contains(@class,'glViewRoot')]//div[contains(@class,'EGSDee')]/div/div[contains(@class,'AH8jVb')]//table[contains(@class,'N2mZ7b')]//tr[starts-with(@class,'xweT7d')]//td[last()]", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        return result;
    },
    
    /** 
    * Return offset of row element provided; starting at 0 for the first row in view. Ignores pagination.
    * function {int} ?
    * @param {Element} elRow - The row element (or any element within the TR) to get the offset for
    * @return The offset for provided row
    */
    /*
    getOffsetFromEl: function(elRow) {
        // Check it's a TR
        while( elRow.parentNode && elRow.nodeName!="TR" ) elRow = elRow.parentNode;
        if( elRow.nodeName!="TR" ) return null;
        
        // Now count back it's position
        var offset = 0;
        while( elRow.previousSibling ) {
            offset++;
            elRow = elRow.previousSibling;
        }
        return offset;
    },
    */
    
    /**
    * Uncheck all visible rows
    * function ?
    */
    uncheckRows: function() {
        var elTL = this._glGmail.page.getViewRoot();
        var elInputs = elTL.getElementsByTagName("INPUT");
        
        for( var i = 0; i < elInputs.length; i++ ) {
            if( elInputs[i].getAttribute("type").toLowerCase()=="checkbox" ) {
                if( elInputs[i].checked==true ) giUIObject.simulateMouseEvent(elInputs[i], "click");	// Uncheck it
            }
        }
    },
    
    /**
     * Check the row at the given offset (indifferent to pagination)
     * @function {} ?
     * @param {int} offset - The index (from 0) of the checkbox to check
     */
    checkRow: function(offset) {                
        var elTable = this.getThreadListTables()[0];
        var elInputs = elTable.getElementsByTagName("INPUT");
        if( elInputs[offset] ) {
            var el = elInputs[offset];
            if( el.checked==false ) giUIObject.simulateMouseEvent(el, "click");	// Check it
        }
    },
    
    
	
    //}
});
