/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Listen to clicks to load Explorer Popup
 *
 * @class gtdEnvironment
 */
var gtdEnvironment = giBase.extend({
    // Properties {
    /**
     * The gtdMainPagePlugin object for the page
     * @property {gtdMainPagePlugin} ?
     */
    _gtd: null,

	/** @property {giEventsContainer} ? */
	_eventsContainer: null,
	
	
	/** @property {private Boolean[]} ? The rows checked before we kick in */
	_lastCheckedRows: null,
	
	
	/** @property {private glThread} */
	_currentThread: null,
	
    //}

    /**
     * The constructor.
     * @constructor ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     */
    constructor: function(gtd) {
        this.base(gtd.gmail);
        this._gtd = gtd;
		this._eventsContainer = new giEventsContainer(this);
		this._attachedListeners = {};
		
		// Add click handler
		this._attachListeners();
		
		// Listen for view changes:
		this._gtd.gmail.page.addEventListener(glPageWrapper.VIEW_CHANGED, this._viewChanged, this);
    },

	/**
	* Cleanup memory usage
	* @function ?
	*/
	unload: function() {
	},

	/**
	* Try to listen for all available (and relevant) documents
	* @function {} ?
	*/
	_attachListeners: function() {
		if(!this._attachedListeners['main']) {
			this._eventsContainer.observe(this._gtd.gmail.canvasDocument.body, 'contextmenu', this._context, true);
			this._eventsContainer.observe(this._gtd.gmail.canvasDocument.body, 'mousedown', this._click, false);
			this._attachedListeners['main'] = true;
		}
		if(!this._attachedListeners['contact']) {
			if( this._gtd.gmail.canvasDocument.defaultView.top.frames[3].frames[0] ) {
				var b = this._gtd.gmail.canvasDocument.defaultView.top.frames[3].frames[0].document.body;
				this._eventsContainer.observe(b, 'contextmenu', this._context, true);
				this._eventsContainer.observe(b, 'mousedown', this._click, false);
				this._attachedListeners['contact'] = true;
			}
		}
	},

	/**
	* Notification the gmail view changed
	* @function {} ?
	*/
	_viewChanged: function(evt) {
		if( this._gtd.gmail.page.getActiveViewType()=="ct" ) {
			setTimeout(giUtil.bind(this, this._attachListeners), 2000);
			setTimeout(giUtil.bind(this, this._attachListeners), 6000);
		}
	},
	
	/**
	* Respond to an environment click
	* @function ?
	*/
	_click: function(event) {
		try {
			if( event.button!=2 ) return;
			var el = event.target;
			var details = this._getDetails(el);
			if( details ) {
				giUtil.stopEvent(event);
						
				if( !details.specificCmd ) {
                    
                    // Check for enabling:
                    if( details.type==gtdBrowser.TYPES.CONTACT ) {
                        if( !this._gtd.prefs.getPref('components.right_click.contact_popup.enabled') ) return;
                    } else {
                        // It's a label
                        if( !this._gtd.prefs.getPref('components.right_click.label_popup.enabled') ) return;
                    }
                    
					// Set the popup to use ThreadsViewer and grab a reference to it
                    var popup = this._gtd.getUIObject('browserPopup');
                    popup.load(gtdBrowser.createBrowser(this._gtd, details.item, details.type));
                    
				} else if( details.specificCmd=='loadTlThread' ) {
                    
                    if( !this._gtd.prefs.getPref('components.right_click.thread_preview.enabled') || this._gtd.gmail.offline.isUnstable() ) return;
                    
                    // Retrieve the gmail threadListView functionality
                    var threadListPreviewer = this._gtd.getUIObject('threadListPreviewer');
                    threadListPreviewer.load( el, event, false );
				}
		
			}
		} catch(e) {
			giLogger.warn(e, 'gtdEnvironment._click', this._gtd.gmail);
		}
	},
	
	
	/**
	* Respond to an environment context click
	* @function ?
	*/
	_context: function(event) {
		try {
			var el = event.target;
			
			var el2 = el;
			var i = 10;
			while( --i>=0 && el2.parentNode ) {			
				if( el2.className=='lightbox' || el2.id=='gtdi-browser-container' ) {
					giUtil.stopEvent(event);
					return;
				}
				el2 = el2.parentNode;
			}
			
			var details = this._getDetails(el);
			if( details ) giUtil.stopEvent(event);
		} catch(e) {
			giLogger.warn(e, 'gtdEnvironment._context', this._gtd.gmail);
		}
	},
	
	
	/**
	* Extract details - if any - about what was clicked
	* @function {Object} ?
	* @param {Element} el - The el clicked
	* @return {Object} The details. Format: {type:String, item:String, specificCmd:String}. Or null if no match.
	*/
	_getDetails: function(el) {
        var loopCount = 5;
        var viewType = this._gtd.gmail.page.getActiveViewType();
        if( viewType=="tl" ) loopCount = 10; // Heavily nested tables
        
        while( loopCount-->=0 && el.parentNode ) {
            // Test for Contact in Gmail Contact List
            if( el.nodeName=="SPAN" && (el.className=="vs" || el.className=="HHshnc") ) {
                return {type:gtdBrowser.TYPES.CONTACT, item: el.innerHTML};
            }
            
            // Test for Contact in Gmail Conversation
            // Test for Contact in Gmail ThreadList
            if( el.getAttribute("email") ) {
                return {type:gtdBrowser.TYPES.CONTACT, item: el.getAttribute("email")};
            }
            
            
            // Test for Label in Gmail Conversation
            if( el.nodeName=="DIV" && el.className=="av" ) {
                return {type:this._gtd.labelsData.getLabelType(el.innerHTML).toLowerCase(), item:el.innerHTML};
            }
            
            // Test for Label in Gmail ThreadList
            if( el.nodeName=="SPAN" && el.className=="hN" ) {
                return {type:this._gtd.labelsData.getLabelType(el.innerHTML).toLowerCase(), item:el.innerHTML};
            }
            
            // Test for Label in Gmail Labels Box
            if( el.nodeName=="A" ) {
                if( el.href.indexOf("#label/")>-1 ) {
                    var m = el.href.match("#label/([^/]+?)$");
                    if( m ) {
                        var label = decodeURIComponent(m[1]);
                        return {type:this._gtd.labelsData.getLabelType(label).toLowerCase(), item:label};
                    }
                }
            }
            
            // Test for Offical Type/Item format:
            if( el.getAttribute("gtdType") ) {
                return {type:el.getAttribute("gtdType"), item:el.getAttribute("gtdItem")};
            }
            
            // Test for Label in Labels Box
            if( el.getAttribute("gtdLabelName") && !el.getAttribute("gtdi-placeholder") ) {
                return {type:this._gtd.labelsData.getLabelType(el.getAttribute("gtdLabelName")).toLowerCase(), item:el.getAttribute("gtdLabelName")};
            }
            
            // Test for Contact in Gmail Contacts List
            if( el.className=='text' && giUtil.testEmailAddress(el.title) ) {
                return {type:gtdBrowser.TYPES.CONTACT, item: el.title};
            }
            
            // Test Email Link in Gmail Contacts List
            if( el.nodeName=='A' && giUtil.testEmailAddress(el.innerHTML) ) {
                return {type:gtdBrowser.TYPES.CONTACT, item: el.innerHTML};
            }		
            
            
            // Test for Row in Gmail ThreadList
            if( viewType=="tl" ) {
                if( el.nodeName=="TR" && el.className.indexOf('zA')>-1 ) {
                    return {specificCmd: 'loadTlThread'};
                }
            }
            
            el = el.parentNode;
        }
		
	},
	

    //}
});
