/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Store hierarchial label data. 
 * 
 * TODO Performance overhaul:
 * * Only build gtdLabelsTree on demand
 * * Teach .getLabelsArray et all to work purely off label data obtained from environment
 *
 * @class ?
 */
var gtdLabelsData = giBase.extend({
    // Properties {

    /** @property {private gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page  */
    _gtd: null,

    
    /** @property {private Object} ? Cache of gtdLabelsTree elements */
    _labelsTrees: null,
	
	/** @property {private Object} ? Collection of labels broken by type. Format  _labels[type] = String[] */
	_labels: null,
    
    /** @property {private String[]} ? Array of lower case labels String[] */
	_labelsLower: null,
	
    //}
	

	// Methods {
    

    /**
    * @constructor ?
    * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
    */
    constructor: function(gtd) {
        this.base(gtd.gmail);
        this._gtd = gtd;
        this._labelsTrees = {};
        this._labels = {};
        this._labelsLower = [];
        
        this._gtd.gmail.environment.addEventListener(glEnvironment.ENVIRONMENT_DATA_RECEIVED, this._refreshLabels, this);
        this._gtd.gmail.environment.addEventListener(glEnvironment.CREATE_LABEL, this._refreshLabels, this);
        this._gtd.gmail.environment.addEventListener(glEnvironment.RENAME_LABEL, this._refreshLabels, this);
        this._gtd.gmail.environment.addEventListener(glEnvironment.DELETE_LABEL, this._refreshLabels, this);
        this._gtd.gmail.environment.update();	// get latest labels
        this._refreshLabels(true);
    },
    
    /**
    * Clean up
    * @function ?
    */
    unload: function() {
        this.base();
        this._gtd = null;

        this._labelsBox.disconnect();
    },
    
    /** 
    * Clear out the gtdLabelsTree cache and rebuild them
    * @function {} ?
    */
    _refreshLabels: function(evt) {
        
        
        var singleLabelChange = false;
        if( evt.name==glEnvironment.ENVIRONMENT_DATA_RECEIVED ) {
            // First load - set up properly
            this._gtd.gmail.environment.removeEventListener(glEnvironment.ENVIRONMENT_DATA_RECEIVED, this._refreshLabels, this);
        } else if( evt.label ) {
            // A label was created/renamed/deleted -
            singleLabelChange = true;
        }
        
        // Clear out old values:
        this._labels = {};
        for( t in this._labelsTrees ) {
            this._labelsTrees[t].disconnect();
        }
        this._labelsTrees = {};
        
        // Prep:
        var typeSettings = this._gtd.prefs.getPref('labels.types');
        var prefixes = {};
        for( t in typeSettings ) if(t!='misc') prefixes[t] = new RegExp("^" + giUtil.prepRE(typeSettings[t].prefix), "i");
        
        var statusDefaultsLower = {};   // Format: {<statusLabelLower>: <statusLabel}
        for( d in typeSettings.statuses.core ) {
            var label = typeSettings.statuses.prefix + typeSettings.statuses.core[d];
            statusDefaultsLower[ label.toLowerCase() ] = {key: d, label:label};
        }
                
        // Process labels
        var type = null;
        var labels = this._gtd.gmail.environment.getUserLabels();
        var added = null;
        var changedStatuses = false;
        if( !singleLabelChange ) added = [];
        for( var i = 0; i < labels.length; i++ ) {
            var label = labels[i].name;
            if( !type || !prefixes[type] || !prefixes[type].test(label) ) {
                // Type has changed
                type = 'misc';
                for( t in typeSettings ) {
                    if( prefixes[t] && prefixes[t].test(label) ) {
                        type = t;
                        break;
                    }
                }
            }
            if( !this._labels[type] ) this._labels[type] = [];
            this._labels[type].push(label);
            this._labelsLower.push(label.toLowerCase());
            
            if( type=='statuses' ) {
                // Do a critical check to ensure our preferences match real labels (case-wise)
                if( statusDefaultsLower[label.toLowerCase()] && statusDefaultsLower[label.toLowerCase()].name!=label ) {
                    // Need to update the setting
                    typeSettings.statuses.core[statusDefaultsLower[label.toLowerCase()].key] = label.replace(typeSettings.statuses.prefix, '');
                    changedStatuses = true;
                }
            }
            
            if( !singleLabelChange ) added.push(label);
        }       
        
        
        if( changedStatuses ) {
            // Write out settings
            this._gtd.prefs.setPref('labels.types.statuses.core', typeSettings.statuses.core, true);
        }
        
        
        if( singleLabelChange ) {
            switch( evt.name ) {
                case glEnvironment.CREATE_LABEL:
                    this.dispatchEvent({name: gtdLabelsData.LABEL_CHANGED, added: [evt.label]});
                    break;
                case glEnvironment.RENAME_LABEL:
                    this.dispatchEvent({name: gtdLabelsData.LABEL_CHANGED, added: [evt.newLabel], removed: [evt.label]});
                    break;
                case glEnvironment.DELETE_LABEL:
                    this.dispatchEvent({name: gtdLabelsData.LABEL_CHANGED, removed: [evt.label]});
                    break;
            }
        } else {
            this.dispatchEvent({name: gtdLabelsData.LABEL_CHANGED, added: added});
        }
        
        
    },
	
	/**
	* Get the type for a given label
	* @function {String} ?
	* @param {String} labelname - The labelname
	* @param {optional String} type - The type to try first (used in label processing loops on assumption labels are grouped by category)
	* @param {optional Object} typeSettings - The typeSettings (again used in label processing loops as cached data)
	* @return {String} The type (e.g. statuses)
	*/
	getLabelType: function(labelname, type, typeSettings) {	// TODO This is generic, should probably live in gtdLabelsBox; which 'pure' (non UI) parts should be abstracted out for
        
		if( !typeSettings ) typeSettings = this._gtd.prefs.getPref('labels.types');
        if( type && type!="misc" && gtdLabelsData.testPrefix(labelname, typeSettings[type].prefix) ) {
            return type;
        }
		var type = "misc";
		for( t in typeSettings ) {
			if( t!="misc" && gtdLabelsData.testPrefix(labelname, typeSettings[t].prefix) ) {
				type = t;
				break;
			}
		}
		return t;
	},
	
	/**
    	* Clone a collection of labels (by type) within the typeBuckets data object (a collection of LabelData objects).
    	* @function ?
    	* @param {string} type - The type (collection of labels) to clone and return
    	* @return {Object} typeObject - Collection of gtdLabelDOM objects and container el for labels. Format: {el:, labels:{<labelname>:gtdLabelDOM}}
    	*/	
	getLabelsEl: function(type) {
	
		var doc = this._gtd.gmail.canvasDocument;
		
		if( this._labelsTrees[type] ) {
			return this._labelsTrees[type].clone();
		} else {
			this._labelsTrees[type] = this._gtd.getUIObject('labelsTree');
			this._labelsTrees[type].init(type);
            if( this._labels[type] ) {
                if( type=="statuses" ) {
                    var statuses = this.getLabelsArray("statuses", true, true);
                    var statusData = this.getStatusData(true);
                    for( var i = 0; i < statuses.length; i++ ) {
                        var key = statusData.reverseLookup[statuses[i]];
                        if( key!="finished" ) {
                            this._labelsTrees[type].addLabel( statuses[i] );
                        }
                    }
                } else {
                    for( var i = 0; i < this._labels[type].length; i++ ) this._labelsTrees[type].addLabel( this._labels[type][i] );
                }
                this._labelsTrees[type].finishAddingLabels();
            }
			return this._labelsTrees[type].clone();	
		}
        
	},
	
	
	/**
    	* Returns an array of labels for a given type (e.g. statuses, projects, etc.)
    	* @function ?
    	* @param {Array} types - The types to find the associated labels of. Can be '_all' (which is default if left blank), and returns all labels.
    	* @param {boolean} sortStatuses - If true and type is 'statuses', then returns core statuses first and in defined order.
    	* @param {boolean} removePlaceHolders - If true, ensures no place holder labels are returned (place holders are interims in hierarchy - e.g. for label 'p/gtdinbox/bugs', 'p/gtdinbox' is a place holder).
    	*/
	getLabelsArray: function(types, sortStatuses) {
	
		if( !types ) types = ["_all"];
		else if( typeof types=="string" ) {
			types = [types];
		}
		
		
		var labelsArray = [];
				
		if( types.length==1 ) {
			if( types[0]=="_all" ) {
				for( t in this._labels ) {
					labelsArray = labelsArray.concat(this._labels[t]);
				}
			} else {
				if( sortStatuses && types[0]=="statuses" ) {
					var statuses = this.getLabelsArray("statuses");
					var defaults = this._gtd.prefs.getPref('labels.types').statuses.core;
					var status = this._gtd.labelsData.getStatusData(true);
					for( d in defaults ) {	// Process the core in order
						for( var i = statuses.length - 1; i >=0; i-- ) {	// Try to find a match in known statuses
							if( statuses[i]==status[d] ) {	// It's a core status; add in order and remove from future consideration
								labelsArray.push(statuses[i]);
								statuses.splice(i, 1);
								break;
							}
						}
					}
					for( var i = 0; i < statuses.length; i++ ) labelsArray.push(statuses[i]);	// Add what statuses remain
				} else {
					if( this._labels[types[0]] ) {
						labelsArray = labelsArray.concat(this._labels[types[0]]);
					}
				}
			}
		} else {
			for( var i = 0; i < types.length; i++ ) {
				labelsArray = labelsArray.concat(this._labels[types[i]]);
			}
		}
		
		
		return labelsArray;
	
	},
	
	
	/**
    * Returns a hash of labels for a given type (e.g. statuses, projects, etc.)
    * @function ?
    * @param {Array} types - The types to find the associated labels of. Can be '_all' (which is default if left blank), and returns all labels.
    * @param {boolean} removePlaceHolders - If true, ensures no place holder labels are returned (place holders are interims in hierarchy - e.g. for label 'p/gtdinbox/bugs', 'p/gtdinbox' is a place holder).
    */
	getLabelsHash: function(types) {
		if( !types ) types = ["_all"];
		else if( typeof types=="string" ) types = [types];
		
		var labelsHash = {};
				
		if( types.length==1 ) {
			if( types[0]=="_all" ) {
				for( t in this._labels ) {
					giUtil.arrayToHash( this._labels[t], false, labelsHash );
				}
			} else {
				if( this._labels[types[0]] ) {
					giUtil.arrayToHash( this._labels[types[0]], false, labelsHash );
				}
			}
		} else {
			for( var i = 0; i < types.length; i++ ) {
				giUtil.arrayToHash( this._labels[types[i]], false, labelsHash );
			}
		}
		return labelsHash;
	},
	
	
	/**
    * Return true with the tested labelname exists
    * @function ?
    * @param {string} labelname - The labelname to test the existence of
    */
	hasLabel: function(labelname) {
		// TODO This is used quite a lot; could cache the labels found into a hash for faster recall. Also cache lower case labels.

        return (this._labelsLower.indexOf(labelname.toLowerCase())>-1);   // True if found
	},
	

	/**
    * Returns a Status object derived from the preferences - for tidier code.
    * @function ?
    * @param {boolean} incReverseLookup - If true, include a .reverseLookup object to find the name of 
    */
	getStatusData: function(incReverseLookup) {
		var typeSettings = this._gtd.prefs.getPref('labels.types');
		
		var status = {};	// Cache for tidier code
		status.prefix = typeSettings.statuses.prefix;
		status["today"] = status.prefix + typeSettings.statuses.core.today;
		status["next_action"] = status.prefix + typeSettings.statuses.core.next_action;
		status["action"] = status.prefix + typeSettings.statuses.core.action;
		status["waiting_on"] = status.prefix + typeSettings.statuses.core.waiting_on;
		status["some_day"] = status.prefix + typeSettings.statuses.core.some_day;
		status["finished"] = status.prefix + typeSettings.statuses.core.finished;
	
		if( incReverseLookup ) {
			status.reverseLookup = {};
			status.reverseLookup[status["today"]] = "today";
			status.reverseLookup[status["next_action"]] = "next_action";
			status.reverseLookup[status["action"]] = "action";
			status.reverseLookup[status["waiting_on"]] = "waiting_on";
			status.reverseLookup[status["some_day"]] = "some_day";
			status.reverseLookup[status["finished"]] = "finished";
		}
		return status;
	},	

},{
    // Static Properties {

    /**
     * Dispatched when new label detected (will not fire on first run)
     * @event ?
     */
    LABEL_CHANGED: 'label_changed',
    

    //}
    
    
    getPrefixRE: function(prefix) {
        return new RegExp("^" + giUtil.prepRE(prefix.toLowerCase()), "i");
    },
    
	testPrefix: function(labelname, prefix) {
		return gtdLabelsData.getPrefixRE(prefix).test(labelname);
	},
	
	removePrefix: function(labelname, prefix) {
		return labelname.replace( new RegExp("^" + giUtil.prepRE(prefix), "i"), "" );
	},
    
});

gtdLabelsData.implement(giEventDispatcher); // Add event dispatching support