/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * GTD plugin for the main Gmail page.
 *
 * @class ? @extends giMainPageUIPlugin
 */
var gtdMainPagePlugin = giMainPageUIPlugin.extend({
    // Properties {
    /**
    * Contains LabelsData
    * @property {gtdLabelsData} ?
    */
    labelsData: null,

    /** @property {gtdEnvironment} ? Listen to clicks */
    environment: null,
    
    /** @property {gtdThreadContainer} */
    statusThreadsContainer: null,
    
    /** @property {gtdThreadRules} */
    threadRules: null,
    
    /** @property {gtdRecent} */
    recent: null,

    /** @property {gtdPreLabeler} */
    preLabeler: null,

    /**
     * True if baseUIReady has completed
     * @property {Boolean} ?
     */
    _loaded: null,
    
    /**
     * True if VIEW_CHANGE fired becaose baseUIReady was complete
     * @property {Boolean} ?
     */
    _missedFirstViewChange: null,
    
    
    //}

    /**
     * The constructor.
     * @constructor ?
     */
    constructor: function() {
        giLogger.log('gtdMainPagePlugin object created', 'gtdMainPagePlugin');
        this.base();
    },

    /**
     * @function {} ?
     * @inheritdesc
     */
    load: function(accountData, pageBase) {
        giLogger.log('Loading gtdMainPagePlugin for gmail', 'gtdMainPagePlugin');
        this.base(accountData, pageBase, 'gtd');

        // Load the international properties for the entire gtd plugin
        giI18N.addStringBundle("gtd");

        // Hook up event listeners
        this.gmail.page.addEventListener(glPageWrapper.UI_READY, this._baseUIReady, this);
        this.gmail.page.addEventListener(glPageWrapper.VIEW_CHANGED, this._viewChanged, this);
        
    },

    /**
     * Called on page unload, it cleans up everything
     * @function ?
     */
    unload: function() {
        if( this.statusThreadsContainer ) {
            this.statusThreadsContainer.preventUnload = false;
            this.statusThreadsContainer.unload();
        }
		
        if( this.preLabeler ) this.preLabeler.unload();
        
        this.labelsData.unload();
        this.environment.unload();

        this.base();
    },


    
    /**
     * Returns a list of the required glGmail components
     * @function {Array} ?
     * @return The required glGmail components
     */
    getRequiredComponents: function() {
        return ['trafficWatcher'];
    },
    

    /**
     * Finishes all the loading that GTDGmail needs. A callback to the
     * {@link glPageWrapper.UI_READY} event.
     * @function ?
     * @param {Event} evt - The UI_READY event
     */
    _baseUIReady: function(evt) {
        
        var loadError = null;
        
        // Remove event listener
        this.gmail.page.removeEventListener(glPageWrapper.UI_READY, this._baseUIReady, this); 

            
        this.gmail.fb = this.gmail.canvasDocument.defaultView.wrappedJSObject.top.console;
        
        
        
        try {
            // Do integrity check:
            if( !this.prefs.getPref('labels.types.statuses.core') || !this.prefs.getPref('labels.types.statuses.prefix') ) {
                // Statuses appear corrupted, quickly repair
                giLogger.log("Status preferences integrity check", "gtdMainPagePlugin._baseUIReady");
                this.prefs.extendPluginPrefsWithDefaults();
            }
            
            
            // Add the CSS // TODO make part of plugin structure
            this.gmail.page.addCSS("chrome://gtdinbox/skin/gtd/c/main.css"); // TODO add to manifest.xml
        
            // Load the Labels data:
            this.labelsData = new gtdLabelsData(this);
            
            // Load the environment box:
            this.environment = new gtdEnvironment(this);
            
            // Load the thread rules
            this.threadRules = new gtdThreadRules(this);
    
            // Load the recent
            this.recent = new gtdRecent(this);
    
            this.preLabeler = new gtdPreLabeler(this);
                        
            // Load control of the gmail labels:
            if( this.prefs.getPref("components.navigation.gmail_labels_box.enabled") ) {
                this.gmailLabelsBox = this.getUIObject("gmailLabelsBox");
            }
    
            // Load the threadContainer for statuses
            this.statusThreadsContainer = new gtdStatusThreadContainer(this);
    
    
            this.views = new gtdViews(this);
    
            // Load gtdMainView, which loads all the rest of the GTD widgets
            var mainView = this.getUIObject('mainView');

        } catch(e) {
            if( !loadError ) loadError = e;
            giLogger.warn(e, 'gtdMainPagePlugin._baseUIReady - main load section');
        }

        try {
            // See if the box plugin exists:
            var box = giPageManager.getPluginInstance(boxMain, this.gmail);
            if( box ) {
                // Box exists. 
                // See if it is loaded
                if( box.isLoaded() ) {
                    this._addGtdToBox(box);
                } else {
                    // Add a call back for when it loads
                    box.addEventListener(boxMain.LOADED, function(evt){ this._addGtdToBox(evt.target) }, this);
                }
            }
        } catch(e) {
            if( !loadError ) loadError = e;
            giLogger.warn(e, 'gtdMainPagePlugin._baseUIReady, box load fail');
        }

        // Try welcome message:
        if( !this.prefs.getPref('shown_welcome') ) {
            // TODO Make this doubly robust by writing a 2nd setting less dependant upon username loading correctly. Nothing is more irritating than a repeated welcome message.
            this.prefs.setPref('shown_welcome', true, true);
            try {
                this.getUIObject('welcome');
            } catch(e) {
                // Likely to fail on linux as it involves local files
                giLogger.warn(e, 'gtdMainPagePlugin._baseUIReady - welcome');
            }
        }
        
        try {
            // Update load status and check flow:
            this._loaded = true;
            if( this._missedFirstViewChange ) {
                // Fired a VIEW_CHANGE while we were loading, which will have been denied. Let it run now.
                this._viewChanged();
            }
        } catch(e) {
            if( !loadError ) loadError = e;
            giLogger.warn(e, 'gtdMainPagePlugin._baseUIReady - viewChanged');
        }
        
        if( loadError ) {
            if( box && box.showLoadError ) {
                box.showLoadError( "Error during gtdMainPagePlugin load:\n" + loadError.toString() );
            }
        }
        
    },
    
    /**
     * Add the GTD sections to the Box
     * @function {} ?
     * @param {boxMain} box
     */
    _addGtdToBox: function(box) {

        // Add Prefs warning if required
        if( !this.prefs.getPref('labels.types.statuses.core') || !this.prefs.getPref('labels.types.statuses.prefix') ) {
            var boxPrefsWarning = this.getUIObject('boxPrefsWarning');
            box.getUIObject('box').addSection(boxPrefsWarning, true);
        }
        
        // Add Dashboard popout icon
        var elDash = this.gmail.canvasDocument.createElement("IMG");
        elDash.className = "gtdi-box-dashboard-popout";
        elDash.src = "chrome://gtdinbox/skin/gtd/popout.PNG";
        elDash.title = giI18N.getString('Box.DashboardShortcut.title');
        var me = this;
        elDash.addEventListener('mousedown', function(evt){me._clickDashPopout(evt)}, true);
        box.getUIObject('box').elHeader.appendChild( elDash );

        // Add setup box if required:
        if( !this.prefs.getPref('components.navigation.gtd_setup.disabled') ) {
            // Set if any status labels missing:
            var sd = this.labelsData.getStatusData();
            var statuses = this.labelsData.getLabelsHash('statuses');
            if( !(statuses[sd.action] && statuses[sd['waiting_on']] && statuses[sd['waiting_on']] && statuses[sd['some_day']] && statuses[sd.finished]) ) {
                var boxSetup = this.getUIObject('boxSetup');
                box.getUIObject('box').addSection(boxSetup, true);
            }
        }
        
        // Add Stats Day. TODO - make optional (and make ToDo within it optional)
        /*
        var boxDayStats = this.getUIObject('statsDayBox');
        box.getUIObject('box').addSection(boxDayStats, true);
        */
        
        // Add box shortcuts:
        if( this.prefs.getPref('components.navigation.types.enabled') ) {
            var boxShortcuts = this.getUIObject('boxShortcuts');
            box.getUIObject('box').addSection(boxShortcuts, true);
        }
        
        // Add box types:
        if( this.prefs.getPref('components.navigation.shortcuts.enabled') ) {
            var boxTypes = this.getUIObject('boxTypes');
            box.getUIObject('box').addSection(boxTypes, true);
        }

    },
    
    /**
     * Load the dashboard
     * @function {} ?
     */
    _clickDashPopout: function(evt) {
        // TODO this was copied and pasted from gtdBox.js - really should have a helper function. Probably best here.
        
        try {
            giUtil.stopEvent(evt);
            // Load the Dashboard
            var popup = this.getUIObject('browserPopup');
            popup.load(gtdBrowser.createBrowser(this));
            
        } catch(e) {
            giLogger.warn(e, 'gtdMainPagePlugin._clickDashboardPopout', this.gmail);
        }
    },
    

    /**
     * Triggered by the {@link glPageWrapper.VIEW_CHANGED} event, it calls
     * getCurrentViewObject to make sure it gets connected.
     * @param {Event} evt - The VIEW_CHANGED event
     */
    _viewChanged: function(evt) {
        
        // Update load status and check flow:
        if( !this._loaded ) {
            this._missedFirstViewChange = false;
            return; // Cannot run until base is loaded. _baseUIReady will ensure this runs when it has loaded.
        }
        
        this.base(evt);
    },
    
    
    
    //}
});