/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd. *
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * The GTD Pre-labeler is responsible for labeling new e-mails after the environment
 * has been updated with the thread id of the outgoing e-mail. When the user clicks
 * send on the compose page, the pre-labeler is notified that the next e-mail should
 * be labeled a certain way, and when it gets the thread id, it labels the e-mail.
 *
 * @class ?
 */
var gtdPreLabeler = giBase.extend({
    // Properties {
    /**
     * The gtdMainPagePlugin object for the page
     * @property {gtdMainPagePlugin} ?
     */
    _gtd: null,

    /**
     * Label request stack
     * @property {Array} ?
     */
    _labelRequests: null,

     /**
      * If there is a setTimeout in process for running the processing of queued
      * labeling requests.
      * @property {Boolean} ?
      */
    _processingScheduled: false,

    /**
     * The label requests to process stack
     * @property {Array} ?
     */
    _toProcessLabelRequests: null,
    
    //}

    /**
     * The constructor.
     * Initializes the object and adds an event listener to glEnvironment so
     * that it is notified about evironment changes.
     * @constructor ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     */
    constructor: function(gtd) {
        this._gtd = gtd;
        this._labelRequests = [];
        this._toProcessLabelRequests = [];

        this._gtd.gmail.environment.addEventListener(glEnvironment.EMAIL_SENT, this._emailSent, this);
        giLogger.log('Connected to the environment', 'gtdPreLabeler');
    },
    
    unload: function() {
        this._gtd = null;
        this._labelRequests = [];
        this._toProcessLabelRequests = [];
    },

    // Methods {
    /**
     * Adds a labeling request to the stack of labeling (although there should only
     * ever be one request on the stack)
     * @function ?
     * @param {String[]} labels - The labels to add to the sent e-mail
     */
    labelNextSent: function(labels) {
        this._labelRequests.length = 0; // Only one at once.
        this._labelRequests.push(labels);
    },

    /**
     * Callback for sent email from the environment. Schedules the delayed labeler
     * to label the thread with the next thread label request if it exists.
     * @function ?
     * @param {Event} evt - The {@link glEnvironment.EMAIL_SENT} event
     */
    _emailSent: function(evt) {
        if(!this._labelRequests.length) return;

        giLogger.log('Email sent with thread id of "'+evt.threadId+'" - labeling...', 'gtdPreLabeler');

        var labels = this._labelRequests.shift();
        this._toProcessLabelRequests.push({labels: labels, threadId: evt.threadId});

        if(this._toProcessLabelRequests.length && !this._processingScheduled) {
            this._processingScheduled = true;
            var me = this;
            setTimeout(function() { me._processLabelRequests(); me._processingScheduled = false; }, 100);
        }
    },

    /**
     * The delayed labeler scheduled in <code>_emailSent</code>. Goes through the
     * <code>_toProcessLabelRequests</code> array and performs all the actions.
     * @function ?
     */
    _processLabelRequests: function() {
        // Queue up actions in actionBuffer
        for each(var request in this._toProcessLabelRequests) {
            for each(var label in request.labels) {
                this._gtd.gmail.actions.performAction(glActionManager.ADD_LABEL, request.threadId, glActionManager.USE_AJAX, label);
            }
        }
        this._toProcessLabelRequests = [];
    },
    
    //}
});
