/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Extends the gtdThreadContainer to include message count functionality
 *
 * @class ?
 */
var gtdStatusThreadContainer = gtdThreadContainer.extend({
    // Properties {
    
    /**
     * Hash of labels, and the number of threads they appear on
     * Format: {<label>:<count>}
     * @property {int{}}
     */
    _labelCount: null,
    
    _isCached: false,
    
    _typeActive: null,
    
    //}

    // Methods {
    
    /**
    * The Construcotr
    * Hook event listeners in
    */
	constructor: function(gtd) {
        
        // Create the search and load the base threadContainer
        this._gtd = gtd;
        var search = this._createSearch();
        this.base(gtd, search);
        
        // Ensure no one tries to clean us up:
        this.preventUnload = true;
        
        this._typeActive = {};
        
        // Add listeners:
        this.addEventListener(gtdThreadContainer.NEW_SEARCH, this._tcNewSearch, this);
        this.addEventListener(gtdThreadContainer.LOADING_UPDATE, this._tcLoadingUpdate, this);
        this.addEventListener(gtdThreadContainer.LOADED, this._tcLoaded, this);
        this.addEventListener(gtdThreadContainer.CHANGED, this._tcChanged, this);
        this._gtd.gmail.environment.addEventListener(glEnvironment.CREATE_LABEL, this._envLabelChanged, this);
        this._gtd.gmail.environment.addEventListener(glEnvironment.RENAME_LABEL, this._envLabelChanged, this);
        this._gtd.gmail.environment.addEventListener(glEnvironment.DELETE_LABEL, this._envLabelChanged, this);
        
        // Start the search automatically upon construction:
        this.search();
	},
    

    /**
     * Disconnects the thread container from the action manager and cleans up
     * all data.
	 * As threadContainers can be shared; it is not impossible this function could be called multiple times on the same object.
     * @function ?
     */
    unload: function() {
        this.removeEventListener(gtdThreadContainer.NEW_SEARCH, this._tcNewSearch, this);
        this.removeEventListener(gtdThreadContainer.LOADING_UPDATE, this._tcLoadingUpdate, this);
        this.removeEventListener(gtdThreadContainer.LOADED, this._tcLoaded, this);
        this._gtd.gmail.environment.removeEventListener(glEnvironment.CREATE_LABEL, this._envLabelChanged, this);
        this._gtd.gmail.environment.removeEventListener(glEnvironment.RENAME_LABEL, this._envLabelChanged, this);
        this._gtd.gmail.environment.removeEventListener(glEnvironment.DELETE_LABEL, this._envLabelChanged, this);
        
		this.base();
    },

    /**
     * Create a search object based on Status labels in environment
     * @function {glSearch}
     * @return An instantiated glSearch object with a query for all active status threads
     */
    _createSearch: function() {
        // Create the query string:
        var typeSettings = this._gtd.prefs.getPref('labels.types');
        var status = this._gtd.labelsData.getStatusData();
        var statuses = this._gtd.labelsData.getLabelsArray("statuses", false, true);
        var orConditions = new glSearchOrConditions();
        var seenStatuses = {};
        
        // Include all core statuses - ensures we have a tight query even if user does not have Status labels created
        for( c in typeSettings.statuses.core ) {
            if( c!="finished" ) {
                seenStatuses[status[c]] = true;
                orConditions.addCondition(glSearch.l(status[c]))
            }
        }
        // Get any unique statuses
        for( var i = 0; i < statuses.length; i++) {
            if( statuses[i]!=status.finished && !seenStatuses[statuses[i]] ) {
                orConditions.addCondition(glSearch.l(statuses[i]));
            }
        }
        var search = new glSearch(this._gtd.gmail);
        search.addCondition(orConditions);
        search.addCondition('-'+glSearch.l(status.finished));
        
        giLogger.log('query string\n' + search.getQuery(), 'gtdStatusThreadContainer');
        
                
        return search;
    },
    
    /**
     * If the label changed was a Status, reload the query
     * @function
     */
    _envLabelChanged: function(evt) {
        var prefix = this._gtd.prefs.getPref('labels.types.statuses.prefix');
        if( gtdLabelsData.testPrefix(evt.label, prefix) ) {
            // It's a status label
            this.update(this._createSearch());
        }
        this._clearCache();
    },

    /**
     * Notification a new search is beginning
     * Reset variables
     * @function {}
     */
    _tcNewSearch: function(evt) {
        this._clearCache();
    },

    /**
     * Notification that some threads have been loaded
     * Extract label counts of new items
     * @function {}
     */
    _tcLoadingUpdate: function(evt) {
        this._clearCache();
    },
    
    /**
     * Notification that all threads have loaded
     * Extract label counts of new items
     * @function {}
     */
    _tcLoaded: function(evt) {
        this._clearCache();
    },
    
    /**
     * Notification that a thread(s) has changed - added/updated or removed
     * Update the counts
     * @function {}
     */
    _tcChanged: function(evt) {
        this._clearCache();
    },
    
    /**
     * Notify that we've cleared the cache
     * @function {} ?
     */
    _clearCache: function() {
        this._isCached = false;
        
    },
    
    /**
     * Extract label count from an array of threads
     * Assume we haven't seen before (i.e. the calling function takes care of it)
     * @function {} ?
     */
    _extractCountFromThreads: function() {
        if( this._isCached ) return;
        
        // Reset:
        this._labelCount = {};
        this._isCached = false;
        
        // Iterate threads:
        var threads = this.threads;
        var labelCount = this._labelCount;
        var labels = null;
        var threadIdx = threads.length;
        while(threadIdx--){
            // Iterate labels on each thread:
            labels = threads[threadIdx].labels;
            for( var j = 0; j < labels.length; j++ ) {
                if( labels[j][0]!="^" ) {
                    // Increment count for each label:
                    if( labelCount[labels[j]] ) {
                        labelCount[labels[j]]++;
                    } else {
                        labelCount[labels[j]] = 1;
                    }
                }
            }
        }
    },
    
    
    
    /**
     * Return count of all status items
     * @function {int} ?
     * @return total count of threads with a status
     */
    getTotalCount: function() {
        return this.threads.length;
    },
    
    /**
     * Return count of all threads with both a status label and the given label
     * @function {int} ?
     * @param {String} label - The label to find. Can be Status too.
     * @return total count of threads that have a status, and this label
     */
    getLabelCount: function(label) {
        this._extractCountFromThreads();    // Refresh if required
        return this._labelCount[label] || 0;
    },
    
    /**
     * Return the labelCount object
     * @function {} ?
     * @param {Boolean} safeClone - If true, clone it first so no original data can be damaged
     */
    getLabelCountObject: function(safeClone) {
        this._extractCountFromThreads();    // Refresh if required
        if( safeClone ) {
            var c = {};
            for( l in this._labelCount ) c[l] = this._labelCount[l];
            return c;
        } else {
            return this._labelCount;
        }
    },
    
    /**
     * Return true if the type (e.g. projects) has an active item
     * @function {Boolean} ?
     * @return True if label with statuses
     */
    isTypeActive: function(type) {
        
        // Test cache first:
        // Works by caching the last found thread and label for the type. If that thread is still in container; and the thread still has the label; we can assume this type is still active.
        var hadLabel = false;
        if( this._typeActive[type] && this._labelCount ) {
            if( this.threads.indexOf(this._typeActive[type].thread)>-1 && this._typeActive[type].thread.labels.indexOf(this._typeActive[type].label)>-1 ) {
                // Still has label:
                return true;
            } else {
                // No longer accurate, but did have a label last time we checked:
                this._typeActive[type] = null;
                hadLabel = true;
            }
        }
        
        
        // Now find the first label of type with an active item:
        this._extractCountFromThreads();  // Refresh if required
        if( type=="misc" ) {
            // Requires us to rule out all but a misc
            var typeSettings = this._gtd.prefs.getPref('labels.types');
            var labelsData = this._gtd.labelsData;
            var t = null;
            
            // Process threads (and their labels)
            var labels = null;
            var threadIdx = this.threads.length;
            while(threadIdx--) {
                labels = this.threads[threadIdx].labels;
                for( var i = 0; i < labels.length; i++ ) {
                    if( labels[i][0]!="^" ) {
                        t = labelsData.getLabelType(labels[i], t, typeSettings);
                        if( t=="misc" ) {
                            // Found a misc
                            this._typeActive[type] = {thread: this.threads[threadIdx], label: labels[i]};   // Cache
                            return true;
                        }
                    }
                }
            }
            
        } else {
            var prefix = this._gtd.prefs.getPref('labels.types.'+type+'.prefix');
            var re = gtdLabelsData.getPrefixRE(prefix);
            
            // Process threads (and their labels)
            var labels = null;
            var threadIdx = this.threads.length;
            while(threadIdx--) {
                labels = this.threads[threadIdx].labels;
                for( var i = 0; i < labels.length; i++ ) {
                    if( re.test(labels[i]) ) {
                        // Found the first label of our type
                        this._typeActive[type] = {thread: this.threads[threadIdx], label: labels[i]};   // Cache
                        return true;
                    }
                }
            }
        }
        
        // No active labels matched
        if( hadLabel ) {
            this.dispatchEvent({name: gtdStatusThreadContainer.LABEL_ZERO, type: type});
        }
        return false;
        
    },

    //}
}, {
    
    
    /**
     * Notification that a label reached zero
     * @event
     */
    LABEL_ZERO: 'label_zero',
});