/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd. *
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Container for managing a set of threads and keeping them up to date with
 * various conditions.
 *
 * @class ?
 */
var gtdThreadContainer = giBase.extend({
    // Properties {
    /**
     * The threads in the thread container
     * @property {glThread[]} ?
     */
    threads: null,
    
	/**
	* If true, stop the unload mechanism working. Used for the shared statuses container; which may try to be unloaded like other containers.
	* @property {Boolean} ?
	*/
	preventUnload: false,
	
    /**
     * The gtdMainPagePlugin object for the page
     * @property {gtdMainPagePlugin} ?
     */
    _gtd: null,

    /**
     * The search object that backs the thread container
     * @property {glSearch} ?
     */
    _search: null,

	/**
	* Whether the thread container is doing its current search
	* @property {Boolean} ?
	*/
	_searching: false,

	/**
	* Whether there are any more threads
	* @property {Boolean} ?
	*/
	_complete: false,
    
	
	/**
	* If true, stop searching after the next search callback
	* @property {Boolean}
	*/	
	_cancel: null,

	/**
	* The desired number of threads, that glSearch works towards.
	* @property {int}
	*/
	_totalDesired: null,
	

    /**
     * Reference to the scheduled setTimeout for the update
     * @property {int} ?
     */
    _updateTimeout: null,
    //}

    // Methods {
    
    /**
    * The Construcotr
    * Hook event listeners in
    */
	constructor: function(gtd, search) {
        // Initialize data
        this._gtd = gtd;
        this._search = search;
        this.threads = [];

        // Connect to action manager
        this._gtd.gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
        this._gtd.gmail.actions.addEventListener(glActionManager.MESSAGE_CHANGED, this._messageChanged, this);

		// Search
	    this._search.addEventListener(glSearch.COMPLETED, this._searchComplete, this);
	},
    


    /**
     * Disconnects the thread container from the action manager and cleans up
     * all data.
	 * As threadContainers can be shared; it is not impossible this function could be called multiple times on the same object.
     * @function ?
     */
    unload: function() {
		if( this.preventUnload ) return;
        if( this.threads ) this.threads = null;
        this._gtd.gmail.actions.removeEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
        this._gtd.gmail.actions.removeEventListener(glActionManager.MESSAGE_CHANGED, this._messageChanged, this);
		if( this._search ) {
        	this._search.removeEventListener(glSearch.COMPLETED, this._searchComplete, this);
        	this._search = null;
		}
    },


	/**
	* Search until we have the number of requested threads.
	* @function ?
	* @param {int} total - the total (as in .threads.length) number of threads desired. If empty, download all. 
	* @param {optional int} num - packet size of search (defaults to 25)
	*/
	search: function(total, num) {
		num = num || 25;
		total = total || gtdThreadContainer.MAX_THREADS_ALLOWED;	// Anymore than max and serious risk of account lock out
		if( num > total ) num = total;
				
				
		// Check if need to continue
		if( this._searching ) return;	// in progress 
		if( this._complete || this.threads.length >= total ) {
			// TODO some kind of notification here? Unwise to be LOADED as a handler my call .search() again on the back of a load; causing a loop.
			return;
		}
		
		this._totalDesired = total;
		this._cancel = false;	// ensure we don't impede request
		this._searching = true;
		
		// TODO restrict num if remaining threads < num. Also on searchComplete
		this.dispatchEvent({ name: gtdThreadContainer.NEW_SEARCH });
		this._search.runBackground( this.threads.length, num, {'num':num} );
	},
	
	/**
	* Callback for search completion. Updates internal threads data from search.
	* If this is the first call, it also dispatches the LOADED event and sets
	* the loaded status to true.
	* @function ?
	*/
	_searchComplete: function(event) {
		giLogger.log('Search completed', 'gtdThreadContainer');
    
        // Extract the threads
        var threads = this._search.getResults();
        // Extract the packet size
        var num = event.contextData.num;
        
        // Add the threads to our collection
        this.threads = this.threads.concat(threads);
        
        // See if we have run out of threads, exceeded our requirements, or it's cancelled:
        if( (threads.length < num) || (this.threads.length >= this._totalDesired) || this._cancel ) {
            // Notify that we have LOADED
            this._searching = false;
            if(threads.length < num) {
                this._complete = true;
            }
            this.dispatchEvent({ name: gtdThreadContainer.LOADED, newThreads: threads });
        } else {
            // Notify that we have LOADING_UPDATE
            this.dispatchEvent( {name: gtdThreadContainer.LOADING_UPDATE, newThreads: threads} );
            // Search again - grab the next packet of threads
            // TODO Need a means to know if the xmlhttprequest call failed (e.g. no connectivity), and if so, to stop recurring
            
            var me = this;
            setTimeout( function(){me._searchCompleteRunNext(num, {'num':num})}, 500 );
            //this._search.runBackground( this.threads.length, num, {'num':num} );
        }
        this._cancel = false;	// ensure we don't impede future requests
	},
    _searchCompleteRunNext: function(num, contextData) {
        this._search.runBackground( this.threads.length, num, contextData );
    },
	
	/**
	* Get the number of threads
	* Depending on type, retrieve number of actual threads we have data for; the number of threads Gmail reports for this search; and the number we are downloading in the current batch
	* @function {int} ?
	* @param {String} type - The information required
	* @return The number of threads
	*/
	getNumberThreads: function(type) {
		switch(type) {
			case gtdThreadContainer.COUNT_TOTAL_DESIRED:
				return this._totalDesired;
				break;
			case gtdThreadContainer.COUNT_TOTAL_CALCULATED:
				return this._search.getNumberResults();
				break;
			default:
				return this.threads.length;
		}
	},
	
	/**
	* Cancel an inprogress search
	* @function ?
	*/
	cancelSearch: function() {
		this._cancel = true;
	},
	
	/**
	* Return true if the thread container has finished it's initial loading
	* @function ?
	* @return {boolean} - True if finished loading
	*/
	isSearching: function() {
		return this._searching;
	},
	
	/**
	* Return true if all threads are downloaded
	* @function ?
	* @return {boolean} - True if all threads downloaded
	*/
	isComplete: function() {
		return this._complete;
	},

    /**
     * Schedules an update for 100 milliseconds from now.
     * @function ?
     * @param {optional glSearch} newSearch - If provided, replace the current search object with this one
     */
    update: function(newSearch) {
        giLogger.log('Update scheduled', 'gtdThreadContainer');
        if( newSearch ) {
        	if( this._search ) {
                this._search.removeEventListener(glSearch.COMPLETED, this._searchComplete, this);
                this._search = null;
        	}
        	this._search = newSearch;
       	    this._search.addEventListener(glSearch.COMPLETED, this._searchComplete, this);
        }
        var me = this;
        clearTimeout(this._updateTimeout);
        this._updateTimeout = setTimeout(giUtil.bind(this, function() {
            try {
                giLogger.log('Update running', 'gtdThreadContainer');
                this.threads.length = 0; 
                this._complete = false;
                this.search();
            } catch(e) {
                giLogger.warn(e, 'gtdThreadContainer.update', this._gtd.gmail);
            }
        }), 100);
    },

    

    /**
     * Callback for the {@link glActionManager.THREAD_CHANGED} event. Updates all
     * stored threads from the event and removes any that no longer fit the search
     * criteria.
     * @function ?
     * @param {Event} evt - The thread changed event
     */
    _threadChanged: function(evt) {
        try { // TODO possiblyh remove
     

            var addedThreads = [];  // Threads that now match the search, where they may not have before
            var updatedThreads = []; // Threads that have been added, removed or in some way modified by this update (e.g. a label added - even if not a key part of criteria)
            var removedThreads = []; // Threads that have been removed
            var thread = null; // The matching thread in our container
                        
            // Look for the thread in our own collection.
            // Must always do this, as only way to know if we're adding the thread.
            for(var i = 0; i < this.threads.length; i++) {
                if( evt.thread==this.threads[i].threadId ) {
                    thread = this.threads[i];
                    break;
                }
            }
            
            if( thread ) {
                // Thread in container
                
                // Found it in our own collection or cache. Update it.
                var matchError = false;
                if(thread.updateFromEvent(evt)) {
                    // Track the change:
                    updatedThreads.push(thread);
                    
                    // Remove if it doesn't match the search anymore or is deleted
                    try {
                        if(thread.isDeleted || !this._search.matchThread(thread)) {
                            // Remove it:
                            this.threads.splice(i--, 1);
                            removedThreads.push(thread);
                        }
                    } catch(e) {
                        matchError = e.toString();
                    }
                }
                
                if( updatedThreads.length > 0 ) {
                    if( matchError ) {
                        // Impossible to match threads but labels changed; means we may be out of date.
                        // TODO Does this trigger a refresh when we send? If so, exclude.
                        giLogger.log("#PERFWATCH matchError causes big update: " + e, 'gtdThreadContainer._threadChanged');
                        this.update();
                    } else {
                        this.dispatchEvent({name: gtdThreadContainer.CHANGED, addedThreads:addedThreads, 'updatedThreads': updatedThreads, 'removedThreads':removedThreads, actionEvent: evt});
                    }
                }
                
            } else {
                // Could not find the thread. Assume any thread could ultimately be added, so test it.
                
                // Can optimise - most adjustments are simple ADD/REMOVE label based; so we can test our search string for the presence of the label before doing anymore computation. If the search criteria is not dependant upon label:X (inc negation) we should not concern ourselves
                var inhibit = false;
                if( evt.action===glActionManager.ADD_LABEL || evt.action===glActionManager.REMOVE_LABEL ) {
                    if( this._search.getQuery().indexOf( glSearch.l(evt.label) )==-1 ) {
                        inhibit = true;
                    }
                }
                
                if( !inhibit ) {
                    
                    // Try to retrieve from the cache. TODO - could use conversation view? Makes less sense as it probably appears in cache
                    thread = this._gtd.gmail.environment.threadListCache.getThreadById(evt.thread);
                    if( thread ) {
                        thread.updateFromEvent(evt);    // Ensure it's given opportunity to update
                        this._threadChanged_NewThread(thread, evt, addedThreads, updatedThreads, removedThreads);
                    } else {
                        // Retrieve from server.
                        // TODO Create a cache for just this purpose? Use ThreadListCache to get it and auto cache it?
                        giLogger.log("#PERFWATCH - Downloading from server: "+evt.thread, 'gtdThreadContainer._threadChanged');
                        var thread = glThread.createFromId(evt.thread, this._gtd.gmail);
                        var me = this;
                        thread.addEventListener(glThread.LOADED, function(tEvt){
                            me._threadChanged_NewThread(tEvt.target, evt, addedThreads, updatedThreads, removedThreads);
                        }, this);
                    }
                }
            }
            
            
        } catch(e) {
            giLogger.warn(e, 'gtdThreadContainer._threadChanged', this._gtd.gmail);
        }
    },
	_threadChanged_NewThread: function(thread, evt, addedThreads, updatedThreads, removedThreads) {
        try {	
            if( this._search.matchThread(thread) ) {
                this.threads.push(thread);
                updatedThreads.push(thread);
                addedThreads.push(thread);
            }
        } catch(e) {
            // Impossible to match
            this.update();
        }
        if(updatedThreads.length > 0) {
            this.dispatchEvent({name: gtdThreadContainer.CHANGED, addedThreads:addedThreads, 'updatedThreads': updatedThreads, 'removedThreads':removedThreads, actionEvent: evt});
        }
    },


    
    /**
    * Get the thread with the id
    * @function ?
    * @param {string} id - The id of the desired thread
    * @return {glThread} - The thread with id
    */
    getThreadById: function(id) {
    	for( var i = 0; i < this.threads.length; i++ ) {
    		if( this.threads[i].threadId==id ) return this.threads[i];
    	}
    },

    /**
     * Callback for the {@link glActionManager.MESSAGE_CHANGED} event. Updates all
     * stored threads from the event and removes any that no longer fit the search
     * criteria.
     * @function ?
     * @param {Event} evt - The message changed event
     */
    _messageChanged: function(evt) {
        for(var i = 0; i < this.threads.length; i++) {
            this.threads[i].updateFromEvent(evt);
        }
    },
    //}
},{
    // Static Properties {
    
    /**
     * Dispatched when the the thread container starts a new search
     * @event ?
     */
    NEW_SEARCH: 'newSearch',
    
    
    /**
     * Dispatched when receive new threads; but not yet full request.
     * @event ?
     */
    LOADING_UPDATE: 'loadingUpdate',
    
    /**
     * Dispatched when the the thread container has completed a search.
     * From then on, it will always be up to date (or as close as is possible anyways).
     * @event ?
     */
    LOADED: 'loaded',
    

    /**
     * Dispatched when the contents of the thread container change (not a loading event - thread changes only)
     * @event ?
     */
    CHANGED: 'changed',
    
    /**
    * Type for getNumberThreads - specifies the #threads downloading
    * @property {String}
    */
    COUNT_TOTAL_DESIRED: 'countTotalDesired',

    /**
    * Type for getNumberThreads - specifies the #threads Gmail reports as known
    * @property {String}
    */    
    COUNT_TOTAL_CALCULATED: 'countTotalCalculated',

	/**
	* A cap on the number of threads to download if no total specified
	* @property {int}
	*/
	MAX_THREADS_ALLOWED: 1000,
    
    //}

    // Static Methods {
    /**
     * Creates and returns a gtdThreadContainer with the give label requirements
     * @function {gtdThreadContainer} ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     * @param {glSearch} search - The search to use to populate the container
     * @return The created thread container
     */
    createWithSearch: function(gtd, search) {
        giLogger.log('createWithSearch called', 'gtdThreadContainer');
        var container = new gtdThreadContainer(gtd, search);
        return container;
    },
    //}
},{
	
	MAX_THREADS_ALLOWED: 1000,
});
gtdThreadContainer.implement(giEventDispatcher); // Add event dispatching support