/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Listen to changes to threads (labels/stars) and optionally enforce rules (e.g. only one status label)
 *
 * @class gtdThreadRules
 */
var gtdThreadRules = giBase.extend({
    // Properties {
    /**
     * The gtdMainPagePlugin object for the page
     * @property {gtdMainPagePlugin} ?
     */
    _gtd: null,

	/** @property {Object[]} ? */
	_recentActions: null,

    //}

    /**
     * The constructor.
     * @constructor ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     */
    constructor: function(gtd) {
        this.base(gtd.gmail);
        this._gtd = gtd;
        this._recentActions = [];
		
    	this._gtd.gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);

    },

    // Methods {
	
	/**
	* Notification from action manager of an event.
	* If environmentGenerated, ensure we have already processed (code-request -> environment-generated-callback). If not, process it as a DOM event (as much be user generated).
	* Otherwise, process the event with whatever serverActionType originally used
	* @function {} ?
	*/
	_threadChanged: function(event) {
        if( !event.thread ) return;
    
    
        this._processAction(event, event.serverActionType )
        /*
        if( event.environmentGenerated ) {
            // Generated by environment. Check not already processed. If not, it must be DOM event, so process as such.
            var alreadyProcessed = false;
            for( var i = this._recentActions.length - 1; i>=0; i-- ) {
                var recentEvent = this._recentActions[i].event;
                if( recentEvent.action==event.action && recentEvent.label==event.label ) {	// TODO be ok for Star events too
                    if( (recentEvent.threadObject && recentEvent.threadObject.hasId(event.thread)) || recentEvent.thread==event.thread ) {	// recently processed
                        alreadyProcessed = true;
                        this._recentActions.splice(i,1);	// Remove all possible versions of this event
                    }
                }
            }
            if( !alreadyProcessed ) {
                this._processAction(event, glActionManager.USE_DOM);
            }
            
        } else {
            // Generated in code (by glActionManager.performAction)
            this._processAction(event, event.serverActionType )
            this._recentActions.push( {event: event, date: (new Date()).valueOf() } );
        }
        */
        
        
        // Clear old recentActions
        var d = (new Date()).valueOf() - (1000*60*2);
        for( var i = this._recentActions.length-1; i>=0; i-- ) {
            if( this._recentActions.date < d ) this._recentActions.splice(i,1);
        }

	},
	
	/**
	* Apply the rules to a given thread - based on what action was taken - and update with serverActionType
	* @function {} ?
	* @param {Object} event - The event from glActionManager
	* @param {String} serverActionType - The server action type to use to update the server
	*/
	_processAction: function(event, serverActionType) {
		var statusData = this._gtd.labelsData.getStatusData(true);
		var typeSettings = this._gtd.prefs.getPref('labels.types');
		
        // Extract crucial action-managing variables:
        var view = this._gtd.gmail.page.getActiveViewType();
        var isCvThisThread = false;
        if( view=='cv' ) {
            if( this._gtd.gmail.canvasDocument.defaultView.window.top.location.href.indexOf(event.thread)>-1 ) {
                isCvThisThread = true;
            } else {
                var conversationThread = this._gtd.gmail.getCurrentViewObject().getThread(true);  // Only get if loaded
                if( conversationThread && conversationThread.hasId(event.thread) ) {
                    isCvThisThread = true;
                }
            }
        }
        
        
		switch(event.action) {
			case glActionManager.ADD_LABEL:
				if( gtdLabelsData.testPrefix(event.label, typeSettings.statuses.prefix) && event.label!=statusData.today ) {	// Remove all other status labels
                    // Use environment to decide best way to control it:
                    var appropriateServerActionType = serverActionType;
                    if( view=='tl' ) {
                        appropriateServerActionType = glActionManager.USE_AJAX;
                    } else if(view=='cv') {
                        if( isCvThisThread && event.label!=statusData.finished ) {
                            appropriateServerActionType = glActionManager.USE_DOM;
                        } else {
                            appropriateServerActionType = glActionManager.USE_AJAX;
                        }
                    }
                    
                    
                    var removeStatuses = true;
                    if( this._gtd.prefs.getPref('allow_multiple_statuses') && statusData.reverseLookup[event.label]!='finished' ) removeStatuses = false;
                    if( this._gtd.prefs.getPref('do_not_remove_statuses_on_finish') && statusData.reverseLookup[event.label]=='finished' ) removeStatuses = false;
                    if( removeStatuses ) {
                        var statuses = this._gtd.labelsData.getLabelsArray('statuses', false, true);
                        if( isCvThisThread ) {
                            // The thread affected is the current cv thread, use faster labelling UI to remove
                            var pos = statuses.indexOf(event.label);
                            if( pos>-1 ) statuses.splice(pos, 1);   // Do not remove newly added
                            // TODO Rapid clicking can disrupt this; better to store 'most recent status label for thread' and maintain that. 
                            this._gtd.getCurrentViewObject().stripLabels(statuses);
                        } else {
                            for( var i = 0; i < statuses.length; i++ ) {
                                if( statuses[i]!=event.label && statuses[i]!=statusData.today ) {
                                    if( !event.threadObject || event.threadObject.hasLabel(statuses[i]) ) { // TODO Make this better at understanding where a thread (event.thread - optional) does not have a status label
                                        this._gtd.gmail.actions.performAction(glActionManager.REMOVE_LABEL, (event.threadObject || event.thread), appropriateServerActionType, statuses[i]);
                                    }
                                }
                            }
                        }
                    }
				}
				
                /*
				if( event.environmentGenerated && statusData.reverseLookup[event.label]=='finished' && !this._gtd.prefs.getPref('do_not_remove_statuses_on_finish') ) {
					// Finish, generated by a user DOM click
                    if( this._gtd.gmail.page.getActiveViewType()!='cv' ) this._gtd.gmail.actions.performAction(glActionManager.MARK_AS_READ, event.thread, glActionManager.USE_DOM);
					this._gtd.gmail.actions.performAction(glActionManager.ARCHIVE, event.thread, glActionManager.USE_DOM);
				}
                */
				
				break;
			case glActionManager.STAR:
				break;
			case glActionManager.UNSTAR:
				break;
		}
	},
	
    //}
});
