/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Store and manage standard glSearchConditions views
 *
 * TODO Views should be able to reference other views, e.g. <!=view.<id>!>
 * TODO Discern whether to constrain to Statuses. This should inherit.
 *          So, view.tasks_to_do (inbuilt) constraints to Statuses.
 *          But, a theoretical view.files_in_tasks, which uses <!=view.tasks_to_do!>, inherits the constraint of Statuses.
 *          But, if it is a problem search (from/to/freetext) such as view.from_bob_in_tasks, it uses <!=view.tasks_to_do!> but creates a new search rather than uses statusThreadsContainer.
 *
 * @class gtdViews
 */
var gtdViews = giBase.extend({
    // Properties {
    /**
     * The gtdMainPagePlugin object for the page
     * @property {gtdMainPagePlugin} ?
     */
    _gtd: null,

    /**
     * Hash of view data.
     * Format: {id: {displayName: String, search: glSearchConditions}}
     * @property {Object}
     */
	_views: null,

    //}

    /**
     * The constructor.
     * @constructor ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     */
    constructor: function(gtd) {
        this.base(gtd.gmail);
        this._gtd = gtd;
        this._views = {};
		
        this._reloadInbuiltViews();
        
    	this._gtd.labelsData.addEventListener(gtdLabelsData.LABEL_CHANGED, this._labelChanged, this);
    },

    // Methods {
	
    /**
     * Reload the inbuilt views
     * @function {} ?
     */
    _reloadInbuiltViews: function() {
        this._views['tasks_to_do'] = {search: new glSearchAndConditions()};
        this._views['tasks_waiting_on'] = {search: new glSearchAndConditions()};
        this._views['tasks_some_day'] = {search: new glSearchAndConditions()};
        
        var typeSettings = this._gtd.prefs.getPref('labels.types');
        var status = this._gtd.labelsData.getStatusData();
        var statuses = this._gtd.labelsData.getLabelsArray("statuses", false, true);
        var seenStatuses = {};
        
        
        var ttdOr = new glSearchOrConditions();
        
        // Include all core statuses - ensures we have a tight query even if user does not have Status labels created
        for( c in typeSettings.statuses.core ) {
            if( c!="finished" ) {
                seenStatuses[status[c]] = true;
                
                if( status[c]==status.waiting_on ) {
                    this._views['tasks_waiting_on'].search.addCondition( glSearch.l(status[c]) );
                } else if( status[c]==status.some_day ) {
                    this._views['tasks_some_day'].search.addCondition( glSearch.l(status[c]) );
                } else {
                    ttdOr.addCondition(glSearch.l(status[c]));
                }
            }
        }
        
        for( var i = 0; i < statuses.length; i++ ) {
            if( statuses[i]!=status.finished && !seenStatuses[statuses[i]]) {
                if( statuses[i]==status.waiting_on ) {
                    this._views['tasks_waiting_on'].search.addCondition( glSearch.l(statuses[i]) );
                } else if( statuses[i]==status.some_day ) {
                    this._views['tasks_some_day'].search.addCondition( glSearch.l(statuses[i]) );
                } else {
                    ttdOr.addCondition(glSearch.l(statuses[i]));
                }
            }
        }
        
        // Progress Tasks To Do:
        this._views['tasks_to_do'].search.addCondition(ttdOr);
        
        // Add -Finished to each one:
        var minusFinished = '-'+glSearch.l(status.finished);
        this._views['tasks_to_do'].search.addCondition(minusFinished);
        this._views['tasks_waiting_on'].search.addCondition(minusFinished);
        this._views['tasks_some_day'].search.addCondition(minusFinished);
        
    },
    
    /**
     * Notification that an environment label was changed.
     * Modify the views.
     * @function {} ?
     */
    _labelChanged: function(evt) {
        this._relatedInbuiltViews();
        
        // TODO Can we refine this to detect only the ones affected by a label change?
        for( i in this._views ) {
            this.dispatchEvent({name: gtdViews.VIEW_EDITED, viewId: i});
        }
    },
    
    /**
     * Return an array of view ids
     * @function {String[]} ?
     * @return An array of view ids
     */
    getViews: function() {
        var ids = [];
        for( i in this._views ) {
            ids.push(i);
        }
        
        return ids;
    },
    
    /**
     * @function {glSearchConditions} ?
     * @param {String} id - The view ID
     * @return The glSearchConditions object for a given view
     */
    getViewSearch: function(id) {
        return this._views[id]? this._views[id].search : null;
    },

    /**
     * @function {String} ?
     * @param {String} id - The view ID
     * @return The display name of this view
     */
    getViewName: function(id) {
        return giI18N.getString("Views."+id+".name", id);
    },
    
    /**
     * Return true if we have a view with the given id
     * @function {Boolean} ?
     * @param {String} id - The id of the view
     * @return True if view known
     */
    hasView: function(id) {
        return this._views[id]? true : false;
    },
    
    //}
}, {
    /**
     * Notification that a view has changed. Includes viewId.
     * @event
     */
    VIEW_EDITED: 'view_edited',
});
gtdViews.implement(giEventDispatcher); // Add event dispatching support