/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Monitor yellow alert boxes and react.
 * Used when send messages to add our Labelling reminder.
 *
 * @class ? @extends giUIObject
 */
var gtdConversationAlerts = giUIObject.extend({
    // Properties {
    

	
	/**
	* The {@link gtdLabellingBox} object
	* @property {gtdLabellingBox} ?
	*/
	_conversationLabels: null,
	
	/** @property {private int} ? The ID of the interval timer, if active */
	_interval: null,
	
	/** @property {private int} ? Time of last interval start (runs for X seconds) */
	_lastIntervalStart: null,
	
	/** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
	
    //}

    // Methods {
    /**
     * Connects the element to the given gtdMainPagePlugin object for the page.
     * @function ?
     * @param {gtdMainPagePlugin} main - The gtdMainPagePlugin object for the page
     */
    connect: function(main) {
        giLogger.log('Connecting gtdConversationAlert object to view', 'gtdConversationAlert');
        this.base(main);
            
        this._eventsContainer = new giEventsContainer(this);
        
        
        // Listen for Sent Messages
        this._main.gmail.trafficWatcher.addEventListener(glTrafficWatcher.SEND_MESSAGE_REQUEST_COMPLETED, this._sendMessage, this);
        
    },


    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting gtdConversationAlert from view', 'gtdConversationAlert');

        
        if( this._conversationLabels ) {
	        this._conversationLabels.disconnect();
        	this._conversationLabels = null;
        }
        
        this._eventsContainer.reset();
        
        if( this._interval ) {
        	clearInterval( this._interval );
        	this._interval = null;
        }
        
        this.base();
    },
    
	
	/**
	* Notification a message was sent. Set up interval to look for alert boxes
	* @function {} ?
	*/
	_sendMessage: function(evt) {
        if( !this._interval ) {
            var me = this;
            this._interval = setInterval(function(){me._intervalEvent()}, 500);
        }
        this._lastIntervalStart = (new Date()).valueOf();
	},
	
	
	/**
	* Notification that a DOM Node was removed. 
	* @function ?
	*/
	_removeNode: function(event) {
		try {
			var elRemoved = event.target;
			
			// See if 
			var i = 3;
			var el = this._elAlert;
			while(--i>=0 && el!=elRemoved ) el = el.parentNode;
			if( el==elRemoved ) {	// Removed alert box
				// Hide Labelling Box
				if( this._conversationLabels ) this._conversationLabels.el.style.display = "none";
				// Remove mods to alert box
				var elContainer = this._elAlert.ownerDocument.evaluate(".//*[@class='gtdi-labelreminder-container']", this._elAlert, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
				if( elContainer ) elContainer.parentNode.removeChild(elContainer);
				// Stop listening to node removals
				this._eventsContainer.stopObserving(this._main.gmail.page.getViewRoot(), 'DOMNodeRemoved', this._removeNode, false);
			}
		} catch(e) {
			giLogger.warn(e, 'gtdConversationAlerts._removeNode', this._main.gmail);
		}
	},
	
	/**
	* Look for alert boxes not previously seen.
	* If find one, clear the interval and add a link to it.
	*/
	_intervalEvent: function() {
		
		try {
			// Get the alert box:
			var doc = this._main.gmail.canvasDocument;
			this._elAlert = doc.evaluate(".//table[@role='alert']", this._main.gmail.page.getViewRoot(), null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
			
			if( this._elAlert ) {
				
				// Remove any previous mods:
				var elContainer = doc.evaluate(".//*[@class='gtdi-labelreminder-container']", this._elAlert, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
				if( elContainer ) elContainer.parentNode.removeChild(elContainer);
			
			
				var statusData = this._main.labelsData.getStatusData(true);
				// Get thread - links are context dependant
				var thread = this._main.gmail.getCurrentViewObject().getThread();
				// Get Status info:
				var hasStatus = null;
				for( var i = 0; i < thread.labels.length; i++ ) {
					if( statusData.reverseLookup[thread.labels[i]] ) {
						hasStatus = thread.labels[i];
						break;
					}
				}
				// Add links to alert box:
				var elText = doc.evaluate(".//td[@class='vh']", this._elAlert, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
				if( hasStatus ) {
					elText.innerHTML += "<div class='gtdi-labelreminder-container'>" + giI18N.getString("Conversation.alerts.labelreminder.hasStatus.prefix") + ": " + hasStatus + " (<a href='#' class='gtdi-labelreminder'>" + giI18N.getString("Conversation.alerts.labelreminder.hasStatus.link") + "</a>)</div>";
				} else {
					elText.innerHTML += "<div class='gtdi-labelreminder-container'>" + giI18N.getString("Conversation.alerts.labelreminder.noStatus.prefix") + " (<a href='#' class='gtdi-labelreminder'>" + giI18N.getString("Conversation.alerts.labelreminder.noStatus.link") + "</a>)</div>";
				}
				
				
				// Listen for clicks:
				this._eventsContainer.stopObserving(this._elAlert, "click", this._click, false);
				this._eventsContainer.observe(this._elAlert, "click", this._click, false);
				
				
				// Listen to node removals:
				this._eventsContainer.stopObserving(this._main.gmail.page.getViewRoot(), 'DOMNodeRemoved', this._removeNode, false);
				this._eventsContainer.observe(this._main.gmail.page.getViewRoot(), 'DOMNodeRemoved', this._removeNode, false);
				
            
                clearInterval(this._interval);
                this._interval = null;
			} else {
				// Check not too old to care:
				if( ((new Date()).valueOf()-this._lastIntervalStart) > 30000 ) {
					clearInterval(this._interval);
					this._interval = null;
				}
			}
		} catch(e) {
			clearInterval(this._interval);
			this._interval = null;
			giLogger.warn(e, 'gtdConversationAlerts._intervalEvent', this._main.gmail);
		}
	},
 
 	/** 
 	* Notification of a click event
 	* @function ?
 	*/
 	_click: function(event) {
 		try {
 	
	 		var el = event.target;
	 		
	 		
	 		if( el.className=="gtdi-labelreminder" ) {
	 			// Clicked label reminder link
	 			giUtil.stopEvent(event);
	 			
	 			// Have we loaded the labelling box?
	 			if( !this._conversationLabels ) {
		 			this._conversationLabels = this._main.getUIObject("labellingBox");
					this._conversationLabels.setServerActionType(glActionManager.USE_DOM);
					this._conversationLabels.load(this._main.gmail.getCurrentViewObject().getThread(), true);
					
					var elCmdButtons = this._conversationLabels.getCmdButtons();
					var elHide = elCmdButtons.ownerDocument.createElement("DIV");
					elHide.className = "gtdi-btn";
					elHide.innerHTML = "<b class='gtdi-btn-t'></b><b class='gtdi-btn-t2'></b><b class='gtdi-btn-m'></b><b class='gtdi-btn-b2'></b><b class='gtdi-btn-b'></b><span class='gtdi-btn-label'><span class='gtdi-hide'>" + giI18N.getString("Labelling.buttons.hide") + "</span></span>";
					elCmdButtons.appendChild( elHide );
					
					this._eventsContainer.observe(elHide, "click", this._click, false);
	 			}
	 			
                var elSpace = this._main.gmail.canvasDocument.createElement("BR");
                elSpace.setAttribute('clear', 'both');
                this._conversationLabels.el.appendChild( elSpace );
                
	 			// Find the place to hook in the labelling box
	 			var elRow = el;
	 			while( elRow.parentNode && elRow.className!="nH" ) {
	 				if( elRow.className=="gtdi-labelreminder-container" ) {
	 					var elC = elRow;
	 					elRow = elRow.parentNode;
	 					elRow.removeChild(elC);
	 				} else {
		 				elRow = elRow.parentNode;
		 			}
	 			}
	 			if( elRow.className=="nH" ) {
	 				if( elRow.nextSibling ) {
	 					elRow.parentNode.insertBefore( this._conversationLabels.el, elRow.nextSibling );
	 				} else {
	 					elRow.parentNode.appendChild( this._conversationLabels.el );
	 				}
	 			}
	 			this._conversationLabels.el.style.display = "";
	 			
	 		}
	 		if( el.className=="gtdi-hide" ) {
	 			this._conversationLabels.el.style.display = "none";
	 		}
 		} catch(e) {
 			giLogger.warn(e, 'gtdConversationAlerts._click', this._main.gmail);
 		}
 	},

    
    //}
});
