/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Provides popup menu that enables a user to add/remove labels to multiple threads. Largely apes the standard Gmail labels menu.
 *
 * @class ? @extends giUIObject
 */
var gtdLabelsMenu = giUIObject.extend({
    // Properties {
    
    
	/** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
    	   	
    
	/**
	* The thread list container
	* @property {Element} ?
	*/
	el: null,
	
	/** @property {private Element} ? The dropped-down menu */
	_elDropdown: null,
	
	/** @property {private Element} ? The actual UL list within the dropped-down menu */
	_elDropdownList: null,
	
	/** @property {private glThreads[]} ? Array of threads to show labels for and add/remove labels from */
	_threads: null,
	
	/** @property {private String} ? Method for server updating */
	_serverActionType: null,
	
    //}


	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} main - The gtdMainPagePlugin object for the page
	*/
	connect: function(main) {
		giLogger.log('Connecting gtdLabelsMenu object to view', 'gtdLabelsMenu');
		
        
		this.base(main);
		this._eventsContainer = new giEventsContainer(this);
		this._threads = [];
		
		// Create container
		var doc = main.gmail.canvasDocument;
		this.el = doc.createElement("DIV");
		this.el.className = "gtdi-btn gtdi-dropdown";
		this.el.innerHTML = "<b class='gtdi-btn-t'></b><b class='gtdi-btn-t2'></b><b class='gtdi-btn-m'></b><b class='gtdi-btn-b2'></b><b class='gtdi-btn-b'></b><span class='gtdi-btn-label'>" + giI18N.getString("LabelsMenu.Title") + "</span>";
		
		this._elDropdown = doc.createElement("DIV");
		this._elDropdown.className = "gtdi-droppeddown";
		this._elDropdown.style.display = "none";
		this.el.appendChild( this._elDropdown );
		
		var elDropdownItems = doc.createElement("DIV");
		elDropdownItems.className = "gtdi-dropdown-items";
		this._elDropdown.appendChild( elDropdownItems );
		
		this._elDropdownList = doc.createElement("UL");
		elDropdownItems.appendChild( this._elDropdownList );
		
		// Populate list with all known labels
		this._reloadLabels();
		
		// Listen for Gmail label changes
		this._main.labelsData.addEventListener(gtdLabelsData.LABEL_CHANGED, this._reloadLabels, this);
		
		// Listen for thread changes
	    this._main.gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
		
		this._showDropdownList(false);
		
	},



	/**
	* Called on page shutdown, it does cleanup
	* @function ?
	*/
	disconnect: function() {
		giLogger.log('Disconnecting gtdLabelsMenu from view', 'gtdLabelsMenu');
		
		
		if( this.el && this.el.parentNode ) this.el.parentNode.removeChild(this.el);
		this.el = null;
		
		// Wipe references to elements
		for( p in this ) {
			if( this[p] && this[p].nodeName && this[p].nodeType ) {
				this[p] = null;
			}
		}
		
		
		this._eventsContainer.reset();
		this._eventsContainer = null;
		
		this._threads = null;
		

		this._main.labelsData.removeEventListener(gtdLabelsData.LABEL_CHANGED, this._reloadLabels, this);
	    this._main.gmail.actions.removeEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
		
		this.base();
	},
	
	
	/**
	* Configure the list for these threads; putting 'labelled' items first (and checked), followed by all other labels alphabetically.
	* Monitor threads for label changes; and update appropriately.
	* @function ?
	* @param {glThread[]} threads - Array of threads to work around (add/remove labels, configure labels menu for).
	*/
	loadThreads: function(threads) {
		this._threads = threads;
		this._reorderLabels();
	},

	/**
	* Organise the labels alphabetically, but display those labels that are on a thread first.
	* It de-activates previously active (checked) labels, restoring them in the hierarchy; before pulling the labels that are now active to the front (and checking them).
	* The current algorith, while efficient, is dependent upon the labels being alphabetically sorted. 
	* A simpler approach would be to create a hash ({<labelname>:el}) of all elements, and two text arrays of inactive and active labelnames; sort the arrays; and append all items in order. But that is both multi-pass (unless cache the labelname hash) and heavy on node-shifting. Users can have 100s of labels; and this is called whenever a thread is changed.
	* @function ?
	*/
	_reorderLabels: function() {	
		// TODO could check active labels vs existing active labels; and not proceed if no change.
		// TODO this is being called by gtdThread presumably all the time - even when out of view


		
		// Load the 'active' labels into a hash for faster querying
		var activeLabelsHash = {};
		for( var i = 0; i < this._threads.length; i++ ) {
			for( var j = 0; j < this._threads[i].labels.length; j++ ) {
				activeLabelsHash[this._threads[i].labels[j]] = true;
			}
		}
				
				
		// Restore any previously active labels into their place in the alphabetical 'off' list:
		
		// Extract previously active labels:
		var activeLabelEls = [];
		var newActiveLabelEls = [];
		while( this._elDropdownList.childNodes.length > 0 && this._elDropdownList.childNodes[0].getElementsByTagName("INPUT")[0].checked ) {	// Process each checked node - which are clustered at the front of the list
			var elLi = this._elDropdownList.childNodes[0];
			elLi.getElementsByTagName("INPUT")[0].checked = false;
			if( activeLabelsHash[elLi.getAttribute("gtdi-labelname")] ) newActiveLabelEls.push(elLi);	// keep out of DOM until ready to re-add at top of list (do not re-insert into order)
			else activeLabelEls.push( elLi );	// to replace
			this._elDropdownList.removeChild( elLi );
		}
		
		
		// Find the first node to insert active elements before:
		var firstInactiveNode = null;
		for( var i = 0; i < this._elDropdownList.childNodes.length; i++ ) {
			if( !activeLabelsHash[ this._elDropdownList.childNodes[i].getAttribute("gtdi-labelname") ] ) {	// This label is the first non-active.
				firstInactiveNode = this._elDropdownList.childNodes[i];
				break;
			}
		}
		
		
		// Process all nodes; restoring previously active nodes that are no longer active, and catching newly active nodes:
		for( var i = 0; i < this._elDropdownList.childNodes.length; i++ ) {
			var elLi = this._elDropdownList.childNodes[i];
			
			// Try to re-add a previous label at the right point:
			if( activeLabelEls.length>0 ) {
				var labelname = elLi.getAttribute("gtdi-labelname").toLowerCase();
				
				// See if we can restore previously active node before it - take the first instance when an existing label no longer fits in the hierarchy
				var matches = [];
				for( var j = activeLabelEls.length - 1; j>=0; j-- ) {
					if( activeLabelEls[j].getAttribute("gtdi-labelname").toLowerCase() < labelname ) {	// This existing label belongs before elLi
						matches.push( activeLabelEls[j] );
						activeLabelEls.splice(j,1);
					}				
				}	
				matches.reverse();	// insertBefore inverts order; so if add multiple labels before single elLi, they'd be out of order
				for( var j = 0; j < matches.length; j++ ) {
					this._elDropdownList.insertBefore( matches[j], elLi );
					i++;	// We're inserting items before current i, so push it forward
					if( activeLabelsHash[matches[j].getAttribute("gtdi-labelname")] ) newActiveLabelEls.push(matches[j]);	// Care to keep adding new ones in sorted order
				}
			}
			
			// See if this node is now/still active:
			if( activeLabelsHash[elLi.getAttribute("gtdi-labelname")] ) newActiveLabelEls.push(elLi);
			
		}
		// If any previously active labels remain, append to the end (because in previous loop, it only catches insertBefore items)
		for( var i = 0; i < activeLabelEls.length; i++ ) {
			if( activeLabelsHash[activeLabelEls[i].getAttribute("gtdi-labelname")] ) newActiveLabelEls.push(activeLabelEls[i]);
			this._elDropdownList.appendChild( activeLabelEls[i] );
		}
		
		// Finally, move active labels above firstInactiveNode:
		for( var i = 0; i < newActiveLabelEls.length; i++ ) {
			newActiveLabelEls[i].getElementsByTagName("INPUT")[0].checked = true;
			if( firstInactiveNode ) {
				this._elDropdownList.insertBefore( newActiveLabelEls[i], firstInactiveNode );
			} else {
				this._elDropdownList.appendChild( newActiveLabelEls[i] );
			}
		}
		
	},
	
	/**
	* Notification from LabelsBox that a label has been added/removed. Reload all labels; then reorder.
	* @function ?
	*/
	_reloadLabels: function(event) {
		// Clear out the existing collection
		this._elDropdownList.innerHTML = "";
		
		// Start creating new entries for each label:
		var labels = this._main.labelsData.getLabelsArray(null, null, true);
		labels.sort(function(x,y) {if( x.toLowerCase() > y.toLowerCase() ) return 1;else return -1});
		
		var doc = this._main.gmail.canvasDocument;
		for( var i = 0; i < labels.length; i++ ) {
			var elLi = doc.createElement("LI");
			elLi.setAttribute("gtdi-labelname", labels[i]);
			elLi.innerHTML = "<input type='checkbox'/><span>" + labels[i] + "</span>";
			this._elDropdownList.appendChild( elLi );
		}
	},
	
	/**
	* Notification that thread data has changed - watch for label add/removed, see if it affects any of our threads, and reorder if neccessary.
	* @function ?
	*/
	_threadChanged: function(event) {
		for( var i = 0; i < this._threads.length; i++ ) {
			if( this._threads[i].threadId==event.thread ) {	// This affects one of our threads
				this._reorderLabels();	// TODO Updates for each selected thread = inefficent. But, no guarenteed way of knowing if every requested thread was updated (as the API code does not necessarily update an individual thread if it never-had/had the label being removed/added to  the CHANGE request). MultiToOne would be a simple approach. Or just allow reorderLabels to decide whether to reorder or not.
				break;
			}
		}
	},
	
	/**
	* Respond to a click on the button, list or window. Show/hide list, add/remove labels.
	* @function ?
	*/
	_click: function(event) {
		try {
			
			var el = event.target;
			
			var elTree = el;
			var clickedWithinEl = false;
			var c = 0;
			while( elTree.parentNode && ++c<5 ) {
				if( elTree.nodeName=="LI" && elTree.parentNode==this._elDropdownList ) {	// Clicked list item
					clickedWithinEl = true;
					
					var labelname = elTree.getAttribute("gtdi-labelname");
					if( elTree.getElementsByTagName("INPUT")[0].checked ) {	// Currently on; turn off on all threads
						for( var i = 0; i < this._threads.length; i++ ) {
                            this._main.gmail.actions.performAction( glActionManager.REMOVE_LABEL, this._threads[i], this._serverActionType, labelname );
                        }
					} else {	// Currently off, turn on on all threads
						for( var i = 0; i < this._threads.length; i++ ) {
                            this._main.gmail.actions.performAction( glActionManager.ADD_LABEL, this._threads[i], this._serverActionType, labelname );
                        }
					}
					
					if( el.nodeName!="INPUT" ) {	// Did not click checkbox, turn list off (in line with Gmail behaviour)
						//this._showDropdownList(false);
					}
                    this._showDropdownList(false);
					
					
					break;
				}
				if( elTree==this.el ) {			// Clicked button - show if hidden
					clickedWithinEl = true;
					if( this._elDropdown.style.display=="none" ) {	// Visible. Turn off.						
						this._showDropdownList(true);	
					}
					break;
				}
				elTree = elTree.parentNode;
			}
			if( !clickedWithinEl && this._elDropdown.style.display=="" ) {	// Clicked out of dropped down list
				this._showDropdownList(false);
			}
		} catch(e) {
			giLogger.warn(e, "gtdLabelsMenu._click", this._main.gmail);
		}
	},
	
	/**
	* Show/hide the dropdown list; and set up event listeners for it's subsequent activation/deactivation
	* @function ?
	* @param {boolean} show - If true, show the list, and listen to clicks on the window that would cause it to be turned off
	*/
	_showDropdownList: function(show) {
		if( show ) {	
			this._elDropdown.style.display = "";
			this._eventsContainer.stopObserving( this.el, "mousedown", this._click, false );
			setTimeout( giUtil.bind( this, function() {
				this._eventsContainer.observe( this.el.ownerDocument.body, "mousedown", this._click, false );
			}), 10);	
		} else {	// Off. Turn on.
			this._elDropdown.style.display = "none";
			this._eventsContainer.stopObserving( this.el.ownerDocument.body, "mousedown", this._click, false );
			setTimeout( giUtil.bind( this, function() {
				this._eventsContainer.observe( this.el, "mousedown", this._click, false );
			}), 10);			
		}
	},
	
	

	/**
	* Set method for updating server. Ensure future created objects can use this info; and existing objects are updated.
	* @function ?
	* @param {String} type - The type. A static variable of glActionManager
	*/
	setServerActionType: function(type) {
		this._serverActionType = type;
	},
	
	

},{
	
	
});
gtdLabelsMenu.implement(giEventDispatcher); // Add event dispatching support