/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
* A container of interactive hierarchial label elements, for a given type, and utility functions to manage them. 
* @class ?
*/

var gtdLabelsTree = giUIObject.extend({
	// Properties {
	
	/** @property {Element} ? The label container */
	el: null,
	
	/** @property {string} ? The type of labels this contains (e.g. 'statuses') */
	type: null,
	
	/** @property {string} ? The prefix for this type */
	_prefix: null, 
		
    /** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
	
	/** @property {Object} ? Hash of labels, pointing to their respective elements */
	labelHash: null,
	
	
	// }

	/**
	* @constructor ?
	* @param {gtdMainPagePlugin} main - The main gtd plugin for this page
	*/
	connect: function(main) {
		giLogger.log("Constructing gtdLabelDOM", "gtdLabelDOM");
		this.base(main);
		
	},
	
    /**
     * @function {} ?
     * @param {string} type - The type of these labels (e.g. statuses)
     * @param {Element} el - Optional. The container el for the labels.
     */
	init: function(type, el) {	// TODO Can we roll this into connect by extending the .getUIObject function to accept additional parameters
		this.type = type;
		this._eventsContainer = new giEventsContainer(this);
		this.labelHash = {};
		
		// Get prefix
		var typeSettings = this._main.prefs.getPref('labels.types');
		this._prefix = typeSettings[type].prefix;
		
		// Create, or hook in, the container el
		var doc = this._main.gmail.canvasDocument;
		if( el ) {
			this.el = el;
		} else {
			this.el = doc.createElement("DIV");
		}
		giUIObject.addClassName( this.el, "gtdi-labeldom" );
		
		// Listen to click events
		this._eventsContainer.observe(this.el, "click", this._click, false );
	},

	/**
	* Clean up memory
	*/
	disconnect: function() {
		this._main = null;
		this.type = null;
		
        if( this.el.parentNode ) this.el.parentNode.removeChild(this.el);
		this.el = null;
		
		this._eventsContainer.reset();
		this._eventsContainer = null;
    	this.base();
	},
	
	/**
	* Add a label, figure out leaf name, ensure it has placeholders if no valid parent, ensure any parents have +/-, and do indentation
	* Can accept either a label element, or a labelname (string)
	* A current restriction is that the labels must be processed in sorted order (P/GTDInbox, P/GTDInbox/Bugs, P/GTDInbox/Ideas, P/LC)
	* @paramset Element
	* @param label {Element} - The main element for this label
	* @param elContainer {Element} - Optional. A child element of 'label' that will contain the expander, and be indented. In normal circumstances, label==elContainer; and when creating artificial lists, label==elContainer==elLeaf.
	* @param elLeaf {Element} - Optional. A child element of 'label' that contains the display text.
	* @param labelname {string} - Optional. Force the labelname
	* @param leafname {string} - Optional. Force the leafname
	* @param fCreatePlaceHolderEl {Function}. Optional. Passed a labelname to create a place holder element.
	* @paramset Labelname
	* @param label {string} - The labelname
	*/
	addLabel: function(label, elContainer, elLeaf, labelname, leafname, fCreatePlaceHolderEl) {
		try {
			var doc = this._main.gmail.canvasDocument;
			var labelDisplayName = null;
			var el = null;
			var depth = 0;
			var leafnameComputed = null;
			
			// If a container, mark it up
			if( elContainer ) elContainer.setAttribute("gtdi-container", "1");
			           
			// Process the basic element and extract labelname:
			if( typeof label=="string" ) {
				// Create a simple element
				labelname = label;
				el = doc.createElement("DIV");
				elLeaf = el;
			} else {
				// Do we need to identify elLeaf?
				el = label;
				labelname = labelname || (elLeaf? elLeaf.innerHTML : (elContainer? elContainer.innerHTML : el.innerHTML));
				labelname = giUtil.decodeEntities( giUtil.stripTags( labelname ), doc);
			}
			giUIObject.addClassName(el, "gtdi-labeldom-label");
			el.setAttribute("gtdLabelName", labelname);
            
			// Add to hash for fast lookup (must be after parents added):	
			this.labelHash[labelname] = el;
			elContainer = this._getContainer(labelname, el);
            		
			// Attach it to the container
			this.el.appendChild( el );
            
			// Strip prefix to get display name:
			labelDisplayName = gtdLabelsData.removePrefix(labelname, this._prefix);
            							
			// Breakdown hierarchy:
			var hierarchy = labelDisplayName.split("/");
			
			
			leafnameComputed = hierarchy[hierarchy.length-1];
			if( hierarchy.length > 1 ) {	// We have a hierarchy
				leafname = leafname || leafnameComputed;
				depth = hierarchy.length-1;
                
				
				// Ensure all parents exist (at least have place holders), and all have +/-
				var lastLabel = labelname;
				var hadParent = false;
				while( hierarchy.length > 1 ) {
					hierarchy.length = hierarchy.length - 1	// evaluate the next one
					
					// Check for existance:
					var parentLabel = this._prefix + hierarchy.join("/");
                    var elParent = this.labelHash[parentLabel];
					hadParent = !!(this.labelHash[parentLabel]);
					if( !hadParent ) {	// Does not exist. Requires place holder
						// Create place holder
						var elParent = null;
						var parentLeaf = gtdLabelsData.removePrefix(hierarchy[hierarchy.length-1], this._prefix);
						if( fCreatePlaceHolderEl ) {
							elParent = fCreatePlaceHolderEl(parentLabel, parentLeaf);
						} else {
							elParent = doc.createElement("DIV");
							elParent.innerHTML = parentLeaf || "/";
							elParent.setAttribute("gtdi-leaf", parentLeaf);
						}					
						giUIObject.addClassName( elParent, "gtdi-labeldom-label" );
						giUIObject.addClassName( elParent, "gtdi-labeldom-placeholder" );
						// Set node-specific attributes
						var elParentContainer = this._getContainer(parentLabel, elParent); //doc.evaluate(".//*[@gtdi-container]", elParent, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
						elParent.setAttribute("gtdi-depth", hierarchy.length-1);
						elParentContainer.style.paddingLeft = (10+((hierarchy.length-1)*10)) + "px";
						elParent.setAttribute("gtdi-placeholder", "1");
						elParent.setAttribute("gtdi-type", this.type);
						elParent.setAttribute("gtdLabelName", parentLabel);
						// Add to hierarchy:
						this.labelHash[lastLabel].parentNode.insertBefore( elParent, this.labelHash[lastLabel] );	// Insert before child node
						this.labelHash[parentLabel] = elParent;	// Add for fast lookup
                        lastLabel = parentLabel;
					}
					// Check for expander:
					if( !this.getExpander(parentLabel, elParent) ) {
						var elExpander = doc.createElement("IMG");
						elExpander.className = "gtdi-labeldom-expander";
						elExpander.src = "chrome://gtdinbox/skin/gtd/collapsetree-closed.png";
						var elExpanderParent = this._getContainer(parentLabel, elParent);
                        elExpander.style.paddingLeft = (parseInt(elExpanderParent.getAttribute('gtdi-depth')) * (gtdLabelsTree.IMG_WIDTH_PX+gtdLabelsTree.IMG_GAP_PX)) + "px";
						elExpanderParent.insertBefore( elExpander, elExpanderParent.firstChild );
					}
                    if( (hierarchy.length-1)>0 ) elParent.style.display = "none"; // It's nested, hide it.
					
					// If had parent, as we add in perfect order, we can now assume all further parents will have been added (note we have checked expander on direct parent; as that is always relevant - even if seen before).
					if( hadParent ) {
						break;
					}
				}
			} else {	// Just take the name
				leafname = leafname || labelDisplayName;
			}
			
			
			// Set Hierarchal Depth:
			el.setAttribute("gtdi-depth", depth );
            elContainer.style.paddingLeft = (depth * (gtdLabelsTree.IMG_WIDTH_PX+gtdLabelsTree.IMG_GAP_PX)) + (gtdLabelsTree.IMG_WIDTH_PX+gtdLabelsTree.IMG_GAP_PX) + "px";
			//elContainer.style.paddingLeft = (10+(depth*10)) + "px"; // First '10' is to indent to allow +/- imgs to take hold. CSS-only solution fails because would need to be on parent DIV (to not interfere with this style mod), and parent DIV is behold to overflow:hidden, which hides the +/- img below. 
			if( depth > 0 ) el.style.display = "none";
						
			// Set display name:
            
			if( elLeaf ) {
				elLeaf.innerHTML = leafname || "/";
				elLeaf.setAttribute("gtdi-leaf", leafnameComputed);
			} else {
				elContainer.innerHTML = leafname || "/";
				elContainer.setAttribute("gtdi-leaf", leafnameComputed);
			}
			el.setAttribute("gtdi-type", this.type);
			
			
			
		} catch(e) {
			giLogger.error(e, "gtdLabelDOM.addLabel");
		}
	},
	
	/**
	* For efficiency, we require labelHash to be order.
	* So, after adding labels, re-organise it
	* This MUST be called after all addLabels done.
	* @function ?
	*/
	finishAddingLabels: function() {
		var labelHash = {};
		var labels = [];
		for( l in this.labelHash ) {
			labels.push(l);
		}
		labels.sort();
		for( var i = 0; i < labels.length; i++ ) {
			labelHash[labels[i]] = this.labelHash[labels[i]];
		}
		this.labelHash = labelHash;
	},
	
	/**
	* Get the element that will be indented, and could container the expander. 
	* Try to find an element with the gtdi-labeldom-container attribute first; if that fails, try the optional failsafe; and if that fails, return the main element.
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {Element}
	*/
	_getContainer: function(labelname, el, elFailsafe) {
		if( !el ) el = this.getLabel(labelname);
		if( el ) {
			var elContainer = el.ownerDocument.evaluate(".//*[@gtdi-container]", el, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
			return elContainer || el;
		}
				
	},
	

	
	/**
	* Response to a click event on the label container
	*/
	_click: function(event) {
		try {
			var el = event.target;
			
			
			if( el.className=="gtdi-labeldom-expander" ) {	// It's an expander. Toggle.
				while( el.parentNode && !el.getAttribute("gtdLabelName" ) ) el = el.parentNode;
				var labelname = el.getAttribute("gtdLabelName");
				if( labelname ) {
					if( this.isExpanded(labelname) ) {// It's open. Close it.
						this.close(labelname);
					} else {	// It's closed. Open it.
						this.open(labelname, true);
					}
				}
			}
		} catch(e) {
			giLogger.error(e, "gtdLabelsTree._click", this._main.gmail);
		}
	},
	
	

	/*
	* Get the label DOM element
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {Element} - The element associated with the label name
	*/
	getLabel: function(labelname) {
		return this.labelHash[labelname];
		
		// TODO Could failsafe at this point and scan the tree; but that would admit the cache is out of date, which shouldn't happen without being caught. How to catch? If Labels box refreshes, we recreate this object (so no problem). If another app injects HTML, our functions are robust enough to react; but we cannot see new labels (nor should we want to - this objects are built on valid label information). 
	},
	
	/** 
	* Get the depth of the label
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {int} - The depth, starting from 0 for root.
	*/
	getDepth: function(labelname) {
		var el = this.getLabel(labelname);
		if( el ) {
			return parseInt(el.getAttribute("gtdi-depth")) || 0;
		} else {
			return 0;
		}
	},
	
	/**
	* Is a label a place holder? (a label we created to fit the hierarchy)
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {boolean} - True if it is a place holder
	*/
	isPlaceHolder: function(labelname) {
		var el = this.getLabel(labelname);
		return ( el && el.getAttribute("gtdi-placeholder") )? true : false;
	},
	
	/**
	* Get the leaf element
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {Element} - The leaf element (can be seperate to the main label element; although not necessarily)
	*/
	getLeaf: function(labelname) {
		var el = this.getLabel(labelname);
		if( el.hasAttribute("gtdi-leaf") ) {
			return el;
		} else {
			return el.ownerDocument.evaluate( ".//*[@gtdi-leaf]", el, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
		}
	},
	
	/**
	* Get the leaf name of the label name
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {string} - The leaf name (e.g. p/gtdinbox/bugs -> bugs)
	*/
	getLeafName: function(labelname) {
		var el = this.getLeaf(labelname);
		if( el ) {
			return el.getAttribute("gtdi-leaf");
		} else {
			return "";
		}
	},
	
	/**
	* Set - or return - a label selected
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @param {boolean} isOn - if true, set it on. if false, set it off. if undefined, return status.
	* @return {boolean} true if on
	*/
	selected: function(labelname, isOn) {
		var el = this.getLabel(labelname);
		if( el ) {
			if( typeof isOn=="undefined" ) {
				return giUIObject.hasClassName(el, "gtdi-active");
			}
			
			if( isOn ) {
				giUIObject.addClassName(el, "gtdi-active");
			} else {
				giUIObject.removeClassName(el, "gtdi-active");			
			}
		}
	},
	
	/**
	* Get all label names that are selected
	* @function ?
	* @return {string[]} - all label names that are selected
	*/
	getSelectedLabelNames: function() {
		var labels = [];
		var result = this.el.ownerDocument.evaluate(".//*[contains(@class,'gtdi-labeldom-label') and contains(@class,'gtdi-active')]", this.el, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
		for( var i = 0; i < result.snapshotLength; i++ ) {
			labels.push( result.snapshotItem(i).getAttribute("gtdLabelName") );
		}
		return labels;
	},
	
	/**
	* Get the expander element (img) for a label
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {Element} - The img expander element
	*/
	getExpander: function(labelname, el) {
		if( !el ) el = this.getLabel(labelname);
		if( el ) {
			return el.ownerDocument.evaluate(".//IMG[contains(@class,'gtdi-labeldom-expander')]", el, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
			//var elExpander = el.getElementsByTagName("IMG")[0];
			//if( elExpander && elExpander.className=="gtdi-labeldom-expander" ) return elExpander;
		}
	},
	
	/** 
	* Is a label expanded to show children?
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @param {Element} elExpander - Optional. The expander img.
	* @return {boolean} - True if label is showing children (expanded)
	*/
	isExpanded: function(labelname, elExpander) {
		elExpander = elExpander || this.getExpander(labelname);
		
		return ((elExpander.src=="chrome://gtdinbox/skin/gtd/collapsetree-open.png")? true : false);
	},
	
	/**
	* Update the expander img to +/-. Does not actually show/hide children (see open/close)
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @param {boolean} open - true if to set to -, false if to set to +
	* @param {Element} elImg - Optional. The expander image.
	*/
	_setExpander: function(labelname, open, elImg) {
		if( !elImg ) elImg = this.getExpander(labelname);
		if( open ) {
			elImg.src = "chrome://gtdinbox/skin/gtd/collapsetree-open.png";
			elImg.title = giI18N.getString("LabelDOM.expander.closer");
		} else {
			elImg.src = "chrome://gtdinbox/skin/gtd/collapsetree-closed.png";
			elImg.title = giI18N.getString("LabelDOM.expander.opener");
		}
	},
	
	/**
	* Get immediate decendant child labels of labelname
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {string[]} Array of child label names one level below specified labelname
	*/
	getChildren: function(labelname) {
		var depth = this.getDepth(labelname) + 1;
		var children = [];
		
		var foundFirstInSortedOrder = false;
		for( l in this.labelHash ) {
			if( l==labelname ) foundFirstInSortedOrder = true;	// Optimised to quickly skip nodes until find desired start point
			
			if( foundFirstInSortedOrder ) {
				var elChild = this.labelHash[l];
				if( elChild.getAttribute("gtdi-depth")==depth && elChild.getAttribute("gtdLabelName").indexOf(labelname)==0 ) {	// Match
					children.push(elChild.getAttribute("gtdLabelName"));
				}
				
				// Check we haven't exceed our label group:
				if( children.length > 0 && elChild.getAttribute("gtdLabelName").indexOf(labelname)!=0 ) break;	// We've gone beyond our labels.
			}
		}
		
		
		return children;
	},
	
	/**
	* Get parent of labelname
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @return {string} - The label name of the parent (or null)
	*/
	getParent: function(labelname) {
		var hierarchy = labelname.split("/");
		if( hierarchy.length > 0 ) {
			hierarchy.length = hierarchy.length - 1;
			var parentLabel = hierarchy.join("/");
			return this.getLabel(parentLabel);
		}
	},
	
	/**
	* Get array of labels contained in this collection
	* @function ?
	* @param {boolean} exclPlaceHolders - If true, remove any place holders (labels that do not exist) from the results
	* @param {string[]} labels - Optional. Existing array to continue to populate.
	* @return {string[]} Array of label names
	*/
	getLabelsArray: function(exclPlaceHolders, labels) {
		var labels = labels || [];
		for( l in this.labelHash ) {
			if( !(exclPlaceHolders && this.isPlaceHolder(l)) ) {
				labels.push(l);
			}
		}
		return labels;
	},
	
	/**
	* Get hash of labels contained in this collection
	* @function ?
	* @param {boolean} exclPlaceHolders - If true, remove any place holders (labels that do not exist) from the results
	* @param {Object} labels - Optional. Hash to continue to populate.
	* @return {Object} Hash of label names, pointing to associated label elements
	*/
	getLabelsHash: function(exclPlaceHolders, labels) {
		var labels = labels || {};
		for( l in this.labelHash ) {
			if( !(exclPlaceHolders && this.isPlaceHolder(l)) ) {
				labels[l] = this.labelHash[l];
			}
		}
		return labels;
	},
	
	/**
	* Open (expand) this label to show children.
	* Unless stated otherwise, this will also ensure all parent nodes leading down to this label (and their direct children - including desired labels' siblings) are also open.
	* @function ?
	* @param {string} labelname - the name of the label to find
	* @param {boolean} ignoreParents - if true, do not check or open any parent nodes. Just this label and it's direct children.
	*/
	open: function(labelname, ignoreParents) {
		// Find labels to open:
		var labels = [labelname];
		if( !ignoreParents ) {
			// Build up list of parents to iteratively process:
			var hierarchy = labelname.split("/");
			while( hierarchy.length > 0 ) {
				hierarchy.length = hierarchy.length - 1;
				var parentLabel = hierarchy.join("/");
				if( this.getLabel(parentLabel) ) labels.push(parentLabel);
			}
		}
        
	
		
		// For each label to process, if has expander; expand it. If it has children, mark them visible.
		for( var i = 0; i < labels.length; i++ ) {
			var el = this.getLabel(labels[i]);	// The label el
			if( el ) {
				// Show it
				el.style.display = "";
				// Open expander if it has one
				if( elExpander = this.getExpander(labels[i]) ) {	// Has expander, ensure set open
					this._setExpander(null, true, elExpander);
				}
				// Open children
				var children = this.getChildren(labels[i]);
				for( var j = 0; j < children.length; j++ ) {
					var elChild = this.getLabel( children[j] );
					if( elChild ) {
						elChild.style.display = "";
					}
				}
			}
		}
	
	},
	
	/**
	* Close (unexpand) this labelname, hide any children, including nested children (which are also closed/unexpanded). 
	* @function ?
	* @param {string} labelname - the name of the label to find. If omitted, close anything that is not a level 0. 
	*/
	close: function(labelname) {
		var foundFirstInSortedOrder = false;
		for( l in this.labelHash ) {
			if( !labelname || l==labelname ) foundFirstInSortedOrder = true;	// Optimised to quickly skip nodes until find desired start point


			if( foundFirstInSortedOrder ) {
				// Get the child node
				var el = this.labelHash[l];
                var elLabelName = el.getAttribute("gtdLabelName");
				if( !labelname || (labelname==elLabelName || elLabelName.indexOf(labelname+'/')==0) ) {	// Match
					if( elExpander = this.getExpander(l) ) this._setExpander(null, false, elExpander);	// Set unexpanded
					if( (!labelname && this.getDepth(l)>0 ) || (labelname && l!=labelname) ) { 
						el.style.display = "none";	// Hide if child of start point
					}
				}
								
				// Check we haven't exceed our label group:
				if( labelname && foundFirstInSortedOrder ) {
                    var elLabelName = el.getAttribute("gtdLabelName");
                    if( elLabelName!=labelname && elLabelName.indexOf(labelname+"/")!=0 ) {
                        break;	// We've gone beyond our labels
                    }
                }
			}
		}
	
	},
	
	/**
	* Clone this object creating new DOM elements. 
	* It purposefully creates a simple structure; ignoring any existing structure. So if you clone the Gmail Labels box, you get a pure labels structure, not Gmail's specific DOM.
	* @function ?
	* @return {gtdLabelDOM} - A clone of this object, with same labels and equivelent (simple, seperate memory space) DOM structure
	*/
	clone: function() {
		var doc = this._main.gmail.canvasDocument;
		// Create a new object, with a blank container.
		var clone = this._main.getUIObject('labelsTree');
		clone.init(this.type);
		
        
		// Clone each existing node in turn; ignoring any previous structural uniqueness.
		var doc = this.el.ownerDocument;
		var result = doc.evaluate(".//*[contains(@class,'gtdi-labeldom-label')]", this.el, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
		for( var i = 0; i < result.snapshotLength; i++ ) {
			var elChild = result.snapshotItem(i);
			
			var elLeaf = elChild;
        
			if( !elChild.hasAttribute("gtdi-leaf") ) elLeaf = doc.evaluate(".//*[@gtdi-leaf]", elChild, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        
			var labelname = elChild.getAttribute("gtdLabelName");
			
			// Create a new label element:
			var elChildClone = doc.createElement("DIV");
			// Copy existing gtd attributes
			for( var j = 0; j < elChild.attributes.length; j++ ) {
				if( elChild.attributes[j].name.indexOf("gtd")==0 ) {
					elChildClone.setAttribute( elChild.attributes[j].name, elChild.attributes[j].value );
				}
			}
			// Set inner text:
			elChildClone.title = labelname;
            var leafDetails = elLeaf.innerHTML || "/";
            leafDetails = leafDetails.replace(/<[^>]+>/g, "");
			elChildClone.setAttribute('gtdi-leaf', leafDetails);
			elChildClone.innerHTML = elLeaf.innerHTML || "/";
			// Set class names:
			elChildClone.className = "gtdi-labeldom-label";
			if( this.isPlaceHolder(labelname) ) elChildClone.className += " gtdi-labeldom-placeholder";
            // Get depth:
            var depth = parseInt(elChildClone.getAttribute("gtdi-depth"));
			// If had expander, clone it - ensure not already present (could have been in elLeaf.innerHTML)
			if( elChildClone.innerHTML.indexOf('gtdi-labeldom-expander')==-1 && this.getExpander(labelname) ) {
				var elExpander = doc.createElement("IMG");
				elExpander.className = "gtdi-labeldom-expander";
				elExpander.src = "chrome://gtdinbox/skin/gtd/collapsetree-closed.png";
				elExpander.title = giI18N.getString("LabelDOM.expander.opener");
                elExpander.style.paddingLeft = (depth * (gtdLabelsTree.IMG_WIDTH_PX+gtdLabelsTree.IMG_GAP_PX)) + "px";
				elChildClone.insertBefore( elExpander, elChildClone.firstChild );
			}
			// Set depth
			elChildClone.style.paddingLeft = (depth * (gtdLabelsTree.IMG_WIDTH_PX+gtdLabelsTree.IMG_GAP_PX)) + (gtdLabelsTree.IMG_WIDTH_PX+gtdLabelsTree.IMG_GAP_PX) + "px";
			// Check depth visibliity
			if( parseInt(elChildClone.getAttribute("gtdi-depth"))>0 ) elChildClone.style.display = "none";
			
			clone.el.appendChild( elChildClone );
			clone.labelHash[labelname] = elChildClone;
		}
		
		return clone;
	},

},{
    
    
    /**
     * The width of the image
     * @property {int}
     */
    IMG_WIDTH_PX: 9,
    
    /**
     * The gap between the right point of the img and the start of the text
     * @property {int}
     */
    IMG_GAP_PX: 4,
    
});
