/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * A UI helper that makes showing/hiding content boxes easy.
 *
 * // TODO Why not extend giUIObject?
 *
 * @interface gtdShowHider
 */
 
var gtdShowHider = giBase.extend({

    // Properties {
    
    /**
    * The gtd object for the page
    * @property {private gtdMainPagePlugin} ?
    */
    _gtd: null,
    
    /** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
    _eventsContainer: null,
    
    /**
    * The ID for this Show/Hider. Used to store show/hide status long-term
    * @property {int} ?
    */
    id: null,
    
    /**
    * The container for the entire ShowHider
    * @property {Element} ?
    */
    el: null,
    
    /**
    * The container for the title elements in the ShowHider (elTriangle and elTitleTxt)
    * @property {Element} ?
    */
    elTitle: null,
    
    /**
    * The image used to show whether the ShowHider is shown or hidden
    * @property {Element} ?
    */
    elTriangle: null,
    
    /**
    * The text that displays the title of this ShowHider
    * @property {Element} ?
    */
    elTitleTxt: null,
    
    /**
    * The content element
    * @property {Element} ?
    */
    elContent: null,


    // }
    
    // Methods {
    
    
    /**
    * The constructor.
    * Initializes the object.
    * @constructor ?
    * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
    * @param {string} id - The unique ID for this ShowHider. Used to remember show/hide status long-term
    * @param {Object} options - A collection of options for how the ShowHider is created. 
    * @param {boolean} show - Force the starting show/hide status of the ShowHider
    */
    constructor: function(gtd, id, options, show) {
        giLogger.log('gtdShowHider object created -> ' + id, 'gtdShowHider');

        this._main = gtd;
        this._eventsContainer = new giEventsContainer(this);
        this.id = id;
                        
        if( options ) this.create(options);
        
        // Decide show status:
        if( typeof show!="undefined" ) {
            this.show(show);
        } else if( this._main ) {	// Read preferences
            var showHiders = this._main.prefs.getPref('showHiders') || {};
            if( typeof showHiders[this.id]!='undefined' ) this.show(showHiders[this.id]);
            else this.show(true);
        }
    },

    /**
    * Create the element structure
    * @function ?
    * @param {Object} options - A collection of options for how the ShowHider is created. 
    */
    create: function(options) {
        // Create the ShowHider:
        var doc = this._main.gmail.canvasDocument;
        if( !options.type ) options.type = "create";

        this.elTitleTxt = doc.createElement("SPAN");
        this.elTriangle = doc.createElement("IMG");
        switch( options.type ) {
            case "create":	// Create a new element structure
                this.el = doc.createElement("DIV");
                this.el.className = "showhider";

                if( options.elParent && options.elBefore ) options.elParent.insertBefore( this.el, options.elBefore );
                else if( options.elParent ) options.elParent.appendChild( this.el );
                
                this.setTitle(options.title);
                this.elTitle = doc.createElement("DIV");
                this.elContent = options.elContent;
                if( !this.elContent ) this.elContent = doc.createElement("DIV");
                this.el.appendChild( this.elTitle );
                this.el.appendChild( this.elContent );
                break;
            case "attach":	// Merge into an existing element structure
                this.el = options.el;
                var els = this.el.getElementsByTagName("*");
                for( var i = 0; i < els.length; i++ ) {
                    if( giUIObject.hasClassName(els[i], options.titleClass) ) {
                        this.elTitle = els[i];
                        break;
                    }
                }
                this.setTitle( this.elTitle.innerHTML.replace(/<.+?>/g,"") );
                this.elTitle.innerHTML = "";
                this.elContent = this.elTitle.nextSibling;
                if( !this.elContent ) {
                    this.elContent = doc.createElement("DIV");
                    this.el.appendChild( this.elContent );
                }
                break;
        }
        this.elTitle.appendChild( this.elTriangle );
        this.elTitle.appendChild( this.elTitleTxt );
                
        giUIObject.addClassName(this.elTitle, "showhider-title");
        giUIObject.addClassName(this.elContent, "showhider-content");
        
        this.elTitle.style.cursor = "pointer";
        if( !options.inline ) {
            this.elTitleTxt.style.marginLeft = "3px;"
            if( !options.noContentIndent ) this.elContent.style.paddingLeft = "14px";
        }
        this._eventsContainer.observe( this.elTitle, "mousedown", this._mouseDownTitle, false );
        
        var elClear = doc.createElement("DIV");
        elClear.style.clear = "both";
        elClear.style.height = "1px";
        this.elTitle.appendChild( elClear );
    },

    /**
    * Update the text of the title element
    * @function ?
    * @param {string} t - The new title
    */
    setTitle: function(t) {
        this.elTitleTxt.innerHTML = t;
    },

    /**
    * Set the content to either visible or invisible; or if no parameter supplied, return the show status.
    * @function ?
    * @param {boolean} v - True: make it shown; false: make it hidden; undefined: return true/false
    */
    show: function(v) {
        if( typeof v=="undefined" ) return this.elContent.style.display=="";
        
        
        if( v ) {
            this.elTriangle.src = "chrome://gtdinbox/skin/opentriangle.gif";
            giUIObject.addClassName(this.el, "showhider-open");
            this.elContent.style.display = "";
            if(this.hasEventListener(gtdShowHider.SHOW)) this.dispatchEvent({name: gtdShowHider.SHOW});
            if(this.hasEventListener(gtdShowHider.TOGGLE)) this.dispatchEvent({name: gtdShowHider.TOGGLE});
        } else {
            giUIObject.removeClassName(this.el, "showhider-open");
            this.elTriangle.src = "chrome://gtdinbox/skin/triangle.gif";
            this.elContent.style.display = "none";
            if(this.hasEventListener(gtdShowHider.HIDE)) this.dispatchEvent({name: gtdShowHider.HIDE});
            if(this.hasEventListener(gtdShowHider.TOGGLE)) this.dispatchEvent({name: gtdShowHider.TOGGLE});
        }
    },

    /**
    * Event in response to a mousedown. Toggle the show/hide status.
    * @function ?
    * @param {Event} event - 
    */	
    _mouseDownTitle: function(event) {
        this.toggle();
        
        if( this._main ) {
            var showHiders = this._main.prefs.getPref('showHiders', true) || {};
            showHiders[this.id] = this.show();
            this._main.prefs.setPref('showHiders', showHiders);
        }
    },

    /**
    * Invert the show/hide status
    * @function ?
    */
    toggle: function() {
        this.show( !this.show() );
    },

    /**
    * Set the entire ShowHider as visible or not.
    * @function ?
    * @param {boolean} v - If true, set visible. If false, set hidden. If undefined, return visible status.
    */
    visible: function(v) {
        if( typeof v=="undefined" ) return (this.el.style.display!="none");
        this.el.style.display = v? "" : "none";
    },
    
    /**
    * Releases items from memory
    * @function ?
    */
    reset: function() {
        this._eventsContainer.reset();
        this.id = null;
        this.el = null;
        this.elTitle = null;
        this.elTriangle = null;
        this.elTitleTxt = null;
        this.elContent = null;
    },
    //}


},{
    // Static Properties {

    /**
     * Dispatched when content made visible
     * @event ?
     */
    SHOW: 'show',

    /**
     * Dispatched when content hidden
     * @event ?
     */
    HIDE: 'hide',
    
    /**
     * Dispatched when ever Show status is modified
     * @event ?
     */
    TOGGLE: 'toggle',
    
    //}
});
gtdShowHider.implement(giEventDispatcher); // Add event dispatching support