/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * A Stats box to put in the sidebar; that builds up stats for the day (storing them in prefs in case of switch off)
 *
 * @class ? @extends glUIView
 */
var gtdStatsDayBox = boxSection.extend({
    // Properties {

    /**
     * Format: {date: 20090101, <type>:{count:<int>, <threadId>:true, <threadId>:true}
     * Types: 
     *  completed.archived_no_action -> ARCHIVE and not present in statusThreadContainer and not in action_finished
     *  completed.action_finished -> statusThreadContainer item removed, on an ADD_LABEL (S/Finished) or REMOVE_LABEL (S/Action). Note this item can be undone by the reappearance of the thread in the statusThreadContainer. Also, if appears here; remove any presence from archived_no_action
     *  completed.spammed -> Detect MARK_NOT_SPAM to remove it
     *  completed.deleted -> Detect UNDELETE?
     *  categorised.actioned -> Detect addition to statusThreadContainer
     *  categorised.labelled -> Detect an ADD_LABEL that is not a status
     * @property {Object}
     */
    _dayStats: null,

    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {gtdMainPagePlugin} main - The gtdMainPagePlugin object for the page
     */
    connect: function(gtd) {
        giLogger.log('Connecting gtdComposeView object to view', 'gtdComposeView');

        this._gtd = gtd;

        // Now get the box plugin: TODO This is extremely ugly. Need to build into plugin framework.
        var box = giPageManager.getPluginInstance(boxMain, this._gtd.gmail);

        this._labelsTrees = {};
        
        this.id = 'gtdstatsdaybox';
        this.base(box);

        // Construct the UI:
        var doc = this._gtd.gmail.canvasDocument;
        this.elRow = doc.createElement("UL");
        this.elRow.className = "gtdi-stats-notifier gtdi-threecol";
        
        this._elRowCategorised = doc.createElement("LI");
        this._elRowCategorised.className = "gtdi-categorised";
        this.elRow.appendChild( this._elRowCategorised );
        this._elRowCategorisedItem = doc.createElement("SPAN");
        this._elRowCategorisedItem.innerHTML = 0;
        this._elRowCategorised.appendChild( this._elRowCategorisedItem );
        
        this._elRowCompleted = doc.createElement("LI");
        this._elRowCompleted.className = "gtdi-done";
        this.elRow.appendChild( this._elRowCompleted );
        this._elRowCompletedItem = doc.createElement("SPAN");
        this._elRowCompletedItem.innerHTML = 0;
        this._elRowCompleted.appendChild( this._elRowCompletedItem );
        
        this._elRowToDo = doc.createElement("LI");
        this._elRowToDo.className = "gtdi-toprocess";
        this.elRow.appendChild( this._elRowToDo );
        this._elRowToDoItem = doc.createElement("SPAN");
        this._elRowToDoItem.innerHTML = 0;
        this._elRowToDo.appendChild( this._elRowToDoItem );
        
        // Render popout:
        this.elPopout = doc.createElement("DIV");
        
        this._popoutRows = {};    // Format {<section>: {<rowname>: {elRow: <Element>, elCount: <Element>}}}
        
        // Popout / categorised:
        this._popoutRows.categorised = {};
        var elCategorisedHeader = doc.createElement("H2");
        elCategorisedHeader.innerHTML = giI18N.getString("stats.daybox.popout.categorised");
        this.elPopout.appendChild( elCategorisedHeader );
        
        var categorised = ['actioned', 'labelled'];
        for( var i = 0; i < categorised.length; i++ ) {
            this._popoutRows.categorised[categorised[i]] = {};
            this._popoutRows.categorised[categorised[i]].elRow = doc.createElement("DIV");
            this._popoutRows.categorised[categorised[i]].elRow.className = "gtdi-stats-daybox-popout-row";
            this._popoutRows.categorised[categorised[i]].elRow.innerHTML = giI18N.getString("stats.daybox.popout.categorised."+categorised[i]);
            
            this._popoutRows.categorised[categorised[i]].elItem = doc.createElement("SPAN");
            this._popoutRows.categorised[categorised[i]].elRow.appendChild( this._popoutRows.categorised[categorised[i]].elItem );
            
            this.elPopout.appendChild( this._popoutRows.categorised[categorised[i]].elRow );
        }
        
        // Add the row:
        this._addRow( 'daybox', this.elRow, this.elPopout );
        

        // Load the prefs:
        this._dayStats = this._gtd.prefs.getPref('stats.day');
        var d = new Date();
        var todayStr = d.getFullYear() + d.getMonth() + d.getDate();
        if( !this._dayStats || this._dayStats.date!=todayStr ){
            this._dayStats = {date: todayStr};
        }
        if( !this._dayStats.categorised ) this._dayStats.categorised = {};
        if( !this._dayStats.categorised.actioned ) this._dayStats.categorised.actioned = {};
        if( !this._dayStats.categorised.labelled ) this._dayStats.categorised.labelled = {};
        if( !this._dayStats.completed ) this._dayStats.completed = {};
        if( !this._dayStats.completed.archived_no_action ) this._dayStats.completed.archived_no_action = {};
        if( !this._dayStats.completed.action_finished ) this._dayStats.completed.action_finished = {};
        if( !this._dayStats.completed.spammed ) this._dayStats.completed.spammed = {};
        if( !this._dayStats.completed.deleted ) this._dayStats.completed.deleted = {};
        // Now check counts of threads for each type:
        var sections = [this._dayStats.categorised, this._dayStats.completed];
        for( var i = 0; i < sections.length; i++ ) {
            var section = sections[i];
            for( t in section ) {
                if( section.hasOwnProperty(t) ) {
                    var count = 0;
                    for( p in section[t] ) {
                        if( p!="count" && section[t].hasOwnProperty(p) ) {
                            count = count + 1;
                        }
                    }
                    section[t].count = count;
                }
            }
        }
        // Write prefs back out:
        this._gtd.prefs.setPref('stats.day', this._dayStats);
        
        

        this._gtd.gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
        this._gtd.statusThreadsContainer.addEventListener(gtdThreadContainer.CHANGED, this._threadChanged, this);
        this._gtd.statusThreadsContainer.addEventListener(gtdStatusThreadContainer.LABEL_ZERO, this._labelZero, this);
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting gtdComposeView from view', 'gtdComposeView');
            
        this._gtd.gmail.actions.removeEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
        this._gtd.statusThreadsContainer.removeEventListener(gtdThreadContainer.CHANGED, this._threadChanged, this);
        this._gtd.statusThreadsContainer.removeEventListener(gtdStatusThreadContainer.LABEL_ZERO, this._labelZero, this);
        
        this.base();
    },
    
    /**
     * Detect an event that affects the stats - gtdThreadContainer.CHANGED and glActionManager.THREAD_CHANGED
     * @function {}
     */
    _threadChanged: function(evt) {
        
        
    // completed.archived_no_action -> ARCHIVE and not present in statusThreadContainer and not in action_finished
    // completed.action_finished -> statusThreadContainer item removed, on an ADD_LABEL (S/Finished) or REMOVE_LABEL (S/Action). Note this item can be undone by the reappearance of the thread in the statusThreadContainer. Also, if appears here; remove any presence from archived_no_action
     
        
        // Test for status labels:
        var statusData = this._gtd.labelsData.getStatusData();
        
        var changed = false;
        
        if( evt.name==glActionManager.THREAD_CHANGED ) {
        
            
            if( evt.action==glActionManager.ARCHIVE ) {
                // Check its not present in statusThreadContainer and not in action finished
                if( !this._dayStats.completed.action_finished[evt.thread] && !this._gtd.statusThreadsContainer.getThreadById(evt.thread) ) {
                    this._dayStats.completed.archived_no_action[evt.thread] = true;
                    changed = true;
                }
            }
            
            
        } else if( evt.name==gtdThreadContainer.CHANGED ) {
            
            if( evt.addedThreads.length>0 ) {
                // Undo any threads that were previously finished and have been re-actioned
                for( var i = 0; i < evt.addedThreads.length; i++ ) {
                    var threadId = evt.addedThreads[i];
                    if( this._dayStats.completed.action_finished[threadId] ) {
                        delete this._dayStats.completed.action_finished[threadId];
                        changed = true;
                    }
                }
            }
            
            if( evt.removedThreads.length>0 ) {
                // These were Status'd items that were finished.
                for( var i = 0; i < evt.removedThreads.length; i++ ) {
                    var threadId = evt.removedThreads[i];
                    this._dayStats.completed.action_finished[threadId] = true;
                    if( !this._dayStats.completed.archived_no_action[threadId] ) delete this._dayStats.completed.archived_no_action[threadId];
                }
                changed = true;
            }
            
            
        }
        
        
        
        if( changed ) {
            // Write out the pref
            this._main.prefs.setPref('stats.day', this._dayStats);
        }
    },
    
    /**
     * Notification a label hit zero
     * @function {}
     */
    _labelZero: function(evt) {

    },
    
    //}
});