/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * Adds command buttons to the thread list
 *
 * @class ? @extends glUIView
 */
var gtdThreadListCommands = giUIObject.extend({
    // Properties {
	
	/** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
		
	/** @property {Element} ? The thread list container */
	elTL: null,
	
	/** @property {Element[]} ? Array of Thread List tables parents*/
	tableContainerEls: null,
	
	/** @property {private giMultiToOne} ? Reduce rapid function calls into a single one per Xms */
    _multiToOne: null,
    	
	/** @property {private Boolean[]} ? The rows checked before we kick in */
	_lastCheckedRows: null,
	
	/** @property {private glThread} */
	_currentThread: null,
    
    /**
     * The popup element
     * @property {Element}
     */
    elPopup: null,
    
    /**
     * The row icon that initiated the popup
     * @property {Element}
     */
    _currentPopupIcon: null,
    
    /**
     * The timeout ID to close the popup
     * @property {int}
     */
    _popupTimeout: null,
    	
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {gtdMainPagePlugin} main - The gtdMainPagePlugin object for the page
     */
    connect: function(main) {
        giLogger.log('Connecting gtdThreadListCommands object to view', 'gtdThreadListCommands');

        this.base(main);
        
        this.elTL = this._main.gmail.page.getViewRoot();
		this.tableContainerEls = [];
        this._eventsContainer = new giEventsContainer(this);
        this._multiToOne = new giMultiToOne(500);
        
        this.elPopup = this._main.gmail.canvasDocument.createElement("DIV");
        this.elPopup.id = "gtdi-tlcmds-popup";
        this.elPopup.style.display = "none";
        this._main.gmail.canvasDocument.body.appendChild( this.elPopup );
        var doc = this._main.gmail.canvasDocument;
        var lineitems = ['preview', 'action', 'finish', 'archive', 'del', 'spam'];
        for( var i = 0; i < lineitems.length; i++ ) {
            var elLi = doc.createElement("DIV");
            elLi.className = "gtdi-tlcmds-popup-lineitem";
            elLi.setAttribute("gtdi-tlcmds-id", lineitems[i]);
            elLi.innerHTML = giI18N.getString('ThreadListUI.GTDBox.Popup.lineitem.'+lineitems[i]);
            this.elPopup.appendChild( elLi );
        }
        
        this._eventsContainer.observe( this.elTL, "mousedown", this._click, true);
        this._eventsContainer.observe( this.elTL, "mouseover", this._mouseOver, true);
        this._eventsContainer.observe( this.elPopup, "mousedown", this._click, true);
        
        if( !this._main.gmail.offline.isUnstable() ) {
            this._refresh();
        }
        
        this._main.gmail.offline.addEventListener(glOffline.GO_ONLINE, this._offlineStateChange, this);
        this._main.gmail.offline.addEventListener(glOffline.GO_UNSTABLE, this._offlineStateChange, this);
    },
	
    
    /**
     * Notification that the Offline state changed
     * @function {} ?
     */
    _offlineStateChange: function(evt) {
        if( evt.name==glOffline.GO_ONLINE ) {
            this._refresh();
        } else if( evt.name==glOffline.GO_UNSTABLE ) {
            this._listenDOM(false);
            var doc = this._main.gmail.canvasDocument;
            var result = doc.evaluate( "//div[contains(@class,'glViewRoot')]//IMG[@class='gtd-row-button']", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
            for( var i = 0; i < result.snapshotLength; i++ ) {
                var el = result.snapshotItem(i);
                el.parentNode.removeChild( el );
            }
        }
    },
    
	/**
	* Listen for changes to the ThreadList; and if changed, do refresh
	* @function ?
	* @param {boolean} v - If true, listen to changes. If not, ignore changes.
	*/
	_listenDOM: function(v) {
		// Remove any listeners regardless
		if( this.tableContainerEls ) {
			for( var i = 0; i < this.tableContainerEls.length; i++ ) this._eventsContainer.stopObserving(this.tableContainerEls[i], "DOMNodeInserted", this._domNodeInserted, false);
		}
		
		// Refresh the list of Tables
        var elTLs = this._main.gmail.getCurrentViewObject().getThreadListTables();
        this.tableContainerEls.length = 0;
        for( var i = 0; i < elTLs.length; i++ ) this.tableContainerEls[i] = elTLs[i].parentNode.parentNode;
        
        if( v ) {
            // Listen to the tables
            for( var i = 0; i < this.tableContainerEls.length; i++ ) {
                this._eventsContainer.observe(this.tableContainerEls[i], "DOMNodeInserted", this._domNodeInserted, false);
            }
        }
	},
	
        
	/**
	* Reduce event-calls for changes to dom-node-inserted; turning any event calls within a 400ms into a single refresh call
	* @function ?
	*/
	_domNodeInserted: function(event) {
		try {
			this._listenDOM(false);
			this._multiToOne.process("threadListDomChange", giUtil.bind(this, this._refresh));
		} catch(e) {
			giLogger.warn(e, 'gtdThreadListCommands._domNodeInserted');
		}
	},
	
	
	/**
	* Ensure we have cmd-icons added to each row in the thread list
	* @function ?
	*/
	_refresh: function() {
					
						
		try {
			this._listenDOM(false);
			
			var doc = this._main.gmail.canvasDocument;
			
			var elImg = doc.createElement("IMG");
			elImg.className = "gtd-row-button";
			elImg.style.position = "absolute";
			elImg.style.right = "0px";
			elImg.src = "chrome://gtdinbox/skin/gtd/gtd_lineitem.png";
			elImg.title = giI18N.getString("ThreadListUI.GTDBox.Tooltip");
			
			
			var previewEnabled = this._main.prefs.getPref('components.thread_list.preview_button.enabled');
						
            if( previewEnabled ) {
                var result = this._main.gmail.getCurrentViewObject().getLastTDInRows();
                if( result.snapshotLength>0 ) {
                    if( (result.snapshotItem(0).innerHTML.indexOf("gtd")==-1 || result.snapshotItem(result.snapshotLength-1).innerHTML.indexOf("gtd")==-1) ) {
                        for( var i = 0; i < result.snapshotLength; i++ ) {
                            var elImgC = elImg.cloneNode(true);
                            result.snapshotItem(i).appendChild( elImgC );
                        }
                    }
                }
            }
			
		} catch(e) {
			giLogger.log('Error in gtdThreadListCommands.refresh -> ' + e.toString(), 'gtdThreadListCommands');
		}
		
		this._listenDOM(true);
		
		
	    giLogger.log('refreshed', 'gtdThreadListCommands');
	},
	
	
	/**
	* If clicked cmd icon, process the click
	* @function ?
	*/
	_click: function(event) {
		try {
			// Retrieve what was clicked, and verify cmd icon
			var elCmd = event.target;
			if( !(elCmd.className=="gtd-row-button" || elCmd.getAttribute('gtdi-tlcmds-id')) ) {
				// Did not click a button
				return;
			}
			
			giUtil.stopEvent(event);
			
			var doc = this._main.gmail.canvasDocument;
			
			// Retrieve the gmail threadListView functionality
			var gmailThreadListView = this._main.gmail.getCurrentViewObject();
			
			// Uncheck all checkboxes
			gmailThreadListView.uncheckRows();
			
            
            if( elCmd.className=='gtd-row-button' || elCmd.getAttribute('gtdi-tlcmds-id')=='preview' ) {
                // Preview box:
                var threadListPreviewer = this._main.getUIObject('threadListPreviewer');
                threadListPreviewer.load( this._currentPopupIcon, this._currentPopupIcon, true );
            } else {
                
                // Retrieve associated checkbox
                var el = this._currentPopupIcon;
                while( el.parentNode && el.nodeName!="TR" ) el = el.parentNode;
                el = el.getElementsByTagName("INPUT")[0];
                if( el.checked==false ) giUIObject.simulateMouseEvent(el, "click");	// Check it
                
                // Get thread id:
                var threadId = gmailThreadListView.getThreadId( this._currentPopupIcon );
                
                // Main container (only one that can archive without a refresh):
                switch( elCmd.getAttribute('gtdi-tlcmds-id') ) {
                    case 'archive':
                        this._main.gmail.actions.performAction(glActionManager.MARK_AS_READ, threadId, glActionManager.USE_AJAX);
                        this._main.gmail.actions.performAction(glActionManager.ARCHIVE, threadId, glActionManager.USE_DOM);
                        break;
                    case 'finish':
                        var status = this._main.labelsData.getStatusData();
						if( this._main.labelsData.hasLabel(status.finished) ) this._main.gmail.actions.performAction(glActionManager.ADD_LABEL, threadId, glActionManager.USE_AJAX, status.finished )
						this._main.gmail.actions.performAction(glActionManager.MARK_AS_READ, threadId, glActionManager.USE_AJAX);
						this._main.gmail.actions.performAction(glActionManager.ARCHIVE, threadId, glActionManager.USE_DOM);
                        break;
                    case 'action':
                        var status = this._main.labelsData.getStatusData();
                        if( this._main.labelsData.hasLabel(status.action) ) this._main.gmail.actions.performAction(glActionManager.ADD_LABEL, threadId, glActionManager.USE_AJAX, status.action )
                        this._main.gmail.actions.performAction(glActionManager.MARK_AS_READ, threadId, glActionManager.USE_AJAX);
                        this._main.gmail.actions.performAction(glActionManager.ARCHIVE, threadId, glActionManager.USE_DOM);
                        break;
                    case 'del':
                        this._main.gmail.actions.performAction(glActionManager.DELETE, threadId, glActionManager.USE_DOM);
                        break;
                    case 'spam':
						this._main.gmail.actions.performAction(glActionManager.REPORT_SPAM, threadId, glActionManager.USE_DOM);
                        break;
                }
            }
            
            this._hidePopup();
            
            
            
			
			/*
			// Choose action
			if( elCmd.className=="gtd-row-button" ) {
				if( event.button==0 && !event.ctrlKey ) {
					if( event.shiftKey ) {	//
					} else {			// Preview
                        var threadListPreviewer = this._main.getUIObject('threadListPreviewer');
                        threadListPreviewer.load( elCmd, elCmd, true );
					}
				} else if( event.button==2 || event.ctrlKey ) {	// Action label + archive
					var threadId = gmailThreadListView.getThreadId( elCmd );
					// add label
					var status = this._main.labelsData.getStatusData();
					if( this._main.labelsData.hasLabel(status.action) ) this._main.gmail.actions.performAction(glActionManager.ADD_LABEL, threadId, glActionManager.USE_AJAX, status.action )
					this._main.gmail.actions.performAction(glActionManager.MARK_AS_READ, threadId, glActionManager.USE_AJAX);
					this._main.gmail.actions.performAction(glActionManager.ARCHIVE, threadId, glActionManager.USE_DOM);
				}
			} else if( elCmd.className=="gtd-row-bin-button" ) {
				if( event.button==0 && !event.ctrlKey ) {
					if( event.shiftKey ) {	// Spam
						var threadId = gmailThreadListView.getThreadId( elCmd );
						this._main.gmail.actions.performAction(glActionManager.REPORT_SPAM, threadId, glActionManager.USE_DOM);
					} else {			// Finish label + archive
						var threadId = gmailThreadListView.getThreadId( elCmd );
						// add label
						var status = this._main.labelsData.getStatusData();
						if( this._main.labelsData.hasLabel(status.finished) ) this._main.gmail.actions.performAction(glActionManager.ADD_LABEL, threadId, glActionManager.USE_AJAX, status.finished )
						this._main.gmail.actions.performAction(glActionManager.MARK_AS_READ, threadId, glActionManager.USE_AJAX);
						this._main.gmail.actions.performAction(glActionManager.ARCHIVE, threadId, glActionManager.USE_DOM);
					}
				} else if( event.button==2 || event.ctrlKey ) {	// Delete
                    var threadId = gmailThreadListView.getThreadId( elCmd );
					this._main.gmail.actions.performAction(glActionManager.DELETE, threadId, glActionManager.USE_DOM);
				}
			}
            */
		} catch(e) {
			giLogger.warn(e, 'gtdThreadListCommands._click', this._main.gmail);
		}
	},
	
	_mouseOver: function(evt) {
        
        // Very quickly test if this is a preview icon:
        if( evt.target.className=='gtd-row-button' ) {
            try {
                var el = evt.target;
                if( this._currentPopupIcon!=el ) {
                    // We have a new one
                    
                    if( this._popupTimeout ) {
                        // Stop the close
                        clearTimeout(this._popupTimeout);
                        this._popupTimeout = null;
                    }
                    
                    // See if we need to apply remover
                    var observeMove = false;
                    if( this._currentPopupIcon ) {
                        this._currentPopupIcon.src = "chrome://gtdinbox/skin/gtd/gtd_lineitem.png";
                    } else {
                        observeMove = true;
                    }
                    this._currentPopupIcon = el;
                    this._currentPopupIcon.src = "chrome://gtdinbox/skin/gtd/gtd_lineitem-hover.png";
                    
                                        
                    // Render the popup
                    var offsetInfo = this._main.gmail.getCurrentViewObject().getOffsetInfo(el);
                    var permitted = null;
                    if( offsetInfo.isMultiple ) {
                        permitted = {preview: true};
                    } else {
                        var status = this._main.labelsData.getStatusData();
                        
                        // Get the thread:
                        var thread = this._main.gmail.getCurrentViewObject().getThread( this._currentPopupIcon )
                        
                        // Can we include action?
                        var permitAction = true;
                        if( !this._main.labelsData.hasLabel(status.action) ) permitAction = false;
                        if( permitAction && thread ) {
                            // Test for no other status labels
                            for( var i = 0; i < thread.labels.length; i++ ) {
                                if( gtdLabelsData.testPrefix( thread.labels[i], status.prefix ) ) {
                                    permitAction = false;
                                    break;
                                }
                            }
                        }
                        
                        permitted = {preview: true, del:true, spam:true};
                        if( offsetInfo.locationHash.indexOf('inbox')>=-1 ) permitted.archive = true;
                        permitted.action = permitAction
                        permitted.finish = this._main.labelsData.hasLabel(status.finished);
                        
                    }
                    for( var i = 0; i < this.elPopup.childNodes.length; i++ ) {
                        var elLi = this.elPopup.childNodes[i];
                        elLi.style.display = permitted[elLi.getAttribute('gtdi-tlcmds-id')]? '' : 'none';
                    }
                    
                    // Show the popup
                    this.elPopup.style.display = "";
                    
                    // Position it
                    var viewport = giUIObject.viewport(this._main.gmail.canvasDocument);
                    var offset = giUIObject.cumulativeOffset(this._currentPopupIcon);
                    var dims = giUIObject.getDimensions(this._currentPopupIcon);
                    var popupDims = giUIObject.getDimensions(this.elPopup);
                    var xy = giUIObject.boundsCheckXY(this.elPopup, 0, offset[1], true);
                    this.elPopup.style.top = xy[1] + "px";
                    if( (dims.width + offset[0] + popupDims.width)>viewport[0] ) {
                        // It is beyond the border, show left
                        this.elPopup.style.left = (offset[0] - popupDims.width) + "px";
                    } else {
                        // It is within the border, show right
                        this.elPopup.style.left = offset[0] + dims.width + "px";
                    }
		            
                    // Apply remover
                    if( observeMove ) {
                        this._eventsContainer.observe(this._main.gmail.canvasDocument.body, 'mousemove', this._mouseMove, false);
                    }
                }
                
            } catch(e) {
                giLogger.warn(e, 'gtdThreadListCommands._mouseOver');
            }
        }
    },
    
    /**
     * Notification of a mousemove in the window; used to detect whether we should close the popup
     * @function {} ?
     */
    _mouseMove: function(evt) {
        // Are we within either the popup icon or the popup?
        var el = evt.target;
        var i = 3;
        while( --i>=0 && el!=this._currentPopupIcon && el!=this.elPopup && el.parentNode ) {
            el = el.parentNode;
        }
        if( el==this._currentPopupIcon || el==this.elPopup ) {
            // Safe
            if( this._popupTimeout ) {
                // Stop the close
                clearTimeout(this._popupTimeout);
                this._popupTimeout = null;
            }
            return;
        }
        
        // Outside of choice, hide it
        if( !this._popupTimeout ) {
            this._popupTimeout = setTimeout(giUtil.bind(this, this._hidePopup), 150);
        }
    },
    
    /**
     * Hide the popup
     * @function {} ?
     */
    _hidePopup: function() {
        this._eventsContainer.stopObserving(this._main.gmail.canvasDocument.body, 'mousemove', this._mouseMove, false);
        this.elPopup.style.display = "none";
        this._currentPopupIcon.src = "chrome://gtdinbox/skin/gtd/gtd_lineitem.png";
        this._currentPopupIcon = null;
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        giLogger.log('Disconnecting gtdThreadListCommands from view', 'gtdThreadListCommands');
        
        this._main.gmail.offline.removeEventListener(glOffline.GO_ONLINE, this._offlineStateChange, this);
        this._main.gmail.offline.removeEventListener(glOffline.GO_UNSTABLE, this._offlineStateChange, this);
        
        if( this._popupTimeout ) clearTimeout( this._popupTimeout );
        
        if( this.elPopup.parentNode ) this.elPopup.parentNode.removeChild( this.elPopup );
        this._currentPopupIcon = null;
        
        this.elTL = null;
        this.tableContainerEls = null;
        this._eventsContainer.reset();
        this._multiToOne.reset();
        this.base();
    },
    
    

    //}
});
