/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 * A list of threads, grouped into containers
 *
 * @class ? @extends giUIObject
 */
var gtdThreadList = giUIObject.extend({
    // Properties {
    
	/**
	* The thread list container
	* @property {Element} ?
	*/
	el: null,

    /**
	* Everything except the optional Action bar
	* @property {Element} ?
	*/	
	_elInner: null,

	/**
	* Container for the GroupBy icons
	* @property {Element} ?
	*/	
	_elGroupByContainer: null,
    
    /**
     * The bar at the bottom of the thread list
     * @property {Element}
     */
    _elActionsBar: null,
    
	
	/** @property {private giEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
	
	
	/**
     * A hash of all glThread objects; keyed on threadId.
     * Format: {threadId: glThread}
     * @property {private Object}
	 */
	_threadsHash: null,
    
    /**
     * Hash of group bys
     * Format: {id: String, data: gtdThreadListGroupBy, el: Element}
     * @property {private Object}
     */
    _groupBys: null,
    	
    /**
     * Array of group objects
     * @property {private gtdThreadListGroup[]}
     */
    _groups: null,
    
    /**
     * A quick look up of groups, based on ID
     * @property {private Object}
     */
    _groupsHash: null,
    
    /**
     * Optional property.
     * Only used to detect changes to threads, so grouping and threadListThread objects can be updated.
     * Otherwise has to run .updateThread event and is less efficient.
     * @property {gtdThreadContainer}
     */
    _threadContainer: null,
    
    /**
     * @property {gtdThreadListDownloadProgress}
     */
    _threadListDownloadProgress: null,
    
    //}


	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} main - The gtdMainPagePlugin object for the page
	*/
	connect: function(main) {
		giLogger.log('Connecting gtdThreadList object to view', 'gtdThreadList');
		
		this.base(main);
		this._eventsContainer = new giEventsContainer(this);
        this._threadsHash = {};
        this._groups = [];
        this._groupsHash = {};
        this._groupBys = {};
		
		var doc = main.gmail.canvasDocument;
        this.el = doc.createElement("DIV");
		this.el.id = "gtdi-threads";
        this.el.setAttribute("gtdi-resizable", "1");
        
		this._elInner = doc.createElement("DIV");
		this._elInner.id = "gtdi-threadlists";
        this._elInner.setAttribute("gtdi-remember-scroll", "1");    // Used by gtdPopup to restore scroll offset
        this._elInner.setAttribute("gtdi-resizable", "1");
        this.el.appendChild( this._elInner );
		
        // Add base bar:
        var elActions = doc.createElement("DIV");
		elActions.id = "gtdi-threadtoolbar";
		this.el.appendChild( elActions );
		
		this._elActionsBar = doc.createElement("DIV");
		this._elActionsBar.className = "gtdi-btns";
		elActions.appendChild( this._elActionsBar );
		
		this._labelsMenu = this._main.getUIObject("labelsMenu");
		this._elActionsBar.appendChild( this._labelsMenu.el );
        
        // Add group by container:
        this._elGroupByContainer = doc.createElement("DIV");
        this._elGroupByContainer.className = "gtdi-btn-grp gtdi-icons gtdi-threadlist-groupby";
        this._elActionsBar.appendChild( this._elGroupByContainer );
        
        
        // Load GroupBy
        this.addGroupBy( new gtdThreadListGroupByProject(this._main) );
        this.addGroupBy( new gtdThreadListGroupByContext(this._main) );
        this.addGroupBy( new gtdThreadListGroupByContact(this._main) );
        this.addGroupBy( new gtdThreadListGroupByList(this._main) );
        this.selectGroupBy('project');
        
		// Add listeners:
		this._eventsContainer.observe(this.el, "click", this._click, false);		
	},

	/**
	* Called on page shutdown, it does cleanup
	* @function ?
	*/
	disconnect: function() {
		giLogger.log('Disconnecting gtdThreadList from view', 'gtdThreadList');
		
        
        if( this._threadContainer ) {
            this._threadContainer.removeEventListener(gtdThreadContainer.CHANGED, this._threadChanged, this);
            this._threadContainer = null;
        } else {
            this._main.gmail.actions.removeEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
        }
        
        if( this._threadListDownloadProgress ) {
            this._threadListDownloadProgress.disconnect();
            this._threadListDownloadProgress = null;
        }
        
        
		if( this.el && this.el.parentNode ) this.el.parentNode.removeChild(this.el);
		this.el = null;
		this._elGroupByContainer = null;
        this._elActionsBar = null;
		
		
        if( this._labelsMenu ) {
            this._labelsMenu.disconnect();
            this._labelsMenu = null;
        }
        
        this._threadsHash = null;
        
        
        for( var i = 0; i < this._groups.length; i++ ) {
            this._groups[i].disconnect();
        }
        this._groups = null;
        this._groupsHash = null;
        
        this._groupBys = null;

		this._threadsHash = null;
        this._eventsContainer.reset();
		this._eventsContainer = null;
		
		this.base();
	},
	
    
    
    /**
     * Set the threads to use
     * Detect changes from what we currently have; only adding/removing what is different
     * @function {} ?
     * @param {glThread[]} threads
     * @param {optional gtdThreadContainer} threadContainer - If present, use it to detect changes to a thread that might affect our grouping
     */
    setThreads: function(threads, threadContainer) {
        
                
        // Decide how we're adding. If currently no threads, add each simply and then sort. If just a handful, sort them as add.
        var liteAdd = true;
        
        // Find what has changed:
        var seenThreads = {}; // Ids of threads checked, used to detect removals
        added = [];
        removed = [];
        for( var i = 0; i < threads.length; i++ ) {
            if( this._threadsHash[threads[i].threadId] ) {
                // We have it
            } else {
                // It is new, add it
                added.push(threads[i]);
            }
            seenThreads[threads[i].threadId] = true;
        }
        // Now detect possible removals (do it before _addThread, otherwise builds up threadsHash too much):
        for( t in this._threadsHash ) {
            liteAdd = false;    // Has current items
            if( !seenThreads[this._threadsHash[t].threadId] ) {
                removed.push( this._threadsHash[t] );
            }
        }
        
                       
        
        // Add new threads:
        for( var i = 0; i < added.length; i++ ) {
            this._threadsHash[added[i].threadId] = added[i];
            
            // Add it to a group:
            this._addThreadToGroups(added[i], liteAdd);
            
        }
        // Remove unwanted threads:
        for( var i = 0; i < removed.length; i++ ) {
            // Remove from groups - can safely try all (and clean group out if necessary):
            for( var j = 0; j < this._groups.length; j++ ) {
                var group = this._groups[j];
                group.removeThread(removed[i]);
                if( group.getNumberThreads()==0 ) {
                    // Remove group
                    group.disconnect();
                    var pos = this._groups.indexOf(group);
                    this._groups.splice(pos, 1);
                    delete this._groupsHash[group.id];
                }
                group = null;
            }
            // Remove the thread from our own records:
            delete this._threadsHash[removed[i].threadId];
        }
        
        if( liteAdd ) {
            // Now sort all affected groups:
            for( var i = 0; i < this._groups.length; i++ ) {
                this._groups[i].sort();
            }
        }
        
        if( added.length>0 || removed.length>0 ) {
            // Something changed, check sort order of groups:
            this.sortGroups();
        }
        
        this._testShowNoThreads();
        
        // Update thread container:
        if( threadContainer!=this._threadContainer ) {
                        
            // Remove old container listeners:
            if( this._threadContainer ) {
                this._threadContainer.removeEventListener(gtdThreadContainer.CHANGED, this._threadChanged, this);
            } else {
                this._main.gmail.actions.removeEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
            }
            
            this._threadContainer = threadContainer;
            
            // Add new container listeners:
            if( this._threadContainer ) {
                this._threadContainer.addEventListener(gtdThreadContainer.CHANGED, this._threadChanged, this);
            } else {
                this._main.gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);
            }
        }
        
        // Update Download Progress:
        if( !this._threadListDownloadProgress && this._threadContainer ) {
            // Create download progress:
            this._threadListDownloadProgress = this._main.getUIObject('threadListDownloadProgress');
            if( this._elInner.firstChild ) {
                this._elInner.insertBefore( this._threadListDownloadProgress.el, this._elInner.firstChild );
            } else {
                this._elInner.appendChild( this._threadListDownloadProgress.el );
            }
        }
        if( this._threadListDownloadProgress ) this._threadListDownloadProgress.setThreadContainer(this._threadContainer);
        
    },
	
    /**
     * Add a thread - find its group(s) and place it
     * @function {} ?
     * @param {glThread} thread
     * @param {optional Boolean} lite - If true, instruct group to do a lite add (unsorted, not shown in UI)
     */
    _addThreadToGroups: function(thread, lite) {
        
        // Get the group ids:
        var groupIds = this._getActiveGroupBy().data.groupThread(thread);
        
        for( var i = 0; i < groupIds.length; i++ ) {
            // Check the group exists:
            var group = this._groupsHash[groupIds[i][0]];
            if( !group ) {
                // Create it:
                group = this._createGroup(groupIds[i][0], groupIds[i][1]);
            }
            // Add the thread to the group:
            group.addThread( thread, lite );
        }
                
    },
    
    /**
     * Create (and return) a new gtdThreadListGroup
     * @function {gtdThreadListGroup} ?
     * @param {String} id - The new group id
     * @return The new Group object
     */
    _createGroup: function(id, displayName) {
        // Create:
        var group = this._main.getUIObject('threadListGroup');
        group.load(id, displayName, this._getActiveGroupBy().data);
        
        // Add meta data:
        var browserTypeItem = this._getActiveGroupBy().data.getBrowserTypeAndItem(group);
        if( browserTypeItem ) {
            // Make it clickable:
            group.elTitle.setAttribute("gtdType", browserTypeItem[0] );
            group.elTitle.setAttribute("gtdItem", browserTypeItem[1] );
            // Add CSS:
            giUIObject.addClassName(group.elTitle, "gtdi-group-has-browser-type");
        }
        
        // Store it:
        this._groups.push(group);
        this._groupsHash[id] = group;
        
        return group;
    },
    
    
    /**
     * Sort the groups into correct order
     * @function {} ?
     */
    sortGroups: function() {
        
        // Sort the groups according to GroupBy:
        var groupBy = this._getActiveGroupBy().data;
        if( groupBy.sortGroups ) this._groups.sort(groupBy.sortGroups);
        
        
        // Simple better - should not be inefficient to move around
        for( var i = 0; i < this._groups.length; i++ ) {
            if( this._groups[i].id=="misc" && i!=this._groups.length-1) {
                // Push it to the end
                this._groups.push(this._groups[i]);
                this._groups.splice(i, 1);
            }
            this._elInner.appendChild( this._groups[i].el );
        }
    },
    
    /**
     * Notification that a thread has changed.
     * @function {} ?
     */
    _threadChanged: function(evt) {
        var threadsChanged = []; 
        if( evt.name==glActionManager.THREAD_CHANGED ) {
            var thread = this._threadsHash[evt.thread];
            if( thread ) {
                thread.updateFromEvent(evt);
                threadsChanged.push(thread);
            }
        } else if( evt.name==gtdThreadContainer.CHANGED ) {
            for( var i = 0; i < evt.updatedThreads.length; i++ ) {
                var thread = this._threadsHash[evt.updatedThreads[i].threadId];
                if( thread ) {
                    threadsChanged.push(thread);
                }
            }
        }
        
        
        // Process each thread, and then each group for it. Remove from no longer relevant; update where found; or add.
        for( var i = 0; i < threadsChanged.length; i++ ) {
            var thread = threadsChanged[i];
           
            // Get the newly relevant groups:
            var groupIdsFull = this._getActiveGroupBy().data.groupThread(thread);
            var groupIds = [];
            for( var j = 0; j < groupIdsFull.length; j++ ) groupIds.push( groupIdsFull[j][0] );
            
            // Test against each group:
            for( var j = 0; j < this._groups.length; j++ ) {
                var group = this._groups[j];
                
                // See if associated with group:
                var groupFound = groupIds.indexOf(group.id);
                var tlThread = group.getThreadListThread(thread);
                if( tlThread ) {
                    // See if group still relevant:
                    if( groupFound==-1 ) {
                        // Remove thread from group:
                        group.removeThread(thread);
                        // TODO Check if only one remaining
                    } else {
                        // Update thread:
                        tlThread.refresh();
                    }
                } else {
                    // Do we need to add it?
                    if( groupFound>-1 ) {
                        group.addThread(thread);
                    }
                }
                // Remove it from groupIds (so we can see if we missed any groups)
                if( groupFound>-1 ) {
                    groupIds.splice(groupFound, 1);
                }
            }
            
            // Create any groups from those that remain:
            for( var j = 0; j < groupIds.length; j++ ) {
                var group = this._createGroup(groupIds[j]);
                group.addThread(thread);
            }
        }
        
        if( threadsChanged.length > 0 ) {
            // TODO Could make this happen only if thread added/removed?
            this.sortGroups();
        }

    },
    
    
    /**
     * Add a GroupBy option
     * @function {} ?
     * @param {gtdThreadListGroupBy} groupBy - The GroupBy data object
     */
    addGroupBy: function(groupBy) {
        var doc = this._main.gmail.canvasDocument;
        // Create the element:
        var el = doc.createElement("DIV");
        el.setAttribute('gtdi-groupby', groupBy.id);
        el.className = "gtdi-btn";
        el.innerHTML = "<b class='gtdi-btn-t'></b><b class='gtdi-btn-t2'></b><b class='gtdi-btn-m'></b><b class='gtdi-btn-b2'></b><b class='gtdi-btn-b'></b><span class='gtdi-btn-label' title='" + groupBy.displayName + "'><img src='"+groupBy.iconSrc+"' /></span>";
        this._elGroupByContainer.appendChild( el );
        // Store the data:
        this._groupBys[groupBy.id] = {id: groupBy.id, data: groupBy, el: el};
    },
    
    /**
     * Get the active group by item
     * @function {Object} ?
     * @return Format: {data: gtdThreadListGroupBy, el: Element, id: String} (same as in this._groupBys)
     */
    _getActiveGroupBy: function() {
        var elGroupBy = this.el.ownerDocument.evaluate(".//DIV[@gtdi-groupby and contains(@class, 'gtdi-active')]", this._elGroupByContainer, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        if( elGroupBy ) {
            return this._groupBys[elGroupBy.getAttribute("gtdi-groupby")];
        }
    },
    
    /**
     * Mark the tab as selected and fire notification event.
     * It is expected that child classes will override this function to provide specific action.
     * @function {} ?
     * @param {String} id - The tab id
     */
    selectGroupBy: function(id) {
        
        
        // Clean up current item:
        var currentGroupBy = this._getActiveGroupBy();
        if( currentGroupBy ) {
            if( currentGroupBy.id==id ) return; // No change
            
            giUIObject.removeClassName( currentGroupBy.el, 'gtdi-active' );
        }
        
        var doc = this._main.gmail.canvasDocument;
        
        // Get the newly selected item and make it active:
        var elGroupBy = doc.evaluate(".//*[@gtdi-groupby='"+id+"']", this._elGroupByContainer, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        if( elGroupBy ) {
            giUIObject.addClassName( elGroupBy, 'gtdi-active');
        }
        
                
        // Enforce the groupby:
        // Disconnect all groups (no longer relevant):
        for( var i = 0; i < this._groups.length; i++ ) {
            this._groups[i].disconnect();
        }
        this._groups.length = 0;
        this._groupsHash = {};
        
        // Rebuild them:
        for( t in this._threadsHash ) {
            this._addThreadToGroups(this._threadsHash[t], true);
        }
        
        // Sort groups:
        for( g in this._groups ) {
            this._groups[g].sort();
        }
        this.sortGroups();
        
    },
    
	/**
	* If conditions right (no threads, finished searching), show the No Threads message.
	* @function {} ?
	*/
	_testShowNoThreads: function() {
		var elNoThreads = this._elInner.ownerDocument.evaluate( ".//*[@class='gtdi-no-threads']", this._elInner, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue
		if( elNoThreads ) elNoThreads.parentNode.removeChild( elNoThreads );
        var foundThread = false;
        for( t in this._threadsHash ) {
            foundThread = true;
            break;
        }
		if( !foundThread ) {
			var elNoThreads = this._elInner.ownerDocument.createElement("DIV");
			elNoThreads.className = "gtdi-no-threads";
			elNoThreads.innerHTML = giI18N.getString('ThreadList.NoThreads');
			this._elInner.appendChild( elNoThreads );
		}
	},
	
	
	/**  
	* Get array of glThread objects that have been checked in the UI
	* @function ?
	* @return {glThread[]} - Array of threads that have been ticked in checkbox
	*/
	getCheckedThreads: function() {
		var checkedThreads = [];
        for( g in this._groups ) {
            var groupThreads = this._groups[g].getThreads();
            for( var i = 0; i < groupThreads.length; i++ ) {
                if( groupThreads[i].isChecked() && checkedThreads.indexOf(groupThreads[i].thread)==-1 ) {
                    checkedThreads.push( groupThreads[i].thread );
                }
            }
        }
		return checkedThreads;
	},
	
	/** 
	* Respond to a click anywhere in the thread list
	* Handle label Xing, star-ing, and clicking links to a thread.
	* @function ?
	*/
	_click: function(event) {
		try {
			var el = event.target;
			

            var thread = this._getThreadFromEl(el);
            while( el.parentNode && el!=this.el ) { 
                if ( el.getAttribute("gtdi-groupby") ) {
                    // It's a groupby item:
                    this.selectGroupBy( el.getAttribute("gtdi-groupby") );
                    break;
                } else if( el.className=="gtdi-labelremove" ) {		// Label-X
                    var labelname = el.parentNode.getAttribute("gtdi-labelname");
                    this._main.gmail.actions.performAction( glActionManager.REMOVE_LABEL, thread, null, labelname );	// Notification of success will be detected the gtdThreadContainer, which will notify us to refresh that thread.
                    break;
                /*
                } else if( el.parentNode.className=="gtdi-star" ) { 	// Star
                    if( el.src=="chrome://gtdinbox/skin/gtd/i/gtdi-star-off.png" ) {	// Turn it on
                        this._main.gmail.actions.performAction( glActionManager.STAR, thread );
                    } else {								// Turn it off
                        this._main.gmail.actions.performAction( glActionManager.UNSTAR, thread );				
                    }
                    break;
                */
                } else if( el.nodeName=="INPUT" ) {	// Check box
                    if( this._labelsMenu ) {
                        // Get the checked threads - TODO, could keep a stack in memory and add/remove items
                        var checkedThreads = this.getCheckedThreads();
                        // Pass to Labels menu
                        this._labelsMenu.loadThreads(checkedThreads);
                    }
                    if( this.hasEventListener(gtdThreadList.CHECK_THREAD)) this.dispatchEvent({name: gtdThreadList.CHECK_THREAD, thread:thread, checked:el.checked});
                    break;
                } else if( el.className=='gtdi-open-in-gmail' ) { // Popout link
                    this.dispatchEvent({name: gtdBrowser.LOAD_TYPE, item: thread, newWindow: event.shiftKey, openInGmail: true, type: gtdBrowser.TYPES.THREAD});
                    break;
                } else if(thread && el.parentNode==this.el) {	// Thread
                    // Broadcast
                    var threads = []; // Generate threads in same order as groups. TODO Can we optimise? Or cache after first execute?
                    for( g in this._groups ) {
                        var groupThreads = this._groups[g].getThreads();
                        for( var i = 0; i < groupThreads.length; i++ ) {
                            threads.push(groupThreads[i].thread);
                        }
                    }
                    this.dispatchEvent({name: gtdBrowser.LOAD_TYPE, threads: threads, item: thread, type: gtdBrowser.TYPES.THREAD});
                    break;
                }
                
                el = el.parentNode;
            }

		} catch(e) {
			giLogger.warn(e, "gtdThreadList._click", this._main.gmail);
		}
	},
	
    /**
     * Using an el from somewhere in our environment, get the thread associated with it
     * @function {gtdThreadListThread} ?
     * @param {element} el - An el within a thread list
     */
    _getThreadFromEl: function(el) {
        while( el.parentNode && el!=this.el && !el.getAttribute("gtdi-thread-id") ) el = el.parentNode;
        if( el.getAttribute("gtdi-thread-id") ) return this._threadsHash[el.getAttribute("gtdi-thread-id")];
    },
	
	/**
	* Set method for updating server. Ensure future created objects can use this info; and existing objects are updated.
	* @function ?
	* @param {String} type - The type. A static variable of glActionManager
	*/
	setServerActionType: function(type) {
        this._labelsMenu.setServerActionType(type);
        // TODO Should our own _click event have access to this?
	},





    //}
},{
    // Static Properties {
    
    /**
     * Notification that a thread was checked
     */
    CHECK_THREAD: 'check_thread',
    
    // }
});
gtdThreadList.implement(giEventDispatcher); // Add event dispatching support
