/*
 * ***** BEGIN LICENSE BLOCK *****
 * GTDInbox source code. Copyright (2009) of The Inbox Foundry Ltd.
 * No reuse without express permission.
 * If you're interested in utilising or working on this code, please contact Andy Mitchell (andy@inboxfoundry.com).
 * ***** END LICENSE BLOCK ***** */


/**
 *
 * @class ? @extends giUIObject
 */
var gtdThreadListGroup = giUIObject.extend({
    // Properties {
    
    /**
     * The ID of the Group
     * @property {String}
     */
    id: null,
    
    /**
     * The display name for the group
     * @property {String}
     */
    displayName: null,
    
	/**
	* The thread list container
	* @property {Element} ?
	*/
	el: null,

    _elThreads: null,
    
    elTitle: null,
	
	/**
     * Array of threadListThread objects
     * @property {private gtdThreadListThread[]}
	 */
	_threads: null,
    
    /**
     * Used as a quick lookup on a threadId
     * Format: {threadId: Boolean}
     * @property {private Object}
     */
    _threadsHash: null,
    
    _groupBy: null,
    
    _eventsContainer: null,
    
    //}


	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} main - The gtdMainPagePlugin object for the page
	*/
	connect: function(main) {
		giLogger.log('Connecting gtdThreadListGroup object to view', 'gtdThreadListGroup');
		
		this.base(main);
		this._eventsContainer = new giEventsContainer(this);
        this._threads = [];
        this._threadsHash = {};
        
		var doc = main.gmail.canvasDocument;
        this.el = doc.createElement("DIV");
		this.el.className = "gtdi-threadlist-group";
        
        
        elHeader = doc.createElement("H2");
        elHeader.className = "gtdi-threadlist-group-title";
        this.el.appendChild( elHeader );
        this.elTitle = doc.createElement("SPAN");
        elHeader.appendChild( this.elTitle );
        
        this._elThreads = doc.createElement("DIV");
        this._elThreads.className = "gtdi-threadlist-group-threads";
        this.el.appendChild( this._elThreads );
		
	},

	/**
	* Called on page shutdown, it does cleanup
	* @function ?
	*/
	disconnect: function() {
		giLogger.log('Disconnecting gtdThreadListGroup from view', 'gtdThreadListGroup');
		
		if( this.el && this.el.parentNode ) this.el.parentNode.removeChild(this.el);
		this.el = null;
        this._elThreads = null;
        this.elTitle = null;
        
        for( var i = 0; i < this._threads.length; i++ ) {
            this._threads[i].disconnect();
        }
        this._threads.length = 0;
        this._threadsHash = null;
		
		this._eventsContainer.reset();
		
        this._groupBy = null;
		this._threads = null;
		this._eventsContainer = null;
		
		this.base();
	},
	
    /**
     * Initiate
     * @function {} ?
     * @param {String} id
     * @param {String} displayName
     * @param {gtdThreadListGroupBy} groupBy 
     */
    load: function(id, displayName, groupBy) {
        this.id = id;
        this.displayName = displayName || id;
        this.elTitle.innerHTML = this.displayName;
        this._groupBy = groupBy;
    },
    
    
    /**
     * Return the number of threads
     * @function {int} ?
     * @return Number of threads
     */
    getNumberThreads: function() {
        return this._threads.length;
    },
    
    /**
     * Return the ThreadListThread object for a thread
     * @function {gtdThreadListThread} ?
     * @param {glThread} thread - The thread to get the corresponding object for
     * @return The threadListThread object or null
     */
    getThreadListThread: function(thread) {
        return this._threadsHash[thread.threadId];
    },
    
    /**
     * Return array of threads
     * @function {} ?
     * @return Live pointer to array of threads
     */
    getThreads: function() {
        return this._threads;
    },
    
    /**
     * Add the thread to our known array.
     * If lite, do not add to UI and do not sort (this will be handled by sortAll later)
     * @function {} ?
     * @param {glThread} thread -
     * @param {Boolean} lite - If true, do not add to UI and do not sort
     */
    addThread: function(thread, lite) {
        
        
        // See if we need to create a ThreadListThread for it:
        if( !this._threadsHash[thread.threadId] ) {
            // Create it:
            var threadListThread = gtdThreadListThread.create(this._main, thread);
            this._threads.push(threadListThread);
            this._threadsHash[thread.threadId] = threadListThread;
        }
        
        if( !lite ) {
            // Sort in place:
            this.sort(this._threadsHash[thread.threadId]);
        }
    },
    
    /**
     * Remove thread from threads array
     * Attempt to remove it from container (although it should do it itself)
     * @function {} ?
     * @param {glThread} thread
     */
    removeThread: function(thread) {
        if( this._threadsHash[thread.threadId] ) {
            var threadListThread = this._threadsHash[thread.threadId];
            threadListThread.disconnect();
            var pos = this._threads.indexOf(threadListThread);
            if( pos>-1 ) this._threads.splice(pos, 1);
            delete this._threadsHash[thread.threadId];
        }
    },
	
    /**
     * Sort threads and display in order
     * @function() ?
     * @param {optional gtdThreadListThread} thread - If present, just sort this into place
     */
    sort: function(thread) {
        if( this._groupBy.sortThreads ) this._threads.sort(this._groupBy.sortThreads);
        if( thread ) {
            var pos = this._threads.indexOf(thread);
            if( pos==this._threads.length-1 || pos>=this._elThreads.childNodes.length ) {
                // Last item - append
                this._elThreads.appendChild(thread.el);
            } else {
                // Insert at point:
                if( this._elThreads.childNodes[pos]!=thread.el ) {
                    this._elThreads.insertBefore(thread.el, this._elThreads.childNodes[pos]);
                }
            }
        } else {
            for( var i = 0; i < this._threads.length; i++ ) {
                this._elThreads.appendChild( this._threads[i].el );
            }
        }
    },
    
	/** 
	* Notification of a click
	* @function ?
	*/
	_click: function(event) {
		try {
			var el = event.target;
			
			
		} catch(e) {
			giLogger.warn(e, "gtdThreadListGroup._click", this._main.gmail);
		}
	},
	
	
    //}
});

