/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Originally developed for GTDInbox for Gmail. http://www.gtdinbox.com
 * Find out more about developing with this codebase at http://www.gtdinbox.com/development.htm
 * Anyone who utilises this code is required to include this license notice in all covered code.
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Base class for a plugin's preferences
 * It has basic handlers for loading/saving simple preferences (any textbox/checkbox with a 'gtdi-preference' attribute)
 * The loading/saving functions are explicitly defined, so inheriting classes have more granular control to override them
 * @class
 */

var gipPluginBase = giBase.extend({
	// Properties {
		
    /**
     * @property {gipPluginContainer}
     */
    _pluginContainer: null,
    
    /**
     * @property {giEventsContainer}
     */
    _eventsContainer: null,
    
	// }
	
	/**
     * Take all the preferences in the page and load them
     * @function {} ?
     * @param {gipPluginContainer} pluginContainer - The plugin container for this account/plugin
	*/
	connect: function(pluginContainer) {
        this._eventsContainer = new giEventsContainer(this);
        
		this._pluginContainer = pluginContainer;
        
        giLogger.log('pluginBase connect - load basic preferences');
        
        // Extract all elements with a preference:
        var result = document.evaluate(".//*[@gtdi-preference]", this._pluginContainer.el, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
        for( var i = 0; i < result.snapshotLength; i++ ) {
            var el = result.snapshotItem(i);
            giLogger.log('load el: ' + el.nodeName + ' [' + el.getAttribute('gtdi-preference') + ']');
            switch( el.nodeName ) {
                case 'checkbox':
                    this._loadCheckbox(el, el.getAttribute('gtdi-preference'));
                    break;
                case 'textbox':
                    this._loadTextbox(el, el.getAttribute('gtdi-preference'));
                    break;
            }
        }
        
        // Listen for changes:
        this._eventsContainer.observe(this._pluginContainer.el, 'keyup', this._save, true); // Text boxes change after update and lose focus
        this._eventsContainer.observe(this._pluginContainer.el, 'mouseup', this._save, true); // Text boxes change after update and lose focus
        this._eventsContainer.observe(this._pluginContainer.el, 'command', this._save, true); // Text boxes change after update and lose focus
        this._eventsContainer.observe(this._pluginContainer.el, 'select', this._save, true); // List/menus are selected
	},
	
    disconnect: function() {
        this._eventsContainer.reset();
        this._pluginContainer = null;
    },
    
    /**
     * Easily over-rided function to laod a preference into a checkbox
     * @function {} ?
     * @param {Element} el - The checkbox
     * @param {String} prefName - the preference name
     */
    _loadCheckbox: function(el, prefName) {
        el.checked = this._pluginContainer.prefManager.getPref(prefName)? true : false;
    },
    
    /**
     * Easily over-rided function to laod a preference into a textbox
     * @function {} ?
     * @param {Element} el - The textbox
     * @param {String} prefName - the preference name
     */
    _loadTextbox: function(el, prefName) {
        el.value = this._pluginContainer.prefManager.getPref(prefName);
    },
    
    /**
     * Fires when an element changes status
     * Pick out preference items of a certain node type and automatically save them.
     * Individual plugins are expected to override this for custom processing.
     * @function {}?
     */
    _save: function(evt) {
        try {
            
            var el = evt.target;
            if( el.hasAttribute('gtdi-preference') ) {
                switch( el.nodeName ) {
                    case 'checkbox':
                        this._saveCheckbox(el, el.getAttribute('gtdi-preference'));
                        break;
                    case 'textbox':
                        this._saveTextbox(el, el.getAttribute('gtdi-preference'));
                        break;
                }
            }
        } catch(e) {
            giLogger.warn(e, 'gipPluginBase._save');
        }
    },
    
    /**
     * Easily over-rided function to save a preference from a checkbox
     * @function {} ?
     * @param {Element} el - The checkbox
     * @param {String} prefName - the preference name
     */
    _saveCheckbox: function(el, prefName) {
        this._pluginContainer.prefManager.setPref(prefName, el.checked, true);
    },
    
    /**
     * Easily over-rided function to save a preference from a textbox
     * @function {} ?
     * @param {Element} el - The textbox
     * @param {String} prefName - the preference name
     */
    _saveTextbox: function(el, prefName) {
        giLogger.log('save textBox: ' + prefName + ', ' + el.value );
        this._pluginContainer.prefManager.setPref(prefName, el.value, true);
    },
	
});